package net.sf.saxon.event;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.StrippedDocument;
import net.sf.saxon.tinytree.TinyDocumentImpl;
import net.sf.saxon.tinytree.TinyBuilder;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.xpath.StandaloneContext;
import net.sf.saxon.dom.DocumentWrapper;
import net.sf.saxon.Configuration;
import net.sf.saxon.tree.TreeBuilder;

import org.xml.sax.*;
import org.w3c.dom.Node;
import org.w3c.dom.Document;

import javax.xml.transform.TransformerException;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import java.util.Date;
import java.io.Writer;
import java.io.PrintWriter;

/**
  * The abstract Builder class is responsible for taking a stream of SAX events
  * and constructing a Document tree. There is one concrete subclass for each
  * tree implementation.
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public abstract class Builder implements Receiver

{
    public final static int STANDARD_TREE = 0;
    public final static int TINY_TREE = 1;

    protected NamePool namePool;
    protected String systemId;
    protected Locator locator;

    // --Recycle Bin (30/06/03 19:53): protected Writer errorOutput = new PrintWriter(System.err);
                                                  // Destination for error messages

    protected DocumentInfo currentDocument;

    protected boolean started = false;
    protected boolean timing = false;

    protected boolean lineNumbering = false;

    private long startTime;

    /**
    * create a Builder and initialise variables
    */

    public Builder() {}

    public void setNamePool(NamePool pool) {
        namePool = pool;
    }

    public NamePool getNamePool() {
        return namePool;
    }

    public void setSystemId(String systemId) {
        this.systemId = systemId;
    }

    public String getSystemId() {
        return systemId;
    }

    public void setDocumentLocator(Locator loc) {
        locator = loc;
    }

    /////////////////////////////////////////////////////////////////////////
    // Methods setting and getting options for building the tree
    /////////////////////////////////////////////////////////////////////////

    /**
    * Set the root (document) node to use. This method is used to support
    * the JAXP facility to attach transformation output to a supplied Document
    * node. It must be called before startDocument(), and the type of document
    * node must be compatible with the type of Builder used.
    */

    public void setRootNode(DocumentInfo doc) {
        currentDocument = doc;
    }


    /**
    * Set timing option on or off
    */

    public void setTiming(boolean on) {
        timing = on;
    }

    /**
    * Get timing option
    */

    public boolean isTiming() {
        return timing;
    }

    /**
    * Set line numbering on or off
    */

    public void setLineNumbering(boolean onOrOff) {
        lineNumbering = onOrOff;
    }

    /**
    * Set output for error messages produced by the default error handler.
    * The default error handler does not throw an exception
    * for parse errors or input I/O errors, rather it returns a result code and
    * writes diagnostics to a user-specified output writer, which defaults to
    * System.err<BR>
    * This call has no effect if setErrorHandler() has been called to supply a
    * user-defined error handler
    * @param writer The Writer to use for error messages
    */

//    public void setErrorOutput(Writer writer) {
//        errorOutput = writer;
//    }

    public void startDocument() throws TransformerException {
        if (timing) {
            System.err.println("Building tree for " + getSystemId() + " using " + getClass());
            startTime = (new Date()).getTime();
        }
    }

    public void endDocument() throws TransformerException {
        if (timing) {
            long endTime = (new Date()).getTime();
            System.err.println("Tree built in " + (endTime-startTime) + " milliseconds");
            if (currentDocument instanceof TinyDocumentImpl) {
                ((TinyDocumentImpl)currentDocument).showSize();
            }
            startTime = endTime;
        }
    }

    /**
    * Get the current document
    * @return the document that has been most recently built using this builder
    */

    public DocumentInfo getCurrentDocument() {
        return currentDocument;
    }

    /**
    * Static method to build a document from any kind of Source object. If the source
     * is already in the form of a tree, it is wrapped as required.
    * @param source Any javax.xml.transform.Source object
    * @param stripper A stripper object, if whitespace text nodes are to be stripped;
     * otherwise null.
     * @param treeModel Either TINY_TREE or STANDARD_TREE. Ignored if a DOMSource or
     * an existing tree is supplied.
    * @return the NodeInfo of the start node in the resulting document object.
    */

    public static NodeInfo build(Source source, Stripper stripper, int treeModel) throws XPathException {
        if (source==null) {
            throw new NullPointerException("Source cannot be null");
        }
        NodeInfo start;
        if (source instanceof DOMSource || source instanceof NodeInfo) {
            if (source instanceof DOMSource) {
                Node dsnode = ((DOMSource)source).getNode();
                if (dsnode instanceof NodeInfo) {
                    start = (NodeInfo)dsnode;
                } else {
                    Document dom;
                    if (dsnode instanceof Document) {
                        dom = (Document)dsnode;
                    } else {
                        dom = dsnode.getOwnerDocument();
                    }
                    DocumentWrapper docWrapper = new DocumentWrapper(dom, source.getSystemId());
                    start = docWrapper.wrap(dsnode);
                }
            } else {
                start = (NodeInfo)source;
            }
            if (stripper != null) {
                DocumentInfo docInfo = start.getDocumentRoot();
                StrippedDocument strippedDoc = new StrippedDocument(docInfo, stripper);
                start = strippedDoc.wrap(start);
            }
            return start;

        } else {
            // we have a SAXSource or StreamSource

            Builder b;
            if (treeModel==Builder.TINY_TREE)  {
                b = new TinyBuilder();
            } else {
                b = new TreeBuilder();
            }
            NamePool pool = NamePool.getDefaultNamePool();
            b.setNamePool(pool);
            Receiver receiver = b;
            if (stripper != null) {
                stripper.setUnderlyingReceiver(b);
                receiver = stripper;
            }
            try {
                new Sender(new Configuration()).send(source, receiver, pool, false);
            } catch (TransformerException err) {
                throw new XPathException.Dynamic(err);
            }
            start = b.getCurrentDocument();
        }
        return start;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
