package net.sf.saxon.event;
import net.sf.saxon.OutputURIResolver;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;
import java.util.Properties;
import java.util.HashMap;
import javax.xml.transform.TransformerException;
import javax.xml.transform.Result;

/**
 * This class is used for generating complex content, that is, the content of an
 * element or document node. It enforces the rules on the order of events within
 * complex content (attributes and namespaces must come first), and it implements
 * part of the namespace fixup rules, in particular, it ensures that there is a
 * namespace node for the namespace used in the element name and in each attribute
 * name.
 *
 * <p>The same ComplexContentOutputter may be used for generating an entire XML
 * document; it is not necessary to create a new outputter for each element node.</p>
 *
 * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
 */



public class ComplexContentOutputter extends Outputter {

	private NamePool namePool;
    private Properties outputProperties;
    private OutputURIResolver outputURIResolver;
    private Receiver receiver;
            // the next receiver in the output pipeline

    private int pendingStartTag = -1;
    private int[] pendingAttCode = new int[20];
    private int[] pendingAttType = new int[20];
    private CharSequence[] pendingAttValue = new String[20];
    private int[] pendingAttProp = new int[20];
    private int pendingAttListSize = 0;

    private int[] pendingNSList = new int[20];
    private int pendingNSListSize = 0;

    private int currentSimpleType = 0;  // non-zero means we are currently writing an
                                        // element of a particular simple type
    private StringBuffer simpleContentBuffer = null;
    private boolean suppressAttributes = false;
                                        // set to true during error recovery only
    private int startElementProperties;


	public void setNamePool(NamePool pool) {
	    namePool = pool;
	}

    public NamePool getNamePool() {
        return namePool;
    }

    public void setSystemId(String systemId) {}

    public String getSystemId() {
        return null;
    }

    public void setOutputURIResolver(OutputURIResolver resolver) {
        outputURIResolver = resolver;
    }
    /**
    * Initialise the outputter for a new output destination, supplying
    * the output format details. <BR>
    * @param props Details of the new output format
    * @param result Details of the new output destination
     * @param characterMapIndex Table of named character maps. May be null.
    */

    public void setOutputDestination(Properties props, Result result, HashMap characterMapIndex)
    throws TransformerException {

        setOutputProperties(props);

        receiver = ResultWrapper.getReceiver(
                result, outputURIResolver, namePool, props, characterMapIndex);

        // add a filter to implement the type-information property

        if (props != null) {
            String typeInfo = props.getProperty(SaxonOutputKeys.TYPE_INFORMATION);
            if ("none".equals(typeInfo)) {
                AnnotationRemover ar = new AnnotationRemover();
                ar.setUnderlyingReceiver(receiver);
                ar.setNamePool(namePool);
                receiver = ar;
            }
        }

		// add a filter to remove duplicate namespaces

		NamespaceReducer ne = new NamespaceReducer();
		ne.setUnderlyingReceiver(receiver);
		ne.setNamePool(namePool);
		receiver = ne;

        startDocument();
    }

    private void setOutputProperties(Properties details) {
        outputProperties = details;
    }

    public Properties getOutputProperties() {
        return outputProperties;
    }

    /**
     * Set the receiver (to handle the next stage in the pipeline) directly
     */

    public void setReceiver(Receiver receiver) {
        this.receiver = receiver;
    }

    /**
     * Start the output process
     */

    public void startDocument() throws TransformerException {
        receiver.startDocument();
        previousAtomic = false;
    }

    /**
    * Produce text content output. <BR>
    * Special characters are escaped using XML/HTML conventions if the output format
    * requires it.
    * @param s The String to be output
    * @exception TransformerException for any failure
    */

    public void characters(CharSequence s, int properties) throws TransformerException {
        if (s==null) return;
        int len = s.length();
        if (len==0) return;
        if (pendingStartTag != -1) {
            startContent();
        }
        receiver.characters(s, properties);
        if (simpleContentBuffer!=null) {
            simpleContentBuffer.append(s);
        }
        previousAtomic = false;
    }

    /**
    * Output an element start tag. <br>
    * The actual output of the tag is deferred until all attributes have been output
    * using attribute().
    * @param nameCode The element name code
    */

    public void startElement(int nameCode, int typeCode, int properties) throws TransformerException {
        // System.err.println("StartElement " + nameCode);
        if (currentSimpleType != 0 && currentSimpleType != Type.UNTYPED_ANY) {
            throw new TransformerException("An element created with a type annotation cannot have child elements");
        }
        if (nameCode == -1) {
            // This is used on error recovery when there is a bad element name. It is needed so that
            // the outputter can suppress any subsequent namespace and attribute nodes
            suppressAttributes = true;
            return;
        } else {
            suppressAttributes = false;
        }

        startElementProperties = properties;

        if (typeCode == 0 || typeCode == Type.UNTYPED_ANY) {
            simpleContentBuffer = null;
        } else {
            simpleContentBuffer = new StringBuffer();
        }
        if (pendingStartTag != -1) {
            startContent();
        }
        pendingAttListSize = 0;
        pendingNSListSize = 0;
        pendingStartTag = nameCode;
        currentSimpleType = typeCode;
        previousAtomic = false;
    }


    /**
    * Output a namespace declaration. <br>
    * This is added to a list of pending namespaces for the current start tag.
    * If there is already another declaration of the same prefix, this one is
    * ignored.
    * Note that unlike SAX2 startPrefixMapping(), this call is made AFTER writing the start tag.
    * @param nscode The namespace code
    * @throws TransformerException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void namespace(int nscode, int properties)
    throws TransformerException {

        if (suppressAttributes) return;

        // System.err.println("Write namespace prefix=" + (nscode>>16) + " uri=" + (nscode&0xffff));
        if (pendingStartTag==-1) {
            throw new NoOpenStartTagException("Cannot write a namespace declaration when there is no open start tag");
        }

        // elimination of namespaces already present on an outer element of the
        // result tree is now done by the NamespaceReducer.

        // Handle declarations whose prefix is duplicated for this element.

        boolean rejectDuplicates = (properties & ReceiverOptions.REJECT_DUPLICATES) != 0;

        for (int i=0; i<pendingNSListSize; i++) {
            if (nscode == pendingNSList[i]) {
                // same prefix and URI: ignore this duplicate
                return;
            }
        	if ((nscode>>16) == (pendingNSList[i]>>16)) {
        	    if (rejectDuplicates) {
        	        throw new TransformerException("Cannot create two namespace nodes with the same name");
        	    } else {
        		    // same prefix, do a quick exit
        		    return;
        		}
        	}
        }

        // It is an error to output a namespace node for the default namespace if the element
        // itself is in the null namespace, as the resulting element could not be serialized

        if (((nscode>>16) == 0) && ((nscode&0xffff)!=0)) {
            if ((namePool.allocateNamespaceCode(pendingStartTag) & 0xffff) == 0) {
                throw new TransformerException("Cannot output a namespace node for the default namespace when the element is in no namespace");
            }
        }

        // if it's not a duplicate namespace, add it to the list for this start tag

        if (pendingNSListSize+1 > pendingNSList.length) {
            int[] newlist = new int[pendingNSListSize * 2];
            System.arraycopy(pendingNSList, 0, newlist, 0, pendingNSListSize);
            pendingNSList = newlist;
        }
        pendingNSList[pendingNSListSize++] = nscode;
        previousAtomic = false;
    }


    /**
    * Output an attribute value. <br>
    * This is added to a list of pending attributes for the current start tag, overwriting
    * any previous attribute with the same name. <br>
    * This method should NOT be used to output namespace declarations.<br>
    * @param nameCode The name of the attribute
    * @param value The value of the attribute
    * @param properties Bit fields containing properties of the attribute to be written
    * @throws TransformerException if there is no start tag to write to (created using writeStartTag),
    * or if character content has been written since the start tag was written.
    */

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties) throws TransformerException {
        //System.err.println("Write attribute " + nameCode + "=" + value + " to Outputter " + this);

        if (suppressAttributes) return;

        if (pendingStartTag==-1) {
            throw new NoOpenStartTagException("Cannot write an attribute when there is no open start tag");
        }

        if (currentSimpleType != 0 && currentSimpleType != Type.UNTYPED_ANY) {
            throw new TransformerException("An element created with a simple type annotation cannot have attributes");
        }

        // if this is a duplicate attribute, overwrite the original, unless
        // the REJECT_DUPLICATES option is set.

        for (int a=0; a<pendingAttListSize; a++) {
            if (pendingAttCode[a] == nameCode) {
                if ((properties & ReceiverOptions.REJECT_DUPLICATES) == 0) {
                    pendingAttType[a] = typeCode;
                    pendingAttValue[a] = value;
                    pendingAttProp[a] = properties;
                    return;
                } else {
                    throw new TransformerException("Duplicate attribute: " +
                            namePool.getDisplayName(nameCode));
                }
            }
        }

        // otherwise, add this one to the list

        if (pendingAttListSize >= pendingAttCode.length) {
            int[] attCode2 = new int[pendingAttListSize*2];
            int[] attType2 = new int[pendingAttListSize*2];
            String[] attValue2 = new String[pendingAttListSize*2];
            int[] attProp2 = new int[pendingAttListSize*2];
            System.arraycopy(pendingAttCode, 0, attCode2, 0, pendingAttListSize);
            System.arraycopy(pendingAttType, 0, attType2, 0, pendingAttListSize);
            System.arraycopy(pendingAttValue, 0, attValue2, 0, pendingAttListSize);
            System.arraycopy(pendingAttProp, 0, attProp2, 0, pendingAttListSize);
            pendingAttCode = attCode2;
            pendingAttType = attType2;
            pendingAttValue = attValue2;
            pendingAttProp = attProp2;
        }

        pendingAttCode[pendingAttListSize] = nameCode;
        pendingAttType[pendingAttListSize] = typeCode;
        pendingAttValue[pendingAttListSize] = value;
        pendingAttProp[pendingAttListSize] = properties;
        pendingAttListSize++;
        previousAtomic = false;
    }

	/**
	* Check that the prefix for an element or attribute is acceptable, allocating a substitute
	* prefix if not. The prefix is acceptable unless a namespace declaration has been
	* written that assignes this prefix to a different namespace URI. This method
	* also checks that the element or attribute namespace has been declared, and declares it
	* if not.
	*/

	private int checkProposedPrefix(int nameCode, int seq) throws TransformerException {
		int nscode = namePool.allocateNamespaceCode(nameCode);
		            // TODO: can we ensure that the namespace code is already in the
		            // namepool, allowing use of getNamespaceCode rather than
		            // allocateNamespaceCode, to avoid synchronization?
		int nsprefix = nscode>>16;

        for (int i=0; i<pendingNSListSize; i++) {
        	if (nsprefix == (pendingNSList[i]>>16)) {
        		// same prefix
        		if ((nscode & 0xffff) == (pendingNSList[i] & 0xffff)) {
        			// same URI
        			return nameCode;	// all is well
        		} else {
        			String prefix = getSubstitutePrefix(nscode, seq);
        			int newCode = namePool.allocate(
        								prefix,
        								namePool.getURI(nameCode),
        								namePool.getLocalName(nameCode));
        			namespace(namePool.allocateNamespaceCode(newCode), 0);
        			return newCode;
        		}
        	}
        }
        // no declaration of this prefix: declare it now
        namespace(nscode, 0);
        return nameCode;
    }

    /**
    * It is possible for a single output element to use the same prefix to refer to different
    * namespaces. In this case we have to generate an alternative prefix for uniqueness. The
    * one we generate is based on the sequential position of the element/attribute: this is
    * designed to ensure both uniqueness (with a high probability) and repeatability
    */

    private String getSubstitutePrefix(int nscode, int seq) {
    	String prefix = namePool.getPrefixFromNamespaceCode(nscode);
        //return prefix + "_" + (nscode&0xffff);
        return prefix + "_" + seq;
        //String uri = namePool.getURIFromNamespaceCode(nscode);
    }

    /**
    * Output an element end tag.
    */

    public void endElement() throws TransformerException {
        //System.err.println("Write end tag " + this + " : " + name);
        if (pendingStartTag != -1) {
            startContent();
        }

        if (simpleContentBuffer != null) {
            try {
                new StringValue(simpleContentBuffer).convert(currentSimpleType);
            } catch (XPathException err) {
                throw new TransformerException("Element content does not match type annotation", err);
            }
        }

        // write the end tag

        receiver.endElement();
        currentSimpleType = 0;
        simpleContentBuffer = null;
        previousAtomic = false;
    }

    /**
    * Write a comment
    */

    public void comment(CharSequence comment, int properties) throws TransformerException {
        if (pendingStartTag != -1) startContent();
        receiver.comment(comment, properties);
        previousAtomic = false;
    }

    /**
    * Write a processing instruction
    */

    public void processingInstruction(String target, CharSequence data, int properties) throws TransformerException {
        if (pendingStartTag != -1) startContent();
        receiver.processingInstruction(target, data, properties);
        previousAtomic = false;
    }

    /**
    * Append an arbitrary item (node or atomic value) to the output
    */

    public void append(Item item) throws TransformerException {
        if (item instanceof AtomicValue) {
            if (previousAtomic) {
                characters(" ", 0);
            }
            characters(item.getStringValue(), 0);
            previousAtomic = true;
        } else if (item instanceof DocumentInfo) {
            SequenceIterator iter = ((DocumentInfo)item).iterateAxis(Axis.CHILD);
            while (iter.hasNext()) {
                append(iter.next());
            }
        } else {
            ((NodeInfo)item).copy(this, NodeInfo.ALL_NAMESPACES, true);
            previousAtomic = false;
        }
    }


    /**
    * Close the output
    */

    public void endDocument() throws TransformerException {
        // System.err.println("Close " + this + " using emitter " + emitter.getClass());
        receiver.endDocument();
        previousAtomic = false;
    }

    /**
    * Flush out a pending start tag
    */

    public void startContent() throws TransformerException {

        if (pendingStartTag == -1) {
            // this can happen if the method is called from outside,
            // e.g. from a SequenceOutputter earlier in the pipeline
            return;
        }

        int elcode = checkProposedPrefix(pendingStartTag, 0);
        receiver.startElement(elcode, currentSimpleType, startElementProperties);

        for (int a=0; a<pendingAttListSize; a++) {
            int attcode = pendingAttCode[a];
            if ((attcode>>20 & 0xff) != 0) {	// non-null prefix
                attcode = checkProposedPrefix(attcode, a+1);
            }
            receiver.attribute( attcode,
                                pendingAttType[a],
                                pendingAttValue[a],
                                pendingAttProp[a] );
        }

        for (int n=0; n<pendingNSListSize; n++) {
            receiver.namespace(pendingNSList[n], 0);
        }

        receiver.startContent();

        pendingAttListSize = 0;
        pendingNSListSize = 0;
        pendingStartTag = -1;
        previousAtomic = false;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
