package net.sf.saxon.event;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.XMLChar;
import net.sf.saxon.value.Type;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;

import javax.xml.transform.TransformerException;

import java.net.URL;
import java.nio.CharBuffer;

/**
  * ContentEmitter is a glue class that provides a standard SAX ContentHandler
  * interface to a Saxon Receiver. To achieve this it needs to map names supplied
  * as strings to numeric name codes, for which purpose it needs access to a name
  * pool. The class also performs the function of assembling adjacent text nodes.
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public class ContentEmitter implements ContentHandler, LexicalHandler, DTDHandler
{
    private NamePool pool;
    private Receiver receiver;
    private boolean inDTD = false;	// true while processing the DTD
    private Locator locator;

    // buffer for accumulating character data, until the next markup event is received

    private char[] buffer = new char[4096];
    private int used = 0;

    // array for accumulating namespace information

    private int[] namespaces = new int[50];
    private int namespacesUsed = 0;

    /**
    * create a ContentEmitter and initialise variables
    */

    public ContentEmitter() {
    }

	public void setReceiver(Receiver e) {
		receiver = e;
	}

	public void setNamePool(NamePool namePool) {
		pool = namePool;
	}

    /**
    * Callback interface for SAX: not for application use
    */

    // Note, all these methods throw SAXException so that a user-defined subclass can throw an exception

    public void startDocument () throws SAXException {
        // System.err.println("ContentEmitter#startDocument");
        try {
            used = 0;
            namespacesUsed = 0;
            receiver.setDocumentLocator(locator);
            receiver.startDocument();
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void endDocument () throws SAXException {
        try {
            flush();
            receiver.endDocument();
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void setDocumentLocator (Locator locator) {
    	this.locator = locator;
        //receiver.setDocumentLocator(locator);
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        //System.err.println("StartPrefixMapping " + prefix + "=" + uri);
    	if (namespacesUsed >= namespaces.length) {
    		int[] n2 = new int[namespacesUsed * 2];
    		System.arraycopy(namespaces, 0, n2, 0, namespacesUsed);
    		namespaces = n2;
    	}
    	namespaces[namespacesUsed++] = pool.allocateNamespaceCode(prefix, uri);
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void endPrefixMapping(String prefix) throws SAXException {}

    /**
    * Callback interface for SAX: not for application use
    */

    public void startElement (String uri, String localname, String rawname, Attributes atts)
    throws SAXException
    {
        // System.err.println("ContentEmitter#startElement " + uri + "," + localname + "," + rawname + " at line " + locator.getLineNumber());
        //for (int a=0; a<atts.getLength(); a++) {
        //     System.err.println("  Attribute " + atts.getURI(a) + "/" + atts.getLocalName(a) + "/" + atts.getQName(a));
        //}
        try {
            flush();

    		int nameCode = getNameCode(uri, localname, rawname);
    		receiver.startElement(nameCode, 0, 0);

    		for (int n=0; n<namespacesUsed; n++) {
    		    receiver.namespace(namespaces[n], 0);
    		}

    		for (int a=0; a<atts.getLength(); a++) {
    		    int attCode = getNameCode(atts.getURI(a), atts.getLocalName(a), atts.getQName(a));

    		    String type = atts.getType(a);
    		    int typeCode = 0;
    		    if (type.equals("ID")) {
    		        typeCode = Type.ID;
    		    } else if (type.equals("IDREF")) {
    		        typeCode = Type.IDREF;
    		    } else if (type.equals("NMTOKEN")) {
    		        typeCode = Type.NMTOKEN;
    		    } else if (type.equals("ENTITY")) {
    		        typeCode = Type.ENTITY;
    		    } else if (type.equals("NOTATION")) {
    		        typeCode = Type.NOTATION;
    		    }
    		    // Ignore CDATA (it basically means "anySimpleType")
    		    // Ignore NMTOKENS, ENTITIES, IDREFS: we can't handle list types yet

    		    receiver.attribute(attCode, typeCode, atts.getValue(a), 0);
    		}

    		receiver.startContent();

            namespacesUsed = 0;

        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    private int getNameCode(String uri, String localname, String rawname) throws SAXException {
        // The XML parser isn't required to report the rawname (qname), though most of them do.
        // If none is provided, we give up
        if (rawname.equals("")) {
            throw new SAXException("Saxon requires an XML parser that reports the QName of each element");
        }
        // It's also possible (especially when using a TransformerHandler) that the parser
        // has been configured to report the QName rather than the localname+URI
        if (localname.equals("")) {
            throw new SAXException("Parser configuration problem: namespace reporting is not enabled");
        }
        String prefix = Name.getPrefix(rawname);
        return pool.allocate(prefix, uri, localname);
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void endElement (String uri, String localname, String rawname) throws SAXException {
        // System.err.println("End element " + rawname);
        try {
            flush();
            // DONE: most Emitters ignore the namecode passed on endElement. It would be more efficient
            // to drop it from the interface, and have those Emitters that need it maintain the stack.
            //int nameCode = namecodes[--namecodesUsed];
            receiver.endElement();
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void characters (char ch[], int start, int length) throws SAXException {
        // System.err.println("characters (" + length + ")");
        // need to concatenate chunks of text before we can decide whether a node is all-white

        while (used + length > buffer.length) {
            char[] newbuffer = new char[buffer.length*2];
            System.arraycopy(buffer, 0, newbuffer, 0, used);
            buffer = newbuffer;
        }
        System.arraycopy(ch, start, buffer, used, length);
        used += length;
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void ignorableWhitespace (char ch[], int start, int length) throws SAXException {
        characters(ch, start, length);
    }

    /**
    * Callback interface for SAX: not for application use<BR>
    */

    public void processingInstruction (String name, String remainder) throws SAXException {
        try {
            flush();
            if (!inDTD) {
                if (name==null) {
                	// trick used by some SAX1 parsers to notify a comment
                	comment(remainder.toCharArray(), 0, remainder.length());
                } else {
                    // some parsers allow through PI names containing colons
                    if (!XMLChar.isValidNCName(name)) {
                        throw new SAXException("Invalid processing instruction name (" + name + ")");
                    }
                	receiver.processingInstruction(name, remainder, 0);
                }
            }
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    /**
    * Callback interface for SAX (part of LexicalHandler interface): not for application use
    */

    public void comment (char ch[], int start, int length) throws SAXException {
        try {
            flush();
            if (!inDTD) {
            	receiver.comment(CharBuffer.wrap(ch, start, length), 0);
            }
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }

    /**
    * Flush buffer for accumulated character data, suppressing white space if appropriate
    */

    private void flush() throws TransformerException {
        if (used > 0) {
            receiver.characters(CharBuffer.wrap(buffer, 0, used), 0);
            used = 0;
        }
    }

    public void skippedEntity(String name) throws SAXException {}

    // No-op methods to satisfy lexical handler interface

	/**
	* Register the start of the DTD. Comments in the DTD are skipped because they
	* are not part of the XPath data model
	*/

    public void startDTD (String name, String publicId, String systemId) throws SAXException {
		inDTD = true;
    }

	/**
	* Register the end of the DTD. Comments in the DTD are skipped because they
	* are not part of the XPath data model
	*/

    public void endDTD () throws SAXException {
		inDTD = false;
    }

    public void startEntity (String name) throws SAXException {};

    public void endEntity (String name)	throws SAXException {};

    public void startCDATA () throws SAXException {};

    public void endCDATA ()	throws SAXException {};

    //////////////////////////////////////////////////////////////////////////////
    // Implement DTDHandler interface
    //////////////////////////////////////////////////////////////////////////////


    public void notationDecl(       String name,
                                    String publicId,
                                    String systemId) throws SAXException
    {}


    public void unparsedEntityDecl( String name,
                                    String publicId,
                                    String systemId,
                                    String notationName) throws SAXException
    {
        //System.err.println("Unparsed entity " + name + "=" + systemId);

        // Some SAX parsers report the systemId as written. We need to turn it into
        // an absolute URL.

        String uri = systemId;
        if (locator!=null) {
            try {
                String baseURI = locator.getSystemId();
                URL absoluteURI = new URL(new URL(baseURI), systemId);
                uri = absoluteURI.toString();
            } catch (Exception err) {}
        }
        try {
            receiver.setUnparsedEntity(name, uri, publicId);
        } catch (TransformerException err) {
            throw new SAXException(err);
        }
    }


}   // end of class ContentEmitter

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
