package net.sf.saxon.event;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;
import javax.xml.transform.Result;
import javax.xml.transform.TransformerException;
import java.util.Stack;
import java.util.Properties;

/**
* A ContentHandlerProxy is an Emitter that filters data before passing it to an
* underlying SAX2 ContentHandler. Relevant events (notably comments) can also be
* fed to a LexicalHandler.

* Note that in general the output passed to an Emitter
* corresponds to an External General Parsed Entity. A SAX2 ContentHandler only expects
* to deal with well-formed XML documents, so we only pass it the contents of the first
* element encountered.
*/

public class ContentHandlerProxy extends Emitter implements Locator
{
    protected ContentHandler handler;
    protected LexicalHandler lexicalHandler;
    protected Locator locator = this;
    private int depth = 0;
    protected boolean requireWellFormed = false;
    private Stack elementStack = new Stack();
    private AttributesImpl pendingAttributes = new AttributesImpl();
    private int pendingElement = -1;

    /**
    * Set the underlying content handler. This call is mandatory before using the Emitter.
    */

    public void setUnderlyingContentHandler(ContentHandler handler) {
        this.handler = handler;
        if (handler instanceof LexicalHandler) {
            this.lexicalHandler = (LexicalHandler)handler;
        }
    }

    /**
    * Set the Lexical Handler to be used. If called, this must be called AFTER
    * setUnderlyingContentHandler()
    */

    public void setLexicalHandler(LexicalHandler handler) {
        this.lexicalHandler = handler;
    }

    /**
    * Set the output details.
    */

    public void setOutputProperties (Properties details) throws TransformerException {
        String prop = details.getProperty(SaxonOutputKeys.REQUIRE_WELL_FORMED);
        if (prop != null) {
            requireWellFormed = prop.equals("yes");
        }
        super.setOutputProperties(details);
    }

    /**
    * Indicate whether the content handler can handle a stream of events that is merely
    * well-balanced, or whether it can only handle a well-formed sequence.
    */

    public void setRequireWellFormed(boolean wellFormed) {
        requireWellFormed = wellFormed;
    }

    /**
    * Set Document Locator
    */

    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    /**
    * Start of document
    */

    public void startDocument() throws TransformerException {
        // System.err.println(this + " startDocument(), handler = " + handler);
        if (handler==null) {
            throw new TransformerException("ContentHandlerProxy.startDocument(): no underlying handler provided");
        }
        try {
            handler.setDocumentLocator(locator);
            handler.startDocument();
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
        depth = 0;
    }

    /**
    * End of document
    */

    public void endDocument() throws TransformerException {
        try {
            handler.endDocument();
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Notify the start of an element
    */

    public void startElement(int nameCode, int typeCode, int properties) throws TransformerException {
        depth++;
        if (depth<=0 && requireWellFormed) {
            notifyNotWellFormed();
        }
        pendingElement = nameCode;
    }

    /**
    * Notify a namespace. Namespaces are notified <b>after</b> the startElement event, and before
    * any children for the element.
    */

    public void namespace(int namespaceCode, int properties) throws TransformerException {
 	    String prefix = namePool.getPrefixFromNamespaceCode(namespaceCode);
    	String uri = namePool.getURIFromNamespaceCode(namespaceCode);
        try {
    	    handler.startPrefixMapping(prefix, uri);
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Notify an attribute. Attributes are notified after the startElement event, and before any
    * children.
    */

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties)
    throws TransformerException {
    	String uri = namePool.getURI(nameCode);
    	String localName = namePool.getLocalName(nameCode);
    	String qname = namePool.getDisplayName(nameCode);
    	int index = pendingAttributes.getIndex(uri, localName);
    	if (index<0) {
    	    // TODO: set the type intelligently
            pendingAttributes.addAttribute(uri, localName, qname, "CDATA", value.toString());
        } else {
            pendingAttributes.setAttribute(index, uri, localName, qname, "CDATA", value.toString());
        }
    }

    /**
    * Notify the start of the content, that is, the completion of all attributes and namespaces.
    * Note that the initial receiver of output from XSLT instructions will not receive this event,
    * it has to detect it itself. Note that this event is reported for every element even if it has
    * no attributes, no namespaces, and no content.
    */

    public void startContent() throws TransformerException {
        try {
            if (depth>0 || !requireWellFormed) {
            	String uri = namePool.getURI(pendingElement);
            	String localName = namePool.getLocalName(pendingElement);
            	String qname = namePool.getDisplayName(pendingElement);

                handler.startElement(
                    uri,
                    localName,
                    qname,
                    pendingAttributes);

                elementStack.push(uri);
                elementStack.push(localName);
                elementStack.push(qname);

                pendingAttributes.clear();
                pendingElement = -1;
            }
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * End of element
    */

    public void endElement() throws TransformerException {
        if (depth>0) {
            try {
                String qname = (String)elementStack.pop();
                String localName = (String)elementStack.pop();
                String uri = (String)elementStack.pop();
                handler.endElement(uri, localName, qname);
            } catch (SAXException err) {
                throw new TransformerException(err);
            }
        }
        depth--;
        // if this was the outermost element, and well formed output is required
        // then no further elements will be processed
        if (requireWellFormed && depth<=0) {
            depth = Integer.MIN_VALUE;     // crude but effective
        }

    }

    /**
    * Character data
    */

    public void characters(CharSequence chars, int properties) throws TransformerException {
        boolean disable = ((properties & ReceiverOptions.DISABLE_ESCAPING) != 0);
        if (disable) setEscaping(false);
        try {
            if (depth<=0 && requireWellFormed) {
                String s = chars.toString();
                boolean isWhite = s.trim().length()==0;
                if (isWhite) {
                    // ignore top-level white space
                } else {
                    notifyNotWellFormed();
                }
            } else {
                handler.characters(chars.toString().toCharArray(), 0, chars.length());
            }
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
        if (disable) setEscaping(true);
    }

    /**
    * The following function is called when it is found that the output is not a well-formed document.
    * Unless the ContentHandler accepts "balanced content", this is a fatal error.
    */

    protected void notifyNotWellFormed() throws TransformerException {
        throw new TransformerException("The result tree cannot be supplied to the ContentHandler because it is not well-formed XML");
    }


    /**
    * Processing Instruction
    */

    public void processingInstruction(String target, CharSequence data, int properties)
    throws TransformerException {
        try {
            handler.processingInstruction(target, data.toString());
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
    }

    /**
    * Output a comment. Passes it on to the ContentHandler provided that the ContentHandler
    * is also a SAX2 LexicalHandler.
    */

    public void comment (CharSequence chars, int properties)
    throws TransformerException {
        try {
            if (lexicalHandler != null) {
                lexicalHandler.comment(chars.toString().toCharArray(), 0, chars.length());
            }
        } catch (SAXException err) {
            throw new TransformerException(err);
        }
    }


    /**
    * Switch escaping on or off. This is called when the XSLT disable-output-escaping attribute
    * is used to switch escaping on or off. It is not called for other sections of output (e.g.
    * element names) where escaping is inappropriate. The action, as defined in JAXP 1.1, is
    * to notify the request to the Content Handler using a processing instruction.
    */

    private void setEscaping(boolean escaping) {
        try {
            handler.processingInstruction(
                (escaping ? Result.PI_ENABLE_OUTPUT_ESCAPING : PI_DISABLE_OUTPUT_ESCAPING),
                "");
        } catch (SAXException err) {}
    }

    ////////////////////////////////////////////////////////////////////
    // dummy implementation of Locator interface
    ////////////////////////////////////////////////////////////////////

    public String getPublicId() {
        return null;
    }

    public int getLineNumber() {
        return -1;
    }

    public int getColumnNumber() {
        return -1;
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
