package net.sf.saxon.event;

import net.sf.saxon.OutputURIResolver;
import net.sf.saxon.dom.DOMEmitter;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.tinytree.TinyBuilder;
import net.sf.saxon.tree.DocumentImpl;
import net.sf.saxon.tree.TreeBuilder;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.stream.StreamResult;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

/**
* Helper class to construct a serialization pipeline for a given result destination
* and a given set of output properties. The pipeline is represented by a Receiver object
* to which result tree events are sent
*/

public class ResultWrapper {

    /**
    * Get a Receiver that wraps a given Result object
    */

    public static Receiver getReceiver( Result result,
                                        OutputURIResolver resolver,
                                        NamePool namePool,
                                        Properties props,
                                        HashMap characterMapIndex)
                                    throws TransformerException {
        if (result instanceof Emitter) {
            ((Emitter)result).setOutputProperties(props);
            return (Emitter)result;
        } else if (result instanceof Receiver) {
            return (Receiver)result;
        } else if (result instanceof SAXResult) {
            ContentHandlerProxy proxy = new ContentHandlerProxy();
            proxy.setUnderlyingContentHandler(((SAXResult)result).getHandler());
            proxy.setNamePool(namePool);
            return proxy;
        } else if (result instanceof StreamResult) {

            // The "target" is the start of the output pipeline, the Receiver that
            // instructions will actually write to (except that other things like a
            // NamespaceReducer may get added in front of it). The "emitter" is the
            // last thing in the output pipeline, the Receiver that actually generates
            // characters or bytes that are written to the StreamResult.

            Receiver target;
            Emitter emitter = null;
            CharacterMapExpander characterMapExpander = null;
            String method = props.getProperty(OutputKeys.METHOD);

            String useMaps = props.getProperty(SaxonOutputKeys.USE_CHARACTER_MAPS);
            if (useMaps != null && characterMapIndex != null) {
                List characterMaps = new ArrayList();
                StringTokenizer st = new StringTokenizer(useMaps);
                while (st.hasMoreTokens()) {
                    String expandedName = st.nextToken();
                    int f = namePool.getFingerprintForExpandedName(expandedName);
                    HashMap map = (HashMap)characterMapIndex.get(new Integer(f));
                    if (map==null) {
                        throw new TransformerException("Character map '" + expandedName + "' has not been defined");
                    }
                    characterMaps.add(map);
                }
                if (characterMaps.size() > 0) {
                    characterMapExpander = new CharacterMapExpander();
                    characterMapExpander.setCharacterMaps(characterMaps);
                }
            }
            if (method==null) {
            	emitter = new UncommittedEmitter();
            	emitter.setNamePool(namePool);
            	target = emitter;
                if (characterMapExpander != null) {
                    characterMapExpander.setUnderlyingReceiver(target);
                    target = characterMapExpander;
                }
            } else if (method.equals("html")) {
                emitter = new HTMLEmitter();
                emitter.setNamePool(namePool);
                target = emitter;
                if (!"no".equals(props.getProperty(OutputKeys.INDENT))) {
                    HTMLIndenter in = new HTMLIndenter();
                    in.setUnderlyingReceiver(target);
                    in.setNamePool(namePool);
                    in.setOutputProperties(props);
                    target=in;
                }
                // TODO: should URI-escape URI-valued HTML attributes without
                // doing character mapping (test charmap009)
                if (characterMapExpander != null) {
                    characterMapExpander.setUnderlyingReceiver(target);
                    target = characterMapExpander;
                }
            } else if (method.equals("xml")) {
                emitter = new XMLEmitter();
                emitter.setNamePool(namePool);
                target = emitter;
                if ("yes".equals(props.getProperty(OutputKeys.INDENT))) {
                    XMLIndenter in = new XMLIndenter();
                    in.setUnderlyingReceiver(target);
                    in.setNamePool(namePool);
                    in.setOutputProperties(props);
                    target=in;
                }
                if (characterMapExpander != null) {
                    characterMapExpander.setUnderlyingReceiver(target);
                    target = characterMapExpander;
                }
                String cdataElements = props.getProperty(OutputKeys.CDATA_SECTION_ELEMENTS);
                if (cdataElements!=null && cdataElements.length()>0) {
                    CDATAFilter filter = new CDATAFilter();
                    filter.setUnderlyingReceiver(target);
                    filter.setNamePool(namePool);
                    filter.setOutputProperties(props);
                    target = filter;
                }
            } else if (method.equals("xhtml")) {
                emitter = new XHTMLEmitter();
                emitter.setNamePool(namePool);
                target = emitter;
                if ("yes".equals(props.getProperty(OutputKeys.INDENT))) {
                    HTMLIndenter in = new HTMLIndenter();
                    in.setUnderlyingReceiver(target);
                    in.setNamePool(namePool);
                    in.setOutputProperties(props);
                    target=in;
                }
                if (characterMapExpander != null) {
                    characterMapExpander.setUnderlyingReceiver(target);
                    target = characterMapExpander;
                }
                String cdataElements = props.getProperty(OutputKeys.CDATA_SECTION_ELEMENTS);
                if (cdataElements!=null && cdataElements.length()>0) {
                    CDATAFilter filter = new CDATAFilter();
                    filter.setUnderlyingReceiver(target);
                    filter.setNamePool(namePool);
                    filter.setOutputProperties(props);
                    target = filter;
                }
            } else if (method.equals("text")) {
                emitter = new TEXTEmitter();
                emitter.setNamePool(namePool);
                target = emitter;
            } else {
                // TODO: externally supplied properties must be validated
                int brace = method.indexOf('}');
                String localName = method.substring(brace+1);
                int colon = localName.indexOf(':');
                localName = localName.substring(colon+1);
//
//                if (localName.equals("fop")) {
//                    // avoid an explicit external reference to avoid build problems
//                    // when FOP is not present on the class path
//                    target = Emitter.makeEmitter("net.sf.saxon.fop.FOPEmitter");
//                } else {
                    emitter = Emitter.makeEmitter(localName);
                    emitter.setNamePool(namePool);
                    target = emitter;
//                }
            }
            if (emitter != null) {
                emitter.setOutputProperties(props);
                StreamResult sr = (StreamResult)result;
                if (sr.getWriter() != null) {
                    emitter.setWriter(sr.getWriter());
                } else if (sr.getOutputStream() != null) {
                    emitter.setOutputStream(sr.getOutputStream());
                } else {
                    String systemId = sr.getSystemId();
                    Result resolvedResult = resolver.resolve(systemId, "");
                    if (resolvedResult instanceof StreamResult) {
                        StreamResult sr2 = (StreamResult)resolvedResult;
                        if (sr2.getWriter() != null) {
                            emitter.setWriter(sr2.getWriter());
                        } else if (sr2.getOutputStream() != null) {
                            emitter.setOutputStream(sr2.getOutputStream());
                        } else {
                            throw new TransformerException(
                                "The systemId supplied in a StreamResult must resolve to a Writer or OutputStream");
                        }
                    } else {
                            throw new TransformerException(
                                "The systemId supplied in a StreamResult must resolve to another StreamResult");
                    }
                }
            }
            return target;

        } else if (result instanceof DOMResult) {
            Node resultNode = ((DOMResult)result).getNode();
            if (resultNode!=null) {
                if (resultNode instanceof NodeInfo) {
                    // Writing to a SAXON tree is handled specially
                    if (resultNode instanceof DocumentInfo) {
                        DocumentInfo doc = (DocumentInfo)resultNode;
                        if (resultNode.getFirstChild() != null) {
                            throw new TransformerException("Target document must be empty");
                        } else {
                            Builder builder;
                            if (doc instanceof DocumentImpl) {
                                builder = new TreeBuilder();
                            } else {
                                builder = new TinyBuilder();
                            }
                            builder.setRootNode(doc);
                            builder.setSystemId(result.getSystemId());
                            builder.setNamePool(namePool);
                            return builder;
                        }
                    } else {
                        throw new TransformerException("Cannot add to an existing Saxon document");
                    }
                } else {
                    // Non-Saxon DOM
                    DOMEmitter emitter = new DOMEmitter();
                    emitter.setNode(resultNode);
                    return emitter;
                }
            } else {
                // no result node supplied; we must create our own
                TinyBuilder builder = new TinyBuilder();
                builder.setSystemId(result.getSystemId());
                builder.setNamePool(namePool);
                builder.createDocument();
                Document resultDoc = (Document)builder.getCurrentDocument();
                ((DOMResult)result).setNode(resultDoc);
                return builder;
            }
        } else {
            throw new IllegalArgumentException("Unknown type of result: " + result.getClass());
        }
    }
}