package net.sf.saxon.event;

import net.sf.saxon.Configuration;
import net.sf.saxon.StandardErrorHandler;
import net.sf.saxon.dom.DOMSender;
import net.sf.saxon.value.Type;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.NodeInfo;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;

import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;


/**
* Sender is a helper class that sends events to a Receiver from any kind of Source object
*/

public class Sender {

    Configuration config;
    public Sender (Configuration config) {
        this.config = config;
    }

    /**
    * Send the contents of a Source to a Receiver. Note that if the Source
     * identifies an element node rather than a document node, only the subtree
     * rooted at that element will be copied.
    * @param source the document or element to be copied
    * @param receiver the destination to which it is to be copied
    * @param namePool the namePool used for the destination
    * @param isStyleSheet true if the destination is a stylesheet document
    */

    public void send(Source source, Receiver receiver, NamePool namePool, boolean isStyleSheet)
    throws TransformerException {

        receiver.setNamePool(namePool);
        receiver.setSystemId(source.getSystemId());

        if (source instanceof NodeInfo) {
            NodeInfo ns = (NodeInfo)source;
            int kind = ns.getItemType();
            if (kind != Type.DOCUMENT && kind != Type.ELEMENT) {
                throw new IllegalArgumentException("DocumentSender can only handle document or element nodes");
            }
            sendDocumentInfo(ns, receiver, namePool);

        } else if (source instanceof SAXSource) {
            sendSAXSource((SAXSource)source, receiver, namePool, isStyleSheet);

        } else if (source instanceof DOMSource) {
            sendDOMSource((DOMSource)source, receiver, namePool);

        } else if (source instanceof StreamSource) {
            StreamSource ss = (StreamSource)source;
            String url = source.getSystemId();
            InputSource is = new InputSource(url);
            is.setCharacterStream(ss.getReader());
            is.setByteStream(ss.getInputStream());
            SAXSource sax = new SAXSource(
                             (isStyleSheet ? config.getStyleParser() : config.getSourceParser()),
                             is);
            sax.setSystemId(source.getSystemId());
            sendSAXSource(sax, receiver, namePool, isStyleSheet);
        } else {
            throw new IllegalArgumentException("Unknown type of source " + source.getClass());
        }
    }

    private void sendDocumentInfo(NodeInfo top, Receiver receiver, NamePool namePool)
    throws TransformerException {
        if (top.getNamePool() != namePool) {
            // This happens for example when turning an arbitrary DocumentInfo tree into a stylesheet
            NamePoolConverter converter = new NamePoolConverter(top.getNamePool(), namePool);
            converter.setUnderlyingReceiver(receiver);
            receiver = converter;
        }
        DocumentSender sender = new DocumentSender(top);
        sender.send(receiver);
    }

    private void sendDOMSource(DOMSource source, Receiver receiver, NamePool namePool)
    throws TransformerException {
        Node startNode = source.getNode();
        if (startNode instanceof DocumentInfo) {
            sendDocumentInfo((DocumentInfo)startNode, receiver, namePool);
        } else {
//            Document doc;
//            if (startNode instanceof Document) {
//                doc = (Document)startNode;
//            } else {
//                doc = startNode.getOwnerDocument();
//            }
            DOMSender driver = new DOMSender();
            driver.setStartNode(startNode);
            driver.setReceiver(receiver);
            driver.setNamePool(namePool);
            driver.setSystemId(source.getSystemId());
            driver.send();
        }
    }

    private void sendSAXSource(SAXSource source, Receiver receiver, NamePool namePool, boolean isStyleSheet)
    throws TransformerException {
        XMLReader parser = source.getXMLReader();
        if (parser==null) {
            SAXSource ss = new SAXSource();
            ss.setInputSource(source.getInputSource());
            ss.setSystemId(source.getSystemId());
            parser = (isStyleSheet ? config.getStyleParser() : config.getSourceParser());
            ss.setXMLReader(parser);
            source = ss;
        }

        if (parser.getErrorHandler()==null) {
            parser.setErrorHandler(new StandardErrorHandler(config.getErrorListener()));
        }

		try {
        	parser.setFeature("http://xml.org/sax/features/namespaces", true);
        	parser.setFeature("http://xml.org/sax/features/namespace-prefixes", false);
        } catch (SAXNotSupportedException err) {    // SAX2 parsers MUST support this feature!
            throw new TransformerException(
                "The SAX2 parser does not recognize a required namespace feature");
    	} catch (SAXNotRecognizedException err) {
            throw new TransformerException(
                "The SAX2 parser does not support a required namespace feature");
    	}

        ContentEmitter ce = new ContentEmitter();
        ce.setReceiver(receiver);
        ce.setNamePool(namePool);
        parser.setContentHandler(ce);
	    parser.setDTDHandler(ce);

        try {
    	    parser.setProperty("http://xml.org/sax/properties/lexical-handler", ce);
        } catch (SAXNotSupportedException err) {    // this just means we won't see the comments
	    } catch (SAXNotRecognizedException err) {
	    }

        try {
            parser.parse(source.getInputSource());
        } catch (Exception err) {
            throw new TransformerException(err);
        }
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
