package net.sf.saxon.event;

import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.UntypedAtomicValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.expr.StaticProperty;

import javax.xml.transform.TransformerException;

/**
 * This is a receiver that can be inserted into a pipeline to do type checking of the
 * nodes and/or atomic values produced by a content constructor, as they are written.
 * It is used, for example, when xsl:sequence has an "as" attribute to indicate the required
 * type, or to enforce the "as" attribute on xsl:template.
 *
 * <p>This filter also atomizes any nodes in the sequence if the required item type
 * is atomic.</p>
 */
public class SequenceChecker extends ProxyReceiver implements SequenceReceiver {

    private int itemType;
    private int contentType;
    private int cardinality;
    private int count = 0;
    private int level = 0;
    private boolean atomize = false;
    private StringBuffer buffer = null;

    public void setRequiredType(SequenceType type) {
        itemType = type.getPrimaryType();
        contentType = type.getContentType();
        cardinality = type.getCardinality();
        atomize = Type.isSubType(itemType, Type.ATOMIC);
    }

    public void startElement(int nameCode, int typeCode, int properties) throws TransformerException {
        if (level++ == 0) {
            if (atomize) {
                buffer = new StringBuffer();
                return;
            } else {
                if (count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.ELEMENT)) {
                    throw new TransformerException("The output sequence is not allowed to contain elements");
                }
                if (contentType!=Type.ITEM && !(Type.isSubType(typeCode, contentType))) {
                    throw new TransformerException(
                            "The required content type for items in the output sequence is " +
                            Type.getTypeName(contentType));
                }
            }
        }
        super.startElement(nameCode, typeCode, properties);
    }

    public void namespace(int namespaceCode, int properties) throws TransformerException {
        if (level==0) {
            if (atomize) {
                return;
                // discard the namespace node when atomizing
            } else {
                if ( count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.NAMESPACE)) {
                    throw new TransformerException("The output sequence is not allowed to contain namespace nodes");
                }
            }
        } else if (buffer!=null) {
            return;
            // discard the namespace node
        }
        super.namespace(namespaceCode, properties);
    }

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties)
            throws TransformerException {
        if (level==0) {
            if (atomize) {
                if (typeCode == 0) {
                    append(new UntypedAtomicValue(value));
                } else {
                    append(new StringValue(value).convert(typeCode));
                }
                return;
            } else {
                if (count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.ATTRIBUTE)) {
                    throw new TransformerException("The output sequence is not allowed to contain attribute nodes");
                }
                if (contentType!=Type.ITEM && !(Type.isSubType(typeCode, contentType))) {
                    throw new TransformerException(
                            "The required content type for items in the output sequence is " +
                            Type.getTypeName(contentType));
                }
            }
        } else if (buffer!=null) {
            // discard the attribute
            return;
        }
        super.attribute(nameCode, typeCode, value, properties);
    }

    public void endElement() throws TransformerException {
        level--;
        if (buffer != null) {
            if (level==0) {
                append(new UntypedAtomicValue(buffer));
                buffer = null;
                return;
            } else {
                return;
            }
        }
        super.endElement();
    }

    public void characters(CharSequence chars, int properties) throws TransformerException {
        if (level==0) {
            if (atomize) {
                append(new UntypedAtomicValue(chars));
                return;
            } else {
                if (count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.TEXT)) {
                    throw new TransformerException("The output sequence is not allowed to contain text nodes");
                }
            }
        } else if (buffer != null) {
            buffer.append(chars);
            return;
        }
        super.characters(chars, properties);
    }

    public void processingInstruction(String target, CharSequence data, int properties) throws TransformerException {
        if (level==0) {
            if (atomize) {
                append(new UntypedAtomicValue(data));
                return;
            } else {
                if (count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.PROCESSING_INSTRUCTION)) {
                    throw new TransformerException("The output sequence is not allowed to contain processing instruction nodes");
                }
            }
        } else if (buffer!=null) {
            // discard the PI
            return;
        }
        super.processingInstruction(target, data, properties);
    }

    public void comment(CharSequence chars, int properties) throws TransformerException {
        if (level == 0) {
            if (atomize) {
                append(new UntypedAtomicValue(chars));
                return;
            } else {
                if (count++==1 && !Cardinality.allowsMany(cardinality)) {
                    throw new TransformerException("The output sequence contains more than one item, only one is allowed");
                }
                if (!(itemType==Type.ITEM || itemType==Type.NODE || itemType==Type.COMMENT)) {
                    throw new TransformerException("The output sequence is not allowed to contain comment nodes");
                }
            }
        } else if (buffer!=null) {
            // discard the comment
            return;
        }
        super.comment(chars, properties);
    }

    public void append(Item item) throws TransformerException {
        if (level == 0) {
            if (atomize && item instanceof NodeInfo) {
                SequenceIterator iter = item.getTypedValue().iterate(null);
                while (iter.hasNext()) {
                    append(iter.next());
                }
                return;
            }
            if (count==1 && !Cardinality.allowsMany(cardinality)) {
                throw new TransformerException("The output sequence contains more than one item, only one is allowed");
            }
            int actualType = item.getItemType();
            if (actualType == Type.UNTYPED_ATOMIC && itemType != Type.UNTYPED_ATOMIC) {
                append(((AtomicValue)item).convert(itemType));
                return;
            } else if (!(Type.isSubType(actualType, itemType))) {
                throw new TransformerException("The output sequence is not allowed to contain items of type " +
                        Type.getTypeName(item.getItemType()));
            }
            // TODO: check the content type of the items
        }
        ((SequenceReceiver)baseReceiver).append(item);
        count++;
    }

    public void finalCheck() throws TransformerException {
        if (count==0 && (cardinality & StaticProperty.CARDINALITY_ALLOWS_ZERO)==0) {
            throw new TransformerException("The output sequence is not allowed to be empty");
        }
        super.endDocument();
    }
}
