package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.om.Item;
import net.sf.saxon.value.*;

/**
* Arithmetic Expression: an expression using one of the operators
* plus, minus, multiply, div, idiv, mod.
*/

class ArithmeticExpression extends BinaryExpression {

    private final static int NUMERIC_ARITHMETIC = 0;
    private final static int DATE_AND_DURATION = 1;
    private final static int DATE_DIFFERENCE = 2;
    private final static int DURATION_ADDITION = 3;
    private final static int DURATION_MULTIPLICATION = 4;

    private final static int UNKNOWN = -1;

    static private int[] plusTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,

        Type.DATE, Type.DURATION, DATE_AND_DURATION, Type.DATE,
        Type.DURATION, Type.DATE, DATE_AND_DURATION, Type.DATE,

        Type.TIME, Type.DURATION, DATE_AND_DURATION, Type.TIME,
        Type.DURATION, Type.TIME, DATE_AND_DURATION, Type.TIME,

        Type.DATE_TIME, Type.DURATION, DATE_AND_DURATION, Type.DATE_TIME,
        Type.DURATION, Type.DATE_TIME, DATE_AND_DURATION, Type.DATE_TIME,

        Type.YEAR_MONTH_DURATION, Type.YEAR_MONTH_DURATION, DURATION_ADDITION, Type.YEAR_MONTH_DURATION,
        Type.DAY_TIME_DURATION, Type.DAY_TIME_DURATION, DURATION_ADDITION, Type.DAY_TIME_DURATION

    };

    static private int[] minusTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,

        Type.DATE, Type.DATE, DATE_DIFFERENCE, Type.DAY_TIME_DURATION,
        Type.DATE, Type.DURATION, DATE_AND_DURATION, Type.DATE,

        Type.TIME, Type.TIME, DATE_DIFFERENCE, Type.DAY_TIME_DURATION,
        Type.TIME, Type.DURATION, DATE_AND_DURATION, Type.TIME,

        Type.DATE_TIME, Type.DATE_TIME, DATE_DIFFERENCE, Type.DAY_TIME_DURATION,
        Type.DATE_TIME, Type.DURATION, DATE_AND_DURATION, Type.DATE_TIME,

        Type.YEAR_MONTH_DURATION, Type.YEAR_MONTH_DURATION, DURATION_ADDITION, Type.YEAR_MONTH_DURATION,
        Type.DAY_TIME_DURATION, Type.DAY_TIME_DURATION, DURATION_ADDITION, Type.DAY_TIME_DURATION

    };

    static private int[] multiplyTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,

        Type.NUMBER, Type.DURATION, DURATION_MULTIPLICATION, Type.DURATION,
        Type.DURATION, Type.NUMBER, DURATION_MULTIPLICATION, Type.DURATION
    };

    static private int[] divideTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,

        Type.DURATION, Type.NUMBER, DURATION_MULTIPLICATION, Type.DURATION
    };

    static private int[] idivTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.INTEGER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.INTEGER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.INTEGER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.INTEGER,
    };

    static private int[] modTable = {
        Type.NUMBER, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.NUMBER, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.NUMBER, NUMERIC_ARITHMETIC, Type.NUMBER,
        Type.UNTYPED_ATOMIC, Type.UNTYPED_ATOMIC, NUMERIC_ARITHMETIC, Type.NUMBER
    };

    private boolean backwardsCompatible = false;

    public ArithmeticExpression(){};

    public ArithmeticExpression(Expression p1, int operator, Expression p2) {
        super(p1, operator, p2);
    }

    /**
    * Type-check the expression statically. We try to work out which particular
    * arithmetic function to use if the types of operands are known an compile time.
    */

    public Expression analyze(StaticContext env) throws XPathException {

        backwardsCompatible = env.isInBackwardsCompatibleMode();

        SequenceType atomicType = new SequenceType(Type.ATOMIC, Type.ITEM, StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE);
        // TODO: this is using the function call rules. Arithetic expressions have slightly different rules.

        RoleLocator role0 = new RoleLocator(RoleLocator.BINARY_EXPR, Tokenizer.tokens[operator], 0);
        operands[0] = TypeChecker.staticTypeCheck(operands[0], atomicType, backwardsCompatible, role0);
        // System.err.println("First operand"); operands[0].display(10);

        RoleLocator role1 = new RoleLocator(RoleLocator.BINARY_EXPR, Tokenizer.tokens[operator], 1);
        operands[1] = TypeChecker.staticTypeCheck(operands[1], atomicType, backwardsCompatible, role1);
        // System.err.println("Second operand"); operands[1].display(10);

        Expression e = super.analyze(env);
        if (e instanceof ArithmeticExpression) {


            final int type0 = operands[0].getItemType();
            final int type1 = operands[1].getItemType();
            final int action = getAction(type0, operator, type1);
            switch(action) {
                case NUMERIC_ARITHMETIC:
                    e = new NumericArithmetic(operands[0], operator, operands[1]);
                    break;
                case DURATION_ADDITION:
                    e = new DurationAddition(operands[0], operator, operands[1]);
                    break;
                case DURATION_MULTIPLICATION:
                    e = new DurationMultiplication(operands[0], operator, operands[1]);
                    break;
                case DATE_AND_DURATION:
                    e = new DateAndDuration(operands[0], operator, operands[1]);
                    break;
                case DATE_DIFFERENCE:
                    e = new DateDifference(operands[0], operator, operands[1]);
                    break;
                default:
                    // either the types are not known yet, or they are wrong
                    if (!backwardsCompatible &&
                            Type.isSubType(type0, Type.ATOMIC) &&
                                type0 != Type.UNTYPED_ATOMIC &&
                                type0 != Type.ATOMIC &&
                            Type.isSubType(type1, Type.ATOMIC) &&
                                type1 != Type.UNTYPED_ATOMIC &&
                                type1 != Type.ATOMIC) {
                        throw new XPathException.Type("Unsuitable operands for arithmetic operation (" +
                                Type.getTypeName(type0) + ", " +
                                Type.getTypeName(type1) + ")");
                    }
                    return e;
            }
            try {
                if (operands[0] instanceof Value && operands[1] instanceof Value) {
                    return ExpressionTool.eagerEvaluate(this, null);
                }
            } catch (XPathException.Dynamic err) {
                // Defer any error reporting until run-time
            }
        }
        return e;
    }

    /**
    * Decide what action is required based on types of operands
    */

    private int getAction(int type1, int operator, int type2) {
        int[] table = getOperatorTable(operator);
        int entry = getEntry(table, type1, type2);
        if (entry<0) return UNKNOWN;
        return table[entry+2];
    }

    private int[] getOperatorTable(int operator) {
        switch (operator) {
            case Tokenizer.PLUS:
                return plusTable;
            case Tokenizer.MINUS:
            case Tokenizer.NEGATE:
                return  minusTable;
            case Tokenizer.MULT:
                return  multiplyTable;
            case Tokenizer.DIV:
                return  divideTable;
            case Tokenizer.IDIV:
                return  idivTable;
            case Tokenizer.MOD:
                return  modTable;
            default:
                throw new IllegalArgumentException("Unknown arithmetic operator");
        }
    }

    private int getEntry(int[] table, int type1, int type2) {
        for (int i=0; i<table.length; i+=4) {
            if (Type.isSubType(type1, table[i]) && Type.isSubType(type2, table[i+1])) {
                return i;
            }
        }
        return -1;
    }

    /**
    * Determine the data type of the expression, if this is known statically
    */

    public int getItemType() {
        int t1 = operands[0].getItemType();
        int t2 = operands[1].getItemType();
        int[] table = getOperatorTable(operator);
        int entry = getEntry(table, t1, t2);

        if (entry<0) return Type.ATOMIC;  // type is not known statically

        int resultType = table[entry+3];
        if (resultType == Type.NUMBER) {
            resultType = NumericValue.promote(t1, t2);
            // exception: integer div integer => decimal
            if (operator == Tokenizer.DIV && resultType == Type.INTEGER) {
                resultType = Type.DECIMAL;
            }
        }
        return resultType;
    }

    /**
    * Evaluate the expression. We only take this path if the type could not be determined
    * statically.
    */

    public Item evaluateItem(XPathContext context) throws XPathException {

        AtomicValue v1 = (AtomicValue)operands[0].evaluateItem(context);
        if (v1==null) return null;

        AtomicValue v2 = (AtomicValue)operands[1].evaluateItem(context);
        if (v2==null) return null;
                    // System.err.println(Type.getTypeName(v1.getItemType())+ " op " + Type.getTypeName(v2.getItemType()));
        int action = getAction(v1.getItemType(), operator, v2.getItemType());

        switch(action) {
            case NUMERIC_ARITHMETIC:
                return new NumericArithmetic(v1, operator, v2).evaluateItem(context);
            case DURATION_ADDITION:
                return new DurationAddition(v1, operator, v2).evaluateItem(context);
            case DURATION_MULTIPLICATION:
                return new DurationMultiplication(v1, operator, v2).evaluateItem(context);
            case DATE_AND_DURATION:
                return new DateAndDuration(v1, operator, v2).evaluateItem(context);
            case DATE_DIFFERENCE:
                return new DateDifference(v1, operator, v2).evaluateItem(context);
            default:
                // types are not known yet. Force to numeric if in 1.0 mode
                if (backwardsCompatible) {
                    NumericValue nv1;
                    NumericValue nv2;
                    try {
                        nv1 = (NumericValue)v1.convert(Type.NUMBER);
                        nv2 = (NumericValue)v2.convert(Type.NUMBER);
                    } catch (XPathException err) {
                        typeError("Unsuitable operands for arithmetic operation (" +
                            Type.getTypeName(v1.getItemType()) + ", " +
                            Type.getTypeName(v2.getItemType()) + ")");
                        return null;
                    }
                    return new NumericArithmetic(nv1, operator, nv2).evaluateItem(context);
                } else {
                    typeError("Unsuitable operands for arithmetic operation (" +
                            Type.getTypeName(v1.getItemType()) + ", " +
                            Type.getTypeName(v2.getItemType()) + ")");
                    return null;
                }
        }

    }

    /**
    * Perform a partial evaluation of the expression, by eliminating specified dependencies
    * on the context.
    * @param dependencies The dependencies to be removed
    * @param context The context to be used for the partial evaluation
    * @return a new expression that does not have any of the specified
    * dependencies
    */

    /**
    * Inner class to handle numeric arithmetic expressions
    */

    public static class NumericArithmetic extends ArithmeticExpression {

        public NumericArithmetic(Expression p1, int operator, Expression p2) {
            super(p1, operator, p2);
        }

        /**
        * Evaluate the expression.
        */

        public Item evaluateItem(XPathContext context) throws XPathException {
            // TODO: try to do more of the work at compile time
            AtomicValue v1 = (AtomicValue)operands[0].evaluateItem(context);
            int reqType = (operator == Tokenizer.IDIV ? Type.INTEGER : Type.DOUBLE);
            if (v1==null) return null;
            if (v1 instanceof UntypedAtomicValue) {
                v1 = v1.convert(reqType);
            }
            AtomicValue v2 = (AtomicValue)operands[1].evaluateItem(context);
            if (v2==null) return null;
            if (v2 instanceof UntypedAtomicValue) {
                v2 = v2.convert(reqType);
            }

            if (operator==Tokenizer.NEGATE) {
                return ((NumericValue)v2).negate();
            } else {
                // TODO: the conversion rules here aren't quite right yet
                if (operator == Tokenizer.IDIV) {
                    if (!(v1 instanceof IntegerValue)) {
                        typeError("First argument to idiv must be an integer");
                        return null;
                    }
                    if (!(v2 instanceof IntegerValue)) {
                        typeError("Second argument to idiv must be an integer");
                        return null;
                    }
                }
                return ((NumericValue)v1).arithmetic(operator, (NumericValue)v2);
            }
        }
    }

    /**
    * Inner class to handle addition and subtraction of two durations
    */

    public static class DurationAddition extends ArithmeticExpression {

        public DurationAddition(Expression p1, int operator, Expression p2) {
            super(p1, operator, p2);
        }

        /**
        * Evaluate the expression.
        */

        public Item evaluateItem(XPathContext context) throws XPathException {
            DurationValue v1 = (DurationValue)operands[0].evaluateItem(context);
            if (v1==null) return null;
            DurationValue v2 = (DurationValue)operands[1].evaluateItem(context);
            if (v2==null) return null;

            // The two durations must be of the same type
            //if (v1.getItemType() != v2.getItemType()) {
            //    throw new XPathException("Durations of different types cannot be combined");
            //}

            if (operator==Tokenizer.PLUS) {
                return v1.add(v2);
            } else if (operator==Tokenizer.MINUS) {
                return v1.subtract(v2);
            } else {
                throw new AssertionError("Unknown operation on durations");
            }

        }
    }

    /**
    * Inner class to handle multiplication (or division) of a duration by a number
    */

    public static class DurationMultiplication extends ArithmeticExpression {

//        public DurationMultiplication() {}

        public DurationMultiplication(Expression p1, int operator, Expression p2) {

            // by the time we get here, we know that one of the operands is a duration,
            // but it might be either one. We make it the first.

            super(p1, operator, p2);
            if (Type.isSubType(p2.getItemType(), Type.DURATION)) {
                operands[0] = p2;
                operands[1] = p1;
            }
        }

        /**
        * Evaluate the expression.
        */

        public Item evaluateItem(XPathContext context) throws XPathException {
            DurationValue v1 = (DurationValue)operands[0].evaluateItem(context);
            if (v1==null) return null;
            NumericValue v2 = (NumericValue)operands[1].evaluateItem(context);

            double d = v2.asDouble();

            if (operator == Tokenizer.DIV) {
                d = 1.0 / d;
            }

            return v1.multiply(d);

        }
    }

    /**
    * Inner class to handle addition or subtraction of a Date (or Time, or DateTime) and a Duration
    */

    public static class DateAndDuration extends ArithmeticExpression {

//        public DateAndDuration() {}

        public DateAndDuration(Expression p1, int operator, Expression p2) {

            // by the time we get here, we know that one of the operands is a duration,
            // but it might be either one. We make it the second.

            super(p1, operator, p2);
            if (Type.isSubType(p1.getItemType(), Type.DURATION)) {
                operands[0] = p2;
                operands[1] = p1;
            }

        }

        /**
        * Evaluate the expression.
        */

        public Item evaluateItem(XPathContext context) throws XPathException {
            CalendarValue v1 = (CalendarValue)operands[0].evaluateItem(context);
            if (v1==null) return null;
            DurationValue v2 = (DurationValue)operands[1].evaluateItem(context);
            if (v2==null) return null;

            if (operator == Tokenizer.MINUS) {
                v2 = v2.multiply(-1.0);
            }

            return v1.add(v2);

        }
    }

    /**
    * Inner class to handle subtraction of a Date (or Time, or DateTime) from another, to return a Duration
    */

    public static class DateDifference extends ArithmeticExpression {

        public DateDifference(Expression p1, int operator, Expression p2) {
            super(p1, operator, p2);
        }
        /**
        * Evaluate the expression.
        */

        public Item evaluateItem(XPathContext context) throws XPathException {
            CalendarValue v1 = (CalendarValue)operands[0].evaluateItem(context);
            if (v1==null) return null;
            CalendarValue v2 = (CalendarValue)operands[1].evaluateItem(context);
            if (v2==null) return null;

            return v1.subtract(v2);

        }
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
