package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;

/**
* Cast Expression: implements "cast as data-type ( expression )". It also allows an internal
* cast, which has the same semantics as a user-requested cast, but maps an empty sequence to
* an empty sequence.
*/

public final class CastExpression extends ComputedExpression  {

    private Expression source;
    private int targetType;
    private boolean allowEmpty = false;

    public CastExpression(Expression source, SequenceType target) throws XPathException {
        this.source = source;
        this.targetType = target.getPrimaryType();
        if (Cardinality.allowsMany(target.getCardinality())) {
            throw new XPathException.Static("Target of cast must not be a sequence type");
        }
        if (Type.isSubType(target.getPrimaryType(), Type.NODE)) {
            throw new XPathException.Static("Target of cast must not be a node type");
        }
    }

    public CastExpression(Expression source, int targetType, boolean allowEmpty) throws XPathException {
        this.source = source;
        this.targetType = targetType;
        this.allowEmpty = allowEmpty;

        if (Type.isSubType(targetType, Type.NODE)) {
            throw new XPathException.Static("Target of cast must not be a node type");
        }
    }

    /**
    * Simplify the expression
    * @return the simplified expression
    */

    public Expression simplify() throws XPathException {
        source = source.simplify();
        if (source instanceof AtomicValue) {
            return (((AtomicValue)source).convert(targetType));
        }
        return this;
    }

    /**
    * Type-check the expression
    */

    public Expression analyze(StaticContext env) throws XPathException {
        source = source.analyze(env);
        SequenceType atomicType =
                new SequenceType(Type.ATOMIC, Type.ITEM, getCardinality());

        RoleLocator role = new RoleLocator(RoleLocator.TYPE_OP, "cast as", 0);
        source = TypeChecker.staticTypeCheck(source, atomicType, false, role);

        if (Type.isSubType(source.getItemType(), targetType)) {
            return source;
        }
        if (Type.isSubType(targetType,Type.QNAME)) {
            return new CastAsQName(source).analyze(env);
        }
        if (source instanceof AtomicValue) {
            return (AtomicValue)evaluateItem(null);
        }
        return this;
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        source = source.promote(offer);
        return this;
    }

    /**
    * Get the immediate subexpressions of this expression
    */

    public Expression[] getSubExpressions() {
        Expression[] exp = new Expression[1];
        exp[0] = source;
        return exp;
    }


    /**
    * Get the static cardinality of the expression
    */

    public int computeCardinality() {
        return (allowEmpty ? StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE : StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    /**
    * Get the static type of the expression
    */

    public int getItemType() {
        return targetType;
    }


    /**
    * Evaluate the expression
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        AtomicValue value = (AtomicValue)source.evaluateItem(context);
        if (value==null) return null;
        try {
            return value.convert(targetType);
        } catch (XPathException err) {
            // add location information
            dynamicError(err.getMessage());
            return null;
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "cast as " +
                Type.getTypeName(targetType));
        source.display(level+1, pool);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
