package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;

/**
* A ConvertToString expression performs the string conversion required for string arguments
* in backwards compatibility mode. If given an empty sequence, it returns an empty string; otherwise,
* it returns the string value of the first item in the sequence.
*/

public final class ConvertToString extends ComputedExpression {

    private Expression base;

    /**
    * Constructor
    * @param base A sequence expression denoting sequence whose first item is to be returned
    */

    public ConvertToString(Expression base) {
        this.base = base;
    }

    /**
    * Get the data type of the items returned
    */

    public int getItemType() {
        return Type.STRING;
    }

    /**
    * Simplify the expression
    */

    public Expression simplify() throws XPathException {
        base = base.simplify();
        return this;
    }

    /**
    * Type-check the expression
    */

    public Expression analyze(StaticContext env) throws XPathException {
        base = base.analyze(env);
        if (base.getCardinality() == StaticProperty.CARDINALITY_EXACTLY_ONE &&
                Type.isSubType(base.getItemType(), Type.STRING)) {
            return base;
        }
        return this;
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        base = base.promote(offer);
        return this;
    }

    /**
    * Get the immediate subexpressions of this expression
    */

    public Expression[] getSubExpressions() {
        Expression[] exp = new Expression[1];
        exp[0] = base;
        return exp;
    }

    /**
    * Get the static cardinality
    */

    public int computeCardinality() {
        return StaticProperty.CARDINALITY_EXACTLY_ONE;
    }

    /**
    * Is this expression the same as another expression?
    */

    public boolean equals(Object other) {
        if (other instanceof ConvertToString) {
            ConvertToString f = (ConvertToString)other;
            return base.equals(f.base);
        }
        return false;
    }

    /**
    * get HashCode for comparing two expressions
    */

    public int hashCode() {
        return "ConvertToString".hashCode() + base.hashCode();
    }

    /**
    * Evaluate the expression
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        SequenceIterator iter = base.iterate(context);
        if (iter.hasNext()) {
            return new StringValue(iter.next().getStringValue());
        } else {
            return StringValue.EMPTY_STRING;
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "convert to string: ");
        base.display(level+1, pool);
    }

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
