package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.sort.NodeOrderComparer;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;


/**
* An enumeration representing a nodeset that is teh difference of two other NodeSets.
* There is an "except" operator in XPath 2.0 to create such an expression.
*/


public class DifferenceEnumeration implements SequenceIterator {


    private SequenceIterator p1;
    private SequenceIterator p2;

    private NodeInfo nextNode1 = null;
    private NodeInfo nextNode2 = null;
    private NodeOrderComparer comparer;

    private NodeInfo nextNode = null;
    private NodeInfo current = null;
    private int position = 0;

    /**
    * Form an enumeration of the difference of two nodesets, that is, the nodes
    * that are in p1 and that are not in p2.
    * @param p1 the first operand
    * @param p2 the second operand
    * @param comparer the comparer
    */

    public DifferenceEnumeration(SequenceIterator p1, SequenceIterator p2,
                                 NodeOrderComparer comparer) throws XPathException {
        this.p1 = p1;
        this.p2 = p2;
        this.comparer = comparer;

        // move to the first node in each input nodeset

        if (p1.hasNext()) {
            nextNode1 = next(p1);
        }
        if (p2.hasNext()) {
            nextNode2 = next(p2);
        }

        // move to the first node in p1 that isn't in p2

        advance();

    }

    /**
    * Get the next item from one of the input sequences,
    * checking that it is a node.
    */

    private NodeInfo next(SequenceIterator iter) throws XPathException {
        return (NodeInfo)iter.next();
        // rely on type-checking to prevent a ClassCastException
    }

    public boolean hasNext() {
        return nextNode!=null;
    }

    public Item next() throws XPathException {
        current = nextNode;
        position++;
        advance();
        return current;
    }

    public Item current() {
        return current;
    }

    public int position() {
        return position;
    }

    private void advance() throws XPathException {

        // main merge loop: if the node in p1 has a lower key value that that in p2, return it;
        // if they are equal, advance both nodesets; if p1 is higher, advance p2.

        while (nextNode1 != null && nextNode2 != null) {
            int c = comparer.compare(nextNode1, nextNode2);
            if (c<0) {                              // p1 is lower
                NodeInfo next = nextNode1;
                if (p1.hasNext()) {
                    nextNode1 = next(p1);
                } else {
                    nextNode1 = null;
                    nextNode = null;
                }
                nextNode = next;
                return;

            } else if (c>0) {                       // p1 is higher
                if (p2.hasNext()) {
                    nextNode2 = next(p2);
                } else {
                    nextNode2 = null;
                    nextNode = null;
                }

            } else {                                // keys are equal

                if (p2.hasNext()) {
                    nextNode2 = next(p2);
                } else {
                    nextNode2 = null;
                }
                if (p1.hasNext()) {
                    nextNode1 = next(p1);
                } else {
                    nextNode1 = null;
                }
            }
        }

        // collect the remaining nodes from the residue of p1

        if (nextNode1!=null) {
            nextNode = nextNode1;
            if (p1.hasNext()) {
                nextNode1 = next(p1);
            } else {
                nextNode1 = null;
            }
            return;
        }

        nextNode = null;

    }

    public SequenceIterator getAnother() throws XPathException {
        return new DifferenceEnumeration(p1.getAnother(), p2.getAnother(), comparer);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
