package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.Value;
import net.sf.saxon.om.NamePool;

/**
* Abstract superclass for calls to system-defined and user-defined functions
*/

public abstract class FunctionCall extends ComputedExpression {

    /**
    * The array of expressions representing the actual parameters
    * to the function call
    */

    protected Expression[] argument;

    /**
    * Determine the number of actual arguments supplied in the function call
    */

    public final int getNumberOfArguments() {
        return argument.length;
    }

    /**
    * Method called by the expression parser when all arguments have been supplied
    */

    public void setArguments(Expression[] args) throws XPathException {
        argument = args;
    }

    /**
    * Simplify the function call. Default method is to simplify each of the supplied arguments and
    * evaluate the function if all are now known.
    */

    public Expression simplify() throws XPathException {
        return simplifyArguments(true);
    }

    /**
    * Simplify the arguments of the function.
    * Called from the simplify() method of each function.
    * @param eval this argument is no longer used
    * @return the result of simplifying the arguments of the expression
    */

    protected final Expression simplifyArguments(boolean eval) throws XPathException {
        for (int i=0; i<getNumberOfArguments(); i++) {
            argument[i] = argument[i].simplify();
        }
        return this;
    }

    /**
    * Type-check the expression. This also calls preEvaluate() to evaluate the function
    * if all the arguments are constant; functions that do not require this behavior
    * can override the preEvaluate method.
    */

    public Expression analyze(StaticContext env) throws XPathException {
        boolean fixed = true;
        for (int i=0; i<getNumberOfArguments(); i++) {
            argument[i] = argument[i].analyze(env);
            if (!(argument[i] instanceof Value)) {
                fixed = false;
            }
        }
        checkArguments(env);
        if (fixed) {
            return preEvaluate(env);
        } else {
            return this;
        }
    }

    /**
    * Pre-evaluate a function at compile time. Functions that do not allow
    * pre-evaluation, or that need access to context information, can override this method.
    */

    public Expression preEvaluate(StaticContext env) throws XPathException {
        return ExpressionTool.eagerEvaluate(this, null);
    }

    /**
    * Promote this expression if possible
    */

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp != null) {
            return exp;
        } else {
            if (offer.action != PromotionOffer.UNORDERED) {
                for (int i=0; i<getNumberOfArguments(); i++) {
                    argument[i] = argument[i].promote(offer);
                }
            }
            return this;
        }
    }

    /**
    * Method supplied by each class of function to check arguments during parsing, when all
    * the argument expressions have been read
    */

    protected abstract void checkArguments(StaticContext env) throws XPathException;

    /**
    * Check number of arguments. <BR>
    * A convenience routine for use in subclasses.
    * @param min the minimum number of arguments allowed
    * @param max the maximum number of arguments allowed
    * @return the actual number of arguments
    * @throws XPathException if the number of arguments is out of range
    */

    protected int checkArgumentCount(int min, int max) throws XPathException {
        int numArgs = getNumberOfArguments();
        if (min==max && numArgs != min) {
            throw new XPathException.Static("Function " + getName() + " must have "
                    + min + pluralArguments(min),
                    ExpressionTool.getLocator(this));
        }
        if (numArgs < min) {
            throw new XPathException.Static("Function " + getName() + " must have at least "
                    + min + pluralArguments(min),
                    ExpressionTool.getLocator(this));
        }
        if (numArgs > max) {
            throw new XPathException.Static("Function " + getName() + " must have no more than "
                    + max + pluralArguments(max),
                    ExpressionTool.getLocator(this));
        }
        return numArgs;
    }

    /**
    * Utility routine used in constructing error messages
    */

    private static String pluralArguments(int num) {
        if (num==1) return " argument";
        return " arguments";
    }

    /**
    * Get the immediate subexpressions of this expression
    */

    public Expression[] getSubExpressions() {
        return argument;
    }


    /**
    * Diagnostic print of expression structure
    */

    public abstract String getName();

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "function " + getName());
        for (int a=0; a<getNumberOfArguments(); a++) {
            argument[a].display(level+1, pool);
        }
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
