package net.sf.saxon.expr;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.sort.NodeOrderComparer;
import net.sf.saxon.xpath.XPathException;

/**
* An enumeration representing a nodeset that is an intersection of two other NodeSets.
* There is currently no operator in XPath to create such an expression, but it is used
* by the extension function intersection(). The code is derived from the analagous UnionEnumeration,
* an inner class of UnionExpression.
*/


public class IntersectionEnumeration implements SequenceIterator {

    private SequenceIterator e1;
    private SequenceIterator e2;
    private NodeInfo nextNode1 = null;
    private NodeInfo nextNode2 = null;
    private NodeOrderComparer comparer;

    private NodeInfo nextNode = null;
    private NodeInfo current = null;
    private int position = 0;

    /**
    * Form an enumeration of the intersection of the nodes in two nodesets
    * @param p1 the first operand: must be in document order
    * @param p2 the second operand: must be in document order
    * @param comparer Comparer to be used for putting nodes in document order
    */

    public IntersectionEnumeration(SequenceIterator p1, SequenceIterator p2,
                                    NodeOrderComparer comparer ) throws XPathException {
        this.e1 = p1;
        this.e2 = p2;
        this.comparer = comparer;

        // move to the first node in each input nodeset

        if (e1.hasNext()) {
            nextNode1 = next(e1);
        }
        if (e2.hasNext()) {
            nextNode2 = next(e2);
        }

        // move to the first node that matches in  both

        advance();

    }

    /**
    * Get the next item from one of the input sequences,
    * checking that it is a node.
    */

    private NodeInfo next(SequenceIterator iter) throws XPathException {
        return (NodeInfo)iter.next();
        // rely on type-checking to prevent a ClassCastException
    }

    public boolean hasNext() {
        return nextNode!=null;
    }

    public Item next() throws XPathException {
        current = nextNode;
        position++;
        advance();
        return current;
    }

    public Item current() {
        return current;
    }

    public int position() {
        return position;
    }


    private void advance() throws XPathException {

        // main merge loop: iterate whichever set has the lower value, returning when a pair
        // is found that match.

        while (nextNode1 != null && nextNode2 !=null) {
            int c = comparer.compare(nextNode1, nextNode2);
            if (c<0) {
                NodeInfo next = nextNode1;
                if (e1.hasNext()) {
                    nextNode1 = next(e1);
                } else {
                    nextNode1 = null;
                    nextNode = null;
                }

            } else if (c>0) {
                NodeInfo next = nextNode2;
                if (e2.hasNext()) {
                    nextNode2 = next(e2);
                } else {
                    nextNode2 = null;
                    nextNode = null;
                }

            } else {            // keys are equal

                nextNode = nextNode2;           // which is the same as nextNode1
                if (e2.hasNext()) {
                    nextNode2 = next(e2);
                } else {
                    nextNode2 = null;
                }
                if (e1.hasNext()) {
                    nextNode1 = next(e1);
                } else {
                    nextNode1 = null;
                }

                return;
            }
        }
        nextNode = null;
    }

    public SequenceIterator getAnother() throws XPathException {
        return new IntersectionEnumeration(e1.getAnother(), e2.getAnother(), comparer);
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
