package net.sf.saxon.expr;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.BooleanValue;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.Value;
import net.sf.saxon.xpath.XPathException;

/**
* A QuantifiedExpression tests whether some/all items in a sequence satisfy
* some condition.
*/

class QuantifiedExpression extends Assignation {

    private int operator;       // Tokenizer.SOME or Tokenizer.EVERY

    public void setOperator(int operator) {
        this.operator = operator;
    }

    /**
    * Determine the static cardinality
    */

    public int computeCardinality() {
        return StaticProperty.CARDINALITY_EXACTLY_ONE;
    }

    /**
    * Type-check the expression
    */

    public Expression analyze(StaticContext env) throws XPathException {

        if (declaration==null) {
            // we've already done the type checking, no need to do it again
            return this;
        }

        // The order of events is critical here. First we ensure that the type of the
        // sequence expression is established. This is used to establish the type of the variable,
        // which in turn is required when type-checking the action part.

        sequence = sequence.analyze(env);

        // "some" and "every" have no ordering constraints

        sequence = ExpressionTool.unsorted(sequence, false);

        SequenceType decl = declaration.getRequiredType();
        SequenceType sequenceType =
                            new SequenceType(decl.getPrimaryType(),
                                                decl.getContentType(),
                                                StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE);
        RoleLocator role = new RoleLocator(RoleLocator.VARIABLE, getVariableName(), 0);
        sequence = TypeChecker.staticTypeCheck(
                                sequence, sequenceType, false, role);
        int actualItemType = sequence.getItemType();
        declaration.refineTypeInformation(actualItemType,
                StaticProperty.CARDINALITY_EXACTLY_ONE,
                null,
                sequence.getSpecialProperties());

        declaration = null;     // let the garbage collector take it

        action = action.analyze(env);

        PromotionOffer offer = new PromotionOffer();
        offer.containingExpression = this;
        offer.action = PromotionOffer.RANGE_INDEPENDENT;
        offer.binding = this;
        action = action.promote(offer);
        if (offer.containingExpression instanceof LetExpression) {
            offer.containingExpression = offer.containingExpression.analyze(env);
        }
        return offer.containingExpression;
    }


    /**
    * Evaluate the expression to return a singleton value
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return BooleanValue.get(effectiveBooleanValue(context));
    }

    /**
    * Get the result as a boolean
    */

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {

        // First create an iteration of the base sequence.

        SequenceIterator base = sequence.iterate(context);

        // Now test to see if some or all of the tests are true. The same
        // logic is used for the SOME and EVERY operators

        boolean some = (operator==Tokenizer.SOME);
        while (base.hasNext()) {
            context.setLocalVariable(slotNumber, Value.asValue(base.next()));
            if (some == action.effectiveBooleanValue(context)) {
                return some;
            }
        }
        return !some;
    }


    /**
    * Determine the data type of the items returned by the expression
    * @return Type.BOOLEAN
    */

	public int getItemType() {
	    return Type.BOOLEAN;
	}

    /**
    * Get data type of the range variable (part of the Binding interface)
    */

    public SequenceType getRequiredType() {
        return new SequenceType(sequence.getItemType(), Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + Tokenizer.tokens[operator] + " $" + getVariableName() + " in");
        sequence.display(level+1, pool);
        System.err.println(ExpressionTool.indent(level) + "satisfies");
        action.display(level+1, pool);
    }

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
