package net.sf.saxon.expr;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.value.IntegerValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.Item;

/**
* A RangeExpression is an expression that represents an integer sequence as
* a pair of end-points (for example "x to y").
* If the end-points are equal, the sequence is of length one.
 * From Saxon 7.8, the sequence must be ascending; if the end-point is less
 * than the start-point, an empty sequence is returned. This is to allow
 * expressions of the form "for $i in 1 to count($seq) return ...."
*/

public class RangeExpression extends BinaryExpression {

    /**
    * Construct a RangeExpression
    */

    public RangeExpression(Expression start, int op, Expression end) {
        super(start, op, end);
    }

    /**
    * Type-check the expression
    */

    public Expression analyze(StaticContext env) throws XPathException {
        operands[0] = operands[0].analyze(env);
        operands[1] = operands[1].analyze(env);

        RoleLocator role0 = new RoleLocator(RoleLocator.BINARY_EXPR, "to", 0);
        operands[0] = TypeChecker.staticTypeCheck(operands[0], INT_TYPE, false, role0);

        RoleLocator role1 = new RoleLocator(RoleLocator.BINARY_EXPR, "to", 1);
        operands[1] = TypeChecker.staticTypeCheck(operands[1], INT_TYPE, false, role1);
        return super.simplify();
    }

    private static SequenceType INT_TYPE =
            new SequenceType(Type.INTEGER, Type.ITEM, StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE);

    /**
    * Get the data type of the items returned
    */

    public int getItemType() {
        return Type.INTEGER;
    }

    /**
    * Determine the static cardinality
    */

    public int computeCardinality() {
        return StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE;
    }

    /**
    * Return an iteration over the sequence
    */

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        IntegerValue v1 = (IntegerValue)operands[0].evaluateItem(context);
        if (v1 == null) {
            return new EmptyIterator();
        }
        IntegerValue v2 = (IntegerValue)operands[1].evaluateItem(context);
        if (v2 == null) {
            return new EmptyIterator();
        }
        long start = v1.getValue();
        long end = v2.getValue();
        if (start > end) {
            return new EmptyIterator();
        }
        return new RangeIterator(v1.getValue(), v2.getValue());
    }

    /**
    * Iterator that produces numeric values in a monotonic sequence,
    * ascending or descending
    */

    private static class RangeIterator implements SequenceIterator,
                                                  ReversibleIterator,
                                                  LastPositionFinder {

        long start;
        long currentValue;
        int increment;
        long limit;

        public RangeIterator(long start, long end) {
            this.start = start;
            increment = (start <= end ? +1 : -1);
            currentValue = start;
            limit = end;
        }

        public boolean hasNext() {
            if (increment>0) {
                return currentValue <= limit;
            } else {
                return currentValue >= limit;
            }
        }

        public Item next() {
            long d = currentValue;
            currentValue += increment;
            return new IntegerValue(d);
        }

        public Item current() {
            return new IntegerValue(currentValue - increment);
        }

        public int position() {
            if (increment>0) {
                return (int)(currentValue - start);
            } else {
                return (int)(start - currentValue);
            }
        }

        public int getLastPosition() {
            return (int)((limit - start) * increment + 1);
        }

        public SequenceIterator getAnother() throws XPathException {
            return new RangeIterator(start, limit);
        }

        public SequenceIterator getReverseIterator() {
            return new RangeIterator(limit, start);
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
