package net.sf.saxon.expr;
import net.sf.saxon.om.Item;
import net.sf.saxon.sort.AtomicComparer;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.BooleanValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import java.util.Comparator;


/**
* Class to handle comparisons of singletons. Unlike ValueComparison, this class
* allows either side to be an empty sequence, and converts untyped atomic values to the
* type of the other argument.
*/

public class SingletonComparison extends BinaryExpression {

    private AtomicComparer comparer;
    private boolean backwardsCompatible = false;

//    public SingletonComparison() {}

    public SingletonComparison(Expression p1, int operator, Expression p2) {
        super(p1, operator, p2);
    }

    public void setComparator(Comparator comp) {
        if (comp instanceof AtomicComparer) {
            comparer = (AtomicComparer)comp;
        } else {
            comparer = new AtomicComparer(comp);
        }
    }

    /**
    * Type-check the expression
    * @return the checked expression
    */

    public Expression analyze(StaticContext env) throws XPathException {
        backwardsCompatible = env.isInBackwardsCompatibleMode();
        return super.analyze(env);
    }

    /**
    * Determine the data type of the expression
    * @return Type.BOOLEAN
    */

    public int getItemType() {
        return Type.BOOLEAN;
    }

    /**
    * Evaluate the expression in a given context
    * @param context the given context for evaluation
    * @return a BooleanValue representing the result of the numeric comparison of the two operands
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        return BooleanValue.get(effectiveBooleanValue(context));
    }

    /**
    * Evaluate the expression in a boolean context
    * @param context the given context for evaluation
    * @return a boolean representing the result of the numeric comparison of the two operands
    */

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        AtomicValue v1 = (AtomicValue)operands[0].evaluateItem(context);
        if (v1==null) return false;
        AtomicValue v2 = (AtomicValue)operands[1].evaluateItem(context);
        if (v2==null) return false;

        try {
            return GeneralComparison.compare(v1, operator, v2, comparer, backwardsCompatible);
        } catch (XPathException.Type e) {
            // re-throw the exception with location information added
            typeError(e.getMessage());
            return false;   // dummy return
        }
    }

    protected String displayOperator() {
        return "singleton " + super.displayOperator();
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
