package net.sf.saxon.expr;

import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.ArrayIterator;
import net.sf.saxon.om.NamePool;

/**
 * A TailExpression represents a FilterExpression of the form EXPR[position() > n]
 * Here n is usually 2, but we allow other values
 */
public class TailExpression extends ComputedExpression {

    Expression base;
    int start;      // 1-based offset of first item from base expression
                    // to be included

    /**
     * Construct a TailExpression, representing a filter expression of the form
     * $base[position() >= $start]
     * @param base    the expression to be filtered
     * @param start   the position (1-based) of the first item to be included
     */

    public TailExpression(Expression base, int start) {
        this.base = base;
        this.start = start;
    }

    public Expression analyze(StaticContext env) {
        // by the time we get here, the analysis has all been done
        return this;
    }

    public Expression promote(PromotionOffer offer) throws XPathException {
        Expression exp = offer.accept(this);
        if (exp != null) {
            return exp;
        } else {
            if (offer.action != PromotionOffer.UNORDERED) {
                base = base.promote(offer);
            }
            return this;
        }
    }

    public int computeSpecialProperties() {
        return base.getSpecialProperties();
    }

    public int getItemType() {
        return base.getItemType();
    }

    public int computeCardinality() {
        return base.getCardinality() | StaticProperty.CARDINALITY_ALLOWS_ZERO;
    }

    public Expression[] getSubExpressions() {
        Expression[] sub = new Expression[1];
        sub[0] = base;
        return sub;
    }

    public Expression getBaseExpression() {
        return base;
    }

    public int getStart() {
        return start;
    }

    public boolean equals(Object other) {
        return other instanceof TailExpression &&
                base.equals(((TailExpression)other).base);
    }

    public int hashCode() {
        return base.hashCode();
    }

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        SequenceIterator baseIter = base.iterate(context);
        if (baseIter instanceof ArrayIterator) {
            return ((ArrayIterator)baseIter).makeSliceIterator(start, Integer.MAX_VALUE);
        } else {
            return new TailIterator(baseIter, start);
        }
    }

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "tail " + start);
        base.display(level+1, pool);
    }

    public static class TailIterator implements SequenceIterator {

        private SequenceIterator base;
        private int start;

        public TailIterator(SequenceIterator base, int start) throws XPathException {
            this.base = base;
            this.start = start;

            // discard the first n-1 items from the underlying iterator
            for (int i=0; i < start-1; i++) {
                if (base.hasNext()) {
                    base.next();
                }
            }
        }

        public boolean hasNext() throws XPathException {
            return base.hasNext();
        }

        public Item next() throws XPathException {
            return base.next();
        }

        public Item current() {
            return base.current();
        }

        public int position() {
            return base.position() - start + 1;
        }

        public SequenceIterator getAnother() throws XPathException {
            return new TailIterator(base.getAnother(), start);
        }
    }
}
