package net.sf.saxon.functions;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.FunctionProxy;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.instruct.NamespaceContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceConstant;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.style.ExpressionContext;
import net.sf.saxon.style.StandardNames;
import net.sf.saxon.style.StyleNodeFactory;
import net.sf.saxon.value.BooleanValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Value;
import net.sf.saxon.xpath.XPathException;

/**
* This class supports the XSLT element-available and function-available functions.
*/

public class Available extends SystemFunction implements XSLTFunction {

    public static final int ELEMENT_AVAILABLE = 0;
    public static final int FUNCTION_AVAILABLE = 1;

    private transient NamespaceContext nsContext;
    private transient StyleNodeFactory styleNodeFactory;

    public void checkArguments(StaticContext env) throws XPathException {
        super.checkArguments(env);
        if (!(argument[0] instanceof Value)) {
            // we need to save the namespace context
            nsContext = ((ExpressionContext)env).getNamespaceContext();
        }
    }

    /**
    * preEvaluate: this method uses the static context to do early evaluation of the function
    * if the argument is known
    */

    public Expression preEvaluate(StaticContext env) throws XPathException {
        String qname = ((StringValue)argument[0]).getStringValue();

        boolean b = false;
        switch(operation) {
            case ELEMENT_AVAILABLE:
                b = ((ExpressionContext)env).isElementAvailable(qname);
                break;
            case FUNCTION_AVAILABLE:
                b = ((ExpressionContext)env).isFunctionAvailable(qname);
                break;
        }
        return BooleanValue.get(b);
    }

    /**
    * Run-time evaluation. This is tricky: we don't keep all the information
    * necessary at run time. Fortunately, this case is rarely used. The function
    * currently returns true only for XSLT and Saxon instructions, and for
    * core functions and Java extensions (not stylesheet functions).
    */

    // TODO: extend this to stylesheet functions and user-defined extension elements

    public Item evaluateItem(XPathContext context) throws XPathException {

        StringValue nameValue = (StringValue)argument[0].evaluateItem(context);
        String name = nameValue.getStringValue();
        String[] parts = null;
        try {
            parts = Name.getQNameParts(name);
        } catch (QNameException e) {
            dynamicError(e.getMessage());
        }
        String prefix = parts[0];
        String lname = parts[1];
        String uri;
        if (prefix.equals("")) {
            if (operation==ELEMENT_AVAILABLE) {
                // Use the default namespace for ELEMENT_AVAILABLE only
                uri = nsContext.getURIForPrefix(prefix,
        	                                    true,
        	                                    context.getController().getNamePool());
            } else {
        	    uri = "";
        	}
        } else {
        	uri = nsContext.getURIForPrefix(prefix,
        	                                false,
        	                                context.getController().getNamePool());
        }

        boolean b = false;
        switch(operation) {
            case ELEMENT_AVAILABLE:
                b = isElementAvailable(uri, lname, context);
                break;
            case FUNCTION_AVAILABLE:
                b = isFunctionAvailable(uri, lname);
                break;
        }
        return BooleanValue.get(b);

    }

    /**
    * Determine at run-time whether a function is available. Returns true only in the case
    * of standard functions, EXSLT and Saxon functions, and Java extension functions found
    * in a class that's implicit from the URI. Currently returns false in the case of
    * stylesheet functions (xsl:function) and Java classes identified using saxon:script
    */

    private static boolean isFunctionAvailable(String uri, String localname) {
        try {
            if (uri.equals("") || uri.equals(NamespaceConstant.FN)) {
                return (StandardFunction.getFunction(localname) != null);
            } else {
                Class theClass = FunctionProxy.getVendorExtensionClass(uri);
                if (theClass == null) {
                    theClass = FunctionProxy.getImplicitJavaClass(uri);
                }
              	if (theClass == null) {
              	    return false;
              	}

            	FunctionProxy fp = new FunctionProxy();
            	fp.setArguments(NO_ARGUMENTS);
            	return fp.setFunctionName(theClass, localname);
            }
        } catch (Exception err) {
            return false;
        }
    }

    /**
    * Determine at run-time whether a particular instruction is available. Returns true
    * only in the case of XSLT instructions and Saxon extension instructions; returns false
    * for user-defined extension instructions
    */

    private boolean isElementAvailable(String uri, String localname, XPathContext context) {

        // This is horribly inefficient. But hopefully it's hardly ever executed, because there
        // is very little point calling element-available() with a dynamically-constructed argument.
        // And the inefficiency is only incurred once, on the first call.

        // Note: this requires the compile-time classes to be available at run-time; it will need
        // changing if we ever want to build a run-time JAR file.

        try {
            if (styleNodeFactory==null) {
                NamePool instructionNamePool = new NamePool();
                StandardNames sn = new StandardNames();
                sn.allocateNames(instructionNamePool);
                styleNodeFactory =
                    new StyleNodeFactory (
                            instructionNamePool,
                            sn,
                            context.getController().getConfiguration().isAllowExternalFunctions());
            }
            return styleNodeFactory.isElementAvailable(uri, localname);
        } catch (Exception err) {
            //err.printStackTrace();
            return false;
        }
    }

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
