package net.sf.saxon.functions;
import net.sf.saxon.Controller;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.style.ExpressionContext;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Value;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import java.io.StringWriter;
import java.util.Properties;


/**
* This class implements the saxon:serialize() extension function,
* which is specially-recognized by the system because it needs access
* to parts of the static context
*/

public class Serialize extends SystemFunction implements XSLTFunction {

    Properties outputProperties;

    /**
    * Method supplied by each class of function to check arguments during parsing, when all
    * the argument expressions have been read
    */

    public void checkArguments(StaticContext env) throws XPathException {
        super.checkArguments(env);
        if (!(env instanceof ExpressionContext)) {
            throw new XPathException.Static("saxon:serialize is available only within XSLT");
        }
        // We require the output format name to be known at compile time. If we allowed it
        // to be dynamic, we would not only need to save the namespace context, but also to
        // save details of all xsl:output declarations for use at run-time.

        if (!(argument[1] instanceof StringValue)) {
            throw new XPathException.Static("Second argument of saxon:serialize must be known at compile time");
        }
        String format = ((StringValue)argument[1]).getStringValue();
        int fingerprint = -1;
        if (!format.equals("")) {
            fingerprint = ((ExpressionContext)env).getFingerprint(format, false);
            if (fingerprint==-1) {
                throw new XPathException.Static("Output format '" + format + "' has not been defined");
            }
        } try {
            outputProperties = ((ExpressionContext)env).getXSLStyleSheet().gatherOutputProperties(fingerprint);
        } catch (TransformerConfigurationException err) {
            throw new XPathException.Static(err);
        }
    }

    /**
    * Evaluate the function
    */

    public Item evaluateItem(XPathContext c) throws XPathException {
        NodeInfo node = (NodeInfo)argument[0].evaluateItem(c);
        if (node==null) {
            return StringValue.EMPTY_STRING;
        }

        try {
            StringWriter result = new StringWriter();
            Controller controller = c.getController();
            SequenceReceiver oldOutputter = controller.getReceiver();

            controller.changeOutputDestination(outputProperties, new StreamResult(result), false);
            node.copy(controller.getReceiver(), NodeInfo.ALL_NAMESPACES, true);
            controller.resetOutputDestination(oldOutputter);
            return new StringValue(result.toString());
        } catch (TransformerException err) {
            throw new XPathException.Dynamic(err);
        }
    }
}




//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
