package net.sf.saxon.functions;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.om.Item;
import net.sf.saxon.value.NumericValue;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.IntegerValue;
import net.sf.saxon.xpath.XPathException;


public class Substring extends SystemFunction {

    /**
    * Evaluate the function
    */

    public Item evaluateItem(XPathContext context) throws XPathException {

        StringValue sv = (StringValue)argument[0].evaluateItem(context);
        if (sv==null) return null;
        String s = sv.getStringValue();

        NumericValue a = (NumericValue)argument[1].evaluateItem(context);

        if (getNumberOfArguments()==2) {
            return new StringValue(substring(s, a));
        } else {
            NumericValue b = (NumericValue)argument[2].evaluateItem(context);
            return new StringValue(substring(s, a, b));
        }
    }

    /**
    * Implement substring function with two arguments.
    */

    private static String substring(String s, NumericValue start) {
        int slength = s.length();

        long lstart;
        if (start instanceof IntegerValue) {
            lstart = ((IntegerValue)start).getValue();
        } else {
            NumericValue rstart = start.round();
            // We need to be careful to handle cases such as plus/minus infinity
            if (rstart.compareTo(IntegerValue.ZERO) <= 0) {
                return s;
            } else if (rstart.compareTo(new IntegerValue(slength)) > 0) {
                // this works even where the string contains surrogate pairs,
                // because the Java length is always >= the XPath length
                return "";
            } else {
                try {
                    lstart = rstart.asLong();
                } catch (XPathException err) {
                    // this shouldn't happen unless the string length exceeds the bounds
                    // of a long
                    throw new AssertionError("string length out of permissible range");
                }
            }
        }

        int pos=1;
        int cpos=0;
        while (cpos<slength) {
            if (pos >= lstart) {
                return s.substring(cpos);
            }

            int ch = (int)s.charAt(cpos++);
            if (ch<55296 || ch>56319) pos++;    // don't count high surrogates, i.e. D800 to DBFF
        }
        return "";
    }

    /**
    * Implement substring function with three arguments.
    */

    private static String substring(String s, NumericValue start, NumericValue len) {
        int slength = s.length();

        long lstart;
        if (start instanceof IntegerValue) {
            lstart = ((IntegerValue)start).getValue();
        } else {
            start = start.round();
            // We need to be careful to handle cases such as plus/minus infinity and NaN
            if (start.compareTo(IntegerValue.ZERO) <= 0) {
                lstart = 0;
            } else if (start.compareTo(new IntegerValue(slength)) > 0) {
                // this works even where the string contains surrogate pairs,
                // because the Java length is always >= the XPath length
                return "";
            } else if (start.isNaN()) {
                return "";
            } else {
                try {
                    lstart = start.asLong();
                } catch (XPathException err) {
                    // this shouldn't happen unless the string length exceeds the bounds
                    // of a long
                    throw new AssertionError("string length out of permissible range");
                }
            }
        }

        NumericValue end;
        try {
            end = start.arithmetic(Tokenizer.PLUS, len.round());
        } catch (XPathException e) {
            throw new AssertionError("Unexpected arithmetic failure in substring");
        }
        long lend;
        if (end instanceof IntegerValue) {
            lend = ((IntegerValue)end).getValue();
        } else {
            // We need to be careful to handle cases such as plus/minus infinity and NaN
            if (end.compareTo(IntegerValue.ZERO) <= 0) {
                return "";
            } else if (end.isNaN()) {
                return "";
            } else if (end.compareTo(new IntegerValue(slength)) > 0) {
                // this works even where the string contains surrogate pairs,
                // because the Java length is always >= the XPath length
                lend = slength+1;
            } else {
                try {
                    lend = end.ceiling().asLong();
                } catch (XPathException err) {
                    // this shouldn't happen unless the string length exceeds the bounds
                    // of a long
                    throw new AssertionError("string length out of permissible range");
                }
            }
        }

        int jstart=-1;
        int jend=-1;
        int pos=1;
        int cpos=0;
        while (cpos<slength) {
            if (pos >= lstart) {
                if (pos < lend) {
                    if (jstart<0) {
                        jstart = cpos;
                    }
                } else {
                    jend = cpos;
                    break;
                }
            }

            int ch = (int)s.charAt(cpos++);
            if (ch<55296 || ch>56319) pos++;    // don't count high surrogates, i.e. D800 to DBFF
        }
        if (jstart<0 || jstart==jend) {
            return "";
        } else if (jend<0) {
            return s.substring(jstart);
        } else {
            return s.substring(jstart, jend);
        }
    }
}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
