package net.sf.saxon.functions;
import net.sf.saxon.Version;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.NamespaceContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceConstant;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.style.ExpressionContext;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.xpath.XPathException;


public class SystemProperty extends SystemFunction implements XSLTFunction {

    private NamespaceContext nsContext;

    public void checkArguments(StaticContext env) throws XPathException {
        super.checkArguments(env);
        if (!(argument[0] instanceof StringValue)) {
            // we need to save the namespace context
            nsContext = ((ExpressionContext)env).getNamespaceContext();
        }
    }

    /**
    * preEvaluate: this method performs compile-time evaluation
    */

    public Expression preEvaluate(StaticContext env) throws XPathException {
        String name = ((StringValue)argument[0]).getStringValue();

        try {
            String[] parts = Name.getQNameParts(name);
            String prefix = parts[0];
            String lname = parts[1];
            String uri;
            if (prefix.equals("")) {
                uri = "";
            } else {
                uri = env.getURIForPrefix(prefix);
            }
            return new StringValue(getProperty(uri, lname));
        } catch (QNameException e) {
            throw new XPathException.Static("Invalid system property name. " + e.getMessage());
        }
    }


    /**
    * Evaluate the function at run-time
    */

    public Item evaluateItem(XPathContext context) throws XPathException {

        String name = argument[0].evaluateItem(context).getStringValue();

        try {
            String[] parts = Name.getQNameParts(name);
            String prefix = parts[0];
            String lname = parts[1];
            String uri;
            if (prefix.equals("")) {
                uri = "";
            } else {
                uri = nsContext.getURIForPrefix(prefix, false, context.getController().getNamePool());
            }
            return new StringValue(getProperty(uri, lname));
        } catch (QNameException e) {
            dynamicError("Invalid system property name. " + e.getMessage());
            return null;
        }
    }

    /**
    * Here's the real code:
    */

    public static String getProperty(String uri, String local) {
        if (uri.equals(NamespaceConstant.XSLT)) {
            if (local.equals("version"))
                return Version.getXSLVersionString();
            if (local.equals("vendor"))
                return Version.getProductTitle();
            if (local.equals("vendor-url"))
                return Version.getWebSiteAddress();
            if (local.equals("product-name"))
                return Version.getProductName();
            if (local.equals("product-version"))
                return Version.getProductVersion();
            if (local.equals("is-schema-aware"))
                return "no";
            if (local.equals("supports-serialization"))
                return "yes";
            if (local.equals("supports-backwards-compatibility"))
                return "yes";
            return "";

        } else if (uri.equals("")) {
	        String val = System.getProperty(local);
	        if (val==null) val="";
	        return val;
	    } else {
	    	return "";
	    }
    }

}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
