package net.sf.saxon.functions;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.MappingIterator;
import net.sf.saxon.expr.MappingFunction;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Value;
import net.sf.saxon.value.IntegerValue;
import net.sf.saxon.xpath.XPathException;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;


/**
* This class implements the tokenize() function for regular expression matching. This returns a
* sequence of strings representing the unmatched substrings: the separators which match the
* regular expression are not returned.
*/

public class Tokenize extends SystemFunction implements MappingFunction {

    private Pattern regexp;

    /**
    * Simplify and validate.
    * This is a pure function so it can be simplified in advance if the arguments are known
    */

    public Expression simplify() throws XPathException {
        Expression e = simplifyArguments(true);

        // compile the regular expression once if possible
        if (!(e instanceof Value)) {
            regexp = Matches.tryToCompile(argument, 1, 2);
        }

        return e;
    }

    /**
    * Iterate over the results of the function
    */

    public SequenceIterator iterate(XPathContext c) throws XPathException {
        StringValue sv = (StringValue)argument[0].evaluateItem(c);
        if (sv==null) return EmptyIterator.getInstance();
        String input = sv.getStringValue();

        Pattern re = regexp;
        if (re == null) {

            sv = (StringValue)argument[1].evaluateItem(c);
            if (sv==null) return EmptyIterator.getInstance();
            String pattern = sv.getStringValue();

            String flags;
            if (getNumberOfArguments()==2) {
                flags = "";
            } else {
                sv = (StringValue)argument[2].evaluateItem(c);
                if (sv==null) return EmptyIterator.getInstance();
                flags =sv.getStringValue();
            }

            try {
                re = Pattern.compile(pattern, Matches.setFlags(flags));
            } catch (PatternSyntaxException err) {
                throw new XPathException.Dynamic(err);
            }
        }
        // There is a special rule if the regex matches an empty sequence
        if (re.matcher("").matches()) {
            return new MappingIterator(sv.iterateCharacters(), this, null, null);
        } else {
            return new TokenIterator(input, re);
        }
    }

    /**
     * Mapping function to map a sequence of Unicode codepoints to a sequence
     * of strings
     */

    public Object map(Item item, XPathContext context, Object info) {
        long next = ((IntegerValue)item).getValue();
        if (next < 65536) {
            return new StringValue("" + (char)next);
        } else {  // output a surrogate pair
            //To compute the numeric value of the character corresponding to a surrogate
            //pair, use this formula (all numbers are hex):
            //(FirstChar - D800) * 400 + (SecondChar - DC00) + 10000
            next -= 65536;
            StringBuffer sb = new StringBuffer(2);
            sb.append((char)((next / 1024) + 55296));
            sb.append((char)((next % 1024) + 56320));
            return new StringValue(sb);
        }
    }


    /**
    * Inner class TokenIterator
    */

    public static class TokenIterator implements SequenceIterator {

        private String input;
        private Pattern pattern;
        private Matcher matcher;
        private String current;
        private String next;
        private int position = 0;
        private int prevEnd = 0;


        /**
        * Construct a TokenIterator.
        */

        public TokenIterator (String input, Pattern pattern) {
            this.input = input;
            this.pattern = pattern;
            matcher = pattern.matcher(input);
            prevEnd = 0;
            advance();
        }

        private void advance() {
            if (prevEnd < 0) {
                next = null;
                return;
            }

            if (matcher.find()) {
                next = input.substring(prevEnd, matcher.start());
                prevEnd = matcher.end();
            } else {
                next = input.substring(prevEnd);
                prevEnd = -1;
            }

        }

        public boolean hasNext() {
            return next != null;
        }

        public Item next() {
            current = next;
            advance();
            position++;
            return new StringValue(current);
        }

        public Item current() {
            return new StringValue(current);
        }

        public int position() {
            return position;
        }

        public SequenceIterator getAnother() {
            return new TokenIterator(input, pattern);
        }


    }

    public static void main(String[] args) throws Exception {
        String in = args[0];
        String[] out = Pattern.compile(args[1]).split(in, 0);
        System.out.println("results");
        for (int i=0; i<out.length; i++) {
            System.out.println("[" + out[i] + "]");
        }
        System.out.println("end results");
    }



}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
