package net.sf.saxon.functions;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Navigator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

/**
* This class supports the XPath 2.0 function trace().
* At present the value is simply traced to the System.err stream
*/


public class Trace extends SystemFunction {

    /**
    * preEvaluate: this method suppresses compile-time evaluation by doing nothing
    */

    public Expression preEvaluate(StaticContext env) {
        return this;
    }

    /**
    * Get the static properties of this expression (other than its type). The result is
    * bit-significant. These properties are used for optimizations. In general, if
    * property bit is set, it is true, but if it is unset, the value is unknown.
    */

    public int computeSpecialProperties() {
        return argument[0].getSpecialProperties();
    }

    /**
    * Get the static cardinality
    */

    public int computeCardinality() {
        return argument[0].getCardinality();
    }

    /**
    * Evaluate the function
    */

    public Item evaluateItem(XPathContext context) throws XPathException {
        Item val = argument[0].evaluateItem(context);
        String label = argument[1].evaluateAsString(context);
        traceItem(val, label);
        return val;
    }

    private void traceItem(Item val, String label) throws XPathException {
        if (val==null) {
            System.err.println(label + ": empty sequence");
        } else {
            if (val instanceof NodeInfo) {
                System.err.println(label + ": " + Type.getTypeName(val.getItemType()) + ": "
                                    + Navigator.getPath((NodeInfo)val));
            } else {
                System.err.println(label + ": " + Type.getTypeName(val.getItemType()) + ": "
                                    + val.getStringValue());
            }
        }
    }

    /**
    * Iterate over the results of the function
    */

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        return new TracingIterator(argument[0].iterate(context), argument[1].evaluateAsString(context));
    }

    /**
    * Tracing Iterator class
    */

    public class TracingIterator implements SequenceIterator {

        SequenceIterator base;
        String label;


        public TracingIterator(SequenceIterator base, String label) throws XPathException {
            this.base = base;
            this.label = label;
            if (!base.hasNext()) {
                traceItem(null, label);
            }
        }

        public boolean hasNext() throws XPathException {
            return base.hasNext();
        }

        public Item next() throws XPathException {
            Item n = base.next();
            traceItem(n, label + " [" + position() + "]");
            return n;
        }

        public Item current() {
            return base.current();
        }

        public int position() {
            return base.position();
        }

        public SequenceIterator getAnother() throws XPathException {
            return new TracingIterator(base.getAnother(), label);
        }

    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
