package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.event.NoOpenStartTagException;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.ReceiverOptions;
import net.sf.saxon.expr.*;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.QNameValue;

import javax.xml.transform.TransformerException;
import java.util.List;

/**
* An instruction derived from an xsl:attribute element in stylesheet, or from
 * an attribute constructor in XQuery
*/

public final class Attribute extends SimpleNodeConstructor {

    private Expression attributeName;
    private Expression namespace=null;
    private NamespaceContext nsContext;
    private int annotation;
    private int options;

    /**
    * Construct an Attribute instruction
    * @param attributeName An expression to calculate the attribute name
    * @param namespace An expression to calculate the attribute namespace
    * @param nsContext a NamespaceContext object containing the static namespace context of the
    * stylesheet instruction
    * @param annotation Integer code identifying the type named in the <code>type</code> attribute
    * of the instruction - zero if the attribute was not present
    * @param disable no longer supported, value is always false
    */

    public Attribute (  Expression attributeName,
                        Expression namespace,
                        NamespaceContext nsContext,
                        int annotation,
                        boolean disable ) {
        this.attributeName = attributeName;
        this.namespace = namespace;
        this.nsContext = nsContext;
        this.annotation = annotation;
        this.options = 0;
//        this.options = (disable ? ReceiverOptions.DISABLE_ESCAPING : 0);
    }

    /**
     * Indicate that two attributes with the same name are not acceptable.
     * (This option is set in XQuery, but not in XSLT)
     */

    public void setRejectDuplicates() {
        this.options |= ReceiverOptions.REJECT_DUPLICATES;
    }

    /**
    * Get the name of this instruction
    * @return the string "xsl:attribute"
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("attribute");
    }

    public SequenceType getResultType() {
        return new SequenceType(Type.ATTRIBUTE, Type.ANY_SIMPLE_TYPE, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    public Expression simplify() throws XPathException {
        attributeName = attributeName.simplify();
        if (namespace!=null) {
            namespace = namespace.simplify();
        }
        return super.simplify();
    }

    public void typeCheck(StaticContext env) throws XPathException {
        attributeName = attributeName.analyze(env);

        RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "attribute/name", 0);

        if (attributeName.getItemType() == Type.QNAME) {
            // Can only happen in XQuery
            attributeName = TypeChecker.staticTypeCheck(attributeName,
                    SequenceType.SINGLE_ITEM, false, role);
        } else {
            attributeName = TypeChecker.staticTypeCheck(attributeName,
                    SequenceType.SINGLE_STRING, false, role);
        }

        if (namespace != null) {
            namespace.analyze(env);

            role = new RoleLocator(RoleLocator.INSTRUCTION, "attribute/namespace", 0);
            namespace = TypeChecker.staticTypeCheck(
                    namespace, SequenceType.SINGLE_STRING, false, role);
        }
    }

    public int getDependencies() {
        int dependencies = 0;
        dependencies |= attributeName.getDependencies();
        if (namespace != null) {
            dependencies |= namespace.getDependencies();
        }
        return dependencies | super.getDependencies();
    }

    public void getXPathExpressions(List list) {
        list.add(attributeName);
        if (namespace != null) {
            list.add(namespace);
        }
        super.getXPathExpressions(list);
    }

   /**
     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
     * By default the offer is not accepted - this is appropriate in the case of simple expressions
     * such as constant values and variable references where promotion would give no performance
     * advantage. This method is always called at compile time.
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception XPathException if any error is detected
     */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        attributeName = attributeName.promote(offer);
        if (namespace != null) {
            namespace = namespace.promote(offer);
        }
        if (select != null) {
            select = select.promote(offer);
        }
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    Expression p = ((ExprInstruction)children[c]).promote(offer);
                    if (p instanceof Instruction) {
                        children[c] = (Instruction)p;
                    } else {
                        children[c] = new Sequence(p, null);
                    }
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
    }

    /**
    * Process this instruction
    * @param context the dynamic context of the transformation
    * @return a TailCall to be executed by the caller, always null for this instruction
    */

    public TailCall processLeavingTail(XPathContext context) throws TransformerException
    {
        Controller controller = context.getController();
        int nameCode = evaluateNameCode(context);
        if (nameCode == -1) {
            return null;
        }
        Receiver out = controller.getReceiver();

    	// we may need to change the namespace prefix if the one we chose is
    	// already in use with a different namespace URI: this is done behind the scenes
    	// by the Outputter

        int typeCode = annotation;
        if (typeCode == Type.ANY_SIMPLE_TYPE) {
            typeCode = 0;
        }
        String value = expandChildren(context).toString();
        if (typeCode != 0) {
            // test whether the value actually conforms to the given type
            new StringValue(value).convert(typeCode);
        }
        try {
            out.attribute(nameCode, typeCode, value, options);
        } catch (NoOpenStartTagException err) {
            recoverableError("Cannot write an attribute node when no element start tag is open",
                controller);
        } catch (TransformerException err) {
            throw styleError(err, context.getController());
        }

        return null;
    }

    protected int evaluateNameCode(XPathContext context) throws XPathException, TransformerException {
        Controller controller = context.getController();
        NamePool pool = controller.getNamePool();

        Item nameValue = attributeName.evaluateItem(context);

        String prefix = null;
        String localName = null;

        if (nameValue instanceof StringValue) {
            // this will always be the case in XSLT
            String rawName = nameValue.getStringValue();
            try {
                String[] parts = Name.getQNameParts(rawName);
                prefix = parts[0];
                localName = parts[1];
            } catch (QNameException err) {
                recoverableError("Invalid attribute name: " + rawName, controller);
                return -1;
            }
            if (rawName.equals("xmlns")) {
                if (namespace==null) {
                    recoverableError("Invalid attribute name: " + rawName, controller);
                    return -1;
                }
            }
            if (prefix.equals("xmlns")) {
                if (namespace==null) {
                    recoverableError("Invalid attribute name: " + rawName, controller);
                    return -1;
                } else {
                    // ignore the prefix "xmlns"
                    prefix = "";
                }
            }

        } else if (attributeName instanceof QNameValue) {
            // this is allowed in XQuery
            localName = ((QNameValue)nameValue).getLocalName();
            String namespaceURI = ((QNameValue)nameValue).getNamespaceURI();
            namespace = new StringValue(namespaceURI);
            if (namespaceURI.equals("")) {
                prefix = "";
            } else {
                // we need to allocate a prefix. Any one will do; if it's a duplicate,
                // a different one will be substituted
                prefix = pool.suggestPrefixForURI(namespaceURI);
                if (prefix == null) {
                    prefix = "nsq0";
                }
            }

        }

        String uri;

        if (namespace==null) {
        	if (prefix.equals("")) {
        		uri = "";
        	} else {
                uri = nsContext.getURIForPrefix(prefix, false, controller.getNamePool());
                if (uri==null) {
                    recoverableError("Undeclared prefix in attribute name: " + prefix, controller);
                    return -1;
      		    }
        	}

        } else {

            // generate a name using the supplied namespace URI

            uri = namespace.evaluateAsString(context);
            if ("".equals(uri)) {
                // there is a special rule for this case in the specification;
                // we force the attribute to go in the null namespace
                prefix = "";

            } else {
                // if a suggested prefix is given, use it; otherwise try to find a prefix
                // associated with this URI; if all else fails, invent one.
                if ("".equals(prefix)) {
                    prefix = pool.suggestPrefixForURI(uri);
                    if (prefix == null) {
                        prefix = "ns0";
                        // this will be replaced later if it is already in use
                    }
                }
            }
        }

        return pool.allocate(prefix, uri, localName);
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "attribute ");
        System.err.println(ExpressionTool.indent(level+1) + "name");
        attributeName.display(level+2, pool);
        super.display(level+1, pool);
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
