package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;

import javax.xml.transform.TransformerException;

/**
* An instruction representing an xsl:comment element in the stylesheet.
*/

public final class Comment extends SimpleNodeConstructor {

    /**
    * Construct the instruction
    */

    public Comment() {}

    /**
    * Get the instruction name, for diagnostics and tracing
    * return the string "xsl:comment"
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("comment");
    }

    public SequenceType getResultType() {
        return new SequenceType(Type.COMMENT, Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    public void typeCheck(StaticContext env) {}


    /**
       * Offer promotion for subexpressions. The offer will be accepted if the subexpression
       * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
       * By default the offer is not accepted - this is appropriate in the case of simple expressions
       * such as constant values and variable references where promotion would give no performance
       * advantage. This method is always called at compile time.
       *
       * @param offer details of the offer, for example the offer to move
       *     expressions that don't depend on the context to an outer level in
       *     the containing expression
       * @exception net.sf.saxon.xpath.XPathException if any error is detected
       */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        if (select != null) {
            select = select.promote(offer);
        }
        if (children != null) {
            for (int c = 0; c < children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    Expression p = ((ExprInstruction) children[c]).promote(offer);
                    if (p instanceof Instruction) {
                        children[c] = (Instruction) p;
                    } else {
                        children[c] = new Sequence(p, null);
                    }
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
    }


    /**
    * Process this instruction, to output a Comment Node
    * @param context the dynamic context for this transformation
    * @return a TailCall representing a call delegated to the caller. Always
    * returns null in this implementation
    */

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        String comment = expandChildren(context).toString();

        while(true) {
            int hh = comment.indexOf("--");
            if (hh < 0) break;
            recoverableError("Invalid characters (--) in comment", controller);
            comment = comment.substring(0, hh+1) + " " + comment.substring(hh+1);
        }
        if (comment.length()>0 && comment.charAt(comment.length()-1)=='-') {
            recoverableError("Invalid character (-) at end of comment", controller);
            comment = comment + " ";
        }
        controller.getReceiver().comment(comment, 0);
        return null;
    }

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "comment");
        super.display(level+1, pool);
    }

}
//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
