package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.event.NoOpenStartTagException;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.Type;

import javax.xml.transform.TransformerException;

/**
* Handler for xsl:copy elements in stylesheet.
*/

public class Copy extends Instruction {

    private AttributeSet[] useAttributeSets;
    private boolean copyNamespaces;

    public Copy( AttributeSet[] useAttributeSets, boolean copyNamespaces ) {
        this.useAttributeSets = useAttributeSets;
        this.copyNamespaces = copyNamespaces;
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    * @return the string "xsl:copy"
    */


    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("copy");
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        SequenceReceiver out = controller.getReceiver();
        Item item = controller.getCurrentItem();
        if (!(item instanceof NodeInfo)) {
            out.append(item);
            return null;
        }
        NodeInfo source = (NodeInfo)item;


        // Processing depends on the node type.

        switch(source.getItemType()) {

        case Type.ELEMENT:
            out.startElement(source.getNameCode(), 0, 0);

            if (copyNamespaces) {
                source.outputNamespaceNodes(out, true);
            }

            if (useAttributeSets != null) {
                AttributeSet.expand(useAttributeSets, context);
            }
            processChildren(context);
            out.endElement();
            break;

        case Type.ATTRIBUTE:
        	int nameCode = source.getNameCode();
        	//if (((nameCode>>20)&0xff) != 0) {	// prefix!=""
        	//	nameCode = out.checkAttributePrefix(nameCode);
        	//}
        	try {
                out.attribute(nameCode, 0, source.getStringValue(), 0);
            } catch (NoOpenStartTagException err) {
                recoverableError("Cannot write an attribute node when no element start tag is open",
                    controller);
            }
            break;

        case Type.TEXT:
            out.characters(source.getStringValue(), 0);
            break;

        case Type.PROCESSING_INSTRUCTION:
            out.processingInstruction(source.getDisplayName(), source.getStringValue(), 0);
            break;

        case Type.COMMENT:
            out.comment(source.getStringValue(), 0);
            break;

        case Type.NAMESPACE:
            try {
                source.copy(out, NodeInfo.NO_NAMESPACES, false);
            } catch (NoOpenStartTagException err) {
                recoverableError("Cannot write a namespace node when no element start tag is open",
                    controller);
            }
            break;

        case Type.DOCUMENT:
            processChildren(context);
            break;

        default:
            throw new IllegalArgumentException("Unknown node type " + source.getItemType());

        }
        return null;
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
