package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SequenceOutputter;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.*;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.value.QNameValue;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import java.util.List;


/**
* An instruction representing xsl:element element in the stylesheet.
*/

public class Element extends ExprInstruction {

    private Expression elementName;
//    private String fixedPrefix = null;
//    private String fixedLocalName = null;
    private Expression namespace = null;
    private NamespaceContext nsContext;
    private AttributeSet[] useAttributeSets;
    private int annotation;

    /**
     * Create an instruction that creates a new element node
     * @param elementName Expression that evaluates to produce the name of the
     * element node as a lexical QName
     * @param namespace Expression that evaluates to produce the namespace URI of
     * the element node. Set to null if the namespace is to be deduced from the prefix
     * of the elementName.
     * @param nsContext Saved copy of the static namespace context for the instruction.
     * Can be set to null if namespace is supplied.
     * @param useAttributeSets Array of attribute sets to be expanded
     * @param annotation Type annotation for the new element node
     * @throws XPathException
     */
    public Element( Expression elementName,
                    Expression namespace,
                    NamespaceContext nsContext,
                    AttributeSet[] useAttributeSets,
                    int annotation ) throws XPathException {
        this.elementName = elementName;
        this.namespace = namespace;
        this.nsContext = nsContext;
        this.useAttributeSets = useAttributeSets;
        this.annotation = annotation;
//        if (elementName instanceof StringValue) {
//            String parts[];
//            try {
//                parts = Name.getQNameParts(((StringValue)elementName).getStringValue());
//            } catch (QNameException err) {
//                throw new XPathException.Static (
//                        "Invalid element name. " + err.getMessage());
//            }
//            fixedPrefix = parts[0];
//            fixedLocalName = parts[1];
//        }
    }

    /**
     * Get the result type returned by this instruction
     * @return a SequenceType representing the type element(*,*)
     */

    public SequenceType getResultType() {
        return new SequenceType(Type.ELEMENT, Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    public Expression simplify() throws XPathException {
        elementName = elementName.simplify();
        if (namespace!=null) {
            namespace = namespace.simplify();
        }
        return super.simplify();
    }

    public Expression analyze(StaticContext env) throws XPathException {
        elementName = elementName.analyze(env);

        RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "element/name", 0);

        if (elementName.getItemType() == Type.QNAME) {
            // Can only happen in XQuery
            elementName = TypeChecker.staticTypeCheck(
                    elementName, SequenceType.SINGLE_ITEM, false, role);
        } else {
            elementName = TypeChecker.staticTypeCheck(
                    elementName, SequenceType.SINGLE_STRING, false, role);
        }
        if (namespace != null) {
            namespace = namespace.analyze(env);

            role = new RoleLocator(RoleLocator.INSTRUCTION, "element/namespace", 0);
            namespace = TypeChecker.staticTypeCheck(
                    namespace, SequenceType.SINGLE_STRING, false, role);
        }
        for (int c=0; c<children.length; c++) {
            if (children[c] instanceof ExprInstruction) {
                children[c] = (ExprInstruction)((ExprInstruction)children[c]).analyze(env);
            } else {
                throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
            }
        }
        return this;
    }


    public void getXPathExpressions(List list) {
        list.add(elementName);
        if (namespace != null) {
            list.add(namespace);
        }
    }

    /**
     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
     * By default the offer is not accepted - this is appropriate in the case of simple expressions
     * such as constant values and variable references where promotion would give no performance
     * advantage. This method is always called at compile time.
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception XPathException if any error is detected
     */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        elementName = elementName.promote(offer);
        if (namespace != null) {
            namespace = namespace.promote(offer);
        }
        for (int c=0; c<children.length; c++) {
            if (children[c] instanceof ExprInstruction) {
                Expression p = ((ExprInstruction)children[c]).promote(offer);
                if (p instanceof Instruction) {
                    children[c] = (Instruction)p;
                } else {
                    children[c] = new Sequence(p, null);
                }
            } else {
                throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
            }
        }
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {

        Controller controller = context.getController();

    	int typeCode = annotation;
    	//if (annotation == Type.ANY_TYPE) {
    	//    typeCode = 0;
    	//}

        // produce (pending) output


        Receiver out = controller.getReceiver();

        int nameCode = evaluateNameCode(context);
        if (nameCode == -1) {
            // indicates an invalid name; we must take the recovery action
            skipElement(context);
            return null;
        } else {
            out.startElement(nameCode, typeCode, 0);

            // apply the content of any attribute sets mentioned in use-attribute-sets
            if (useAttributeSets != null) {
                AttributeSet.expand(useAttributeSets, context);
            }

            // process subordinate elements in stylesheet
            processChildren(context);

            // output the element end tag
            out.endElement();
            return null;
        }
    }

    private int evaluateNameCode(XPathContext context) throws XPathException, TransformerException {

        Controller controller = context.getController();
        NamePool pool = controller.getNamePool();

        String prefix = null;
        String localName = null;
        String uri;

//        if (fixedLocalName == null) {
            // name needs to be evaluated at run-time
            Item nameValue = elementName.evaluateItem(context);
            if (nameValue instanceof StringValue) {
                // this will always be the case in XSLT
                String rawName = nameValue.getStringValue();
                try {
                    String[] parts = Name.getQNameParts(rawName);
                    prefix = parts[0];
                    localName = parts[1];
                } catch (QNameException err) {
                    recoverableError("Invalid element name. " + err.getMessage(),
                            controller);
                    return -1;
                }
            } else if (nameValue instanceof QNameValue) {
                // this is allowed in XQuery
                localName = ((QNameValue)nameValue).getLocalName();
                uri = ((QNameValue)nameValue).getNamespaceURI();
                namespace = new StringValue(uri);
                // we need to allocate a prefix. Any one will do; if it's a duplicate,
                // a different one will be substituted
                prefix = pool.suggestPrefixForURI(uri);
                if (prefix == null) {
                    prefix = "nsq0";
                }
            }
//        }

        if (namespace==null) {
            uri = nsContext.getURIForPrefix(prefix, true, pool);
            if (uri==null) {
                recoverableError("Undeclared prefix in element name: " + prefix, controller);
                return -1;
  		    }

        } else {
            uri = namespace.evaluateAsString(context);
            if (uri.equals("")) {
                // there is a special rule for this case in the specification;
                // we force the element to go in the null namespace
                prefix = "";
            }
            if (prefix.equals("xmlns")) {
                // this isn't a legal prefix so we mustn't use it
                prefix = "x-xmlns";
            }
        }

        return pool.allocate(prefix, uri, localName);

    }

    /**
    * Recovery action when the element name is invalid. We need to tell the outputter
    * about this, so that it can ignore attributes in the content
    */

    private void skipElement(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        controller.getReceiver().startElement(-1, 0, 0);
        // Sending a namecode of -1 to the Outputter is a special signal to ignore
        // this element and the attributes that follow it
        processChildren(context);
        // Note, we don't bother with an endElement call
    }

   /**
     * Evaluate as an expression. We rely on the fact that when these instructions
     * are generated by XQuery, there will always be a valueExpression to evaluate
     * the content
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        try {
            Controller controller = context.getController();
            NamePool pool = controller.getNamePool();
            SequenceReceiver old = controller.getReceiver();
            SequenceOutputter seq = new SequenceOutputter();
            //seq.setSystemId(baseURI);
            seq.setNamePool(pool);

            seq.startDocument();
            seq.startElement(evaluateNameCode(context), 0, 0);
            // ignore attribute sets for now

            controller.changeToSequenceOutputDestination(seq);
            processChildren(context);

            seq.endElement();
            seq.endDocument();
            controller.resetOutputDestination(old);
            return seq.getFirstItem();
        } catch (TransformerException err) {
            if (err instanceof XPathException) {
                throw (XPathException)err;
            } else {
                throw new XPathException.Dynamic(err);
            }
        }
    }


    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("element");
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "element ");
        System.err.println(ExpressionTool.indent(level+1) + "name");
        elementName.display(level+2, pool);
        if (children.length==0) {
            System.err.println(ExpressionTool.indent(level+1) + "empty content");
        } else {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    ((Expression)children[c]).display(level+1, pool);
                } else {
                    System.err.println(ExpressionTool.indent(level+1) + children[c].getInstructionName());
                }
            }
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
