package net.sf.saxon.instruct;
import net.sf.saxon.Configuration;
import net.sf.saxon.Controller;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.event.Stripper;
import net.sf.saxon.sort.CodepointCollator;
import net.sf.saxon.trans.DecimalFormatManager;
import net.sf.saxon.trans.KeyManager;
import net.sf.saxon.trans.Mode;
import net.sf.saxon.trans.RuleManager;

import javax.xml.transform.TransformerException;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Properties;

/**
* A compiled stylesheet in executable form.
* Note that the original stylesheet tree is no longer retained.
*/

public class Executable extends Instruction {

                // the Configuration options
    private Configuration config;

                // definitions of strip/preserve space action
    private Mode stripperRules;

                // boolean indicating whether any whitespace is stripped
    private boolean stripsWhitespace;

                // definitions of template rules
    private RuleManager ruleManager;

                // definitions of keys
    private KeyManager keyManager;

                // definitions of decimal formats
    private DecimalFormatManager decimalFormatManager;

                // the number of slots needed for global variables and params
    private int numberOfVariables = 0;

                // default collating sequence
    private String defaultCollationName;

                // named sort key definitions
    private HashMap sortKeyTable;

                // default output properties (for the unnamed output format)
    private Properties defaultOutputProperties;

                // index of global variables and parameters, by fingerprint
                // (overridden variables are excluded)
    //private HashMap variableMap = new HashMap();

                // index of named templates. Available at run-time only if saxon:allow-avt is used
    private HashMap namedTemplateTable = new HashMap();

                // count of the maximum number of local variables in any template
    private int largestStackFrame = 0;

                // table of named collations defined in the stylesheet
    private HashMap collationTable = new HashMap();

                // table of system IDs for stylesheet modules
    private String[] moduleArray;

                // table of character maps
    private HashMap characterMapIndex;

    //private HashMap outputDefinitionTable = new HashMap();

    /**
     * Set the configuration
     */

    public void setConfiguration(Configuration config) {
        this.config = config;
    }

    /**
     * Get the configuration
     */

    public Configuration getConfiguration() {
        return config;
    }

    /**
    * Get the name of this instruction
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("transform");
    }

    /**
    * Set the list of stylesheet modules, for use in error messages and tracing
    * @param modules an array of strings containing the URIs of the stylesheet modules
    */

    public void setModuleArray(String[] modules) {
        moduleArray = modules;
    }

    /**
    * Get the ID of the stylesheet module with a given index
    * @param index an integer identifying the module within the array of modules passed
    * to the setModuleArray() method
    * @return the URI of the index'th module
    */

    public String getSystemId(int index) {
        return moduleArray[index];
    }


    /**
    * Set the RuleManager that handles template rules
    * @param rm the RuleManager containing details of all the template rules
    */

    public void setRuleManager(RuleManager rm) {
        ruleManager = rm;
    }

    /**
    * Get the RuleManager which handles template rules
    * @return the RuleManager registered with setRuleManager
    */

    public RuleManager getRuleManager() {
        return ruleManager;
    }

    /**
    * Set the named template table
    * @param table a hash table containing entries that map the names of named
    * templates to the instructions representing the xsl:template instruction
    */

    public void setNamedTemplateTable(HashMap table) {
        namedTemplateTable = table;
    }

    /**
    * Get the named template table
    * @return a hash table containing entries that map the names of named
    * templates to the instructions representing the xsl:template instruction
    */

    public HashMap getNamedTemplateTable() {
        if (namedTemplateTable==null) {
            namedTemplateTable = new HashMap();
        }
        return namedTemplateTable;
    }

    /**
    * Set the named Sort Key Definitions
    * @param sk a hash table that maps the names of sort key definitions
    * to the SortKeyDefinition objects defining them
    */

    public void setSortKeyTable(HashMap sk) {
        sortKeyTable = sk;
    }

    /**
    * Get the named Sort Key Definitions
    * @return the hash table containing named sort key definitions
    */

    public HashMap getSortKeyTable() {
        if (sortKeyTable==null) {
            sortKeyTable = new HashMap();
        }
        return sortKeyTable;
    }

    /**
    * Set the index of named character maps
    * @param cmi a hash table that maps the names of character maps
    * to the HashMap objects representing the character maps
    */

    public void setCharacterMapIndex(HashMap cmi) {
        characterMapIndex = cmi;
    }

    /**
    * Get the index of named character maps
    * @return the hash table that maps the names of character maps
    * to the HashMap objects representing the character maps
    */

    public HashMap getCharacterMapIndex() {
        if (characterMapIndex==null) {
            characterMapIndex = new HashMap();
        }
        return characterMapIndex;
    }

    /**
    * Set the rules determining which nodes are to be stripped from the tree
    * @param rules a Mode object containing the whitespace stripping rules. A Mode
    * is generally a collection of template rules, but it is reused here to represent
    * a collection of stripping rules.
    */

    public void setStripperRules(Mode rules) {
        stripperRules = rules;
    }

    /**
    * Indicate that the stylesheet does some whitespace stripping
    * @param strips true if the stylesheet performs whitespace stripping
    * of one or more elements.
    */

    public void setStripsWhitespace(boolean strips) {
        stripsWhitespace = strips;
    }

    /**
    * Create a Stripper which handles whitespace stripping definitions
    * @return the constructed Stripper object
    */

    public Stripper newStripper() {
        return new Stripper(stripperRules);
    }

    /**
    * Determine whether this stylesheet does any whitespace stripping
    * @return true if the stylesheet performs whitespace stripping
    * of one or more elements.
    */

    public boolean stripsWhitespace() {
        return stripsWhitespace;
    }

    /**
    * Set the KeyManager which handles key definitions
    * @param km the KeyManager containing the xsl:key definitions
    */

    public void setKeyManager(KeyManager km) {
        keyManager = km;
    }

    /**
    * Get the KeyManager which handles key definitions
    * @return the KeyManager containing the xsl:key definitions
    */

    public KeyManager getKeyManager() {
        if (keyManager==null) {
            keyManager = new KeyManager();
        }
        return keyManager;
    }

    /**
    * Set the default output properties (the properties for the unnamed output format)
    * @param properties the output properties to be used when the unnamed output format
    * is selected
    */

    public void setDefaultOutputProperties(Properties properties) {
        defaultOutputProperties = properties;
    }

    /**
    * Get the default output properties
    * @return the properties for the unnamed output format
    */

    public Properties getDefaultOutputProperties() {
        if (defaultOutputProperties==null) {
            defaultOutputProperties = new Properties();
        }
        return defaultOutputProperties;
    }

    /**
    * Set the DecimalFormatManager which handles decimal-format definitions
    * @param dfm the DecimalFormatManager containing the named xsl:decimal-format definitions
    */

    public void setDecimalFormatManager(DecimalFormatManager dfm) {
        decimalFormatManager = dfm;
    }

    /**
    * Get the DecimalFormatManager which handles decimal-format definitions
    * @return the DecimalFormatManager containing the named xsl:decimal-format definitions
    */

    public DecimalFormatManager getDecimalFormatManager() {
        if (decimalFormatManager==null) {
            decimalFormatManager = new DecimalFormatManager();
        }
        return decimalFormatManager;
    }

    /**
    * Set the default collation
    * @param name the name of the default collation
    */

    public void setDefaultCollationName(String name) {
        defaultCollationName = name;
    }

    /**
    * Get the default collation
    * @return a Comparator that implements the default collation
    */

    public Comparator getDefaultCollation() {
        if (defaultCollationName==null) {
            return CodepointCollator.getInstance();
        } else {
            return getNamedCollation(defaultCollationName);
        }
    }

    /**
    * Set the table of collations
    * @param table a hash table that maps collation names (URIs) to objects representing the
    * collation information
    */

    public void setCollationTable(HashMap table) {
        collationTable = table;
    }

    /**
    * Find a named collation.
    * @param name identifies the name of the collation required; null indicates that the default
    * collation is required
    * @return the requested collation, or null if the collation is not found
    */

    public Comparator getNamedCollation(String name) {
        if (collationTable==null) {
            collationTable = new HashMap();
        }
        return (Comparator)collationTable.get(name);
    }

    /**
    * Set the space requirements for global and local variables
    * @param globals The number of global variables and parameters
    * @param locals The largest number of local variables used by any template or function (etc)
    */

    public void setSlotSpace(int globals, int locals) {
        numberOfVariables = globals;
        largestStackFrame = locals;
    }

    /**
    * Allocate space in bindery for all the variables needed
    * This has to be done early to accommodate preview mode
    * @param bindery The bindery to be initialized
    */

    public void initialiseBindery(Bindery bindery) {
       // ensure enough slots are available for global variables and for the largest stackframe

        bindery.allocateGlobals(numberOfVariables+1);
        bindery.allocateLocals(largestStackFrame + Bindery.FRAME_VARIABLES);
    }

    /**
    * Dummy processLeavingTail() method. This is called to initialize processing, but currently
    * does nothing. The controller takes charge of finding the initial template.
    * @return null (always)
    */

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        return null;
    }



}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
