package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SequenceOutputter;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import java.util.List;


/**
* An instruction that creates an element node whose name is known statically.
 * Used for literal results elements in XSLT, for direct element constructors
 * in XQuery, and for xsl:element in cases where the name and namespace are
 * known statically.
*/

public class FixedElement extends ExprInstruction {

    private int nameCode;
    private AttributeSet[] useAttributeSets;
    private int annotation;
    private int[] namespaceCodes;

    /**
     * Create an instruction that creates a new element node
     * @param nameCode Represents the name of the element node
     * @param namespaceCodes List of namespaces to be added to the element node.
     *                       May be null if none are required.
     * @param useAttributeSets Array of attribute sets to be expanded. May be null
     *                       if none are required.
     * @param annotation Type annotation for the new element node
     */
    public FixedElement( int nameCode,
                         int[] namespaceCodes,
                         AttributeSet[] useAttributeSets,
                         int annotation ) {
        this.nameCode = nameCode;
        this.namespaceCodes = namespaceCodes;
        this.useAttributeSets = useAttributeSets;
        this.annotation = annotation;
    }

    /**
     * Get the result type returned by this instruction
     * @return a SequenceType representing the type element(*,*)
     */

    public SequenceType getResultType() {
        return new SequenceType(Type.ELEMENT, Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    public Expression analyze(StaticContext env) throws XPathException {
        for (int c=0; c<children.length; c++) {
            if (children[c] instanceof ExprInstruction) {
                children[c] = (ExprInstruction)((ExprInstruction)children[c]).analyze(env);
            } else {
                throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
            }
        }
        return this;
    }


    public void getXPathExpressions(List list) {
        // do nothing
    }

    /**
     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
     * By default the offer is not accepted - this is appropriate in the case of simple expressions
     * such as constant values and variable references where promotion would give no performance
     * advantage. This method is always called at compile time.
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception XPathException if any error is detected
     */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        for (int c=0; c<children.length; c++) {
            if (children[c] instanceof ExprInstruction) {
                Expression p = ((ExprInstruction)children[c]).promote(offer);
                if (p instanceof Instruction) {
                    children[c] = (Instruction)p;
                } else {
                    children[c] = new Sequence(p, null);
                }
            } else {
                throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
            }
        }
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {

        Controller controller = context.getController();

        Receiver out = controller.getReceiver();
        out.startElement(nameCode, annotation, 0);

        // output the namespace nodes
        if (namespaceCodes != null) {
            for (int i=0; i<namespaceCodes.length; i++) {
                out.namespace(namespaceCodes[i], 0);
            }
        }

        // apply the content of any attribute sets mentioned in use-attribute-sets
        if (useAttributeSets != null) {
            AttributeSet.expand(useAttributeSets, context);
        }

        // process subordinate instructions to generate attributes and content
        processChildren(context);

        // output the element end tag
        out.endElement();
        return null;
    }


   /**
     * Evaluate as an expression. We rely on the fact that when these instructions
     * are generated by XQuery, there will always be a valueExpression to evaluate
     * the content
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        try {
            Controller controller = context.getController();
            NamePool pool = controller.getNamePool();
            SequenceReceiver old = controller.getReceiver();
            SequenceOutputter seq = new SequenceOutputter();
            //seq.setSystemId(baseURI);
            seq.setNamePool(pool);

            seq.startDocument();
            seq.startElement(nameCode, 0, 0);
            // ignore attribute sets for now
            // output the namespace nodes
            if (namespaceCodes != null) {
                for (int i=0; i<namespaceCodes.length; i++) {
                    seq.namespace(namespaceCodes[i], 0);
                }
            }
            controller.changeToSequenceOutputDestination(seq);
            processChildren(context);

            seq.endElement();
            seq.endDocument();
            controller.resetOutputDestination(old);

            // the constructed element is the first and only item in the sequence
            return seq.getFirstItem();

        } catch (TransformerException err) {
            if (err instanceof XPathException) {
                throw (XPathException)err;
            } else {
                throw new XPathException.Dynamic(err);
            }
        }
    }


    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("element");
        if (namePool != null) {
            details.setProperty("name", namePool.getDisplayName(nameCode));
        }
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "element ");
        System.err.println(ExpressionTool.indent(level+1) + "name " +
                (pool==null ? nameCode+"" : pool.getDisplayName(nameCode)));
        if (children.length==0) {
            System.err.println(ExpressionTool.indent(level+1) + "empty content");
        } else {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    ((Expression)children[c]).display(level+1, pool);
                } else {
                    System.err.println(ExpressionTool.indent(level+1) + children[c].getInstructionName());
                }
            }
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
