package net.sf.saxon.instruct;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.Controller;
import net.sf.saxon.Configuration;
import net.sf.saxon.ParameterSet;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.trace.InstructionInfo;
import net.sf.saxon.trace.TraceListener;

import javax.xml.transform.TransformerException;
import java.io.Serializable;

/**
* Abstract superclass for all instructions in the compiled stylesheet.
* This represents a compiled instruction, and as such, the minumum information is
* retained from the original stylesheet. <br>
* Note: this class implements SourceLocator.
*/

public abstract class Instruction implements Serializable {

    // The instructions that are children of this instruction
    Instruction[] children;

    // The location of this instruction. The top 12 bits represent the systemId of the
    // source stylesheet module, as an index into the systemID map held by the Executable object.
    // The bottom 20 bits represent the line number within that module
    private int sourceLocation;

    /**
    * Constructor
    */

    public Instruction() {}

    /**
    * Set the children of this instruction
    * @param children The instructions that are children of this instruction
    */

    public void setChildren(Instruction[] children) {
        this.children = children;
    }

    /**
    * Get the children of this instruction
    * @return the children of this instruction, as an array of Instruction objects
    */

    public Instruction[] getChildren() {
        return children;
    }

    /**
    * Set the system ID and the line number of this instruction.
    * @param module integer identifying the module in which this instruction occurs,
    * This is an index into the module array owned by the Executable object
    * @param lineNumber the line number of this instruction within its module
    */

    public void setSourceLocation(int module, int lineNumber) {
        sourceLocation = (module<<20) | (lineNumber & 0x7ffff);
    }

    /**
    * Get the system ID of the source stylesheet module containing this instruction.
    * @param exec The Executable object that owns the index of modules
    * @return the URI of the containing module
    */

    public String getSystemId(Executable exec) {
        int module = (sourceLocation>>20) & 0x3ff;
        return exec.getSystemId(module);
    }

    /**
    * Get the line number of the source stylesheet instruction within its module.
    * @return the line number of the instruction
    */

    public int getLineNumber() {
        return sourceLocation & 0x7ffff;
    }

    /**
    * Contruct details of an InstructionInfo for use by the TraceListener.
    * @param controller the Controller for the transformation. This is supplied because
    * (to save space) the Executable is not reachable from an individual Instruction
    * @return an object containing information about the instruction, designed for use
    * by a TraceListener
    */

    public InstructionDetails getInstructionDetails(Controller controller) {
        InstructionDetails details = new InstructionDetails();
        setInstructionDetails(controller, controller.getNamePool(), details);
        details.setSystemId(getSystemId(controller.getExecutable()));
        details.setLineNumber(getLineNumber());
        details.setProperty("controller", controller);
        return details;
    }


    /**
    * Get the name of the instruction for use in diagnostics
    */

    public abstract void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details);

    public String getInstructionName() {
        InstructionDetails details = new InstructionDetails();
        setInstructionDetails(null, null, details);
        return details.getInstructionName();
    }

    /**
     * Get the static type of the value returned by the instruction
     * @return the SequenceType that the returned value will conform to.
     */

    public SequenceType getResultType() {
        // TODO: this method is not properly implemented on all subclasses
        // the default implementation assumes the worst case: we don't know anything
        // about the result type
        return SequenceType.ANY_SEQUENCE;
    }

    /**
    * ProcessLeavingTail: called to do the real work of this instruction. This method
    * must be implemented in each subclass. The results of the instruction are written
    * to the current Receiver, which can be obtained via the Controller.
    * @param context The dynamic context of the transformation, giving access to the current node,
    * the current variables, etc.
    * @return null if the instruction has completed execution; or a TailCall indicating
    * a function call or template call that is delegated to the caller, to be made after the stack has
    * been unwound so as to save stack space.
    */

    public abstract TailCall processLeavingTail(XPathContext context) throws TransformerException;

    // TODO: provide an iterate() method that iterates over the results of the instruction.
    // This is needed to achieve lazy evaluation, especially of function calls.

    /**
    * Process the instruction, without returning any tail calls
    * @param context The dynamic context, giving access to the current node,
    * the current variables, etc.
    */

    public void process(XPathContext context) throws TransformerException {
        TailCall tc = processLeavingTail(context);
        while (tc != null) {
            tc = tc.processLeavingTail(context);
        }
    }

    /**
    * Process the children of this instruction, including any tail calls
    * @param context The dynamic context for the transformation
     * @throws TransformerException if a dynamic error occurs
    */

    protected void processChildren(XPathContext context) throws TransformerException {
        if (children==null) {
            return;
        }
        Controller controller = context.getController();
    	if (controller.isTracing()) {
            TraceListener listener = controller.getTraceListener();
    	    for (int i=0; i<children.length; i++) {
    		    try {

    	            InstructionInfo info = children[i].getInstructionDetails(controller);
        	        listener.enter(info);
     	            children[i].process(context);
      		        listener.leave(info);

    		    } catch (TransformerException err) {
    		        if (err instanceof SkipInstructionException) {
    		            // report the error and continue
    		            children[i].recoverableError(err.getMessage(), controller);
    		        } else {
    		            // terminate execution
    			        throw children[i].styleError(err, controller);
    			    }
    		    }
            }

    	} else {

    	    for (int i=0; i<children.length; i++) {
    		    try {
        	        children[i].process(context);
    		    } catch (TransformerException err) {
    		        if (err instanceof SkipInstructionException) {
    		            // report the error and continue
    		            children[i].recoverableError(err.getMessage(), controller);
    		        } else {
    		            // terminate execution
    			        throw children[i].styleError(err, controller);
    			    }
    		    }
    	    }
    	}
    }

    /**
    * Process the children of this node in the stylesheet, returning any tail call made by
    * the last child instruction
    * @param context The dynamic context of the transformation, giving access to the current node,
    * the current variables, etc.
    * @return null if the instruction has completed execution; or a TailCall indicating
    * a function call or template call that is delegated to the caller, to be made after the stack has
    * been unwound so as to save stack space.
    */

    protected TailCall processChildrenLeavingTail(XPathContext context) throws TransformerException {

        if (children==null) return null;

        Controller controller = context.getController();
        TailCall tc = null;
    	if (controller.isTracing()) {
            TraceListener listener = controller.getTraceListener();
    	    for (int i=0; i<children.length; i++) {
    		    try {

    	            InstructionInfo info = children[i].getInstructionDetails(controller);
        	        listener.enter(info);
        	        //if (i != (children.length - 1)) {
        	            // This is not the last child, don't return any tail calls
        	        //    children[i].process(context);
        	        //} else {
        	            // This is the last child, do return any tail calls
        	            tc = children[i].processLeavingTail(context);
        	        //}
      		        listener.leave(info);

    		    } catch (TransformerException err) {
    		        if (err instanceof SkipInstructionException) {
    		            // report the error and continue
    		            children[i].recoverableError(err.getMessage(), controller);
    		        } else {
    		            // terminate execution
    			        throw children[i].styleError(err, controller);
    			    }
    		    }
            }

    	} else {

    	    for (int i=0; i<children.length; i++) {
    		    try {

        	       // if (i != (children.length - 1)) {
        	            // This is not the last child, don't return any tail calls
        	       //     children[i].process(context);
        	        //} else {
        	            // This is the last child, do return any tail calls
        	            tc = children[i].processLeavingTail(context);
        	        //}

    		    } catch (TransformerException err) {
    		        if (err instanceof SkipInstructionException) {
    		            // report the error and continue
    		            children[i].recoverableError(err.getMessage(), controller);
    		        } else {
    		            // terminate execution
    			        throw children[i].styleError(err, controller);
    			    }
    		    }
    	    }
    	}
    	return tc;
    }

    /**
    * Process the instruction with tracing
    */
/*
    public TailCall processWithTrace(Controller controller) throws TransformerException {
    	if (controller.isTracing()) {
    	    TraceListener listener = controller.getTraceListener();
    	    InstructionInfo info = getInstructionDetails(controller);
        	listener.enter(info);
            TailCall tc = processLeavingTail(controller);
      		listener.leave(info);
      		return tc;
   	    } else {
   	        return processLeavingTail(controller);
   	    }
    }
*/
    /**
    * Construct an exception with diagnostic information. Note that this method
    * returns the exception, it does not throw it: that is up to the caller.
    * @param error The exception containing information about the error
    * @param controller The controller of the transformation
    * @return an exception based on the supplied exception, but with location information
    * added relating to this instruction
    */

    protected TransformerException styleError(TransformerException error, Controller controller) {
        //if (error instanceof StyleException) return error;
        if (error instanceof TerminationException) return error;
        if (error.getException() instanceof TransformerException) {
            // This happens when the instruction executes an XPath expression and the XPath
            // expression invokes XSLT instructions via a variable reference or a function call.
            // We return the underlying exception in order to show the original line number.
            return styleError((TransformerException)error.getException(), controller);
        }
        if (error.getLocator()==null) {
            // If the exception has no location information, construct a new
            // exception containing the required information
            try {
                InstructionDetails info = getInstructionDetails(controller);
                return new TransformerException(error.getMessage(),
                                                info,
                                                error.getException());
            } catch (Exception secondaryError) {
                // currently happens when running XQuery
                return error;
            }
        } else if (error instanceof XPathException) {
            // a locator within an XPathException is useful in XQuery, but less
            // useful in XSLT, so the action here depends on the host language
            if (controller.getConfiguration().getHostLanguage() == Configuration.XQUERY) {
                return error;
            } else {
                try {
                    InstructionDetails info = getInstructionDetails(controller);
                    return new TransformerException(error.getMessage(),
                                                    info,
                                                    error.getException());
                } catch (Exception secondaryError) {
                    // ? can no longer happen ?
                    return error;
                }
            }
        }
        return error;
    }

    /**
    * Raise a dynamic error
    * @param message An English text error message
    * @param controller The controller of the transformation
    * @return an exception containing details of the dynamic error
    */

    protected TransformerException styleError(String message, Controller controller) {
        InstructionDetails info = getInstructionDetails(controller);
        return new TransformerException(message, info);
    }

    /**
    * Signal a recoverable error. The error may be reported if the error handling
    * policy indicates that recoverable errors should be reported
    * @param message an English-language error message
    * @param controller the controller for the transformation
    * @throws TransformerException - the method throws an exception if the selected
    * policy is not to recover from recoverable errors
    */

    protected void recoverableError(String message, Controller controller)
    throws TransformerException {
        InstructionDetails info = getInstructionDetails(controller);
        controller.recoverableError(message, info);
    }

    /**
     * Assemble a ParameterSet. Method used by instructions that have xsl:with-param
     * children. This method is used for the non-tunnel parameters.
     */

    protected ParameterSet assembleParams(XPathContext context,
                                          WithParam[] actualParams)
    throws TransformerException {
        if (actualParams == null || actualParams.length == 0) {
            return null;
        }
        ParameterSet params = new ParameterSet();
        for (int i=0; i<actualParams.length; i++) {
            params.put(actualParams[i].getVariableFingerprint(),
                       actualParams[i].getSelectValue(context));
        }
        return params;
    }

    /**
     * Assemble a ParameterSet. Method used by instructions that have xsl:with-param
     * children. This method is used for the tunnel parameters.
     */

    protected ParameterSet assembleTunnelParams(XPathContext context,
                                          WithParam[] actualParams)
    throws TransformerException {
        ParameterSet existingParams = context.getController().getBindery().getTunnelParameters();
        if (existingParams == null) {
            return assembleParams(context, actualParams);
        }
        ParameterSet newParams = new ParameterSet(existingParams);
        if (actualParams == null || actualParams.length == 0) {
            return newParams;
        }
        for (int i=0; i<actualParams.length; i++) {
            newParams.put(actualParams[i].getVariableFingerprint(),
                          actualParams[i].getSelectValue(context));
        }
        return newParams;
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
