package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.event.Emitter;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.value.Cardinality;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerException;
import java.io.OutputStreamWriter;
import java.util.Properties;

/**
* An xsl:message element in the stylesheet.
*/

public class Message extends Instruction {

    private Expression terminate;

    public Message( Expression terminate ) {
        this.terminate = terminate;
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    * @return the string "xsl:message"
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("message");
    }

    /**
     * Get the result type of the instruction. The result is always an empty sequence.
     * @return A SequenceType denoting the type EMPTY
     */

    public SequenceType getResultType() {
        return new SequenceType(Type.EMPTY, Type.ITEM, StaticProperty.EMPTY);
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        Emitter emitter = controller.getMessageEmitter();
        if (emitter==null) {
            emitter = controller.makeMessageEmitter();
        }
        if (emitter.getWriter()==null) {
            emitter.setWriter(new OutputStreamWriter(System.err));
        }

        SequenceReceiver old = controller.getReceiver();
        Properties props = new Properties();
        props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
        controller.changeOutputDestination(props, emitter, false);

        processChildren(context);

        controller.resetOutputDestination(old);

        if (terminate != null) {
            String term = terminate.evaluateAsString(context);
            if (term.equals("no")) {
                // no action
            } else if (term.equals("yes")) {
                throw new TerminationException("Processing terminated by xsl:message at line " + getLineNumber());
            } else {
                throw new TransformerException("The terminate attribute of xsl:message must be 'yes' or 'no'");
            }
        }
        return null;
    }

}
//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
