package net.sf.saxon.instruct;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.xpath.XPathException;

import java.io.Serializable;

/**
  * An object representing a list of Namespaces. Used when the namespace
  * controller in the stylesheet needs to be kept for use at run-time. The list of namespaces
  * is maintained in the form of numeric prefix/uri codes, which are only meaningful
  * in the controller of a name pool; however, in order to save space, the NamespaceContext
  * object does not keep a reference to the namepool, it requires this to be supplied
  * by the caller.
  */

public final class NamespaceContext implements Serializable {

    private int[] namespaceCodes;

    /**
    * Create a NamespaceContext object
    * @param nscodes an array of namespace codes. Each namespace code is an integer
    * in which the first 16 bits represent the prefix (zero if it's the default namespace)
    * and the next 16 bits represent the uri. These are codes held in the NamePool
    */

    public NamespaceContext(int[] nscodes) {
        namespaceCodes = nscodes;
    }

    /**
    * Get the list of in-scope namespaces held in this NamespaceContext
    * @return the list of namespaces
    */

    public int[] getNamespaceCodes() {
        return namespaceCodes;
    }

    /**
    * Get the namespace URI corresponding to a given prefix. Return null
    * if the prefix is not in scope.
    * @param prefix the namespace prefix
    * @param useDefault true if the default namespace is to be used when the
    * prefix is ""
    * @param pool the namepool to be used for looking up the codes. This must be the
    * run-time namepool.
    * @return the uri for the namespace, or null if the prefix is not in scope
    */

    public String getURIForPrefix(String prefix, boolean useDefault, NamePool pool) {

        if (prefix.equals("") && !useDefault) {
            return "";
        }

        //NamePool pool = controller.getNamePool();
        for (int i=namespaceCodes.length-1; i>=0; i--) {
            try {
                if (pool.getPrefixFromNamespaceCode(namespaceCodes[i]).equals(prefix)) {
                    return pool.getURIFromNamespaceCode(namespaceCodes[i]);
                }
            } catch (NullPointerException err) {
                System.err.println("Crash! i=" + i + ", prefix=" + prefix + ";");
                System.err.println("namespacecodes[i] = " + namespaceCodes[i]);
                System.err.println("pool = " + pool);
                pool.diagnosticDump();
            }

        }

        if (prefix.equals("") && useDefault) {
            // use the "default default namespace" - namely ""
            return "";
        } else {
            return null;
        }
    }

    /**
    * Use this NamespaceContext to resolve a lexical QName
    * @param qname the lexical QName; this must have already been lexically validated
    * @param useDefault true if the default namespace is to be used to resolve an unprefixed QName
    * @param pool the NamePool to be used
    * @return the integer fingerprint that uniquely identifies this name
    */

    public int getFingerprint(String qname, boolean useDefault, NamePool pool)
    throws XPathException {
        String[] parts;
        try {
            parts = Name.getQNameParts(qname);
        } catch (QNameException err) {
            throw new XPathException.Dynamic(err.getMessage());
        }
        String uri = getURIForPrefix(parts[0], useDefault, pool);
        if (uri == null) {
            throw new XPathException.Dynamic(
                    "Namespace prefix " + parts[0] + " has not been declared");
        }
        return pool.getFingerprint(uri, parts[1]);
    }




}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
