package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.OutputURIResolver;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.SaxonOutputKeys;
import net.sf.saxon.event.StandardOutputResolver;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;

import javax.xml.transform.TransformerException;
import java.util.Properties;

/**
* An xsl:result-document (formerly saxon:output) element in the stylesheet. <BR>
* The xsl:result-document element takes an attribute href="filename". The filename will
* often contain parameters, e.g. {position()} to ensure that a different file is produced
* for each element instance. <BR>
* There is a further attribute "name" which determines the format of the
* output file, it identifies the name of an xsl:output element containing the output
* format details.
*/

public class ResultDocument extends Instruction {

    private Expression href;
    private Properties outputProperties;
    private String baseURI;     // needed only for saxon:next-in-chain

    public ResultDocument(Properties outputProperties, Expression href, String baseURI) {
        this.outputProperties = outputProperties;
        this.href = href;
        this.baseURI = baseURI;
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    * @return the string "xsl:result-document"
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("result-document");
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Controller controller = context.getController();
        SequenceReceiver oldReceiver = controller.getReceiver();

        OutputURIResolver resolver = controller.getOutputURIResolver();

        String hrefValue = href.evaluateAsString(context);
        javax.xml.transform.Result result =
            resolver.resolve(hrefValue, controller.getPrincipalResultURI());
        if (result==null) {
            resolver = StandardOutputResolver.getInstance();
            result = resolver.resolve(hrefValue, controller.getPrincipalResultURI());
        }

        boolean timing = controller.getConfiguration().isTiming();
        if (timing) {
            System.err.println("Writing to " + result.getSystemId());
        }

        String nextInChain = outputProperties.getProperty(SaxonOutputKeys.NEXT_IN_CHAIN);
        if (nextInChain != null) {
            result = controller.prepareNextStylesheet(nextInChain, baseURI, result);
        }

        controller.changeOutputDestination(outputProperties, result, true);
        processChildren(context);
        controller.resetOutputDestination(oldReceiver);
        resolver.close(result);
        return null;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Additional Contributor(s): Brett Knights [brett@knightsofthenet.com]
//
