package net.sf.saxon.instruct;
import net.sf.saxon.Controller;
import net.sf.saxon.event.SequenceChecker;
import net.sf.saxon.event.SequenceReceiver;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.RoleLocator;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import java.util.List;


/**
* An xsl:sequence element in the stylesheet.
*/

public class Sequence extends ExprInstruction {

    private Expression select;
    private SequenceType requiredType;

    public Sequence(Expression select, SequenceType requiredType) {
        this.select = select;
        this.requiredType = requiredType;
    }

    public Expression getSelectExpression() {
        return select;
    }

    public Expression simplify() throws XPathException {
        if (select!=null) {
            select = select.simplify();
        }
        return super.simplify();
    }

    public Expression analyze(StaticContext env) throws XPathException {
        if (select != null) {
            select = select.analyze(env);
            if (requiredType != null) {
                RoleLocator role =
                    new RoleLocator(RoleLocator.INSTRUCTION, "sequence/select", 0);
                select = TypeChecker.staticTypeCheck(select, requiredType, false, role);
            }
        }

        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    children[c] = (ExprInstruction)((ExprInstruction)children[c]).analyze(env);
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
        return this;

    }

    public void getXPathExpressions(List list) {
        if (select != null) {
            list.add(select);
            return;
        }
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    ((ExprInstruction)children[c]).getXPathExpressions(list);
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
    }

   /**
     * Offer promotion for subexpressions. The offer will be accepted if the subexpression
     * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
     * By default the offer is not accepted - this is appropriate in the case of simple expressions
     * such as constant values and variable references where promotion would give no performance
     * advantage. This method is always called at compile time.
     *
     * @param offer details of the offer, for example the offer to move
     *     expressions that don't depend on the context to an outer level in
     *     the containing expression
     * @exception XPathException if any error is detected
     */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        if (select != null) {
            select = select.promote(offer);
        }
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof ExprInstruction) {
                    Expression p = ((ExprInstruction)children[c]).promote(offer);
                    if (p instanceof Instruction) {
                        children[c] = (Instruction)p;
                    } else {
                        children[c] = new Sequence(p, null);
                    }
                } else {
                    throw new IllegalStateException("Children of an ExprInstruction must themselves be ExprInstructions");
                }
            }
        }
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("sequence");
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {

        Controller controller = context.getController();
    	SequenceReceiver out = controller.getReceiver();

        if (select != null) {
            SequenceIterator iter = select.iterate(context);
            while (iter.hasNext()) {
                out.append(iter.next());
            }
            return null;

        } else if (requiredType==null) {

    	    // No type-check necessary, so write the values directly to the current output destination

            TailCall tc = null;
            if (children != null) {
                tc = processChildrenLeavingTail(context);
            }
            return tc;

        } else {

            // Type check requested, so we need to create a type-checking filter

            SequenceChecker checker = new SequenceChecker();
            checker.setRequiredType(requiredType);
            checker.setUnderlyingReceiver(out);
            controller.setReceiver(checker);

            if (select != null) {
                SequenceIterator iter = select.iterate(context);
                while (iter.hasNext()) {
                    checker.append(iter.next());
                }
            }

            if (children != null) {
                processChildren(context);
            }

            checker.finalCheck();

            controller.resetOutputDestination(out);

            return null;
        }

    }

    /**
     * Iterate the results, treating this instruction as an expression
     */

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        if (children != null) {
            throw new IllegalStateException("Cannot process xsl:sequence with children as an expression");
        }
        return select.iterate(context);
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "sequence");
        if (select != null) {
            select.display(level+1, pool);
        }
        if (children != null) {
            for (int c=0; c<children.length; c++) {
                if (children[c] instanceof Expression) {
                    ((Expression)children[c]).display(level+1, pool);
                } else {
                    System.err.println(ExpressionTool.indent(level+1) + children[c].getInstructionName());
                }
            }
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
