package net.sf.saxon.instruct;

import net.sf.saxon.Controller;
import net.sf.saxon.trace.InstructionInfo;
import net.sf.saxon.trace.TraceListener;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.value.Value;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import java.util.HashMap;
import java.io.Serializable;

/**
 * This object represents the compiled form of a user-written function
 * (the source can be either an XSLT stylesheet function or an XQuery function).
 *
 * <p>It is assumed that type-checking, of both the arguments and the results,
 * has been handled at compile time. That is, the expression supplied as the body
 * of the function must be wrapped in code to check or convert the result to the
 * required type, and calls on the function must be wrapped at time to check or
 * convert the supplied arguments.
*/

public final class UserFunction implements CallableFunction, Serializable, InstructionInfo {

    private Expression body;
    private String functionName;
    private boolean memoFunction = false;
    private int lineNumber;


    public UserFunction (Expression body) {
        this.body = body;
    };

    public void setLineNumber(int lineNumber) {
        this.lineNumber = lineNumber;
    }

    public int getLineNumber() {
        return lineNumber;
    }

    public String getSystemId() {
        return null;
    }

    public String getInstructionName() {
        return "function";
    }

    public Object getProperty(String propertyName) {
        if (propertyName.equals("name")) {
            return functionName;
        } else {
            return null;
        }
    }

    public HashMap getProperties() {
        HashMap map = new HashMap();
        map.put("name", functionName);
        return map;
    }

    public void setMemoFunction(boolean isMemo) {
        memoFunction = isMemo;
    }

    public void setFunctionName(String name) {
        functionName = name;
    }

    public String getFunctionName() {
        return functionName;
    }

    /**
    * Call this function
    */

    public Value call(Value[] actualArgs, Controller controller) throws TransformerException {

        // If this is a memo function, see if the result is already known
        if (memoFunction) {
            Value value = getCachedValue(controller, actualArgs);
            if (value!=null) return value;
        }

        // Otherwise evaluate the function

    	if (controller.isTracing()) {
    	    TraceListener listener = controller.getTraceListener();
        	listener.enter(this);
   	    }

        Bindery bindery = controller.getBindery();
        bindery.openStackFrame(actualArgs);
  	    Value resultValue;
        try {
            resultValue = ExpressionTool.eagerEvaluate(body, controller.newXPathContext());
        } catch (XPathException err) {
            // TODO: this isn't very nice error handling!
            System.err.println("Dynamic error in function " + functionName);
            throw err;
        }
        bindery.closeStackFrame();

    	if (controller.isTracing()) {
    	    TraceListener listener = controller.getTraceListener();
      		listener.leave(this);
   	    }

        // If this is a memo function, save the result in the cache
        if (memoFunction) {
            putCachedValue(controller, actualArgs, resultValue);
        }

        return resultValue;
    }

    /**
    * For memo functions, get a saved value from the cache.
    * @return the cached value, or null if no value has been saved for these parameters
    */

    private Value getCachedValue(Controller controller, Value[] params) {

        try {
            HashMap map = (HashMap)controller.getUserData(this, "memo-function-cache");
            if (map==null) {
                return null;
            }
            String key = getCombinedKey(params);
            return (Value)map.get(key);
        } catch (XPathException err) {
            return null;
        }
    }

    /**
    * For memo functions, put the computed value in the cache.
    */

    private void putCachedValue(Controller controller, Value[] params, Value value) {
        try {
            HashMap map = (HashMap)controller.getUserData(this, "memo-function-cache");
            if (map==null) {
                map = new HashMap();
                controller.setUserData(this, "memo-function-cache", map);
            }
            String key = getCombinedKey(params);
            map.put(key, value);
        } catch (XPathException err) {
            // it doesn't matter if we fail to cache the result
        }
    }

    /**
    * Get a key value representing the values of all the supplied arguments
    */

    private static String getCombinedKey(Value[] params) throws XPathException {
        StringBuffer sb = new StringBuffer();

        for (int i=0; i<params.length; i++) {
            Value val = params[i];
            SequenceIterator iter = val.iterate(null);
            while (iter.hasNext()) {
                Item item = iter.next();
                if (item instanceof NodeInfo) {
                    NodeInfo node = (NodeInfo)item;
                    //sb.append(""+node.getDocumentRoot().getDocumentNumber());
                    //sb.append('/');
                    sb.append(node.generateId());
                } else {
                    sb.append(""+item.getItemType());
                    sb.append('/');
                    sb.append(item.getStringValue());
                }
                sb.append('\u0001');
            }
            sb.append('\u0002');
        }
        return sb.toString();
    }



}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//

