package net.sf.saxon.instruct;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Orphan;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.event.ReceiverOptions;
import net.sf.saxon.expr.*;
import net.sf.saxon.Controller;

import javax.xml.transform.TransformerException;
import java.util.List;

/**
* An xsl:value-of element in the stylesheet. <br>
* The xsl:value-of element takes attributes:<ul>
* <li>a mandatory attribute select="expression".
* This must be a valid String expression</li>
* <li>an optional disable-output-escaping attribute, value "yes" or "no"</li>
* <li>an optional separator attribute. This is handled at compile-time: if the separator attribute
* is present, the select expression passed in here will be a call to the string-join() function.</li>
* </ul>
*/

public final class ValueOf extends ExprInstruction {

    // TODO: merge this class with Text, and make it a SimpleNodeConstructor

    private Expression select;
    private int options;

    public ValueOf( Expression select,
                    boolean disable ) {
        this.select = select;
        this.options = (disable ? ReceiverOptions.DISABLE_ESCAPING : 0);
    }

    /**
    * Get the name of this instruction for diagnostic and tracing purposes
    */

    public void setInstructionDetails(Controller controller, NamePool namePool, InstructionDetails details) {
        details.setInstructionName("value-of");
    }

    public Expression analyze(StaticContext env) throws XPathException {
        select = select.analyze(env);
        RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "xsl:value-of/select", 0);
        select = TypeChecker.staticTypeCheck(select, SequenceType.SINGLE_STRING, false, role);
        return this;
    }

    public int getDependencies() {
        return select.getDependencies();
    }

    public void getXPathExpressions(List list) {
        list.add(select);
    }

    /**
    * Offer promotion for subexpressions. The offer will be accepted if the subexpression
    * is not dependent on the factors (e.g. the context item) identified in the PromotionOffer.
    * By default the offer is not accepted - this is appropriate in the case of simple expressions
    * such as constant values and variable references where promotion would give no performance
    * advantage. This method is always called at compile time.
    *
    * @param offer details of the offer, for example the offer to move
    *     expressions that don't depend on the context to an outer level in
    *     the containing expression
    * @exception XPathException if any error is detected
    */

    public void promoteInst(PromotionOffer offer) throws XPathException {
        select = select.promote(offer);
    }


    public SequenceType getResultType() {
         return new SequenceType(Type.TEXT, Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
    }

    public TailCall processLeavingTail(XPathContext context) throws TransformerException {
        Receiver out = context.getController().getReceiver();
        out.characters(select.evaluateAsString(context), options);
        return null;
    }

    public Item evaluateItem(XPathContext context) throws XPathException {
        String val = select.evaluateAsString(context);
        Orphan o = new Orphan(context.getController().getNamePool());
        o.setNodeKind(Type.TEXT);
        o.setStringValue(val);
        return o;
    }

    /**
     * Display this instruction as an expression, for diagnostics
     */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "value-of");
        select.display(level+1, pool);
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
