package net.sf.saxon.jdom;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.om.*;
import net.sf.saxon.pattern.AnyNodeTest;
import net.sf.saxon.pattern.NodeTest;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.UntypedAtomicValue;
import net.sf.saxon.value.Value;
import org.jdom.*;
import org.jdom.Namespace;

import javax.xml.transform.TransformerException;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.HashMap;

/**
  * A node in the XML parse tree representing an XML element, character content, or attribute.<P>
  * This is the implementation of the NodeInfo interface used as a wrapper for JDOM nodes.
  * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
  */

public class NodeWrapper implements NodeInfo, VirtualNode, SiblingCountingNode {

    protected Object node;
    protected short nodeKind;
    private NodeWrapper parent;     // null means unknown
    protected DocumentWrapper docWrapper;
    protected int index;            // -1 means unknown

    /**
     * This constructor is protected: nodes should be created using the wrap
     * factory method on the DocumentWrapper class
     * @param node    The JDOM node to be wrapped
     * @param parent  The NodeWrapper that wraps the parent of this node
     * @param index   Position of this node among its siblings
     */
    protected NodeWrapper(Object node, NodeWrapper parent, int index) {
        this.node = node;
        this.parent = parent;
        this.index = index;
    }

    /**
     * Factory method to wrap a JDOM node with a wrapper that implements the Saxon
     * NodeInfo interface.
     * @param node        The JDOM node
     * @param docWrapper  The wrapper for the Document containing this node
     * @return            The new wrapper for the supplied node
     */
    protected NodeWrapper makeWrapper(Object node, DocumentWrapper docWrapper) {
        return makeWrapper(node, docWrapper, null, -1);
    }

    /**
     * Factory method to wrap a JDOM node with a wrapper that implements the Saxon
     * NodeInfo interface.
     * @param node        The JDOM node
     * @param docWrapper  The wrapper for the Document containing this node
     * @param parent      The wrapper for the parent of the JDOM node
     * @param index       The position of this node relative to its siblings
     * @return            The new wrapper for the supplied node
     */

    protected NodeWrapper makeWrapper(Object node, DocumentWrapper docWrapper,
                                      NodeWrapper parent, int index) {
        NodeWrapper wrapper;
        if (node instanceof Document) {
            return docWrapper;
        } else if (node instanceof Element) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.ELEMENT;
        } else if (node instanceof Attribute) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.ATTRIBUTE;
        } else if (node instanceof String || node instanceof Text) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.TEXT;
        } else if (node instanceof CDATA) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.TEXT;
        } else if (node instanceof Comment) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.COMMENT;
        } else if (node instanceof ProcessingInstruction) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.PROCESSING_INSTRUCTION;
        } else if (node instanceof Namespace) {
            wrapper = new NodeWrapper(node, parent, index);
            wrapper.nodeKind = Type.NAMESPACE;
        } else {
            throw new IllegalArgumentException("Bad node type in JDOM! " + node.getClass() + " instance " + node.toString());
        }
        wrapper.docWrapper = docWrapper;
        return wrapper;
    }

    /**
    * Get the underlying DOM node, to implement the VirtualNode interface
    */

    public Object getUnderlyingNode() {
        return node;
    }


    /**
     * Get the name pool for this node
     * @return the NamePool
     */

    public NamePool getNamePool() {
        return docWrapper.getNamePool();
    }

    /**
    * Return the type of node.
    * @return one of the values Node.ELEMENT, Node.TEXT, Node.ATTRIBUTE, etc.
    */

    public int getItemType() {
        return nodeKind;
    }

    /**
    * Get the typed value of the item
    */

    public Value getTypedValue() {
        return new UntypedAtomicValue(getStringValue());
    }

    /**
    * Get the type annotation
    * @return 0 (there is no type annotation)
    */

    public int getTypeAnnotation() {
        return 0;
    }

    /**
    * Determine whether this is the same node as another node. <br />
    * Note: a.isSameNode(b) if and only if generateId(a)==generateId(b)
    * @return true if this Node object and the supplied Node object represent the
    * same node in the tree.
    */

    public boolean isSameNode(NodeInfo other) {
        if (!(other instanceof NodeWrapper)) {
            return false;
        }
        NodeWrapper ow = (NodeWrapper)other;
        return node.equals(ow.node);
    }

    /**
    * Get the System ID for the node.
    * @return the System Identifier of the entity in the source document containing the node,
    * or null if not known. Note this is not the same as the base URI: the base URI can be
    * modified by xml:base, but the system ID cannot.
    */

    public String getSystemId() {
        return docWrapper.baseURI;
    }

    public void setSystemId(String uri) {
        docWrapper.baseURI = uri;
    }

    /**
    * Get the Base URI for the node, that is, the URI used for resolving a relative URI contained
     * in the node. In the JDOM model, base URIs are held only an the document level. We don't
     * currently take any account of xml:base attributes.
    */

    public String getBaseURI() {
        if (getItemType() == Type.NAMESPACE) {
            return null;
        }
        NodeInfo n = this;
        if (getItemType() != Type.ELEMENT) {
            n = getParent();
        }
        // Look for an xml:base attribute
        while (n != null) {
            String xmlbase = n.getAttributeValue(NamespaceConstant.XML, "base");
            if (xmlbase != null) {
                return xmlbase;
            }
        }
        // if not found, return the base URI of the document node
        return docWrapper.baseURI;
    }

    /**
    * Get line number
    * @return the line number of the node in its original source document; or -1 if not available
    */

    public int getLineNumber() {
        return -1;
    }

    /**
    * Determine the relative position of this node and another node, in document order.
    * The other node will always be in the same document.
    * @param other The other node, whose position is to be compared with this node
    * @return -1 if this node precedes the other node, +1 if it follows the other
    * node, or 0 if they are the same node. (In this case, isSameNode() will always
    * return true, and the two nodes will produce the same result for generateId())
    */

    public int compareOrder(NodeInfo other) {
        return Navigator.compareOrder(this, (SiblingCountingNode)other);
    }

    /**
    * Return the string value of the node. The interpretation of this depends on the type
    * of node. For an element it is the accumulated character content of the element,
    * including descendant elements.
    * @return the string value of the node
    */

    public String getStringValue() {
        switch (nodeKind) {
            case Type.DOCUMENT:
                List children1 = ((Document)node).getContent();
                StringBuffer sb1 = new StringBuffer();
                expandStringValue(children1, sb1);
                return sb1.toString();

            case Type.ELEMENT:
                List children2 = ((Element)node).getContent();
                StringBuffer sb2 = new StringBuffer();
                expandStringValue(children2, sb2);
                return sb2.toString();

            case Type.ATTRIBUTE:
                return ((Attribute)node).getValue();

            case Type.TEXT:
                if (node instanceof String) return (String)node;
                if (node instanceof Text) return ((Text)node).getText();
                if (node instanceof CDATA) return ((CDATA)node).getText();
                return "";

            case Type.COMMENT:
                return ((Comment)node).getText();

            case Type.PROCESSING_INSTRUCTION:
                return ((ProcessingInstruction)node).getData();

            case Type.NAMESPACE:
                return ((Namespace)node).getURI();

            default:
                return "";
        }
    }

    private void expandStringValue(List list, StringBuffer sb) {
        Iterator iter = list.iterator();
        while (iter.hasNext()) {
            Object obj = iter.next();
            if (obj instanceof Element) {
                expandStringValue(((Element)obj).getContent(), sb);
            } else if (obj instanceof Text) { // beta 0.8
                sb.append(((Text)obj).getText());
            } else if (obj instanceof CDATA) {
                sb.append(((CDATA)obj).getText());
            } else if (obj instanceof EntityRef) {
                throw new IllegalStateException("Unexpanded entity in JDOM tree");
            } else {
                throw new AssertionError("Unknown JDOM node type");
            }
        }
    }

	/**
	* Get name code. The name code is a coded form of the node name: two nodes
	* with the same name code have the same namespace URI, the same local name,
	* and the same prefix. By masking the name code with &0xfffff, you get a
	* fingerprint: two nodes with the same fingerprint have the same local name
	* and namespace URI.
    * @see net.sf.saxon.om.NamePool#allocate allocate
	*/

	public int getNameCode() {
	    switch (nodeKind) {
	        case Type.ELEMENT:
	        case Type.ATTRIBUTE:
	        case Type.PROCESSING_INSTRUCTION:
            case Type.NAMESPACE:
	            return docWrapper.namePool.allocate(getPrefix(),
	                                                getURI(),
	                                                getLocalPart());
	        default:
	            return -1;
	    }
	}

	/**
	* Get fingerprint. The fingerprint is a coded form of the expanded name
	* of the node: two nodes
	* with the same name code have the same namespace URI and the same local name.
	* A fingerprint of -1 should be returned for a node with no name.
	*/

	public int getFingerprint() {
	    return getNameCode()&0xfffff;
	}

    /**
    * Get the local part of the name of this node. This is the name after the ":" if any.
    * @return the local part of the name. For an unnamed node, returns "".
    */

    public String getLocalPart() {
        switch (nodeKind) {
            case Type.ELEMENT:
                return ((Element)node).getName();
            case Type.ATTRIBUTE:
                return ((Attribute)node).getName();
            case Type.TEXT:
            case Type.COMMENT:
            case Type.DOCUMENT:
                return "";
            case Type.PROCESSING_INSTRUCTION:
                return ((ProcessingInstruction)node).getTarget();
            case Type.NAMESPACE:
                return ((Namespace)node).getPrefix();
            default:
                return null;
        }
    }

    /**
    * Get the prefix part of the name of this node. This is the name before the ":" if any.
     * (Note, this method isn't required as part of the NodeInfo interface.)
    * @return the prefix part of the name. For an unnamed node, return an empty string.
    */

    public String getPrefix() {
        switch (nodeKind) {
            case Type.ELEMENT:
                return ((Element)node).getNamespacePrefix();
            case Type.ATTRIBUTE:
                return ((Attribute)node).getNamespacePrefix();
            default:
                return "";
        }
    }

    /**
    * Get the URI part of the name of this node. This is the URI corresponding to the
    * prefix, or the URI of the default namespace if appropriate.
    * @return The URI of the namespace of this node. For an unnamed node, return null.
    * For a node with an empty prefix, return an empty string.
    */

    public String getURI() {
        switch (nodeKind) {
            case Type.ELEMENT:
                return ((Element)node).getNamespaceURI();
            case Type.ATTRIBUTE:
                return ((Attribute)node).getNamespaceURI();
            default:
                return "";
        }
    }

    /**
    * Get the display name of this node. For elements and attributes this is [prefix:]localname.
    * For unnamed nodes, it is an empty string.
    * @return The display name of this node.
    * For a node with no name, return an empty string.
    */

    public String getDisplayName() {
        switch (nodeKind) {
            case Type.ELEMENT:
                return ((Element)node).getQualifiedName();
            case Type.ATTRIBUTE:
                return ((Attribute)node).getQualifiedName();
            case Type.PROCESSING_INSTRUCTION:
            case Type.NAMESPACE:
                return getLocalPart();
            default:
                return "";

        }
    }

    /**
    * Get the NodeInfo object representing the parent of this node
    */

    public NodeInfo getParent() {
        if (parent==null) {
            if (node instanceof Element) {
                if (((Element)node).isRootElement()) {
                    parent = makeWrapper(((Element)node).getDocument(), docWrapper);
                } else {
                    parent = makeWrapper(((Element)node).getParent(), docWrapper);
                }
            } else if (node instanceof Text) {
                parent = makeWrapper(((Text)node).getParent(), docWrapper);
            } else if (node instanceof CDATA) {
                parent = makeWrapper(((CDATA)node).getParent(), docWrapper);
            } else if (node instanceof Comment) {
                parent = makeWrapper(((Comment)node).getParent(), docWrapper);
            } else if (node instanceof ProcessingInstruction) {
                parent = makeWrapper(((ProcessingInstruction)node).getParent(), docWrapper);
            } else if (node instanceof Attribute) {
                parent = makeWrapper(((Attribute)node).getParent(), docWrapper);
            } else if (node instanceof Document) {
                parent = null;
            } else if (node instanceof Namespace) {
                throw new UnsupportedOperationException("Cannot find parent of JDOM namespace node");
            } else {
                throw new IllegalStateException("Unknown JDOM node type " + node.getClass());
            }
        }
        return parent;
    }

    /**
     * Get the index position of this node among its siblings (starting from 0)
     */

    public int getSiblingPosition() {
        if (index == -1) {
            int ix = 0;
            getParent();
            AxisIterator iter;
            switch (nodeKind) {
                case Type.ELEMENT:
                case Type.TEXT:
                case Type.COMMENT:
                case Type.PROCESSING_INSTRUCTION:
                    iter = parent.iterateAxis(Axis.CHILD);
                    break;
                case Type.ATTRIBUTE:
                    iter = parent.iterateAxis(Axis.ATTRIBUTE);
                    break;
                case Type.NAMESPACE:
                    iter = parent.iterateAxis(Axis.NAMESPACE);
                    break;
                default:
                    index = 0;
                    return index;
            }
            while (iter.hasNext()) {
                NodeInfo n = (NodeInfo)iter.next();
                if (n.isSameNode(this)) {
                    index = ix;
                    return index;
                }
                ix++;
            }
            throw new IllegalStateException("JDOM node not linked to parent node");
        }
        return index;
    }

    /**
    * Return an iteration over the nodes reached by the given axis from this node
    * @param axisNumber the axis to be used
    * @return a SequenceIterator that scans the nodes reached by the axis in turn.
    */

    public AxisIterator iterateAxis(byte axisNumber) {
        return iterateAxis(axisNumber, AnyNodeTest.getInstance());
    }

    /**
    * Return an iteration over the nodes reached by the given axis from this node
    * @param axisNumber the axis to be used
    * @param nodeTest A pattern to be matched by the returned nodes
    * @return a SequenceIterator that scans the nodes reached by the axis in turn.
    */

    public AxisIterator iterateAxis(byte axisNumber, NodeTest nodeTest) {
        switch (axisNumber) {
            case Axis.ANCESTOR:
                if (nodeKind==Type.DOCUMENT) return EmptyIterator.getInstance();
                return new Navigator.AxisFilter(
                            new Navigator.AncestorEnumeration(this, false),
                            nodeTest);

            case Axis.ANCESTOR_OR_SELF:
                if (nodeKind==Type.DOCUMENT) return EmptyIterator.getInstance();
                return new Navigator.AxisFilter(
                            new Navigator.AncestorEnumeration(this, true),
                            nodeTest);

            case Axis.ATTRIBUTE:
                if (nodeKind!=Type.ELEMENT) return EmptyIterator.getInstance();
                return new Navigator.AxisFilter(
                            new AttributeEnumeration(this),
                            nodeTest);

            case Axis.CHILD:
                if (hasChildNodes()) {
                    return new Navigator.AxisFilter(
                            new ChildEnumeration(this, true, true),
                            nodeTest);
                } else {
                    return EmptyIterator.getInstance();
                }

            case Axis.DESCENDANT:
                if (hasChildNodes()) {
                    return new Navigator.AxisFilter(
                            new Navigator.DescendantEnumeration(this, false, true),
                            nodeTest);
                } else {
                    return EmptyIterator.getInstance();
                }

            case Axis.DESCENDANT_OR_SELF:
                 return new Navigator.AxisFilter(
                            new Navigator.DescendantEnumeration(this, true, true),
                            nodeTest);

            case Axis.FOLLOWING:
                 return new Navigator.AxisFilter(
                            new Navigator.FollowingEnumeration(this),
                            nodeTest);

            case Axis.FOLLOWING_SIBLING:
                 switch (nodeKind) {
                    case Type.DOCUMENT:
                    case Type.ATTRIBUTE:
                    case Type.NAMESPACE:
                        return EmptyIterator.getInstance();
                    default:
                        return new Navigator.AxisFilter(
                            new ChildEnumeration(this, false, true),
                            nodeTest);
                 }

            case Axis.NAMESPACE:
                 if (nodeKind!=Type.ELEMENT) {
                     return EmptyIterator.getInstance();
                 }
                 return new Navigator.AxisFilter(
                                new NamespaceEnumeration(this),
                                nodeTest);

            case Axis.PARENT:
                 getParent();
                 if (parent==null) return EmptyIterator.getInstance();
                 if (nodeTest.matches(parent.getItemType(), parent.getFingerprint(), Type.ITEM)) {
                    return new SingletonIterator(parent);
                 }
                 return EmptyIterator.getInstance();

            case Axis.PRECEDING:
                 return new Navigator.AxisFilter(
                            new Navigator.PrecedingEnumeration(this, false),
                            nodeTest);

            case Axis.PRECEDING_SIBLING:
                 switch (nodeKind) {
                    case Type.DOCUMENT:
                    case Type.ATTRIBUTE:
                    case Type.NAMESPACE:
                        return EmptyIterator.getInstance();
                    default:
                        return new Navigator.AxisFilter(
                            new ChildEnumeration(this, false, false),
                            nodeTest);
                 }

            case Axis.SELF:
                 if (nodeTest.matches(getItemType(), getFingerprint(), Type.ITEM)) return new SingletonIterator(this);
                 return EmptyIterator.getInstance();

            case Axis.PRECEDING_OR_ANCESTOR:
                 return new Navigator.AxisFilter(
                            new Navigator.PrecedingEnumeration(this, true),
                            nodeTest);

            default:
                 throw new IllegalArgumentException("Unknown axis number " + axisNumber);
        }
    }

    /**
     * Find the value of a given attribute of this node. <BR>
     * This method is defined on all nodes to meet XSL requirements, but for nodes
     * other than elements it will always return null.
     * @param uri the namespace uri of an attribute ("" if no namespace)
     * @param localName the local name of the attribute
     * @return the value of the attribute, if it exists, otherwise null
     */

    public String getAttributeValue(String uri, String localName) {
        if (nodeKind==Type.ELEMENT) {
            Namespace ns = Namespace.getNamespace(uri);
            return ((Element)node).getAttributeValue(localName, ns);
        } else {
            return "";
        }
    }

    /**
    * Get the value of a given attribute of this node
    * @param fingerprint The fingerprint of the attribute name
    * @return the attribute value if it exists or null if not
    */

    public String getAttributeValue(int fingerprint) {
        if (nodeKind==Type.ELEMENT) {
            Iterator list = ((Element)node).getAttributes().iterator();
            NamePool pool = docWrapper.getNamePool();
            while (list.hasNext()) {
                Attribute att = (Attribute)list.next();
                int nameCode = pool.allocate(att.getNamespacePrefix(),
                                             att.getNamespaceURI(),
                                             att.getName());
                if (fingerprint == (nameCode & 0xfffff)) {
                    return att.getValue();
                }
            }
        }
        return null;
    }

    /**
    * Get the root node - always a document node with this tree implementation
    * @return the NodeInfo representing the containing document
    */

    public NodeInfo getRoot() {
        return docWrapper;
    }

    /**
    * Get the root (document) node
    * @return the DocumentInfo representing the containing document
    */

    public DocumentInfo getDocumentRoot() {
        return docWrapper;
    }

    /**
    * Determine whether the node has any children. <br />
    * Note: the result is equivalent to <br />
    * getEnumeration(Axis.CHILD, AnyNodeTest.getInstance()).hasNext()
    */

    public boolean hasChildNodes() {
        switch (nodeKind) {
            case Type.DOCUMENT:
                return true;
            case Type.ELEMENT:
                return !((Element)node).getContent().isEmpty();
            default:
                return false;
        }
    }
    /**
    * Get a character string that uniquely identifies this node.
    * Note: a.isSameNode(b) if and only if generateId(a)==generateId(b)
    * @return a string that uniquely identifies this node, within this
    * document. The calling code prepends information to make the result
    * unique across all documents.
    */

    public String generateId() {
        return "j" + node.hashCode();
    }

    /**
    * Copy this node to a given outputter (deep copy)
    */

    public void copy(Receiver out, int whichNamespaces, boolean copyAnnotations) throws TransformerException {
        Navigator.copy(this, out, docWrapper.namePool, whichNamespaces, copyAnnotations);
    }

    /**
    * Output all namespace nodes associated with this element. Does nothing if
    * the node is not an element.
    * @param out The relevant outputter
    * @param includeAncestors True if namespaces declared on ancestor elements must
    * be output; false if it is known that these are already on the result tree
    */

    public void outputNamespaceNodes(Receiver out, boolean includeAncestors)
        throws TransformerException {
        if (nodeKind==Type.ELEMENT) {
            NamePool pool = docWrapper.getNamePool();
            AxisIterator enum = iterateAxis(Axis.NAMESPACE);
            while (enum.hasNext()) {
                NodeWrapper wrapper = (NodeWrapper)enum.next();
                if (!includeAncestors && !wrapper.getParent().isSameNode(this)) {
                    break;
                }
                Namespace ns = (Namespace)wrapper.node;
                int nscode = pool.allocateNamespaceCode(
                                ns.getPrefix(),
                                ns.getURI());
                out.namespace(nscode, 0);
            }
        }
    }

    ///////////////////////////////////////////////////////////////////////////////
    // Axis enumeration classes
    ///////////////////////////////////////////////////////////////////////////////

//    private class AxisFilter extends AxisIteratorImpl {
//        private BaseEnumeration base;
//        private NodeTest nodeTest;
//        private NodeInfo next;
//        private int last = -1;
//
//        public AxisFilter(BaseEnumeration base, NodeTest test) {
//            this.base = base;
//            this.nodeTest = test;
//            position = 0;
//            advance();
//        }
//
//        public void advance() {
//            while (base.hasNext()) {
//    	        next = (NodeInfo)base.next();
//    	        if (nodeTest.matches(next.getItemType(), next.getFingerprint(), Type.ITEM)) {
//    	            return;
//    	        }
//    	    }
//    	    next = null;
//    	}
//
//    	public Item next() {
//    	    current = next;
//    	    advance();
//    	    position++;
//    	    return current;
//        }
//
//    	public boolean hasNext() {
//    	    return next!=null;
//    	}
//
//    	public int getLastPosition() {
//
//    	    // To find out how many nodes there are in the axis, we
//    	    // make a copy of the original node enumeration, and run through
//    	    // the whole thing again, counting how many nodes match the filter.
//
//    	    if (last>=0) {
//    	        return last;
//    	    }
//    	    last = 0;
//    	    BaseEnumeration b = (BaseEnumeration)base.getAnother();
//    	    while (b.hasNext()) {
//    	        NodeInfo n = (NodeInfo)b.next();
//    	        if (nodeTest.matches(n.getItemType(), n.getFingerprint(), Type.ITEM)) {
//    	            last++;
//    	        }
//    	    }
//    	    return last;
//    	}
//
//    	public SequenceIterator getAnother() {
//    	    return new AxisFilter(base, nodeTest);
//    	}
//	}

//	private abstract class BaseEnumeration extends AxisIteratorImpl {
//
//        protected NodeWrapper next;
//
//        public final boolean hasNext() {
//            return next!=null;
//        }
//
//        public final Item next() {
//            current = next;
//            advance();
//            return current;
//        }
//
//        public abstract void advance();
//
//	    public abstract SequenceIterator getAnother();
//
//	    public final int getLastPosition() {// not used
//	        return 1;
//	    }
//
//	}


//    final private class AncestorEnumeration extends BaseEnumeration {
//
//        private boolean includeSelf;
//        private NodeWrapper start;
//
//        public AncestorEnumeration(NodeWrapper start, boolean includeSelf) {
//            this.start = start;
//            this.includeSelf = includeSelf;
//            this.next = start;
//            if (!includeSelf) {
//                advance();
//            }
//        }
//
//        public void advance() {
//            next=(NodeWrapper)next.getParent();
//        }
//
//        public SequenceIterator getAnother() {
//            return new AncestorEnumeration(start, includeSelf);
//        }
//
//    } // end of class AncestorEnumeration

    private final class AttributeEnumeration extends Navigator.BaseEnumeration {

        private Iterator atts;
        private int ix = 0;
        private NodeWrapper start;

        public AttributeEnumeration(NodeWrapper start) {
            this.start = start;
            atts = ((Element)start.node).getAttributes().iterator();
            advance();
        }

        public void advance() {
            if (atts.hasNext()) {
                next = makeWrapper(atts.next(), docWrapper, start, ix++);
            } else {
                next = null;
            }
        }

        public SequenceIterator getAnother() {
            return new AttributeEnumeration(start);
        }

    }  // end of class AttributeEnumeration

    private final class NamespaceEnumeration extends Navigator.BaseEnumeration {

        private HashMap nslist = new HashMap();
        private Iterator prefixes;
        private int ix = 0;
        private NodeWrapper start;

        public NamespaceEnumeration(NodeWrapper start) {
            this.start = start;
            NodeWrapper curr = start;

            // build the complete list of namespaces

            do {
                Element elem = (Element)curr.node;
                Namespace ns = elem.getNamespace();
                String prefix = ns.getPrefix();
                String uri = ns.getURI();
                if (!(prefix.equals("") && uri.equals(""))) {
                    if (!nslist.containsKey(prefix)) {
                        nslist.put(ns.getPrefix(), ns);
                    }
                }
                List addl = elem.getAdditionalNamespaces();
                if (addl.size() > 0) {
                    Iterator itr = addl.iterator();
                    while (itr.hasNext()) {
                        ns = (Namespace) itr.next();
                        if (!nslist.containsKey(ns.getPrefix())) {
                            nslist.put(ns.getPrefix(), ns);
                        }
                    }
                }
                curr = (NodeWrapper)curr.getParent();
            } while (curr.getItemType()==Type.ELEMENT);

            nslist.put("xml", Namespace.XML_NAMESPACE);
            prefixes = nslist.keySet().iterator();
            advance();
        }

        public void advance() {
            if (prefixes.hasNext()) {
                String prefix = (String)prefixes.next();
                Namespace ns = (Namespace)nslist.get(prefix);
                next = makeWrapper(ns, docWrapper, start, ix++);
            } else {
                next = null;
            }
        }

        public SequenceIterator getAnother() {
            return new NamespaceEnumeration(start);
        }

        // NB: namespace nodes in the JDOM implementation do not support all
        // XPath functions, for example namespace nodes have no parent.


    }  // end of class NamespaceEnumeration


    /**
    * The class ChildEnumeration handles not only the child axis, but also the
    * following-sibling and preceding-sibling axes. It can also iterate the children
    * of the start node in reverse order, something that is needed to support the
    * preceding and preceding-or-ancestor axes (the latter being used by xsl:number)
    */

    private final class ChildEnumeration extends Navigator.BaseEnumeration {

        private NodeWrapper start;
        private NodeWrapper commonParent;
        private ListIterator children;
        private int ix = 0;
        private boolean downwards;  // iterate children of start node (not siblings)
        private boolean forwards;   // iterate in document order (not reverse order)

        public ChildEnumeration(NodeWrapper start,
                                boolean downwards, boolean forwards) {
            this.start = start;
            this.downwards = downwards;
            this.forwards = forwards;

            if (downwards) {
                commonParent = start;
            } else {
                commonParent = (NodeWrapper)start.getParent();
            }

            if (commonParent.getItemType()==Type.DOCUMENT) {
                children = ((Document)commonParent.node).getContent().listIterator();
            } else {
                children = ((Element)commonParent.node).getContent().listIterator();
            }

            if (downwards) {
                if (!forwards) {
                    // backwards enumeration: go to the end
                    while (children.hasNext()) {
                        children.next();
                        ix++;
                    }
                }
            } else {
                ix = start.getSiblingPosition();
                // find the start node among the list of siblings
                if (forwards) {
                    for (int i=0; i<=ix; i++) {
                        children.next();
                    }
                    ix++;
                } else {
                    for (int i=0; i<ix; i++) {
                        children.next();
                    }
                    ix--;
                }
            }
            advance();
        }



        public void advance() {
            if (forwards) {
                if (children.hasNext()) {
                    Object nextChild = children.next();
                    if (nextChild instanceof EntityRef) {
                        throw new IllegalStateException("Unexpanded entity in JDOM tree");
                    } else {
                        next = makeWrapper(nextChild, docWrapper, commonParent, ix++);
                    }
                } else {
                    next = null;
                }
            } else {    // backwards
                if (children.hasPrevious()) {
                    Object nextChild = children.previous();
                    if (nextChild instanceof EntityRef) {
                        throw new IllegalStateException("Unexpanded entity in JDOM tree");
                    } else {
                        next = makeWrapper(nextChild, docWrapper, commonParent, ix--);
                    }
                } else {
                    next = null;
                }
            }
        }

        public SequenceIterator getAnother() {
            return new ChildEnumeration(start, downwards, forwards);
        }

    } // end of class ChildEnumeration

    /**
    * The DescendantEnumeration class supports the XPath descendant axis.
    * But it also has the option to return the descendants in reverse document order;
    * this is used when evaluating the preceding axis. Note that the includeSelf option
    * should not be used when scanning in reverse order, as the self node will always be
    * returned first.
    */

//    private final class DescendantEnumeration extends BaseEnumeration {
//
//        private AxisIterator children = null;
//        private AxisIterator descendants = null;
//        private NodeWrapper start;
//        private boolean includeSelf;
//        private boolean forwards;
//        private boolean atEnd = false;
//
//        public DescendantEnumeration(NodeWrapper start,
//                                 boolean includeSelf, boolean forwards) {
//            this.start = start;
//            this.includeSelf = includeSelf;
//            this.forwards = forwards;
//            advance();
//        }
//
//        public void advance() {
//            if (descendants!=null) {
//                if (descendants.hasNext()) {
//                    next = (NodeWrapper)descendants.next();
//                    return;
//                } else {
//                    descendants = null;
//                }
//            }
//            if (children!=null) {
//                if (children.hasNext()) {
//                    NodeWrapper n = (NodeWrapper)children.next();
//                    if (n.hasChildNodes()) {
//                        if (forwards) {
//                            descendants = new DescendantEnumeration(n,
//                                             false, forwards);
//                            next = n;
//                        } else {
//                            descendants = new DescendantEnumeration(n, true, forwards);
//                            advance();
//                        }
//                    } else {
//                        next = n;
//                    }
//                } else {
//                    if (forwards || !includeSelf) {
//                        next = null;
//                    } else {
//                        atEnd = true;
//                        children = null;
//                        next = start;
//                    }
//                }
//            } else if (atEnd) {
//                // we're just finishing a backwards scan
//                next = null;
//            } else {
//                // we're just starting...
//                if (start.hasChildNodes()) {
//                    children = new ChildEnumeration(start, true, forwards);
//                } else {
//                    children = EmptyIterator.getInstance();
//                }
//                if (forwards && includeSelf) {
//                    next = start;
//                } else {
//                    advance();
//                }
//            }
//        }
//
//        public SequenceIterator getAnother() {
//            return new DescendantEnumeration(start, includeSelf, forwards);
//        }
//
//    } // end of class DescendantEnumeration

//    private final class FollowingEnumeration extends BaseEnumeration {
//
//        private NodeWrapper start;
//        private AxisIterator ancestorEnum = null;
//        private AxisIterator siblingEnum = null;
//        private AxisIterator descendEnum = null;
//
//        public FollowingEnumeration(NodeWrapper start) {
//            this.start = start;
//            ancestorEnum = new AncestorEnumeration(start, false);
//            switch (start.getItemType()) {
//                case Type.ELEMENT:
//                case Type.TEXT:
//                case Type.COMMENT:
//                case Type.PROCESSING_INSTRUCTION:
//                    siblingEnum = new ChildEnumeration(start, false, true);
//                        // gets following siblings
//                    break;
//                case Type.ATTRIBUTE:
//                case Type.NAMESPACE:
//                    siblingEnum = new ChildEnumeration((NodeWrapper)start.getParent(), true, true);
//                        // gets children of the attribute's parent node
//                    break;
//                default:
//                    siblingEnum = EmptyIterator.getInstance();
//            }
//            advance();
//        }
//
//        public void advance() {
//            if (descendEnum!=null) {
//                if (descendEnum.hasNext()) {
//                    next = (NodeWrapper)descendEnum.next();
//                    return;
//                } else {
//                    descendEnum = null;
//                }
//            }
//            if (siblingEnum!=null) {
//                if (siblingEnum.hasNext()) {
//                    next = (NodeWrapper)siblingEnum.next();
//                    if (next.hasChildNodes()) {
//                        descendEnum = new DescendantEnumeration(next, false, true);
//                    } else {
//                        descendEnum = null;
//                    }
//                    return;
//                } else {
//                    descendEnum = null;
//                    siblingEnum = null;
//                }
//            }
//            if (ancestorEnum.hasNext()) {
//                next = (NodeWrapper)ancestorEnum.next();
//                if (next.getItemType() == Type.DOCUMENT) {
//                    siblingEnum = EmptyIterator.getInstance();
//                } else {
//                    siblingEnum = new ChildEnumeration(next, false, true);
//                }
//                advance();
//            } else {
//                next = null;
//            }
//        }
//
//        public SequenceIterator getAnother() {
//            return new FollowingEnumeration(start);
//        }
//
//    } // end of class FollowingEnumeration

//    private final class PrecedingEnumeration extends BaseEnumeration {
//
//        private NodeWrapper start;
//        private AxisIterator ancestorEnum = null;
//        private AxisIterator siblingEnum = null;
//        private AxisIterator descendEnum = null;
//        private boolean includeAncestors;
//
//        public PrecedingEnumeration(NodeWrapper start, boolean includeAncestors) {
//            this.start = start;
//            this.includeAncestors = includeAncestors;
//            ancestorEnum = new AncestorEnumeration(start, false);
//            switch (start.getItemType()) {
//                case Type.ELEMENT:
//                case Type.TEXT:
//                case Type.COMMENT:
//                case Type.PROCESSING_INSTRUCTION:
//                    // get preceding-sibling enumeration
//                    siblingEnum = new ChildEnumeration(start, false, false);
//                    break;
//                default:
//                    siblingEnum = EmptyIterator.getInstance();
//            }
//            advance();
//        }
//
//        public void advance() {
//            if (descendEnum!=null) {
//                if (descendEnum.hasNext()) {
//                    next = (NodeWrapper)descendEnum.next();
//                    return;
//                } else {
//                    descendEnum = null;
//                }
//            }
//            if (siblingEnum!=null) {
//                if (siblingEnum.hasNext()) {
//                    NodeWrapper sib = (NodeWrapper)siblingEnum.next();
//                    if (sib.hasChildNodes()) {
//                        descendEnum = new DescendantEnumeration(sib, true, false);
//                        advance();
//                    } else {
//                        descendEnum = null;
//                        next = sib;
//                    }
//                    return;
//                } else {
//                    descendEnum = null;
//                    siblingEnum = null;
//                }
//            }
//            if (ancestorEnum.hasNext()) {
//                next = (NodeWrapper)ancestorEnum.next();
//                if (next.getItemType() == Type.DOCUMENT) {
//                    siblingEnum = EmptyIterator.getInstance();
//                } else {
//                    siblingEnum = new ChildEnumeration(next, false, false);
//                }
//                if (!includeAncestors) {
//                    advance();
//                }
//            } else {
//                next = null;
//            }
//        }
//
//        public SequenceIterator getAnother() {
//            return new PrecedingEnumeration(start, includeAncestors);
//        }
//
//    } // end of class PrecedingEnumeration

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
