package net.sf.saxon.om;
import net.sf.saxon.xpath.XPathException;

/**
 * A SequenceIterator is used to iterate over any XPath 2 sequence (of values or nodes).
 * It is modelled on the JDK 1.2 Iterator class, but differs in two main respects:
 * a SequenceIterator can be cloned (to create a new iterator that iterates over the
 * same sequence of values, but with the initial position reset to the start); and
 * the methods can throw exceptions if access to the underlying data fails.
 * Also, a SequenceIterator keeps track of the current Item and the current position.
 * The objects returned by the SequenceIterator will always be either nodes
 * (class NodeInfo) or singleton values (class AtomicValue), these are represented
 * by the interface Item.
 */

public interface SequenceIterator {

    /**
     * Determine whether there are more items to come. <BR>
     * This method must be called before next() is called to get the next item, and next()
     * must not be called unless hasNext() returns true. The hasNext() method must not
     * change the current position, it must be possible to call hasNext() repeatedly and
     * get the same result each time.
     *
     * @exception XPathException if any error occurs while determining whether
     *      there are more items in the sequence
     * @return true if there are more items available
     */

    public boolean hasNext() throws XPathException;

    // TODO: get rid of the hasNext() method, and make next() return null when the
    // sequence is exhausted. This will halve the number of calls to the iterator,
    // and it will avoid the need for iterators to look-ahead unnecessarily, which is
    // wasteful if the pipeline is stopped prematurely, and also creates unnecessary
    // objects on the heap. It will also work much better in cases such as regex
    // iterators that don't allow the lookahead.

    /**
     * Get the next item in the sequence. <BR>
     * This must not be called unless hasNext() has been called to test whether
     * there is a next item.
     *
     * @exception XPathException if an error occurs retrieving the next item
     * @return the next item
     */

    public Item next() throws XPathException;

    /**
     * Get the current value in the sequence (the one returned by the
     * most recent call on next()). This will be null before the first
     * call of next().
     *
     * @return the current item, the one most recently returned by a call on
     *     next(); or null, if next() has not been called
     */

    public Item current();

    /**
     * Get the current position. This will be zero before the first call
     * on next(), otherwise it will be the number of times that next() has
     * been called.
     *
     * @return the current position, the position of the item returned by the
     *     most recent call of next()
     */

    public int position();

    /**
     * Get another SequenceIterator that iterates over the same items as the original,
     * but which is repositioned at the start of the sequence.
     *
     * @exception XPathException if any error occurs
     * @return a SequenceIterator that iterates over the same items,
     *     positioned before the first item
     */

    public SequenceIterator getAnother() throws XPathException;

}



//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
