package net.sf.saxon.query;

import net.sf.saxon.Controller;
import net.sf.saxon.Configuration;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.instruct.Executable;
import net.sf.saxon.instruct.Bindery;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.xpath.XPathEvaluator;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerException;
import javax.xml.transform.ErrorListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * XQueryExpression represents a compiled query. This object is immutable and thread-safe,
 * the same compiled query may be executed many times in series or in parallel. The object
 * can be created only by using the compileQuery method of the QueryProcessor class.
 *
 * <p>Various methods are provided for evaluating the query, with different options for
 * delivery of the results.</p>
 */
public class XQueryExpression {

    private Expression expression;
    private Executable executable;

    /**
    * The constructor is protected, to ensure that instances can only be
    * created using the compileQuery() methods of QueryProcessor
    */

    protected XQueryExpression(Expression exp, StaticQueryContext staticEnv, Configuration config) {
        expression = exp;
        executable = new Executable();
        executable.setConfiguration(config);
        executable.setDefaultCollationName(staticEnv.getDefaultCollationName());
        executable.setCollationTable(staticEnv.getAllCollations());
        executable.setSlotSpace(staticEnv.getNumberOfGlobalVariables(),
                staticEnv.getNumberOfLocalVariables());
        String[] modules = new String[1];
        modules[0] = staticEnv.getSystemId();
        executable.setModuleArray(modules);
    }

    /**
     * Execute a the compiled Query, returning the results as a List.
     * @param env Provides the dynamic query evaluation context
     * @return The results of the expression, as a List. The List represents the sequence
     * of items returned by the expression. Each item in the list will either be an
     * object representing a node, or an object representing an atomic value.
     * For the types of Java object that may be returned, see the description of the evaluate method
     * of class XPathProcessor
     */

    public List evaluate(DynamicQueryContext env) throws TransformerException {
        SequenceIterator iterator = iterator(env);
        ArrayList list = new ArrayList();
        while (iterator.hasNext()) {
            Item item = iterator.next();
            list.add(XPathEvaluator.convert(item));
        }
        return list;
    }

    /**
     * Execute the compiled Query, returning the first item in the result.
     * This is useful where it is known that the expression will only return
     * a singleton value (for example, a single node, or a boolean).
     * @param env Provides the dynamic query evaluation context
     * @return The first item in the sequence returned by the expression. If the expression
     * returns an empty sequence, this method returns null. Otherwise, it returns the first
     * item in the result sequence, represented as a Java object using the same mapping as for
     * the evaluate() method
     */

    public Object evaluateSingle(DynamicQueryContext env) throws TransformerException {
        SequenceIterator iterator = iterator(env);
        if (iterator.hasNext()) {
            return XPathEvaluator.convert(iterator.next());
        } else {
            return null;
        }
    }

    /**
     * Get an iterator over the results of the expression. This returns results without
     * any conversion of the returned items to "native" Java classes. The iterator will
     * deliver a sequence of Item objects, each item being either a NodeInfo (representing
     * a node) or an AtomicValue (representing an atomic value).
     *
     * <p>To get the results of the query in the form of an XML document in which each
     * item is wrapped by an element indicating its type, use:</p>
     *
     * <p><code>QueryResult.wrap(iterator(env))</code></p>
     *
     * <p>To serialize the results to a file, use the QueryResult.serialize() method.</p>
     *
     * @param env Provides the dynamic query evaluation context
     * @return an iterator over the results of the query. The class SequenceIterator
     * is modeled on the standard Java Iterator class, but has extra functionality
     * and can throw exceptions when errors occur.
     * @throws XPathException if a dynamic error occurs in evaluating the query. Some
     * dynamic errors will not be reported by this method, but will only be reported
     * when the individual items of the result are accessed using the returned iterator.
     */

    public SequenceIterator iterator(DynamicQueryContext env) throws TransformerException {
        Controller controller = new Controller(executable.getConfiguration());
        controller.setExecutable(executable);
        executable.initialiseBindery(controller.getBindery());

        HashMap parameters = env.getParameters();
        if (parameters != null) {
            Iterator iter = parameters.keySet().iterator();
            while (iter.hasNext()) {
                String paramName = (String)iter.next();
                Object paramValue = parameters.get(paramName);
                controller.setParameter(paramName, paramValue);
            }
        }

        controller.setURIResolver(env.getURIResolver());
        controller.setErrorListener(env.getErrorListener());

        try {
            NodeInfo node = env.getContextNode();
            if (node != null) {
                controller.makeContext(node);
                controller.setPrincipalSourceDocument(node.getDocumentRoot());
            }
            Bindery bindery = controller.getBindery();
            bindery.openStackFrame();
            controller.defineGlobalParameters(bindery);
            XPathContext context = controller.newXPathContext();
            SequenceIterator iterator = expression.iterate(context);
            //controller.getBindery().closeStackFrame();
            return new ErrorReportingIterator(iterator, controller.getErrorListener());
        } catch (XPathException err) {
            controller.getErrorListener().error(err);
            // if the error handler suppresses the exception, return an empty sequence
            return EmptyIterator.getInstance();
        }
    }

    /**
     * Diagnostic method: display a representation of the compiled query on the
     * System.err output stream.
     */

    public void explain(NamePool pool) {
        System.err.println("============ Compiled Expression ============");
        expression.display(10, pool);
        System.err.println("=============================================");
    }

    /**
     * ErrorReportingIterator is an iterator that wraps a base iterator and reports
     * any exceptions that are raised to the ErrorListener
     */

    private static class ErrorReportingIterator implements SequenceIterator {
        private SequenceIterator base;
        private ErrorListener listener;

        public ErrorReportingIterator(SequenceIterator base, ErrorListener listener) {
            this.base = base;
            this.listener = listener;
        }

        public boolean hasNext() throws XPathException {
            try {
                return base.hasNext();
            } catch (XPathException e1) {
                try {
                    listener.error(e1);
                } catch (TransformerException e2) {}
                throw e1;
            }
        }

        public Item next() throws XPathException {
            try {
                return base.next();
            } catch (XPathException e1) {
                try {
                    listener.error(e1);
                } catch (TransformerException e2) {}
                throw e1;
            }
        }

        public Item current() {
            return base.current();
        }

        public int position() {
            return base.position();
        }

        public SequenceIterator getAnother() throws XPathException {
            return new ErrorReportingIterator(base.getAnother(), listener);
        }

    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.

// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// The line marked PB-SYNC is by Peter Bryant (pbryant@bigfoot.com). All Rights Reserved.
//
// Contributor(s): Michael Kay, Peter Bryant, David Megginson
//

