package net.sf.saxon.sort;

import net.sf.saxon.expr.ComputedExpression;
import net.sf.saxon.expr.PromotionOffer;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.Value;
import net.sf.saxon.value.ObjectValue;

/**
 * A tuple expression is an expression that returns a tuple. Specifically,
 * it is a list of n expressions, which are evaluated to create a list of n items.
 * Tuple expressions are used during the evaluation of a FLWR expression. A tuple
 * is not a value within the XPath/XQuery type system, so it is represented as
 * an external object, specifically as a Java array wrapped inside an ObjectValue.
 *
 */
public class TupleExpression extends ComputedExpression {

    Expression[] components;

    public TupleExpression(int width) {
        components = new Expression[width];
    }

    public void setExpression(int i, Expression exp) {
        components[i] = exp;
    }

    public Expression simplify() throws XPathException {
        for (int i=0; i<components.length; i++) {
            components[i] = components[i].simplify();
        }
        return this;
    }

    public Expression promote(PromotionOffer offer) throws XPathException {
        for (int i=0; i<components.length; i++) {
            components[i] = components[i].promote(offer);
        }
        return this;
    }

    public int getItemType() {
        return Type.OBJECT;
    }

    public Expression analyze(StaticContext env) throws XPathException {
        for (int i=0; i<components.length; i++) {
            components[i] = components[i].analyze(env);
        }
        return this;
    }

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "Tuple");
        for (int i=0; i<components.length; i++) {
            components[i].display(level+1, pool);
        }
    }


    public Item evaluateItem(XPathContext context) throws XPathException {
        Item[] tuple = new Item[components.length];
        for (int i=0; i<components.length; i++) {
            tuple[i] = components[i].evaluateItem(context);
        }
        return new ObjectValue(tuple);
    }

    /**
     * Get the cardinality of the expression. This is exactly one, in the sense
     * that evaluating the TupleExpression returns a single tuple.
     * @return
     */

    public int computeCardinality() {
        return StaticProperty.CARDINALITY_EXACTLY_ONE;
    }

    public int getIntrinsicDependencies() {
        return 0;
    }

    public Expression[] getSubExpressions() {
        return components;
    }

    public SequenceIterator iterate(XPathContext context) throws XPathException {
        return super.iterate(context);
    }

}
