package net.sf.saxon.style;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.Atomizer;
import net.sf.saxon.expr.AtomicSequenceConverter;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.functions.SystemFunction;
import net.sf.saxon.functions.StringJoin;
import net.sf.saxon.functions.Concat;

/**
* This class represents an attribute value template. The class allows an AVT to be parsed, and
* can construct an Expression that returns the effective value of the AVT.
*
* This is an abstract class that is never instantiated, it contains static methods only.
*/

public abstract class AttributeValueTemplate {

    private AttributeValueTemplate() {}


    /**
    * Static factory method to create an AVT from an XSLT string representation
    */

    public static Expression make(String avt, StaticContext env) throws XPathException {

        if ( avt.indexOf("{") < 0 && avt.indexOf("}") < 0) {        // fast path: no embedded expressions
            return new StringValue(avt);

        } else {

            int nr = 0;
            Expression[] components = new Expression[avt.length()];     // for once, we'll set a limit!
                                             // process embedded expressions within the avt
            int i0, i1, i2, i8, i9, len, last;
            char inquote = ' ';
            last = 0;
            len = avt.length();
            while (last<len) {
                i0 = avt.indexOf("{", last);
                i1 = avt.indexOf("{{", last);
                i8 = avt.indexOf("}", last);
                i9 = avt.indexOf("}}", last);
                //System.err.println("AVT:" + avt + ": " + i0 + "," + i1 + "," + i8 + "," + i9);

                if (i8>=0 && (i0<0 || i8<i0)) {             // found a "}"
                    if (i8 != i9) {                        // a "}" that isn't a "}}"
                        throw new XPathException.Static("Closing curly brace in attribute value template \"" + avt + "\" must be doubled");
                    }
                    components[nr++] = new StringValue(avt.substring(last, i8+1));
                    //components[nr++] = avt.substring(last, i8+1);
                    last = i8+2;
                } else if (i1>=0 && i1==i0) {              // found a doubled "{{"
                    components[nr++] = new StringValue(avt.substring(last, i1+1));
                    //components[nr++] = avt.substring(last, i1+1);
                    last = i1+2;
                } else if (i0>=0) {                        // found a single "{"
                    if (i0>last) {
                        components[nr++] = new StringValue(avt.substring(last, i0));
                        //components[nr++] = avt.substring(last, i0);
                    }
                    for (i2=i0+1; i2<len; i2++) {
                        if (avt.charAt(i2)=='\"') inquote = '\"';
                        if (avt.charAt(i2)=='\'') inquote = '\'';
                        if (inquote!=' ') {
                            i2++;
                            while (i2<len && avt.charAt(i2)!=inquote) i2++;
                            inquote = ' ';
                        } else {
                            if (avt.charAt(i2)=='}') {
                               // we're in an expression so we don't need to check for a doubled "}}"
                               break;
                            }
                        }
                    }
                    if (i2>=len) {
                        throw new XPathException.Static("No closing \"}\" in attribute value template " + avt);
                    }

                    String expr = avt.substring(i0+1, i2);
                    Expression ex = makeStringJoin(expr, env);
                    components[nr++] = ex;
                    last=i2+1;
                } else {                // didn't find anything
                    components[nr++] = new StringValue(avt.substring(last));
                    //components[nr++] = avt.substring(last);
                    last=len;
                }
            }

            // is it empty?

            if (nr==0) {
                return StringValue.EMPTY_STRING;
            }

            // is it a single component?

            if (nr==1) {
                return components[0].simplify();
            }

            // otherwise, return an expression that concatenates the components

    		Concat fn = (Concat)SystemFunction.makeSystemFunction("concat");
    		Expression[] args = new Expression[nr];
    		System.arraycopy(components, 0, args, 0, nr);
    		fn.setArguments(args);
    		return fn.simplify();

        }
    }

    /**
    * Make a string-join expression that concatenates the string-values of items in
    * a sequence with intervening spaces. This may be simplified later as a result
    * of type-checking.
    */

    private static Expression makeStringJoin(String expression, StaticContext env) throws XPathException {
        Expression exp = ExpressionTool.make(expression, env);

        exp = new Atomizer(exp);
        exp = new AtomicSequenceConverter(exp, Type.STRING);

		StringJoin fn = (StringJoin)SystemFunction.makeSystemFunction("string-join");
		Expression[] args = new Expression[2];
		args[0] = exp;
		args[1] = new StringValue(" ");
		fn.setArguments(args);
		return fn;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
