package net.sf.saxon.style;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.tree.NodeFactory;
import net.sf.saxon.tree.ElementImpl;
import net.sf.saxon.om.*;
import net.sf.saxon.Loader;

import javax.xml.transform.*;
import org.xml.sax.Locator;
import java.util.HashMap;

/**
  * Class StyleNodeFactory. <br>
  * A Factory for nodes in the stylesheet tree. <br>
  * Currently only allows Element nodes to be user-constructed.
  * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
  */

public class StyleNodeFactory implements NodeFactory {

    HashMap userStyles = new HashMap();
    NamePool namePool;
    StandardNames sn;
    boolean allowExtensions;


    public StyleNodeFactory(NamePool pool, StandardNames standardNames, boolean allowExtensions) {

		namePool = pool;
		sn = standardNames;
		this.allowExtensions = allowExtensions;
    }

    public StandardNames getStandardNames() {
        return sn;
    }

    /**
    * Create an Element node. Note, if there is an error detected while constructing
    * the Element, we add the element anyway, and return success, but flag the element
    * with a validation error. This allows us to report more than
    * one error from a single compilation.
    * @param nameCode The element name
    * @param attlist the attribute list
    */

    public ElementImpl makeElementNode(
                        NodeInfo parent,
                        int nameCode,
                        AttributeCollection attlist,
                        int[] namespaces,
                        int namespacesUsed,
                        Locator locator,
                        int sequence)
    {
        boolean toplevel = (parent instanceof XSLStyleSheet);
        String baseURI = null;
        int lineNumber = -1;

        if (locator!=null) {
            baseURI = locator.getSystemId();
            lineNumber = locator.getLineNumber();
        }

        if (parent instanceof DataElement) {
            DataElement d = new DataElement();
            d.setNamespaceDeclarations(namespaces, namespacesUsed);
            d.initialise(nameCode, attlist, parent, baseURI, lineNumber, sequence);
            return d;
        }

    	int f = nameCode&0xfffff;

    	// Try first to make an XSLT element

    	StyleElement e = makeXSLElement(f);

		if (e != null) {  // recognized as an XSLT element
			try {
	        	e.setNamespaceDeclarations(namespaces, namespacesUsed);
                e.setLineNumber(lineNumber);
	            e.initialise(nameCode, attlist, parent, baseURI, -1, sequence);
	            e.processExtensionElementAttribute(sn.EXTENSION_ELEMENT_PREFIXES);
	            e.processExcludedNamespaces(sn.EXCLUDE_RESULT_PREFIXES);
	            e.processVersionAttribute(sn.VERSION);
	            e.processDefaultXPathNamespaceAttribute(sn.XPATH_DEFAULT_NAMESPACE);
	        } catch (TransformerException err) {
	            e.setValidationError(err, StyleElement.REPORT_ALWAYS);
	        }
            return e;

        } else {   // not recognized as an XSLT element

	        short uriCode = namePool.getURICode(nameCode);
	        String localname = namePool.getLocalName(nameCode);

	        Class assumedClass = LiteralResultElement.class;

	        // We can't work out the final class of the node until we've examined its attributes
	        // such as version and extension-element-prefixes; but we can have a good guess, and
	        // change it later if need be.

			StyleElement temp = null;
			boolean assumedSaxonElement = false;

			// map EXSLT func:function and func:result onto Saxon equivalents

	        if (uriCode == NamespaceConstant.SAXON_CODE) {
	        	temp = makeSaxonElement(f);
	        	if (temp!=null) {
	        		assumedClass = temp.getClass();
	        		assumedSaxonElement = true;
	        	}
	        } else if (toplevel && uriCode != 0) {
                DataElement d = new DataElement();
                d.setNamespaceDeclarations(namespaces, namespacesUsed);
                d.initialise(nameCode, attlist, parent, baseURI, lineNumber, sequence);
                return d;
            }

			if (temp==null) {
		        temp = new LiteralResultElement();
		    }

	        temp.setNamespaceDeclarations(namespaces, namespacesUsed);

	        try {
	            temp.initialise(nameCode, attlist, parent, baseURI, lineNumber, sequence);
                temp.setLineNumber(lineNumber);
	            temp.processExtensionElementAttribute(sn.XSL_EXTENSION_ELEMENT_PREFIXES);
	            temp.processExcludedNamespaces(sn.XSL_EXCLUDE_RESULT_PREFIXES);
	            temp.processVersionAttribute(sn.XSL_VERSION);
	            temp.processDefaultXPathNamespaceAttribute(sn.XSL_XPATH_DEFAULT_NAMESPACE);
	        } catch (TransformerException err) {
	            temp.setValidationError(err, StyleElement.REPORT_UNLESS_FORWARDS_COMPATIBLE);
	        }

	        // Now we work out what class of element we really wanted, and change it if necessary

	        TransformerException reason = null;
	        Class actualClass = LiteralResultElement.class;

	        if (uriCode == NamespaceConstant.XSLT_CODE) {
                reason = new TransformerConfigurationException("Unknown XSLT element: " + localname);
                actualClass = AbsentExtensionElement.class;
                temp.setValidationError(reason, StyleElement.REPORT_UNLESS_FORWARDS_COMPATIBLE);
	        } else if (uriCode == NamespaceConstant.SAXON_CODE || uriCode == NamespaceConstant.EXSLT_FUNCTIONS_CODE) {
	        	if (toplevel || temp.isExtensionNamespace(uriCode)) {
	        		if (assumedSaxonElement) {
	        			// all is well
	        			actualClass = assumedClass;
	        		} else {
	        			actualClass = AbsentExtensionElement.class;
	        			reason = new TransformerConfigurationException(
	        			                "Unknown Saxon extension element: " + localname);
                        temp.setValidationError(reason, StyleElement.REPORT_IF_INSTANTIATED);
	        		}
	        	} else {
	        		actualClass = LiteralResultElement.class;
	        	}
	        } else if (temp.isExtensionNamespace(uriCode) && !toplevel) {
            	Integer nameKey = new Integer(nameCode&0xfffff);

                actualClass = (Class)userStyles.get(nameKey);
                if (actualClass==null) {
                    if (!allowExtensions) {
                        actualClass = AbsentExtensionElement.class;
	        			reason = new TransformerConfigurationException(
	        			                "Extension elements are disabled");
                        temp.setValidationError(reason, StyleElement.REPORT_IF_INSTANTIATED);
                    } else {

                        ExtensionElementFactory factory = getFactory(uriCode);
                        if (factory != null) {
                            actualClass = factory.getExtensionClass(localname);
                            if (actualClass != null) {
                                userStyles.put(nameKey, actualClass);             // for quicker access next time
                            }
                        }
                    }

                    if (actualClass == null) {

                        // if we can't instantiate an extension element, we don't give up
                        // immediately, because there might be an xsl:fallback defined. We
                        // create a surrogate element called AbsentExtensionElement, and
                        // save the reason for failure just in case there is no xsl:fallback

                        actualClass = AbsentExtensionElement.class;
                        reason = new TransformerConfigurationException("Unknown extension element", temp);
                        temp.setValidationError(reason, StyleElement.REPORT_IF_INSTANTIATED);
                    }
                }
	        } else {
	        	actualClass = LiteralResultElement.class;
	        }

	        StyleElement node;
	        if (!actualClass.equals(assumedClass)) {
	            try {
	                node = (StyleElement)actualClass.newInstance();
	                //if (reason!=null) {
	                //    node.setValidationError(reason);
	                //}
	            } catch (java.lang.InstantiationException err1) {
	                throw new TransformerFactoryConfigurationError(err1, "Failed to create instance of " + actualClass.getName());
	            } catch (java.lang.IllegalAccessException err2) {
	                throw new TransformerFactoryConfigurationError(err2, "Failed to access class " + actualClass.getName());
	            }
	            node.substituteFor(temp);   // replace temporary node with the new one
	        } else {
	            node = temp;    // the original element will do the job
	        }
	        return node;
	    }
    }

	/**
	* Make an XSL element node
	*/

	private StyleElement makeXSLElement(int f) {

		StyleElement e = null;

		if (f==sn.XSL_ANALYZE_STRING) 			e=new XSLAnalyzeString();
		else if (f==sn.XSL_APPLY_IMPORTS) 		e=new XSLApplyImports();
		else if (f==sn.XSL_APPLY_TEMPLATES) 	e=new XSLApplyTemplates();
		else if (f==sn.XSL_ATTRIBUTE) 			e=new XSLAttribute();
		else if (f==sn.XSL_ATTRIBUTE_SET) 		e=new XSLAttributeSet();
		else if (f==sn.XSL_CALL_TEMPLATE) 		e=new XSLCallTemplate();
		else if (f==sn.XSL_CHARACTER_MAP) 		e=new XSLCharacterMap();
		else if (f==sn.XSL_CHOOSE) 				e=new XSLChoose();
		else if (f==sn.XSL_COMMENT) 			e=new XSLComment();
		else if (f==sn.XSL_COPY) 				e=new XSLCopy();
		else if (f==sn.XSL_COPY_OF) 			e=new XSLCopyOf();
		else if (f==sn.XSL_DECIMAL_FORMAT) 		e=new XSLDecimalFormat();
		else if (f==sn.XSL_ELEMENT) 			e=new XSLElement();
		else if (f==sn.XSL_FALLBACK) 			e=new XSLFallback();
		else if (f==sn.XSL_FOR_EACH) 			e=new XSLForEach();
		else if (f==sn.XSL_FOR_EACH_GROUP)		e=new XSLForEachGroup();
		else if (f==sn.XSL_FUNCTION) 			e=new XSLFunction();
		else if (f==sn.XSL_IF) 					e=new XSLIf();
		else if (f==sn.XSL_IMPORT) 				e=new XSLImport();
		else if (f==sn.XSL_INCLUDE) 			e=new XSLInclude();
		else if (f==sn.XSL_KEY) 				e=new XSLKey();
		else if (f==sn.XSL_MATCHING_SUBSTRING)	e=new XSLMatchingSubstring();
		else if (f==sn.XSL_MESSAGE) 			e=new XSLMessage();
		else if (f==sn.XSL_NEXT_MATCH)          e=new XSLNextMatch();
		else if (f==sn.XSL_NON_MATCHING_SUBSTRING)
		                                        e=new XSLMatchingSubstring();	//sic
		else if (f==sn.XSL_NUMBER) 				e=new XSLNumber();
		else if (f==sn.XSL_NAMESPACE) 	        e=new XSLNamespace();
		else if (f==sn.XSL_NAMESPACE_ALIAS) 	e=new XSLNamespaceAlias();
		else if (f==sn.XSL_OTHERWISE) 			e=new XSLOtherwise();
		else if (f==sn.XSL_OUTPUT)   			e=new XSLOutput();
		else if (f==sn.XSL_OUTPUT_CHARACTER)    e=new XSLOutputCharacter();
		else if (f==sn.XSL_PARAM) 				e=new XSLParam();
		else if (f==sn.XSL_PERFORM_SORT) 		e=new XSLPerformSort();
		else if (f==sn.XSL_PRESERVE_SPACE) 		e=new XSLPreserveSpace();
		else if (f==sn.XSL_PROCESSING_INSTRUCTION) e=new XSLProcessingInstruction();
		else if (f==sn.XSL_RESULT_DOCUMENT)		e=new XSLResultDocument();
		else if (f==sn.XSL_SEQUENCE)			e=new XSLSequence();
		else if (f==sn.XSL_SORT) 				e=new XSLSort();
//		else if (f==sn.XSL_SORT_KEY) 			e=new XSLSortKey();
		else if (f==sn.XSL_STRIP_SPACE) 		e=new XSLPreserveSpace();
		else if (f==sn.XSL_STYLESHEET) 			e=new XSLStyleSheet();
		else if (f==sn.XSL_TEMPLATE) 			e=new XSLTemplate();
		else if (f==sn.XSL_TEXT) 				e=new XSLText();
		else if (f==sn.XSL_TRANSFORM) 			e=new XSLStyleSheet();
		else if (f==sn.XSL_VALUE_OF) 			e=new XSLValueOf();
		else if (f==sn.XSL_VARIABLE) 			e=new XSLVariable();
		else if (f==sn.XSL_WITH_PARAM) 			e=new XSLWithParam();
		else if (f==sn.XSL_WHEN) 				e=new XSLWhen();

		return e;
	}

	/**
	* Make a SAXON extension element
	*/

	private StyleElement makeSaxonElement(int f) {

		StyleElement e = null;

		if (f==sn.SAXON_ASSIGN) 			e=new SaxonAssign();
		else if (f==sn.SAXON_ENTITY_REF) 	e=new SaxonEntityRef();
		else if (f==sn.SAXON_CALL_TEMPLATE)	e=new SaxonCallTemplate();
		else if (f==sn.SAXON_COLLATION) 	e=new SaxonCollation();
		else if (f==sn.SAXON_DOCTYPE) 		e=new SaxonDoctype();
		else if (f==sn.SAXON_IMPORT_QUERY)	e=new SaxonImportQuery();
		else if (f==sn.SAXON_SCRIPT) 		e=new SaxonScript();
		else if (f==sn.SAXON_WHILE) 		e=new SaxonWhile();
		return e;
	}

    /**
    * Get the factory class for user extension elements
    * If there is no appropriate class, return null
    */

    private ExtensionElementFactory getFactory(short uriCode) {
    	String uri = namePool.getURIFromNamespaceCode(uriCode);
        int lastSlash = uri.lastIndexOf('/');
        if (lastSlash<0 || lastSlash==uri.length()-1) {
            return null;
        }
        String factoryClass = uri.substring(lastSlash+1);
        ExtensionElementFactory factory;

        try {
            factory = (ExtensionElementFactory)Loader.getInstance(factoryClass);
        } catch (Exception err) {
            return null;
        }
        return factory;
    }

    /**
    * Method to support the element-available() function
    */

    public boolean isElementAvailable(String uri, String localName) {
    	int fingerprint = namePool.getFingerprint(uri, localName);
    	if (uri.equals(NamespaceConstant.XSLT)) {
    		if (fingerprint==-1) return false; 	// all names are pre-registered
    		StyleElement e = makeXSLElement(fingerprint);
    		if (e!=null) return e.isInstruction();
    	}

		if (uri.equals(NamespaceConstant.SAXON)) {
			if (fingerprint==-1) return false;	// all names are pre-registered
	    	StyleElement e = makeSaxonElement(fingerprint);
	    	if (e!=null) return e.isInstruction();
	    }
        if (!allowExtensions) {
            // extension elements are disabled
            return false;
        }
    	short uriCode = namePool.getCodeForURI(uri);
        ExtensionElementFactory factory = getFactory(uriCode);
        if (factory==null) return false;
        Class actualClass = factory.getExtensionClass(localName);
        return (actualClass != null);

    }




}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
