package net.sf.saxon.style;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.Attribute;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.NamespaceContext;
import net.sf.saxon.instruct.FixedAttribute;
import net.sf.saxon.om.*;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;

import javax.xml.transform.TransformerConfigurationException;

/**
* xsl:attribute element in stylesheet. <br>
*/

public final class XSLAttribute extends XSLStringConstructor {

    private Expression attributeName;
    private Expression namespace = null;
    private int validationAction = Validation.PRESERVE;
    private int annotation = Type.ANY_SIMPLE_TYPE;
//    private boolean disable = false;

    public void prepareAttributes() throws TransformerConfigurationException {

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		String nameAtt = null;
		String namespaceAtt = null;
        String selectAtt = null;
		String validationAtt = null;
		String typeAtt = null;
//		String disableAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		nameAtt = atts.getValue(a).trim();
        	} else if (f==sn.NAMESPACE) {
        		namespaceAtt = atts.getValue(a).trim();
        	} else if (f==sn.SELECT) {
        		selectAtt = atts.getValue(a).trim();
        	} else if (f==sn.VALIDATION) {
        		validationAtt = atts.getValue(a).trim();
        	} else if (f==sn.TYPE) {
        		typeAtt = atts.getValue(a).trim();
//        	} else if (f==sn.DISABLE_OUTPUT_ESCAPING) {
//        		disableAtt = atts.getValue(a).trim();
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (nameAtt==null) {
            reportAbsence("name");
            return;
        } else {
            attributeName = makeAttributeValueTemplate(nameAtt);
            if (attributeName instanceof StringValue) {
                if (!Name.isQName(nameAtt)) {
                    compileError("Attribute name is not a valid QName");
                    // prevent a duplicate error message...
                    attributeName = new StringValue("saxon-error-attribute");
                }
            }
        }

        if (namespaceAtt!=null) {
            namespace = makeAttributeValueTemplate(namespaceAtt);
        }

        if (selectAtt!=null) {
            select = makeExpression(selectAtt);
        }

        if (validationAtt!=null) {
            validationAction = Validation.getCode(validationAtt);
            if (validationAction == Validation.STRICT) {
                compileError("validation=strict is not yet supported");
            }
            if (validationAction == Validation.LAX) {
                compileError("validation=lax is not yet supported");
            }
            if (validationAction == Validation.PRESERVE) {
                compileError("validation=preserve is not yet supported");
            }
            if (validationAction == Validation.STRIP) {
                compileError("validation=strip is not yet supported");
            }
            if (validationAction == Validation.INVALID) {
                compileError("invalid value of validation attribute");
            }
        }

        if (typeAtt!=null) {
            annotation = getTypeAnnotationCode(typeAtt);
        }

        if (typeAtt != null && validationAtt != null) {
            compileError("validation and type attributes are mutually exclusive");
        }

//        if (disableAtt!=null) {
//            if (disableAtt.equals("yes")) {
//                disable = true;
//            } else if (disableAtt.equals("no")) {
//                // no-op
//            } else {
//                compileError("disable-output-escaping must be 'yes' or 'no'");
//            }
//        }

    }

    public void validate() throws TransformerConfigurationException {
        if (!(getParentNode() instanceof XSLAttributeSet)) {
            checkWithinTemplate();
        }
        attributeName = typeCheck("name", attributeName);
        namespace = typeCheck("namespace", namespace);
        select = typeCheck("select", select);
        super.validate();
    }

    public Instruction compile() throws TransformerConfigurationException {
        NamespaceContext nsContext = null;

        // deal specially with the case where the attribute name is known statically

        if (attributeName instanceof StringValue) {
            String qName = ((StringValue)attributeName).getStringValue().trim();
            String[] parts;
            try {
                parts = Name.getQNameParts(qName);
            } catch (QNameException e) {
                compileError("Invalid attribute name: " + qName);
                return null;
            }

            if (qName.equals("xmlns")) {
                if (namespace==null) {
                    compileError("Invalid attribute name: " + qName);
                    return null;
                }
            }
            if (parts[0].equals("xmlns")) {
                if (namespace==null) {
                    compileError("Invalid attribute name: " + qName);
                    return null;
                } else {
                    // ignore the prefix "xmlns"
                    parts[0] = "";
//                    parts[0] = null;
//                    if (namespace instanceof StringValue) {
//                        parts[0] = getTargetNamePool().suggestPrefixForURI(
//                                ((StringValue)namespace).getStringValue());
//                    }
//                    if (parts[0] == null) {
//                        parts[0] = "ns0";
//                    }
                }
            }
            if (namespace==null) {
                String nsuri = "";
                if (!parts[0].equals("")) {
                    try {
                        nsuri = getURIForPrefix(parts[0], false);
                    } catch (NamespaceException err) {
                        compileError(err.getMessage());
                        return null;
                    }
                }
                int nameCode = getTargetNamePool().allocate(parts[0], nsuri, parts[1]);
                FixedAttribute inst = new FixedAttribute(nameCode, annotation);
                compileContent(inst);
                return inst;
            } else if (namespace instanceof StringValue) {
                String nsuri = ((StringValue)namespace).getStringValue();
                if (nsuri.equals("")) {
                    parts[0] = "";
                } else if (parts[0].equals("")) {
                    // Need to choose an arbitrary prefix
                    // First see if the requested namespace is declared in the stylesheet
                    AxisIterator iter = iterateAxis(Axis.NAMESPACE);
                    while (iter.hasNext()) {
                        NodeInfo ns = (NodeInfo)iter.next();
                        if (ns.getStringValue().equals(nsuri)) {
                            parts[0] = ns.getLocalPart();
                            break;
                        }
                    }
                    // Otherwise see the URI is known to the namepool
                    if (parts[0].equals("")) {
                        String p = getTargetNamePool().suggestPrefixForURI(
                                ((StringValue)namespace).getStringValue());
                        if (p != null) {
                            parts[0] = p;
                        }
                    }
                    // Otherwise choose something arbitrary. This will get changed
                    // if it clashes with another attribute
                    if (parts[0].equals("")) {
                        parts[0] = "ns0";
                    }
                }
                int nameCode = getTargetNamePool().allocate(parts[0], nsuri, parts[1]);
                FixedAttribute inst = new FixedAttribute(nameCode, annotation);
                compileContent(inst);
                return inst;
            }
        } else {
            // if the namespace URI must be deduced at run-time from the attribute name
            // prefix, we need to save the namespace context of the instruction

            if (namespace==null) {
                nsContext = makeNamespaceContext();
            }
        }

        Attribute inst = new Attribute( attributeName,
                                        namespace,
                                        nsContext,
                                        annotation,
                                        false );
//                                        disable );
        compileContent(inst);
        return inst;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
