package net.sf.saxon.style;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceException;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

/**
* An xsl:character-map declaration in the stylesheet. <br>
*/

public class XSLCharacterMap extends StyleElement {

    int fingerprint;
                // the name of this character map

    String use;
                // the value of the use-character-maps attribute, as supplied

    List characterMapElements = null;
                // list of XSLCharacterMap objects referenced by this one

    boolean validated = false;
                // set to true once validate() has been called

    boolean redundant = false;
                // set to true if another character-map overrrides this one

    /**
     * Get the fingerprint of the name of this character map
     * @return the fingerprint value
     */

    public int getCharacterMapFingerprint() {
        return fingerprint;
    }

    /**
     * Test whether this character map is redundant (because another with the
     * same name has higher import precedence). Note that a character map is not
     * considered redundant simply because it is not referenced in an xsl:output
     * declaration; we allow character-maps to be selected at run-time using the
     * setOutputProperty() API.
     */

    public boolean isRedundant() {
        return redundant;
    }

    /**
     * Validate the attributes on this instruction
     * @throws TransformerConfigurationException
     */

    public void prepareAttributes() throws TransformerConfigurationException {

		String name = null;
		use = null;

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		name = atts.getValue(a).trim();
        	} else if (f==sn.USE_CHARACTER_MAPS) {
        		use = atts.getValue(a);
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (name==null) {
            reportAbsence("name");
            return;
        }

        try {
            int nameCode = makeNameCode(name);
            fingerprint = nameCode & 0xfffff;
        } catch (NamespaceException err) {
            compileError(err.getMessage());
        } catch (XPathException err) {
            compileError(err.getMessage());
        }

    }

    public void validate() throws TransformerConfigurationException {

        if (validated) return;

        // check that this is a top-level declaration

        checkTopLevel();

        // check that the only children are xsl:output-character elements

        AxisIterator kids = iterateAxis(Axis.CHILD);
        while (kids.hasNext()) {
            if (!(kids.next() instanceof XSLOutputCharacter)) {
                compileError("Only xsl:output-character is allowed within xsl:character-map");
            }
        }

        // check that there isn't another character-map with the same name and import
        // precedence

        XSLStyleSheet principal = getPrincipalStyleSheet();
        XSLCharacterMap other = principal.getCharacterMap(fingerprint);
        if (other != this) {
            if (this.getPrecedence() == other.getPrecedence()) {
                compileError("There are two character-maps with the same name and import precedence");
            } else if (this.getPrecedence() < other.getPrecedence()) {
                redundant = true;
            }
        }

        // validate the use-character-maps attribute

        if (use!=null) {

            // identify any character maps that this one refers to

            characterMapElements = new ArrayList();
            StringTokenizer st = new StringTokenizer(use);

            while (st.hasMoreTokens()) {
                String displayname = st.nextToken();
                try {
                    String[] parts = Name.getQNameParts(displayname);
                    String uri = getURIForPrefix(parts[0], true);
                    int nameCode = getTargetNamePool().allocate(parts[0], uri, parts[1]);
                    XSLCharacterMap ref =
                            principal.getCharacterMap(nameCode & 0xfffff);
                    if (ref == null) {
                        compileError("No character-map named '" + displayname + "' has been defined");
                    } else {
                        characterMapElements.add(ref);
                    }
                } catch (QNameException err) {
                    compileError("Invalid character-map name. " + err.getMessage());
                } catch (NamespaceException err) {
                    compileError(err.getMessage());
                }
            }

            // check for circularity

            for (Iterator it=characterMapElements.iterator(); it.hasNext();) {
                ((XSLCharacterMap)it.next()).checkCircularity(this);
            }
        }

        validated = true;
    }

    /**
    * Check for circularity: specifically, check that this attribute set does not contain
    * a direct or indirect reference to the one supplied as a parameter
    */

    private void checkCircularity(XSLCharacterMap origin) throws TransformerConfigurationException {
        if (this==origin) {
            compileError("The definition of the character map is circular");
            characterMapElements = null;    // for error recovery
        } else {
            if (!validated) {
                // if this attribute set isn't validated yet, we don't check it.
                // The circularity will be detected when the last attribute set in the cycle
                // gets validated
                return;
            }
            if (characterMapElements != null) {
                for (Iterator it=characterMapElements.iterator(); it.hasNext();) {
                    ((XSLCharacterMap)it.next()).checkCircularity(origin);
                }
            }
        }
    }

    /**
     * Assemble all the mappings defined by this character map, adding them to a
     * HashMap that maps integer codepoints to strings
     */

    public void assemble(HashMap map) {
        if (characterMapElements != null) {
            for (int i = 0; i < characterMapElements.size(); i++) {
                XSLCharacterMap charmap = (XSLCharacterMap) characterMapElements.get(i);
                charmap.assemble(map);
            }
        }
        AxisIterator kids = iterateAxis(Axis.CHILD);
        while (kids.hasNext()) {
            XSLOutputCharacter oc = (XSLOutputCharacter)kids.next();
            map.put(new Integer(oc.getCodePoint()), oc.getReplacementString());
        }
    }

    /**
    * Get a pattern that will match the context node
    * @return null if no such pattern can be identified
    */

//    public Pattern getContextNodePattern() {
//        // Context node is not known statically
//        return null;
//    }

    public Instruction compile() throws TransformerConfigurationException {

            return null;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
