package net.sf.saxon.style;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.AttributeSet;
import net.sf.saxon.instruct.Element;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.NamespaceContext;
import net.sf.saxon.instruct.FixedElement;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceException;
import net.sf.saxon.om.Validation;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;


/**
* An xsl:element element in the stylesheet. <br>
*/

public class XSLElement extends StyleElement {

    private Expression elementName;
    private Expression namespace = null;
    private String use;
    private AttributeSet[] attributeSets = null;
    private int validationAction = Validation.PRESERVE;
    private int annotation = Type.UNTYPED_ANY;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws TransformerConfigurationException {

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		String nameAtt = null;
		String namespaceAtt = null;
		String validationAtt = null;
		String typeAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		nameAtt = atts.getValue(a).trim();
        	} else if (f==sn.NAMESPACE) {
        		namespaceAtt = atts.getValue(a);
        	} else if (f==sn.VALIDATION) {
        		validationAtt = atts.getValue(a).trim();
        	} else if (f==sn.TYPE) {
        		typeAtt = atts.getValue(a).trim();
        	} else if (f==sn.USE_ATTRIBUTE_SETS) {
        		use = atts.getValue(a);
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (nameAtt==null) {
            reportAbsence("name");
        } else {
            elementName = makeAttributeValueTemplate(nameAtt);
            if (elementName instanceof StringValue) {
                if (!Name.isQName(nameAtt)) {
                    compileError("Element name is not a valid QName");
                    // to prevent duplicate error messages:
                    elementName = new StringValue("saxon-error-element");
                }
            }
        }

        if (namespaceAtt!=null) {
            namespace = makeAttributeValueTemplate(namespaceAtt);
        }

        if (validationAtt!=null) {
            validationAction = Validation.getCode(validationAtt);
            if (validationAction == Validation.STRICT) {
                compileError("validation=strict is not yet supported");
            }
            if (validationAction == Validation.LAX) {
                compileError("validation=lax is not yet supported");
            }
            if (validationAction == Validation.PRESERVE) {
                compileError("validation=preserve is not yet supported");
            }
            if (validationAction == Validation.STRIP) {
                compileError("validation=strip is not yet supported");
            }
            if (validationAction == Validation.INVALID) {
                compileError("invalid value of validation attribute");
            }
        }

        if (typeAtt!=null) {
            annotation = getTypeAnnotationCode(typeAtt);
        }

        if (typeAtt != null && validationAtt != null) {
            compileError("validation and type attributes are mutually exclusive");
        }
    }

    public void validate() throws TransformerConfigurationException {
        checkWithinTemplate();
        if (use!=null) {
            attributeSets = getAttributeSets(use, null);        // find any referenced attribute sets
        }
        elementName = typeCheck("name", elementName);
        namespace = typeCheck("namespace", namespace);
    }

    public Instruction compile() throws TransformerConfigurationException {

        NamespaceContext nsContext = null;

        // deal specially with the case where the element name is known statically

        if (elementName instanceof StringValue) {
            String qName = ((StringValue)elementName).getStringValue();

            String[] parts;
            try {
                parts = Name.getQNameParts(qName);
            } catch (QNameException e) {
                compileError("Invalid element name: " + qName);
                return null;
            }

            String nsuri = null;
            if (namespace instanceof StringValue) {
                nsuri = ((StringValue)namespace).getStringValue();
                if (nsuri.equals("")) {
                    parts[0] = "";
                }
            } else if (namespace==null) {
                try {
          		    nsuri = getURIForPrefix(parts[0], true);
          		} catch (NamespaceException err) {
          		    compileError(err.getMessage());
          		    return null;
          		}
            }
            if (nsuri != null) {
                int nameCode = getTargetNamePool().allocate(parts[0], nsuri, parts[1]);
                FixedElement inst =
                        new FixedElement(nameCode, null, attributeSets, annotation);
                compileChildren(inst);
                return inst;
            }
        } else {
            // if the namespace URI must be deduced at run-time from the element name
            // prefix, we need to save the namespace context of the instruction

            if (namespace==null) {
                nsContext = makeNamespaceContext();
            }
        }

        try {
            Instruction inst = new Element( elementName,
                                            namespace,
                                            nsContext,
                                            attributeSets,
                                            annotation );
            compileChildren(inst);
            return inst;
        } catch (XPathException e) {
            compileError(e.getMessage());
            return null;
        }
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
