package net.sf.saxon.style;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.RoleLocator;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.instruct.DocumentInstr;
import net.sf.saxon.instruct.GeneralVariable;
import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceException;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Axis;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;

/**
* This class defines common behaviour across xsl:variable, xsl:param, and xsl:with-param
*/

public abstract class XSLGeneralVariable extends StyleElement  {

    protected int variableFingerprint = -1;
    protected Expression select = null;
    protected SequenceType requiredType = null;
    protected String constantText = null;
    protected boolean global;
    protected Procedure procedure = null;  // used only for global variables
    protected boolean assignable = false;
    protected boolean redundant = false;
    protected boolean requiredParam = false;
    protected boolean tunnel = false;
    //protected Class validationClass = null;
        // TODO: this class uses a different mechanism from xsl:result-document
        // to set type-information. Both could do it the same way, using OutputProperties.
    private boolean textonly;

    //private static final int[] treeSizeParameters = {50, 10, 5, 200};
        // estimated size of a temporary tree: {nodes, attributes, namespaces, characters}

    /**
     * Determine the type of item returned by this instruction (only relevant if
     * it is an instruction).
     * @return the item type returned. This is EMPTY for a variable: we are not
     * interested in the type of the variable, but in what the xsl:variable constributes
     * to the result of the sequence constructor it is part of.
     */

    protected int getReturnedItemType() {
        return Type.EMPTY;
    }
    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    protected boolean allowsNameAttribute() {
        return true;
    }

    protected boolean allowsAsAttribute() {
        return true;
    }

    protected boolean allowsTunnelAttribute() {
        return false;
    }

    protected boolean allowsValue() {
        return true;
    }

    protected boolean allowsRequired() {
        return false;
    }

    /**
    * Test whether it is permitted to assign to the variable using the saxon:assign
    * extension element. This will only be true if the extra attribute saxon:assignable="yes"
    * is present.
    */

    public boolean isAssignable() {
        return assignable;
    }

    public boolean isTunnelParam() {
        return tunnel;
    }

    /**
    * Get the display name of the variable.
    */

    public String getVariableName() {
    	return getAttributeValue("", "name");
    }

    protected Expression getSelectExpression() {
        return select;
    }

    protected void setSelectExpression(Expression select) {
        this.select = select;
    }

    /**
    * Mark this global variable as redundant. This is done before prepareAttributes is called.
    */

    public void setRedundant() {
        redundant = true;
    }

    /**
    * Get the fingerprint of the variable name
    */

    public int getVariableFingerprint() {

        // if an expression has a forwards reference to this variable, getVariableFingerprint() can be
        // called before prepareAttributes() is called. We need to allow for this. But we'll
        // deal with any errors when we come round to processing this attribute, to avoid
        // duplicate error messages

        // TODO: this won't establish the requiredType in time to optimize an expression containing
        // a forwards reference to the variable

        if (variableFingerprint==-1) {
        	StandardNames sn = getStandardNames();
            String nameAttribute = getAttributeValue(sn.NAME & 0xfffff);
            if (nameAttribute==null) {
                return -1;              // we'll report the error later
            }
            try {
                variableFingerprint = makeNameCode(nameAttribute.trim()) & 0xfffff;
            } catch (NamespaceException err) {
                variableFingerprint = -1;
            } catch (XPathException err) {
                variableFingerprint = -1;
            }
        }
        return variableFingerprint;
    }

    public void prepareAttributes() throws TransformerConfigurationException {

        getVariableFingerprint();

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		String selectAtt = null;
        String assignAtt = null;
        String nameAtt = null;
        String asAtt = null;
        String requiredAtt = null;
        String tunnelAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		nameAtt = atts.getValue(a).trim();
        	} else if (f==sn.SELECT) {
        		selectAtt = atts.getValue(a);
        	} else if (f==sn.AS && allowsAsAttribute()) {
        		asAtt = atts.getValue(a);
        	} else if (f==sn.REQUIRED && allowsRequired()) {
        		requiredAtt = atts.getValue(a).trim();
            } else if (f==sn.TUNNEL && allowsTunnelAttribute()) {
        		tunnelAtt = atts.getValue(a).trim();
        	} else if (f==sn.SAXON_ASSIGNABLE && this instanceof XSLVariableDeclaration) {
        		assignAtt = atts.getValue(a).trim();
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (nameAtt==null) {
            if (allowsNameAttribute()) {
                reportAbsence("name");
            }
        } else if (!Name.isQName(nameAtt)) {
            compileError("Variable name must be a valid QName");
        }

        if (selectAtt!=null) {
            if (!allowsValue()) {
                compileError("Function parameters cannot have a default value");
            }
            select = makeExpression(selectAtt);
        }

        if (assignAtt!=null && assignAtt.equals("yes")) {
            assignable=true;
        }

        if (requiredAtt!=null) {
            if (requiredAtt.equals("yes")) {
                requiredParam = true;
            } else if (requiredAtt.equals("no")) {
                requiredParam = false;
            } else {
                compileError("The attribute 'required' must be set to 'yes' or 'no'");
            }
        }

        if (tunnelAtt!=null) {
            if (tunnelAtt.equals("yes")) {
                tunnel = true;
            } else if (tunnelAtt.equals("no")) {
                tunnel = false;
            } else {
                compileError("The attribute 'tunnel' must be set to 'yes' or 'no'");
            }
        }

        if (asAtt!=null) {
            requiredType = makeSequenceType(asAtt);
        }
    }

    public void validate() throws TransformerConfigurationException {
        global = (getParentNode() instanceof XSLStyleSheet);

        if (global) {
            procedure = new Procedure();
        }
        if (select!=null && hasChildNodes()) {
            compileError("An " + getDisplayName() + " element with a select attribute must be empty");
        }

        checkAgainstRequiredType(requiredType);

        if (select==null && allowsValue()) {
            textonly = true;
            AxisIterator kids = iterateAxis(Axis.CHILD);
            if (!kids.hasNext()) {
                if (requiredType == null) {
                    select = StringValue.EMPTY_STRING;
                } else {
                    try {
                        select = StringValue.EMPTY_STRING.convert(requiredType.getPrimaryType());
                    } catch (XPathException err) {
                        // this error isn't fatal if we're dealing with the default of an
                        // xsl:param, because the default might never be used
                        if (this instanceof XSLParam) {
                            select = StringValue.EMPTY_STRING;
                        } else {
                            compileError(err);
                        }
                    }
                }
            } else {
                NodeInfo first = (NodeInfo)kids.next();
                if (!kids.hasNext()) {
                    // there is exactly one child node
                    if (first.getItemType() == Type.TEXT) {
                        // it is a text node: optimize for this case
                        constantText = first.getStringValue();
                    }
                }

                // Determine if the temporary tree can only contain text nodes
                textonly = (getCommonChildItemType() == Type.TEXT);
            }
        }
        select = typeCheck("select", select);
    }

    /**
     * Method called for parameters of call-template to check the type of the actual
     * parameter against the type of the required parameter
     * @param required The type required by the signature of the called template
     */

    protected void checkAgainstRequiredType(SequenceType required)
    throws TransformerConfigurationException {
        try {
            RoleLocator role =
                    new RoleLocator(RoleLocator.VARIABLE, getVariableName(), 0);
            if (requiredType!=null) {
                // check that the expression is consistent with the required type

                if (select != null) {
                    select = TypeChecker.staticTypeCheck(select, requiredType, false, role);
                } else {
                    // TODO: check the type of the instruction sequence statically
                }
            }
        } catch (XPathException err) {
            compileError(err);
        }
    }

    /**
    * Initialize - common code called from the compile() method of all subclasses
    */

    protected void initializeInstruction(GeneralVariable var)
    throws TransformerConfigurationException {

        var.init(   select,
                    requiredType,
                    variableFingerprint);
        var.setGlobal(global);
        var.setAssignable(assignable);
        var.setRequiredParam(requiredParam);
        var.setTunnel(tunnel);
        var.setContainsLocals(global && procedure.getNumberOfVariables()>0);

        // handle the "temporary tree" case by creating a Document sub-instruction
        // to construct and return a document node.
        if (hasChildNodes() && requiredType==null) {
            DocumentInstr doc = new DocumentInstr(textonly, constantText, getBaseURI());
            compileChildren(doc);
            var.setSelect(doc);
        } else {
            compileChildren(var);
        }
    }

    /**
    * Get associated Procedure (for details of stack frame, if this is a global variable containing
    * local variable declarations)
    */

    public Procedure getProcedure() {
        return procedure;
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
