package net.sf.saxon.style;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.trans.KeyManager;
import net.sf.saxon.trans.KeyDefinition;
import net.sf.saxon.om.Name;
import net.sf.saxon.om.NamespaceException;
import net.sf.saxon.om.QNameException;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.pattern.Pattern;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.expr.RoleLocator;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.xpath.XPathException;
import javax.xml.transform.TransformerConfigurationException;
import java.util.Comparator;
import java.text.Collator;

/**
* Handler for xsl:key elements in stylesheet. <br>
*/

public class XSLKey extends StyleElement  {

    private int keyFingerprint = -1;
    private Pattern match;
    private Expression use;
    private int itemType;
    private String collationName;

    public void prepareAttributes() throws TransformerConfigurationException {

        String nameAtt = null;
        String matchAtt = null;
        String useAtt = null;
        String typeAtt = null;

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.NAME) {
        		nameAtt = atts.getValue(a).trim();
        	} else if (f==sn.USE) {
        		useAtt = atts.getValue(a);
        	} else if (f==sn.MATCH) {
        		matchAtt = atts.getValue(a);
        	} else if (f==sn.AS) {
        		typeAtt = atts.getValue(a);
        	} else if (f==sn.COLLATION) {
        		collationName = atts.getValue(a).trim();
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (nameAtt==null) {
            reportAbsence("name");
            return;
        }
        try {
            keyFingerprint = makeNameCode(nameAtt) & 0xfffff;
        } catch (NamespaceException err) {
            compileError(err.getMessage());
        } catch (XPathException err) {
            compileError(err.getMessage());
        }

        if (matchAtt==null) {
            reportAbsence("match");
            matchAtt = "*";
        }
        match = makePattern(matchAtt);

        if (useAtt==null) {
            reportAbsence("use");
            useAtt = ".";
        }
        use = makeExpression(useAtt);

        if (typeAtt==null) {
            itemType = Type.STRING;
        } else {
            if (!Name.isQName(typeAtt)) {
                compileError("Data type '" + typeAtt + "'is not a valid QName'");
            }
            try {
                String[] parts = Name.getQNameParts(typeAtt);
                String uri = getURIForPrefix(parts[0], false);
                itemType = Type.getNamedType(uri, parts[1]);
                if (!Type.isSubType(itemType, Type.ATOMIC)) {
                    compileError("The type of the key values must be an atomic type");
                }
            } catch (NamespaceException err) {
                compileError("Invalid type. " + err.getMessage());
            } catch (QNameException err) {
                compileError("Invalid type. " + err.getMessage());
            } catch (XPathException err) {
                compileError(err);
            }
        }
    }

    public void validate() throws TransformerConfigurationException {
        checkTopLevel();
        checkEmpty();
        try {
            if (use==null) return;  // error already reported
            RoleLocator role =
                new RoleLocator(RoleLocator.INSTRUCTION, "xsl:key/use", 0);
            use = TypeChecker.staticTypeCheck(
                            use,
                            new SequenceType(itemType, Type.ITEM, StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE),
                            false, role);
        } catch (XPathException err) {
            compileError(err);
        }

        use = typeCheck("use", use);
        match = typeCheck("match", match);

        // TODO: check that the collation for this key is consistent
        // with other key definitions of the same name

        // TODO: allow keys of any type including mixed types; remove "as" attribute
    }

    public Instruction compile() throws TransformerConfigurationException {

        Collator collator = null;
        if (collationName != null) {
            Comparator comp = getPrincipalStyleSheet().findCollation(collationName);
            if (comp==null) {
                 compileError("The collation name '" + collationName + "' has not been defined");
            }
            if (!(comp instanceof Collator)) {
                compileError("The collation used for xsl:key must be a java.text.Collator");
            }
            collator = (Collator)comp;
        }

        KeyManager km = getPrincipalStyleSheet().getKeyManager();
        KeyDefinition keydef = new KeyDefinition(match, use, itemType, collator);
        km.setKeyDefinition(keyFingerprint, keydef);
        return null;
    }

}
//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
