package net.sf.saxon.style;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.Sequence;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.Type;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Navigator;
import net.sf.saxon.sort.SortKeyDefinition;
import net.sf.saxon.sort.SortExpression;

import javax.xml.transform.TransformerConfigurationException;


/**
* Handler for xsl:perform-sort elements in stylesheet (XSLT 2.0). <br>
*/

public class XSLPerformSort extends StyleElement {

    Expression select = null;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain an xsl:fallback
    * instruction
    */

    public boolean mayContainFallback() {
        return true;
    }

    /**
     * Determine the type of item returned by this instruction (only relevant if
     * it is an instruction).
     * @return the item type returned
     */

    protected int getReturnedItemType() {
        if (select==null) {
            return Type.ITEM;
        } else {
            return select.getItemType();
        }
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return false;
    }

    public void prepareAttributes() throws TransformerConfigurationException {

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		String selectAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.SELECT) {
        		selectAtt = atts.getValue(a);
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (selectAtt==null) {
            reportAbsence("select");
        } else {
            select = makeExpression(selectAtt);
        }

    }

    public void validate() throws TransformerConfigurationException {
        checkWithinTemplate();

        // only xsl:sort and xsl:fallback are allowed in the content

        AxisIterator kids = iterateAxis(Axis.CHILD);
        boolean sortFound = false;
        while (kids.hasNext()) {
            NodeInfo child = (NodeInfo)kids.next();
            if (child instanceof XSLSort) {
                sortFound = true;
            } else if(child instanceof XSLFallback) {
                // no-op
            } else if (child.getItemType() == Type.TEXT) {
                    // with xml:space=preserve, white space nodes may still be there
                if (!Navigator.isWhite(child.getStringValue())) {
                    compileError(
                        "Character data is not allowed within xsl:perform-sort");
                }
            } else {
                compileError("Invalid element within xsl:perform-sort");
            }
        }
        if (!sortFound) {
            compileError("xsl:perform-sort must have at least one xsl:sort child");
        }
        select = typeCheck("select", select);
    }

    public Instruction compile() throws TransformerConfigurationException {
        SortKeyDefinition[] sortKeys = makeSortKeys();
        Instruction inst = new Sequence(new SortExpression(select, sortKeys),
                                        SequenceType.ANY_SEQUENCE);
        compileChildren(inst);
        return inst;
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
