package net.sf.saxon.style;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.event.SaxonOutputKeys;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.ResultDocument;
import net.sf.saxon.om.NamespaceException;
import net.sf.saxon.om.Validation;
import net.sf.saxon.tree.AttributeCollection;

import javax.xml.transform.TransformerConfigurationException;
import java.util.Properties;

/**
* An xsl:result-document (formerly saxon:output) element in the stylesheet. <BR>
* The xsl:result-document element takes an attribute href="filename". The filename will
* often contain parameters, e.g. {position()} to ensure that a different file is produced
* for each element instance. <BR>
* There is a further attribute "name" which determines the format of the
* output file, it identifies the name of an xsl:output element containing the output
* format details.
*/

public class XSLResultDocument extends StyleElement {

    private Expression href;
    private int format = -1;     // fingerprint of required xsl:output element
    private String typeInformation = null;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws TransformerConfigurationException {
		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

        String formatAttribute = null;
        String hrefAttribute = null;
        String validationAtt = null;

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.FORMAT) {
        		formatAttribute = atts.getValue(a).trim();
        	} else if (f==sn.HREF) {
        		hrefAttribute = atts.getValue(a).trim();
            } else if (f==sn.VALIDATION) {
                validationAtt = atts.getValue(a).trim();
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (hrefAttribute==null) {
            reportAbsence("href");
        } else {
            href = makeAttributeValueTemplate(hrefAttribute);
        }

        if (formatAttribute!=null) {
            try {
                format = makeNameCode(formatAttribute) & 0xfffff;
            } catch (NamespaceException err) {
                compileError(err.getMessage());
            } catch (XPathException err) {
                compileError(err.getMessage());
            }
        }

        if (validationAtt==null) {
            typeInformation = "none";
        } else {
            int validationAction = Validation.getCode(validationAtt);
            if (validationAction == Validation.STRICT) {
                compileError("validation=strict is not yet supported");
            }
            if (validationAction == Validation.LAX) {
                compileError("validation=lax is not yet supported");
            }
            if (validationAction == Validation.PRESERVE) {
                typeInformation = "preserve";
            }
            if (validationAction == Validation.STRIP) {
                typeInformation = "none";
            }
            if (validationAction == Validation.INVALID) {
                compileError("invalid value of validation attribute");
            }
        }
    }

    public void validate() throws TransformerConfigurationException {
        checkWithinTemplate();
        if (!getPreparedStyleSheet().getConfiguration().isAllowExternalFunctions()) {
            compileError("xsl:result-document is disabled when extension functions are disabled");
        }
        href = typeCheck("href", href);
    }

    public Instruction compile() throws TransformerConfigurationException {
        Properties props = null;
        try {
            props = getPrincipalStyleSheet().gatherOutputProperties(format);
        } catch (TransformerConfigurationException err) {
            compileError("Named output format has not been defined");
            return null;
        }
        if (!typeInformation.equals("preserve")) {
            props.setProperty(SaxonOutputKeys.TYPE_INFORMATION, typeInformation);
        }
        Instruction inst = new ResultDocument(props, href, getBaseURI());
        compileChildren(inst);
        return inst;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Additional Contributor(s): Brett Knights [brett@knightsofthenet.com]
//
