package net.sf.saxon.style;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.RoleLocator;
import net.sf.saxon.expr.TypeChecker;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.Sequence;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.xpath.XPathException;

import javax.xml.transform.TransformerConfigurationException;


/**
 * An xsl:sequence element in the stylesheet. <br>
 * The xsl:sequence element takes attributes:<ul>
 * <li>an optional attribute select="expression".</li>
 * <li>an optional attribute as="sequence type".</li>
 * </ul>
 */

public final class XSLSequence extends StyleElement {

    private Expression select;
    SequenceType requiredType = null;

    /**
    * Determine whether this node is an instruction.
    * @return true - it is an instruction
    */

    public boolean isInstruction() {
        return true;
    }

    /**
     * Determine the type of item returned by this instruction (only relevant if
     * it is an instruction).
     * @return the item type returned
     */

    protected int getReturnedItemType() {
        if (requiredType==null) {
            if (select!=null) {
                return select.getItemType();
            } else {
                return getCommonChildItemType();
            }
            //return Type.ITEM;
        } else {
            return requiredType.getPrimaryType();
        }
    }

    /**
    * Determine whether this type of element is allowed to contain a template-body
    * @return true: yes, it may contain a template-body
    */

    public boolean mayContainTemplateBody() {
        return true;
    }

    public void prepareAttributes() throws TransformerConfigurationException {

		String selectAtt = null;
		String asAtt = null;

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();

		for (int a=0; a<atts.getLength(); a++) {
			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.SELECT) {
        		selectAtt = atts.getValue(a);
        	} else if (f==sn.AS) {
        		asAtt = atts.getValue(a);
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }

        if (selectAtt!=null) {
            select = makeExpression(selectAtt);
        }

        if (asAtt != null) {
            requiredType = makeSequenceType(asAtt);
        }
    }

    public void validate() throws TransformerConfigurationException {
        checkWithinTemplate();
        if (select!=null && hasChildNodes()) {
            compileError("If xsl:sequence has a select attribute, it must not have child instructions");
        }
        select = typeCheck("select", select);
        if (select != null && requiredType != null) {
            try {
                RoleLocator role =
                        new RoleLocator(RoleLocator.INSTRUCTION, "xsl:sequence/select", 0);
                select = TypeChecker.staticTypeCheck(select, requiredType, false, role);
            } catch (XPathException err) {
                compileError(err);
            }
        }
    }

    /**
    * Mark tail-recursive calls on templates and functions.
    */

    public void markTailCalls() {
        if (requiredType != null) {
                // don't attempt tail call optimization if the return type needs checking
            StyleElement last = getLastChildInstruction();
            if (last != null) {
                last.markTailCalls();
            } else if (select != null) {
                ExpressionTool.markTailFunctionCalls(select);
            }
        }
    }


    public Instruction compile() throws TransformerConfigurationException {

        Instruction inst = new Sequence(select, requiredType);
        compileChildren(inst);
        return inst;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
