package net.sf.saxon.style;
import net.sf.saxon.instruct.Instruction;
import net.sf.saxon.instruct.Executable;
import net.sf.saxon.tree.AttributeCollection;
import net.sf.saxon.PreparedStyleSheet;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.query.XQueryFunction;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.Navigator;
import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.Axis;
import net.sf.saxon.trans.KeyManager;
import net.sf.saxon.trans.DecimalFormatManager;
import net.sf.saxon.trans.RuleManager;
import net.sf.saxon.trans.Mode;

import net.sf.saxon.sort.SortKeyDefinition;
import net.sf.saxon.sort.CodepointCollator;
import net.sf.saxon.sort.CollationFactory;

import net.sf.saxon.value.Type;
import javax.xml.transform.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Properties;
import java.util.Iterator;
/**
* An xsl:stylesheet or xsl:transform element in the stylesheet. <br>
* Note this element represents a stylesheet module, not necessarily
* the whole stylesheet.
*/

public class XSLStyleSheet extends StyleElement {

    // TODO: create the Executable object early, and maintain all these variables in the Executable,
    // instead of copying them to the Executable at the end.


                // the name pool used for names that will be needed at run-time, notably
                // the names used in XPath expressions and patterns, but also key names, parameter names, etc
    private NamePool targetNamePool;

                // true if this stylesheet was included by xsl:include, false if it is the
                // principal stylesheet or if it was imported
    private boolean wasIncluded = false;

                // the import precedence for top-level elements in this stylesheet
    private int precedence = 0;

                // the lowest precedence of any stylesheet imported by this one
    private int minImportPrecedence = 0;

                // the StyleSheet that included or imported this one; null for the principal stylesheet
    private XSLStyleSheet importer = null;

                // the PreparedStyleSheet object used to load this stylesheet
    private PreparedStyleSheet stylesheet;

                // the top-level elements in this logical stylesheet (after include/import)
    private List topLevel;

                // definitions of strip/preserve space action
    private Mode stripperRules = null;

                // definitions of template rules
    private RuleManager ruleManager;

                // definitions of keys
    private KeyManager keyManager;

                // definitions of decimal formats
    private DecimalFormatManager decimalFormatManager;

                // table of named collations defined in the stylesheet
    private HashMap collationTable;

                // default collating sequence
    private String defaultCollationName = null;

                // table of named sort key definitions
    private HashMap sortKeyTable;

                // table of character maps
    private HashMap characterMapIndex;

                // table of named templates
    private HashMap templateIndex = new HashMap();

                // table of functions imported from XQuery library modules
    private HashMap xqueryFunctionIndex;

                // media type (MIME type) of principal output
    //private String mediaType;

                // namespace aliases. This information is needed at compile-time only
    private int numberOfAliases = 0;
    private ArrayList namespaceAliasList = new ArrayList(5);
    private short[] aliasSCodes;
    //private short[] aliasRCodes;
    private int[] aliasNCodes;

                // count of the number of global parameters and variables
    private int numberOfVariables = 0;
                // index of global variables and parameters, by fingerprint
                // (overridden variables are excluded). Used at compile-time only
    private HashMap variableMap = new HashMap();

                // count of the maximum number of local variables in any template
    private int largestStackFrame = 0;

                // flag to indicate whether template names are needed at run-time
    private boolean runtimeTemplateMapRequired = false;

                // map of system ids for all stylesheet modules
    private HashMap stylesheetModules = new HashMap();

    /**
    * Create link to the owning PreparedStyleSheet object
    */

    public void setPreparedStyleSheet(PreparedStyleSheet sheet) {
        stylesheet = sheet;
        targetNamePool = sheet.getTargetNamePool();
        ruleManager = new RuleManager();
    }

    /**
    * Get the owning PreparedStyleSheet object
    */

    public PreparedStyleSheet getPreparedStyleSheet() {
        if (importer!=null) return importer.getPreparedStyleSheet();
        return stylesheet;
    }

    /**
    * Get the namepool to be used at run-time, this namepool holds the names used in
    * all XPath expressions and patterns
    */

    public NamePool getTargetNamePool() {
        return targetNamePool;
    }

    /**
    * Add a stylesheet module to the list of modules if it is not already there,
    * returning its integer code.
    */

    protected int putModuleNumber(String module) {
        Integer index = (Integer)stylesheetModules.get(module);
        if (index==null) {
            index = new Integer(stylesheetModules.size());
            stylesheetModules.put(module, index);
        }
        return index.intValue();
    }

    /**
    * Get the index number of a particular stylesheet module.
    * @throws NullPointerException if not found
    */

    protected int getModuleNumber(String module) {
        Integer index = (Integer)stylesheetModules.get(module);
        return index.intValue();
    }


    /**
    * Construct an array of stylesheet modules indexed by position
    */

    private String[] getModuleArray() {
        String[] array = new String[stylesheetModules.size()];
        for (Iterator it=stylesheetModules.keySet().iterator(); it.hasNext();) {
            String module = (String)it.next();
            int index = getModuleNumber(module);
            array[index] = module;
        }
        return array;
    }

    /**
    * Get the RuleManager which handles template rules
    */

    public RuleManager getRuleManager() {
        return ruleManager;
    }

    /**
    * Set a flag to indicate that template names are needed at run-time
    */

    public void setRequireRuntimeTemplateMap(boolean required) {
        runtimeTemplateMapRequired = required;
    }

    /**
    * Get the rules determining which nodes are to be stripped from the tree
    */

    protected Mode getStripperRules() {
        if (stripperRules==null) {
            stripperRules = new Mode();
        }
        return stripperRules;
    }


    /**
    * Create a Stripper which handles whitespace stripping definitions
    */

//    public Stripper newStripper() {
//        return new Stripper(stripperRules);
//    }

    /**
    * Determine whether this stylesheet does any whitespace stripping
    */

    public boolean stripsWhitespace() {
        StandardNames sn = getStandardNames();
        for (int i=0; i<topLevel.size(); i++) {
            NodeInfo s = (NodeInfo)topLevel.get(i);
            if (s.getFingerprint()==sn.XSL_STRIP_SPACE) {
                return true;
            }
        }
        return false;
    }

    /**
    * Get the KeyManager which handles key definitions
    */

    public KeyManager getKeyManager() {
        if (keyManager==null) {
            keyManager = new KeyManager();
        }
        return keyManager;
    }

    /**
    * Get the DecimalFormatManager which handles decimal-format definitions
    */

    public DecimalFormatManager getDecimalFormatManager() {
        if (decimalFormatManager==null) {
            decimalFormatManager = new DecimalFormatManager();
        }
        return decimalFormatManager;
    }

    /**
    * Register a named collation (actually a Comparator)
    */

    public void setCollation(String name, Comparator collation, boolean isDefault) {
        if (collationTable==null) {
            collationTable = new HashMap();
        }
        collationTable.put(name, collation);
        if (isDefault) {
            defaultCollationName = name;
        }
    }

    /**
    * Find a named collation. Note this method should only be used at compile-time, before declarations
    * have been pre-processed. After that time, use getCollation().
    * @param name identifies the name of the collation required; null indicates that the default
    * collation is required
    * @return null if the collation is not found
    */

    protected Comparator findCollation(String name) {

        // First try to find it in the table

        Comparator c = null;
        if (name==null) {
            name = defaultCollationName;
        }
        if (collationTable != null) {
            c = (Comparator)collationTable.get(name);
        }
        if (c!=null) return c;

        // At compile-time, the collation might not yet be in the table. So search for it

        XSLStyleSheet stylesheet = getPrincipalStyleSheet();
        List toplevel = stylesheet.getTopLevel();

        // search for a matching collation name, starting at the end in case of duplicates.
        // this also ensures we get the one with highest import precedence.
        for (int i=toplevel.size()-1; i>=0; i--) {
            if (toplevel.get(i) instanceof SaxonCollation) {
                SaxonCollation t = (SaxonCollation)toplevel.get(i);
                if (name==null && t.isDefaultCollation()) {
                    defaultCollationName = t.getCollationName();
                    return t.getCollator();
                } else if (t.getCollationName().equals(name)) {
                    return t.getCollator();
                }
            }
        }

        // if it's not defined in the stylesheet, it might be a standard URI

        if (name==null) {
            return null;
        }

        try {
            return CollationFactory.makeCollationFromURI(name);
        } catch (TransformerException e) {
            return null;
        }
    }

    /**
    * Get the name of the default collation
    */

    public String getDefaultCollationName() {
        if (defaultCollationName != null) {
            return defaultCollationName;
        }
        findCollation(null);
        if (defaultCollationName != null) {
            return defaultCollationName;
        }
        return CodepointCollator.URI;
    }

    /**
    * Register a named sort key
    */

    public void setSortSpecification(String uri, String localName, SortKeyDefinition[] sortKeys) {
        if (sortKeyTable==null) {
            sortKeyTable = new HashMap();
        }
        sortKeyTable.put(localName + "#" + uri, sortKeys);
    }

    /**
     * Get a character map, identified by the fingerprint of its name.
     * Search backwards through the stylesheet.
     * @param fingerprint The fingerprint of the character map name,
     * in the target namepool.
     * @return the identified character map, or null if not found
     */

    public XSLCharacterMap getCharacterMap(int fingerprint) {
        for (int i=topLevel.size()-1; i>=0; i--) {
            if (topLevel.get(i) instanceof XSLCharacterMap) {
                XSLCharacterMap t = (XSLCharacterMap)topLevel.get(i);
                if (t.getCharacterMapFingerprint() == fingerprint) {
                    return t;
                }
            }
        }
        return null;
    }

    /**
    * Set the import precedence of this stylesheet
    */

    public void setPrecedence(int prec) {
        precedence = prec;
    }

    /**
    * Get the import precedence of this stylesheet
    */

    public int getPrecedence() {
        if (wasIncluded) return importer.getPrecedence();
        return precedence;
    }

    /**
    * Get the minimum import precedence of this stylesheet, that is, the lowest precedence
    * of any stylesheet imported by this one
    */

    public int getMinImportPrecedence() {
        return minImportPrecedence;
    }

    /**
    * Set the minimum import precedence of this stylesheet, that is, the lowest precedence
    * of any stylesheet imported by this one
    */

    public void setMinImportPrecedence(int precedence) {
        minImportPrecedence = precedence;
    }

    /**
    * Set the StyleSheet that included or imported this one.
    */

    public void setImporter(XSLStyleSheet importer) {
        this.importer = importer;
    }

    /**
    * Get the StyleSheet that included or imported this one.
    * @return null if this is the principal stylesheet
    */

    public XSLStyleSheet getImporter() {
        return importer;
    }

    /**
    * Indicate that this stylesheet was included (by its "importer") using an xsl:include
    * statement as distinct from xsl:import
    */

    public void setWasIncluded() {
        wasIncluded = true;
    }

    /**
    * Determine whether this stylesheet was included (by its "importer") using an xsl:include
    * statement as distinct from xsl:import.
    */

//    public boolean wasIncluded() {
//        return wasIncluded;
//    }

    /**
    * Get the top level elements in this stylesheet, after applying include/import
    */

    public List getTopLevel() {
        return topLevel;
    }

    /**
    * Allocate a slot number for a global variable or parameter
    */

    public int allocateSlotNumber() {
        return numberOfVariables++;
    }

    /**
    * Ensure there is enough space for local variables or parameters in any template
    */

    public void allocateLocalSlots(int n) {
        if (n > largestStackFrame) {
            largestStackFrame = n;
        }
    }

    /**
    * Prepare the attributes on the stylesheet element
    */

    public void prepareAttributes() throws TransformerConfigurationException {

		StandardNames sn = getStandardNames();
		AttributeCollection atts = getAttributeList();
		for (int a=0; a<atts.getLength(); a++) {

			int nc = atts.getNameCode(a);
			int f = nc & 0xfffff;
			if (f==sn.VERSION) {
                processVersionAttribute(f);
			} else if (f==sn.ID) {
        		//
			} else if (f==sn.EXTENSION_ELEMENT_PREFIXES) {
        		//
			} else if (f==sn.EXCLUDE_RESULT_PREFIXES) {
        		//
        	} else {
        		checkUnknownAttribute(nc);
        	}
        }
        if (version==null) {
            reportAbsence("version");
        }
    }

    /**
    * Get the declared namespace alias for a given namespace URI code if there is one.
    * If there is more than one, we get the last.
    * @param uriCode The code of the uri used in the stylesheet.
    * @return The namespace code to be used (prefix in top half, uri in bottom half): return -1
    * if no alias is defined
    */

    protected int getNamespaceAlias(short uriCode) {

		// if there are several matches, the last in stylesheet takes priority;
		// but the list is in reverse stylesheet order
		for (int i=0; i<numberOfAliases; i++) {
			if (uriCode==aliasSCodes[i]) {
				return aliasNCodes[i];
			}
		}
		return uriCode;
	}

    /**
    * Validate this element
    */

    public void validate() throws TransformerConfigurationException {
        if (validationError != null) {
            compileError(validationError);
        }
        if (!(getParentNode() instanceof DocumentInfo)) {
            throw new TransformerConfigurationException(
                        getDisplayName() + " must be the outermost element");
        }
    }

    /**
    * Preprocess does all the processing possible before the source document is available.
    * It is done once per stylesheet, so the stylesheet can be reused for multiple source
    * documents.
    */

    public void preprocess() throws TransformerConfigurationException {

        // process any xsl:include and xsl:import elements

        spliceIncludes();

        // build indexes for selected top-level elements

        buildIndexes();

        // process the attributes of every node in the tree

        processAllAttributes();

        // collect any namespace aliases

        collectNamespaceAliases();

        // fix up references from XPath expressions to variables and functions, for static typing

        for (int i=0; i<topLevel.size(); i++) {
            Object node = topLevel.get(i);
            if (node instanceof StyleElement) {
                ((StyleElement)node).fixupReferences();
            }
        }

        // Validate the whole logical style sheet (i.e. with included and imported sheets)

        validate();
        for (int i=0; i<topLevel.size(); i++) {
            Object node = topLevel.get(i);
            if (node instanceof StyleElement) {
                ((StyleElement)node).validateSubtree();
            }
        }

        // Preprocess definitions of top-level elements.

//        for (int i=0; i<topLevel.size(); i++) {
//            Object s = topLevel.get(i);
//            if (s instanceof StyleElement) {
//                try {
//                    ((StyleElement)s).preprocess();
//                } catch (TransformerConfigurationException err) {
//                    ((StyleElement)s).compileError(err);
//                }
//            }
//        }

    }

    /**
    * Process xsl:include and xsl:import elements.
    */

    public void spliceIncludes() throws TransformerConfigurationException {

        boolean foundNonImport = false;
        topLevel = new ArrayList();
        minImportPrecedence = precedence;
        StyleElement previousElement = this;

        AxisIterator kids = iterateAxis(Axis.CHILD);

        while(kids.hasNext()) {
            NodeInfo child = (NodeInfo)kids.next();
            if (child.getItemType() == Type.TEXT) {
                // in an embedded stylesheet, white space nodes may still be there
                if (!Navigator.isWhite(child.getStringValue())) {
                    previousElement.compileError(
                        "No character data is allowed between top-level elements");
                }

            } else if (child instanceof DataElement) {
                // do nothing
            } else {
                previousElement = (StyleElement)child;
                if (child instanceof XSLGeneralIncorporate) {
                    XSLGeneralIncorporate xslinc = (XSLGeneralIncorporate)child;
                    xslinc.processAttributes();

                    if (xslinc.isImport()) {
                        if (foundNonImport) {
                            xslinc.compileError("xsl:import elements must come first");
                        }
                    } else {
                        foundNonImport = true;
                    }

                    // get the included stylesheet. This follows the URL, builds a tree, and splices
                    // in any indirectly-included stylesheets.

                    XSLStyleSheet inc =
                        xslinc.getIncludedStyleSheet(this, precedence);
                    if (inc==null) return;  // error has been reported

                    // after processing the imported stylesheet and any others it brought in,
                    // adjust the import precedence of this stylesheet if necessary

                    if (xslinc.isImport()) {
                        precedence = inc.getPrecedence() + 1;
                    } else {
                        precedence = inc.getPrecedence();
                        inc.setMinImportPrecedence(minImportPrecedence);
                        inc.setWasIncluded();
                    }

                    // copy the top-level elements of the included stylesheet into the top level of this
                    // stylesheet. Normally we add these elements at the end, in order, but if the precedence
                    // of an element is less than the precedence of the previous element, we promote it.
                    // This implements the requirement in the spec that when xsl:include is used to
                    // include a stylesheet, any xsl:import elements in the included document are moved
                    // up in the including document to after any xsl:import elements in the including
                    // document.

                    List incchildren = inc.topLevel;
                    for (int j=0; j<incchildren.size(); j++) {
                        StyleElement elem = (StyleElement)incchildren.get(j);
                        int last = topLevel.size() - 1;
                        if (last < 0 || elem.getPrecedence() >= ((StyleElement)topLevel.get(last)).getPrecedence()) {
                            topLevel.add(elem);
                        } else {
                            while (last >=0 && elem.getPrecedence() < ((StyleElement)topLevel.get(last)).getPrecedence()) {
                                last--;
                            }
                            topLevel.add(last+1, elem);
                        }
                    }
                } else {
                    foundNonImport = true;
                    topLevel.add(child);
                }
            }
        }
    }

	/**
	* Build indexes for selected top-level declarations
	*/

	private void buildIndexes() throws TransformerConfigurationException {
        // Scan the declarations in reverse order
        for (int i=topLevel.size()-1; i>=0; i--) {
            Object node = topLevel.get(i);
            if (node instanceof XSLTemplate) {
                indexNamedTemplate((XSLTemplate)node);
            } else if (node instanceof XSLVariableDeclaration) {
                indexVariableDeclaration((XSLVariableDeclaration)node);
            } else if (node instanceof XSLNamespaceAlias) {
                namespaceAliasList.add(node);
                numberOfAliases++;
            } else if (node instanceof SaxonImportQuery) {
                ((SaxonImportQuery)node).importModule();
            }
        }
    }

    /**
     * Index a global xsl:variable or xsl:param element
     * @param var The XSLVariable or XSLParam element
     * @throws TransformerConfigurationException
     */

    private void indexVariableDeclaration(XSLVariableDeclaration var) throws TransformerConfigurationException {
        int fingerprint = var.getVariableFingerprint();
        //System.err.println("fingerprint = " + fingerprint);
        if (fingerprint != -1) {
            Integer key = new Integer(fingerprint);
            // see if there is already a global variable with this precedence
            XSLVariableDeclaration other = (XSLVariableDeclaration)variableMap.get(key);
            if (other==null) {
                // this is the first
                variableMap.put(key, var);
            } else {
                // check the precedences
                int thisPrecedence = var.getPrecedence();
                int otherPrecedence = other.getPrecedence();
                if (thisPrecedence == otherPrecedence) {
                    var.compileError("Duplicate global variable declaration (see line " +
                            other.getLineNumber() + " of " + other.getSystemId() + ")");
                } else if (thisPrecedence < otherPrecedence) {
                    var.setRedundant();
                } else {
                    // can't happen, but we'll play safe
                    other.setRedundant();
                    variableMap.put(key, var);
                }
            }
        }
    }

    /**
     * Add a named template to the index
     * @param template The XSLTemplate element
     * @throws TransformerConfigurationException
     */
    private void indexNamedTemplate(XSLTemplate template) throws TransformerConfigurationException {
        int fingerprint = template.getTemplateFingerprint();
        if (fingerprint != -1) {
            Integer key = new Integer(fingerprint);
            // see if there is already a named template with this precedence
            XSLTemplate other = (XSLTemplate)templateIndex.get(key);
            if (other==null) {
                // this is the first
                templateIndex.put(key, template);
            } else {
                // check the precedences
                int thisPrecedence = template.getPrecedence();
                int otherPrecedence = other.getPrecedence();
                if (thisPrecedence == otherPrecedence) {
                    template.compileError("Duplicate named template (see line " +
                            other.getLineNumber() + " of " + other.getSystemId() + ")");
                } else if (thisPrecedence < otherPrecedence) {
                    template.setRedundantNamedTemplate();
                } else {
                    // can't happen, but we'll play safe
                    other.setRedundantNamedTemplate();
                    templateIndex.put(key, template);
                }
            }
        }
    }

    /**
	* Collect any namespace aliases
	*/

	private void collectNamespaceAliases() throws TransformerConfigurationException {
	    aliasSCodes = new short[numberOfAliases];
	    //aliasRCodes	= new short[numberOfAliases];
	    aliasNCodes = new int[numberOfAliases];
	    int precedenceBoundary = 0;
	    int currentPrecedence = -1;
	    // Note that we are processing the list in reverse stylesheet order,
	    // that is, highest precedence first.
        for (int i=0; i<numberOfAliases; i++) {
            XSLNamespaceAlias xna = (XSLNamespaceAlias)namespaceAliasList.get(i);
            short scode = xna.getStylesheetURICode();
            int ncode = xna.getResultNamespaceCode();
            int prec = xna.getPrecedence();

            // check that there isn't a conflict with another xsl:namespace-alias
            // at the same precedence

            if (currentPrecedence != prec) {
                currentPrecedence = prec;
                precedenceBoundary = i;
            }

            for (int j=precedenceBoundary; j<i; j++) {
                if (scode == aliasSCodes[j]) {
                    if ((ncode&0xffff) != (aliasNCodes[j]&0xffff)) {
                        xna.compileError("Inconsistent namespace aliases");
                    }
                }
            }

            aliasSCodes[i] = scode;
            aliasNCodes[i] = ncode;
        }
        namespaceAliasList = null;  // throw it in the garbage
    }

    protected boolean hasNamespaceAliases() {
    	return numberOfAliases>0;
    }

    /**
    * Process the attributes of every node in the stylesheet
    */

    public void processAllAttributes() throws TransformerConfigurationException {
        prepareAttributes();
        for (int i=0; i<topLevel.size(); i++) {
            Object s = topLevel.get(i);
            if (s instanceof StyleElement) {
                try {
                    ((StyleElement)s).processAllAttributes();
                } catch (TransformerConfigurationException err) {
                    ((StyleElement)s).compileError(err);
                }
            }
        }
    }

    /**
    * Get the global variable or parameter with a given fingerprint (taking
    * precedence rules into account)
    */

    public XSLVariableDeclaration getGlobalVariable(int fingerprint) {
        return (XSLVariableDeclaration)variableMap.get(new Integer(fingerprint));
    }

    /**
    * Create an output properties object representing the xsl:output elements in the stylesheet.
    * @param fingerprint The name of the output format required. If set to -1, gathers
    * information for the unnamed output format
    * @return the Properties object containing the details of the specified output format
    * @throws TransformerConfigurationException if a named output format does not exist in
     * the stylesheet
    */

    public Properties gatherOutputProperties(int fingerprint) throws TransformerConfigurationException {
        boolean found = (fingerprint == -1);
        Properties details = new Properties();
        for (int i=0; i<topLevel.size(); i++) {
            Object s = topLevel.get(i);
            if (s instanceof XSLOutput) {
                XSLOutput xo = (XSLOutput)s;
                if (xo.getOutputFingerprint()==fingerprint) {
                    found = true;
                    xo.gatherOutputProperties(details);
                }
            }
        }
        if (!found) {
            throw new TransformerConfigurationException("Requested output format has not been defined");
        }
        return details;
    }

    /**
    * Get a Java class for a given namespace URI, if possible.
    * @return the Class referred to by the URI (as a Class object),
     * or null if none is found.
    */

    public Class getExternalJavaClass(String uri) {
        if (!getPreparedStyleSheet().getConfiguration().isAllowExternalFunctions()) {
            return null;
        }
        for (int i=topLevel.size() - 1; i>=0; i--) {
            Object s = topLevel.get(i);
            if (s instanceof SaxonScript) {
                SaxonScript script = (SaxonScript)s;
                Class c = script.getJavaClass(uri);
                if (c != null) {
                    return c;
                }
            }
        }
        return null;
    }

    /**
     * Declare an imported XQuery function
     */

    protected void declareXQueryFunction(XQueryFunction function) {
        if (xqueryFunctionIndex == null) {
            xqueryFunctionIndex = new HashMap();
        }
        xqueryFunctionIndex.put(new Integer(function.getFunctionFingerprint()),
                                function);
    }

    /**
     * Get an imported XQuery function with a given name, if there is one
     * @return the XQueryFunction object if it exists, or null otherwise
     */

    protected XQueryFunction bindXQueryFunction(int fingerprint) {
        if (xqueryFunctionIndex == null) {
            return null;
        }
        return (XQueryFunction)xqueryFunctionIndex.get(new Integer(fingerprint));
    }

    /**
    * Compile the stylesheet to create an executable.
    */

    public Instruction compile() throws TransformerConfigurationException {

        Executable exec = new Executable();

        try {

            // If any XQuery functions were imported, call the "compile" method
            // on these functions. This will cause all references to be fixed up properly

            if (xqueryFunctionIndex != null) {
                Iterator qf = xqueryFunctionIndex.values().iterator();
                while (qf.hasNext()) {
                    XQueryFunction f = (XQueryFunction)qf.next();
                    try {
                        f.fixupReferences();
                    } catch (XPathException e) {
                        compileError(e);
                    }
                }
            }

            // Call compile method for each top-level object in the stylesheet

            for (int i=0; i<topLevel.size(); i++) {
    	        NodeInfo node = (NodeInfo)topLevel.get(i);
                if (node instanceof StyleElement) {
        		    StyleElement snode = (StyleElement)node;
          			Instruction inst = snode.compile();
          			if (inst != null) {
          			    inst.setSourceLocation(putModuleNumber(snode.getSystemId()),
          			                           snode.getLineNumber());
          			}
        		}
    	    }

            // Fix up references to the default default decimal format

            if (decimalFormatManager != null) {
                decimalFormatManager.fixupDefaultDefault();
            }

            exec.setModuleArray(getModuleArray());
            exec.setRuleManager(ruleManager);
            exec.setKeyManager(keyManager);
            exec.setStripperRules(stripperRules);
            exec.setStripsWhitespace(stripsWhitespace());   // TODO: inefficient
            exec.setDecimalFormatManager(decimalFormatManager);
            exec.setSortKeyTable(sortKeyTable);
            exec.setDefaultCollationName(defaultCollationName);
            exec.setCollationTable(collationTable);
            exec.setDefaultOutputProperties(gatherOutputProperties(-1));
            exec.setSlotSpace(numberOfVariables, largestStackFrame);

            if (runtimeTemplateMapRequired) {
                HashMap compiledTemplateIndex = new HashMap(templateIndex.size());
                Iterator iter = templateIndex.keySet().iterator();
                while (iter.hasNext()) {
                    Integer key = (Integer)iter.next();
                    XSLTemplate t = (XSLTemplate)templateIndex.get(key);
                    compiledTemplateIndex.put(key, t.getCompiledTemplate());
                }
                exec.setNamedTemplateTable(compiledTemplateIndex);
            }

            // Build the index of named character maps

            for (int i=0; i<topLevel.size(); i++) {
                if (topLevel.get(i) instanceof XSLCharacterMap) {
                    XSLCharacterMap t = (XSLCharacterMap)topLevel.get(i);
                    if (!t.isRedundant()) {
                        int fp = t.getCharacterMapFingerprint();
                        HashMap map = new HashMap();
                        t.assemble(map);
                        if (characterMapIndex==null) {
                            characterMapIndex = new HashMap();
                        }
                        characterMapIndex.put(new Integer(fp), map);
                    }
                }
            }
            exec.setCharacterMapIndex(characterMapIndex);

            return exec;

        } catch (RuntimeException err) {
            // if syntax errors were reported earlier, then exceptions may occur during this phase
            // due to inconsistency of data structures. We can ignore these exceptions as they
            // will go away when the user corrects the stylesheet
            if (getPreparedStyleSheet().getErrorCount() > 0) {
                // do nothing
                return exec;
            } else {
                // rethrow the exception
                throw err;
            }
        }
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s):
// Portions marked "e.g." are from Edwin Glaser (edwin@pannenleiter.de)
//
