package net.sf.saxon.tinytree;
import net.sf.saxon.om.AxisIteratorImpl;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.pattern.NodeTest;

/**
* This class supports both the child:: and following-sibling:: axes, which are
* identical except for the route to the first candidate node.
* It enumerates either the children or the following siblings of the specified node.
* In the case of children, the specified node must always
* be a node that has children: to ensure this, construct the enumeration
* using NodeInfo#getEnumeration()
*/

final class SiblingEnumeration extends AxisIteratorImpl {

    private TinyDocumentImpl document;
    private int nextNodeNr;
    private NodeTest test;
    private TinyNodeImpl startNode;
    private TinyNodeImpl parentNode;
    private boolean getChildren;
    private boolean needToAdvance = false;

    /**
     * Return an enumeration over children or siblings of the context node
     * @param doc The Document node containing the context node
     * @param node The context node, the start point for the iteration
     * @param nodeTest Test that the selected nodes must satisfy, or null indicating
     * that all nodes are selected
     * @param getChildren True if children of the context node are to be returned, false
     * if following siblings are required
     */

    protected SiblingEnumeration(TinyDocumentImpl doc, TinyNodeImpl node,
                              NodeTest nodeTest, boolean getChildren) {
        document = doc;
        test = nodeTest;
        startNode = node;
        this.getChildren = getChildren;
        if (getChildren) {          // child:: axis
            parentNode = node;

            // move to first child
            // ASSERT: we don't invoke this code unless the node has children
            nextNodeNr = node.nodeNr + 1;

        } else {                    // following-sibling:: axis
            parentNode = (TinyNodeImpl)node.getParent();

            // move to next sibling
            nextNodeNr = doc.next[node.nodeNr];
            if (nextNodeNr < node.nodeNr) {         // *O*
                nextNodeNr = -1;                    // *O*
            }                                       // *O*
        }

        // check if this matches the conditions
        if (nextNodeNr >= 0 && nodeTest != null) {
            if (!nodeTest.matches(document.nodeKind[nextNodeNr],
                                  document.nameCode[nextNodeNr],
                                  document.getElementAnnotation(nextNodeNr))) {
                needToAdvance = true;
            }
        }
    }

    public boolean hasNext() {
        if (needToAdvance) {
            // inlined advance() method
            final int thisNode = nextNodeNr;
            if (test==null) {
                nextNodeNr = document.next[nextNodeNr];
            } else {
                do {
                    nextNodeNr = document.next[nextNodeNr];
                } while ( nextNodeNr >= thisNode &&
                        !test.matches(document.nodeKind[nextNodeNr],
                                      document.nameCode[nextNodeNr],
                                      document.getElementAnnotation(nextNodeNr)));
            }
            needToAdvance = false;
            if (nextNodeNr < thisNode) {    // *O*: indicates we've found an owner pointer
                nextNodeNr = -1;
                return false;
            } else {
                return true;
            }
        }
        return nextNodeNr >= 0;
    }

    public Item next() {
//        if (needToAdvance) {
//            // shouldn't happen if the user calls hasNext() properly, but we'll play safe
//            advance();
//        }
        position++;
        current = document.getNode(nextNodeNr);
        ((TinyNodeImpl)current).setParentNode(parentNode);
        needToAdvance = true;
        return current;

        // DONE: the lookahead approach is widely used, but it can be quite inefficent.
        // With a path expression such as child::TITLE, where we are only fetching a single node,
        // it means we look beyond the selected node to find the next one, which might mean searching the
        // rest of the document for no useful purpose.

        // TODO: apply the same logic to other iterators.
    }

    /**
    * Get another enumeration of the same nodes
    */

    public SequenceIterator getAnother() {
        return new SiblingEnumeration(document, startNode, test, getChildren);
    }

}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
