package net.sf.saxon.tinytree;
import net.sf.saxon.event.Receiver;
import net.sf.saxon.om.*;
import net.sf.saxon.tree.LineNumberMap;
import net.sf.saxon.tree.SystemIdMap;
import net.sf.saxon.value.Type;
import org.w3c.dom.Document;

import javax.xml.transform.TransformerException;
import java.util.ArrayList;
import java.util.HashMap;


/**
  * A node in the XML parse tree representing the Document itself (or equivalently, the root
  * node of the Document).<P>
  * @author <A HREF="mailto:michael.h.kay@ntlworld.com>Michael H. Kay</A>
  * @version 26 April 1999
  */

public final class TinyDocumentImpl extends TinyParentNodeImpl
    implements DocumentInfo, Document {

    private HashMap idTable = null;
    private NamePool namePool;
    private int documentNumber;
    private HashMap elementList = null;
    private boolean usesNamespaces = false;
    private HashMap entityTable = null;
    private HashMap elementTypeMap = null;

    // This structure supports trees whose root is an element node rather than a document node.
    // The document node still exists, for implementation reasons, but it is not regarded as
    // part of the tree. The variable rootNode identifies the actual root of the tree, which
    // is the document node by default.

    protected int rootNode = 0;

    // the contents of the document

    protected char[] charBuffer;
    protected int charBufferLength = 0;
    protected StringBuffer commentBuffer = null; // created when needed

    protected int numberOfNodes = 0;    // excluding attributes and namespaces
//    protected int lastLevelOneNode = -1;

    // The following arrays contain one entry for each node other than attribute
    // and namespace nodes, arranged in document order.

    // nodeKind indicates the kind of node, e.g. element, text, or comment
    protected byte[] nodeKind;

    // depth is the depth of the node in the hierarchy, i.e. the number of ancestors
    protected short[] depth;

    // next is the node number of the next sibling
    // *O* unless it points backwards, in which case it is the node number of the parent
    protected int[] next;

    // alpha holds a value that depends on the node kind. For text nodes, it is the offset
    // into the text buffer. For comments and processing instructions, it is the offset into
    // the comment buffer. For elements, it is the index of the first attribute node, or -1
    // if this element has no attributes.
    protected int[] alpha;

    // beta holds a value that depends on the node kind. For text nodes, it is the length
    // of the text. For comments and processing instructions, it is the length of the text.
    // For elements, it is the index of the first namespace node, or -1
    // if this element has no namespaces.
    protected int[] beta;

    // nameCode holds the name of the node, as an index into the name pool
    protected int[] nameCode;

    // the prior array indexes preceding-siblings; it is constructed only when required
    protected int[] prior = null;

    // the owner array gives fast access from a node to its parent; it is constructed
    // only when required
    // protected int[] parentIndex = null;

    // the following arrays have one entry for each attribute.
    protected int numberOfAttributes = 0;

    // attParent is the index of the parent element node
    protected int[] attParent;

    // attCode is the nameCode representing the attribute name
    protected int[] attCode;

    // attValue is the string value of the attribute
    protected CharSequence[] attValue;

    // attTypeCode holds type annotations. The array is created only if any nodes have a type annotation
    private int[] attTypeCode;

    // The following arrays have one entry for each namespace declaration
    protected int numberOfNamespaces = 0;

    // namespaceParent is the index of the element node owning the namespace declaration
    protected int[] namespaceParent;

    // namespaceCode is the namespace code used by the name pool: the top half is the prefix
    // code, the bottom half the URI code
    protected int[] namespaceCode;

    private LineNumberMap lineNumberMap;
    private SystemIdMap systemIdMap = null;

    public TinyDocumentImpl() {
        this(4000, 100, 20, 4000);
    }

    public TinyDocumentImpl(int nodes, int attributes, int namespaces, int characters) {
        nodeNr = 0;
        document = this;
        nodeKind = new byte[nodes];
        depth = new short[nodes];
        next = new int[nodes];
        alpha = new int[nodes];
        beta = new int[nodes];
        nameCode = new int[nodes];

        numberOfAttributes = 0;
        attParent = new int[attributes];
        attCode = new int[attributes];
        attValue = new String[attributes];

        numberOfNamespaces = 0;
        namespaceParent = new int[namespaces];
        namespaceCode = new int[namespaces];

        charBuffer = new char[characters];
        charBufferLength = 0;
    }

	/**
	* Set the name pool used for all names in this document
	*/

	public void setNamePool(NamePool pool) {
		namePool = pool;
		addNamespace(0, pool.getNamespaceCode("xml", NamespaceConstant.XML));
		documentNumber = pool.allocateDocumentNumber(this);
	}

	/**
	* Get the name pool used for the names in this document
	*/

	public NamePool getNamePool() {
		return namePool;
	}

	/**
	* Get the unique document number
	*/

	public int getDocumentNumber() {
	    return documentNumber;
	}


    protected void ensureNodeCapacity() {
        if (nodeKind.length < numberOfNodes+1) {
            int k = numberOfNodes*2;

            byte[] nodeKind2 = new byte[k];
            int[] next2 = new int[k];
            short[] depth2 = new short[k];
            int[] alpha2 = new int[k];
            int[] beta2 = new int[k];
            int[] nameCode2 = new int[k];

            System.arraycopy(nodeKind, 0, nodeKind2, 0, numberOfNodes);
            System.arraycopy(next, 0, next2, 0, numberOfNodes);
            System.arraycopy(depth, 0, depth2, 0, numberOfNodes);
            System.arraycopy(alpha, 0, alpha2, 0, numberOfNodes);
            System.arraycopy(beta, 0, beta2, 0, numberOfNodes);
            System.arraycopy(nameCode, 0, nameCode2, 0, numberOfNodes);

            nodeKind = nodeKind2;
            next = next2;
            depth = depth2;
            alpha = alpha2;
            beta = beta2;
            nameCode = nameCode2;
        }
    }

    protected void ensureAttributeCapacity() {
        if (attParent.length < numberOfAttributes+1) {
            int k = numberOfAttributes*2;

            int[] attParent2 = new int[k];
            int[] attCode2 = new int[k];
            String[] attValue2 = new String[k];

            System.arraycopy(attParent, 0, attParent2, 0, numberOfAttributes);
            System.arraycopy(attCode, 0, attCode2, 0, numberOfAttributes);
            System.arraycopy(attValue, 0, attValue2, 0, numberOfAttributes);

            attParent = attParent2;
            attCode = attCode2;
            attValue = attValue2;

            if (attTypeCode != null) {
                int[] attTypeCode2 = new int[k];
                System.arraycopy(attTypeCode, 0, attTypeCode2, 0, numberOfAttributes);
                attTypeCode = attTypeCode2;
            }
        }
    }

    protected void ensureNamespaceCapacity() {
        if (namespaceParent.length < numberOfNamespaces+1) {
            int k = numberOfNamespaces*2;

            int[] namespaceParent2 = new int[k];
            int[] namespaceCode2 = new int[k];

            System.arraycopy(namespaceParent, 0, namespaceParent2, 0, numberOfNamespaces);
            System.arraycopy(namespaceCode, 0, namespaceCode2, 0, numberOfNamespaces);

            namespaceParent = namespaceParent2;
            namespaceCode = namespaceCode2;
        }
    }

    /**
    * Set the root node. Parentless elements are implemented using a full tree structure
    * containing a document node, but the document node is not regarded as part of the tree
    */

    public void setRootNode(NodeInfo root) {
        rootNode = ((TinyNodeImpl)root).nodeNr;
    }

    /**
     * Add a node to the document
     * @param type0         The kind of the node
     * @param depth0        The depth in the tree
     * @param alpha0        Pointer to attributes or text
     * @param beta0         Pointer to namespaces or text
     * @param nameCode0     The name of the node
     * @return the node number of the node that was added
     */
    protected int addNode(short type0, int depth0, int alpha0, int beta0, int nameCode0) {
        ensureNodeCapacity();
        nodeKind[numberOfNodes] = (byte)type0;
        depth[numberOfNodes] = (short)depth0;
        alpha[numberOfNodes] = alpha0;
        beta[numberOfNodes] = beta0;
        nameCode[numberOfNodes] = nameCode0;
        next[numberOfNodes] = -1;      // safety precaution, esp for preview mode

//        if (depth0 == 1) lastLevelOneNode = numberOfNodes;

        return numberOfNodes++;
    }

    protected void appendChars(CharSequence chars) {
        while (charBuffer.length < charBufferLength + chars.length()) {
            char[] ch2 = new char[charBuffer.length * 2];
            System.arraycopy(charBuffer, 0, ch2, 0, charBufferLength);
            charBuffer = ch2;
        }
        char[] newchars = chars.toString().toCharArray();
        System.arraycopy(newchars, 0, charBuffer, charBufferLength, chars.length());
        // TODO: this is inefficient. There is no absolute reason that the
        // character data for the whole document needs to be in one array.
        charBufferLength += chars.length();
    }

    /**
    * Condense the tree: release unused memory. This is done after the full tree has been built.
    * The method makes a pragmatic judgement as to whether it is worth reclaiming space; this is
    * only done when the constructed tree is very small compared with the space allocated.
    */

    public void condense() {
        if (numberOfNodes * 5 < nodeKind.length) {
            int k = numberOfNodes;

            byte[] nodeKind2 = new byte[k];
            int[] next2 = new int[k];
            short[] depth2 = new short[k];
            int[] alpha2 = new int[k];
            int[] beta2 = new int[k];
            int[] nameCode2 = new int[k];

            System.arraycopy(nodeKind, 0, nodeKind2, 0, numberOfNodes);
            System.arraycopy(next, 0, next2, 0, numberOfNodes);
            System.arraycopy(depth, 0, depth2, 0, numberOfNodes);
            System.arraycopy(alpha, 0, alpha2, 0, numberOfNodes);
            System.arraycopy(beta, 0, beta2, 0, numberOfNodes);
            System.arraycopy(nameCode, 0, nameCode2, 0, numberOfNodes);

            nodeKind = nodeKind2;
            next = next2;
            depth = depth2;
            alpha = alpha2;
            beta = beta2;
            nameCode = nameCode2;
        }

        if (numberOfAttributes * 5 < attParent.length) {
            int k = numberOfAttributes;

            int[] attParent2 = new int[k];
            int[] attCode2 = new int[k];
            String[] attValue2 = new String[k];

            System.arraycopy(attParent, 0, attParent2, 0, numberOfAttributes);
            System.arraycopy(attCode, 0, attCode2, 0, numberOfAttributes);
            System.arraycopy(attValue, 0, attValue2, 0, numberOfAttributes);

            attParent = attParent2;
            attCode = attCode2;
            attValue = attValue2;

            if (attTypeCode != null) {
                int[] attTypeCode2 = new int[k];
                System.arraycopy(attTypeCode, 0, attTypeCode2, 0, numberOfAttributes);
                attTypeCode = attTypeCode2;
            }
        }

        if (numberOfNamespaces * 5 < namespaceParent.length) {
            int k = numberOfNamespaces;
            int[] namespaceParent2 = new int[k];
            int[] namespaceCode2 = new int[k];

            System.arraycopy(namespaceParent, 0, namespaceParent2, 0, numberOfNamespaces);
            System.arraycopy(namespaceCode, 0, namespaceCode2, 0, numberOfNamespaces);

            namespaceParent = namespaceParent2;
            namespaceCode = namespaceCode2;
        }
    }

    /**
    * Set the type annotation of an element node
    */

    protected void setElementAnnotation(int nodeNr, int typeCode) {
        // The implementation here is not very economical. It is optimized for the case
        // where annotating an element with a type is unusual.
        if (elementTypeMap == null) {
            elementTypeMap = new HashMap();
        }
        elementTypeMap.put(new Integer(nodeNr), new Integer(typeCode));
    }

    /**
    * Get the type annotation of an element node.
     * Type.UNTYPED_ANY if there is no type annotation
    */

    protected int getElementAnnotation(int nodeNr) {
        if (elementTypeMap == null) {
            return Type.UNTYPED_ANY;
        }
        Integer i = (Integer)elementTypeMap.get(new Integer(nodeNr));
        if (i == null) {
            return Type.UNTYPED_ANY;
        } else {
            return i.intValue();
        }
    }

    /**
    * On demand, make an index for quick access to preceding-sibling nodes
    */

    protected void ensurePriorIndex() {
        if (prior==null) {
            makePriorIndex();
        }
    }

    private synchronized void makePriorIndex() {
        prior = new int[numberOfNodes];
        for (int i=0; i<numberOfNodes; i++) {
            prior[i] = -1;
        }
        for (int i=0; i<numberOfNodes; i++) {
            int nextNode = next[i];
            //if (nextNode!=-1) {
            if (nextNode > i) {       // *O*
                prior[nextNode] = i;
            }
        }
    }

    /*
    * The following code to create parent pointers works, but doesn't
    * appear to give any performance benefit. In fact it slows things down

        protected synchronized void makeParentIndex() {
            int[] stack = new int[20];
            int stackSize = 0;
            parentIndex = new int[numberOfNodes];
            for (int i=1; i<numberOfNodes; i++) {
                int step = depth[i] - depth[i-1];
                if (step == 0) {
                    // we are on the same level
                    parentIndex[i] = parentIndex[i-1];
                } else if (step > 0) {
                    // we've gone down a level
                    if (depth[i] >= stack.length) {
                        int[] stack2 = new int[stack.length * 2];
                        System.arraycopy(stack, 0, stack2, 0, stackSize);
                        stack = stack2;
                    }
                    stack[stackSize++] = i-1;
                    parentIndex[i] = i-1;
                } else {
                    // we've gone up a level
                    stackSize--;
                    parentIndex[i] = stack[stackSize];
                }
            }
        }
    */

    protected void addAttribute(int parent0, int code0, int type0, CharSequence value0) {
        // System.err.println("addAttribute(" + parent0 + "," + code0 + "," + type0 + "," + value0 + ")");
        ensureAttributeCapacity();
        attParent[numberOfAttributes] = parent0;
        attCode[numberOfAttributes] = code0;
        attValue[numberOfAttributes] = value0;

        if (type0 == 0) {
            type0 = Type.UNTYPED_ATOMIC;
        }
        if (type0 != Type.UNTYPED_ATOMIC) {
            if (attTypeCode==null) {
                attTypeCode = new int[attParent.length];
                for (int i=0; i<numberOfAttributes; i++) {
                    attTypeCode[i] = Type.UNTYPED_ATOMIC;
                }
            }
        }
        if (attTypeCode != null) {
            attTypeCode[numberOfAttributes] = type0;
        }

        if (alpha[parent0] == -1) {
            alpha[parent0] = numberOfAttributes;
        }

        if (type0 == Type.ID) {
            // The attribute is marked as being an ID. But we don't trust it - it
            // might come from a non-validating parser. Before adding it to the index, we
            // check that it really is an ID.

            if (XMLChar.isValidNCName(value0.toString())) {
            	if (idTable==null) {
            		idTable = new HashMap();
            	}
    			NodeInfo e = getNode(parent0);
                registerID(e, value0.toString());
            } else if (attTypeCode != null) {
                attTypeCode[numberOfAttributes] = Type.UNTYPED_ATOMIC;
            }
        }

        numberOfAttributes++;
    }

    protected void addNamespace(int parent, int nscode ) {
        usesNamespaces = true;
        ensureNamespaceCapacity();
        namespaceParent[numberOfNamespaces] = parent;
        namespaceCode[numberOfNamespaces] = nscode;

        if (beta[parent] == -1) {
            beta[parent] = numberOfNamespaces;
        }
        numberOfNamespaces++;
    }

    public TinyNodeImpl getNode(int nr) {

        switch ((short)nodeKind[nr]) {
            case Type.DOCUMENT:
                return this;
            case Type.ELEMENT:
                return new TinyElementImpl(this, nr);
            case Type.TEXT:
                return new TinyTextImpl(this, nr);
            case Type.COMMENT:
                return new TinyCommentImpl(this, nr);
            case Type.PROCESSING_INSTRUCTION:
                return new TinyProcInstImpl(this, nr);
        }

        return null;
    }

    /**
    * Get the node sequence number (in document order). Sequence numbers are monotonic but not
    * consecutive.
    */

    public long getSequenceNumber() {
        return 0;
    }

    /**
    * Make a (transient) attribute node from the array of attributes
    */

    protected TinyAttributeImpl getAttributeNode(int nr) {
        return new TinyAttributeImpl(this, nr);
    }

    /**
    * Get the type annotation of an attribute node.
     * @return Type.UNTYPED_ATOMIC if there is no annotation
    */

    protected int getAttributeAnnotation(int nr) {
        if (attTypeCode == null) {
            return Type.UNTYPED_ATOMIC;
        } else {
            return attTypeCode[nr];
        }
    }

    /**
    * determine whether this document uses namespaces
    */

    protected boolean isUsingNamespaces() {
        return usesNamespaces;
    }

    /**
    * Make a (transient) namespace node from the array of namespace declarations
    */

    protected TinyNamespaceImpl getNamespaceNode(int nr) {
        return new TinyNamespaceImpl(this, nr);
    }

    /**
    * Set the system id of this node
    */

    public void setSystemId(String uri) {
        if (uri==null) {
            uri = "";
        }
        if (systemIdMap==null) {
            systemIdMap = new SystemIdMap();
        }
        systemIdMap.setSystemId(nodeNr, uri);
    }

    /**
    * Get the system id of this root node
    */

    public String getSystemId() {
        if (systemIdMap==null) {
            return null;
        }
        return systemIdMap.getSystemId(nodeNr);
    }

    /**
    * Get the base URI of this root node. For a root node the base URI is the same as the
    * System ID.
    */

    public String getBaseURI() {
        return getSystemId();
    }

    /**
    * Set the system id of an element in the document
    */

    protected void setSystemId(int seq, String uri) {
        if (uri==null) {
            uri = "";
        }
        if (systemIdMap==null) {
            systemIdMap = new SystemIdMap();
        }
        systemIdMap.setSystemId(seq, uri);
    }


    /**
    * Get the system id of an element in the document
    */

    protected String getSystemId(int seq) {
        if (systemIdMap==null) {
            return null;
        }
        return systemIdMap.getSystemId(seq);
    }


    /**
    * Set line numbering on
    */

    public void setLineNumbering() {
        lineNumberMap = new LineNumberMap();
        lineNumberMap.setLineNumber(0, 0);
    }

    /**
    * Set the line number for an element. Ignored if line numbering is off.
    */

    protected void setLineNumber(int sequence, int line) {
        if (lineNumberMap != null) {
            lineNumberMap.setLineNumber(sequence, line);
        }
    }

    /**
    * Get the line number for an element. Return -1 if line numbering is off.
    */

    protected int getLineNumber(int sequence) {
        if (lineNumberMap != null) {
            return lineNumberMap.getLineNumber(sequence);
        }
        return -1;
    }

    /**
    * Get the line number of this root node.
    * @return 0 always
    */

    public int getLineNumber() {
        return 0;
    }

    /**
    * Return the type of node.
    * @return Type.DOCUMENT (always)
    */

    public final int getItemType() {
        return Type.DOCUMENT;
    }

    /**
     * Find the parent node of this node.
     * @return The Node object describing the containing element or root node.
     */

    public NodeInfo getParent()  {
        return null;
    }

    /**
    * Get the root node
    * @return the NodeInfo that is the root of the tree - not necessarily a document node
    */

    public NodeInfo getRoot() {
        return (rootNode==nodeNr ? this : getNode(rootNode));
    }

    /**
    * Get the root (document) node
    * @return the DocumentInfo representing the document node, or null if the
    * root of the tree is not a document node
    */

    public DocumentInfo getDocumentRoot() {
        return (rootNode==nodeNr ? this : null);
    }

    /**
    * Get a character string that uniquely identifies this node
    * @return an identifier based on the document number
    */

    public String generateId() {
        return "d"+documentNumber;
    }

    /**
    * Get a list of all elements with a given name. This is implemented
    * as a memo function: the first time it is called for a particular
    * element type, it remembers the result for next time.
    */

    protected AxisIterator getAllElements(int fingerprint) {
    	Integer key = new Integer(fingerprint);
    	if (elementList==null) {
    	    elementList = new HashMap();
    	}
        ArrayList list = (ArrayList)elementList.get(key);
        if (list==null) {
            list = new ArrayList();
            for (int i=1; i<numberOfNodes; i++) {
                if (nodeKind[i]==Type.ELEMENT &&
                        (nameCode[i] & 0xfffff ) == fingerprint) {
                    list.add(getNode(i));
                }
            }
            elementList.put(key, list);
        }
        return new ArrayListIterator(list);
    }

    /**
    * Register a unique element ID. Fails if there is already an element with that ID.
    * @param e The NodeInfo (always an element) having a particular unique ID value
    * @param id The unique ID value
    */

    private void registerID(NodeInfo e, String id) {
        // the XPath spec (5.2.1) says ignore the second ID if it's not unique
        NodeInfo old = (NodeInfo)idTable.get(id);
        if (old==null) {
            idTable.put(id, e);
        }

    }

    /**
    * Get the element with a given ID.
    * @param id The unique ID of the required element, previously registered using registerID()
    * @return The NodeInfo (always an Element) for the given ID if one has been registered,
    * otherwise null.
    */

    public NodeInfo selectID(String id) {
        if (idTable==null) return null;			// no ID values found
        return (NodeInfo)idTable.get(id);
    }

    /**
    * Set an unparsed entity URI associated with this document. For system use only, while
    * building the document.
    */

    protected void setUnparsedEntity(String name, String uri, String publicId) {
        // System.err.println("setUnparsedEntity( " + name + "," + uri + ")");
        if (entityTable==null) {
            entityTable = new HashMap();
        }
        String[] ids = new String[2];
        ids[0] = uri;
        ids[1] = publicId;
        entityTable.put(name, ids);
    }

    /**
    * Get the unparsed entity with a given nameID if there is one, or null if not. If the entity
    * does not exist, return null.
    * @param name the name of the entity
    * @return if the entity exists, return an array of two Strings, the first holding the system ID
    * of the entity, the second holding the public
    */

    public String[] getUnparsedEntity(String name) {
        if (entityTable==null) {
            return null;
        }
        return (String[])entityTable.get(name);
    }

    /**
    * Copy this node to a given outputter
    */

    public void copy(Receiver out, int whichNamespaces, boolean copyAnnotations) throws TransformerException {

        // output the children

        AxisIterator children = iterateAxis(Axis.CHILD);
        while (children.hasNext()) {
            ((NodeInfo)children.next()).copy(out, whichNamespaces, copyAnnotations);
        }
    }

	/**
	* Produce diagnostic print of main tree arrays
	*/

	public void diagnosticDump() {
		System.err.println("    node    type   depth    next   alpha    beta    name");
		for (int i=0; i<numberOfNodes; i++) {
			System.err.println(n8(i) + n8(nodeKind[i]) + n8(depth[i]) + n8(next[i]) +
									 n8(alpha[i]) + n8(beta[i]) + n8(nameCode[i]));
		}
		System.err.println("    attr  parent    name    value");
		for (int i=0; i<numberOfAttributes; i++) {
		    System.err.println(n8(i) + n8(attParent[i]) + n8(attCode[i]) + "    " + attValue[i]);
		}
		System.err.println("      ns  parent  prefix     uri");
		for (int i=0; i<numberOfNamespaces; i++) {
		    System.err.println(n8(i) + n8(namespaceParent[i]) + n8(namespaceCode[i]>>16) + n8(namespaceCode[i]&0xffff));
		}
	}

    /**
    * Output a number as a string of 8 characters
    */

    private String n8(int val) {
        String s = "        " + val;
        return s.substring(s.length()-8);
    }

    public void showSize() {
        System.err.println("Tree size: " + numberOfNodes + " nodes, " + charBufferLength + " characters, " +
                                numberOfAttributes + " attributes");
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file except PB-SYNC section.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions marked PB-SYNC are Copyright (C) Peter Bryant (pbryant@bigfoot.com). All Rights Reserved.
//
// Contributor(s): Michael Kay, Peter Bryant.
//
