package net.sf.saxon.trans;
import net.sf.saxon.Controller;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.om.ArrayListIterator;
import net.sf.saxon.om.Axis;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.pattern.NodeTest;
import net.sf.saxon.pattern.Pattern;
import net.sf.saxon.sort.LocalOrderComparer;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.DoubleValue;
import net.sf.saxon.value.FloatValue;
import net.sf.saxon.value.NumericValue;
import net.sf.saxon.value.Type;
import net.sf.saxon.xpath.XPathException;

import java.io.Serializable;
import java.lang.ref.WeakReference;
import java.text.Collator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.WeakHashMap;

/**
  * KeyManager manages the set of key definitions in a stylesheet, and the indexes
  * associated with these key definitions. It handles xsl:sort-key as well as xsl:key
  * definitions.
  *
  * <p>The memory management in this class is subtle, with extensive use of weak references.
  * The idea is that an index should continue to exist in memory so long as both the compiled
  * stylesheet and the source document exist in memory: if either is removed, the index should
  * go too. The document itself holds no reference to the index. The compiled stylesheet (which
  * owns the KeyManager) holds a weak reference to the index. The index, of course, holds strong
  * references to the nodes in the document. The Controller holds a strong reference to the
  * list of indexes used for each document, so that indexes remain in memory for the duration
  * of a transformation even if the documents themselves are garbage collected.</p>
  *
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public class KeyManager implements Serializable {

    private HashMap keyList;         // one entry for each named key; the entry contains
                                     // a list of key definitions with that name
    private transient WeakHashMap docIndexes;
                                     // one entry for each document that is in memory;
                                     // the entry contains a HashMap mapping the fingerprint of
                                     // the key name to the HashMap that is the actual index
                                     // of key/value pairs.

    /**
    * create a KeyManager and initialise variables
    */

    public KeyManager() {
        keyList = new HashMap();
        docIndexes = new WeakHashMap();
    }

    /**
    * Register a key definition. Note that multiple key definitions with the same name are
    * allowed
    * @param fingerprint Integer representing the name of the key
    * @param keydef The details of the key's definition
    */

    public void setKeyDefinition(int fingerprint, KeyDefinition keydef) {
        Integer keykey = new Integer(fingerprint);
        ArrayList v = (ArrayList)keyList.get(keykey);
        if (v==null) {
            v = new ArrayList();
            keyList.put(keykey, v);
        }
        v.add(keydef);
    }

    /**
    * Get all the key definitions that match a particular fingerprint
    * @param fingerprint The fingerprint of the name of the required key
    * @return The key definition of the named key if there is one, or null otherwise.
    */

    public ArrayList getKeyDefinitions(int fingerprint) {
        return (ArrayList)keyList.get(new Integer(fingerprint));
    }

    /**
    * Build the index for a particular document for a named key
    * @param fingerprint The fingerprint of the name of the required key
    * @param doc The source document in question
    * @param controller The controller
    * @return the index in question, as a HashMap mapping a key value onto a ArrayList of nodes
    */

    private synchronized HashMap buildIndex(int fingerprint,
                                           DocumentInfo doc,
                                           Controller controller) throws XPathException {

        ArrayList definitions = getKeyDefinitions(fingerprint);
        if (definitions==null) {
            throw new XPathException.Dynamic("Key " +
            		controller.getNamePool().getDisplayName(fingerprint) +
            							" has not been defined");
        }

        HashMap index = new HashMap();

        for (int k=0; k<definitions.size(); k++) {
            constructIndex( doc,
                            index,
                            (KeyDefinition)definitions.get(k),
                            controller,
                            k==0);
        }

        return index;

    }

    /**
    * Process one key definition to add entries to an index
    */

    private void constructIndex(    DocumentInfo doc,
                                    HashMap index,
                                    KeyDefinition keydef,
                                    Controller controller,
                                    boolean isFirst) throws XPathException {

        Pattern match = keydef.getMatch();
        Expression use = keydef.getUse();
        int itemType = keydef.getItemType();
        Collator collator = keydef.getCollation();
        NodeInfo sourceRoot = doc;
        NodeInfo curr = sourceRoot;
        SequenceIterator save = controller.getCurrentIterator();
        Object[] savedContext = controller.saveContext();
        XPathContext xc = controller.newXPathContext();

        int nodeType = match.getItemType();

        if (nodeType==Type.ATTRIBUTE || nodeType==Type.NODE || nodeType==Type.DOCUMENT) {
            // If the match pattern allows attributes to appear, we must visit them.
            // We also take this path in the ridiculous event that the pattern can match
            // document nodes.
            SequenceIterator all = doc.iterateAxis(Axis.DESCENDANT_OR_SELF);
            while(all.hasNext()) {
                curr = (NodeInfo)all.next();
                if (curr.getItemType()==Type.ELEMENT) {
                    SequenceIterator atts =
                        curr.iterateAxis(Axis.ATTRIBUTE);
                    while (atts.hasNext()) {
                        NodeInfo att = (NodeInfo)atts.next();
                        if (match.matches(att, controller)) {
                            processKeyNode(att, use, itemType,
                                            collator, index, xc, isFirst);
                        }
                    }
                    if (nodeType==Type.NODE) {
                        // index the element as well as its attributes
                        if (match.matches(curr, controller)) {
                            processKeyNode(curr, use, itemType,
                                        collator, index, xc, isFirst);
                        }
                    }
                } else {
                    if (match.matches(curr, controller)) {
                        processKeyNode(curr, use, itemType,
                                     collator, index, xc, isFirst);
                    }
                }
            }

        } else {
            SequenceIterator all =
                doc.iterateAxis( Axis.DESCENDANT,
                                    match.getNodeTest());
            // If the match is a nodetest, we avoid testing it again
            while(all.hasNext()) {
                curr = (NodeInfo)all.next();
                if (match instanceof NodeTest || match.matches(curr, controller)) {
                    processKeyNode(curr, use, itemType,
                                collator, index, xc, isFirst);
                }
            }
        }
        // reset the Controller's context
        controller.setCurrentIterator(save);
        controller.restoreContext(savedContext);
    }

    /**
    * Process one matching node, adding entries to the index if appropriate
    */

    private void processKeyNode(    NodeInfo curr,
                                    Expression use,
                                    int itemType,
                                    Collator collation,
                                    HashMap index,
                                    XPathContext xc,
                                    boolean isFirst) throws XPathException {


        // Make the node we are testing the context node and the current node,
        // with context position and context size set to 1

        SingletonIterator si = new SingletonIterator(curr);
        if (si.hasNext()) {
            si.next();    // need to position iterator at first node
        }
        xc.setCurrentIterator(si);
        xc.getController().setCurrentIterator(si);

        // Evaluate the "use" expression against this context node

        SequenceIterator useval = use.iterate(xc);
        while (useval.hasNext()) {
            // the static type checking ensures that we get an atomic value of the right type
            AtomicValue item = (AtomicValue)useval.next();
            Object val;
            if (itemType==Type.STRING) {
                if (collation==null) {
                    val = item.getStringValue();
                } else {
                    val = collation.getCollationKey(item.getStringValue());
                }
            } else if (itemType==Type.FLOAT && Float.isNaN(((FloatValue)item).getValue())) {
                val = "NaN";
            } else if (itemType==Type.DOUBLE && Double.isNaN(((DoubleValue)item).getValue())) {
                val = "NaN";
            } else {
                val = item;
            }

            ArrayList nodes = (ArrayList)index.get(val);
            if (nodes==null) {
                // this is the first node with this key value
                nodes = new ArrayList();
                index.put(val, nodes);
                nodes.add(curr);
            } else {
                // this is not the first node with this key value.
                // add the node to the list of nodes for this key,
                // unless it's already there
                if (isFirst) {
                    // if this is the first index definition that we're processing,
                    // then this node must be after all existing nodes in document
                    // order, or the same node as the last existing node
                    if (nodes.get(nodes.size()-1)!=curr) {
                        nodes.add(curr);
                    }
                } else {
                    // otherwise, we need to insert the node at the correct
                    // position in document order.
                    LocalOrderComparer comparer = LocalOrderComparer.getInstance();
                    for (int i=0; i<nodes.size(); i++) {
                        int d = comparer.compare(curr, (NodeInfo)nodes.get(i));
                        if (d<=0) {
                            if (d==0) {
                                // node already in list; do nothing
                            } else {
                                // add the node at this position
                                nodes.add(i, curr);
                            }
                            return;
                        }
                        // else continue round the loop
                    }
                    // if we're still here, add the new node at the end
                    nodes.add(curr);
                }
            }
        }

    }

    /**
    * Get the nodes with a given key value
    * @param fingerprint The fingerprint of the name of the required key
    * @param doc The source document in question
    * @param value The required key value
    * @param controller The controller, needed only the first time when the key is being built
    * @return an enumeration of nodes, always in document order
    */

    public SequenceIterator selectByKey(
                                int fingerprint,
                                DocumentInfo doc,
                                AtomicValue value,
                                Controller controller) throws XPathException {

        Object indexObject = getIndex(doc, fingerprint);
        if (indexObject instanceof String) {
            // index is under construction
            throw new XPathException.Dynamic("Key definition is circular");
        }
        HashMap index = (HashMap)indexObject;
        if (index==null) {
            putIndex(doc, fingerprint, "Under Construction", controller);
            index = buildIndex(fingerprint, doc, controller);
            putIndex(doc, fingerprint, index, controller);
        }

        KeyDefinition definition = (KeyDefinition)getKeyDefinitions(fingerprint).get(0);
                    // the itemType and collation will be the same for all keys with the same name
        int itemType = definition.getItemType();
        Collator collation = definition.getCollation();

        Object val;
        if (itemType==Type.STRING) {
            if (collation==null) {
                val = value.getStringValue();
            } else {
                val = collation.getCollationKey(value.getStringValue());
            }
        } else if (itemType==Type.FLOAT || itemType==Type.DOUBLE ) {
            // special case: ensure NaN=NaN for indexing purposes
            val = value.convert(itemType);
            if (Double.isNaN(((NumericValue)val).asDouble())) {
                val = "NaN";
            }
        } else {
            val = value.convert(itemType);
        }

        ArrayList nodes = (ArrayList)index.get(val);
        if (nodes==null) {
            return EmptyIterator.getInstance();
        } else {
            return new ArrayListIterator(nodes);
        }
    }

    /**
    * Save the index associated with a particular key and a particular document. This
    * needs to be done in such a way that the index is discarded by the garbage collector
    * if the document is discarded. We therefore use a WeakHashMap indexed on the DocumentInfo,
    * which returns HashMap giving the index for each key fingerprint. This index is itself another
    * HashMap.
    * The methods need to be synchronized because several concurrent transformations (which share
    * the same KeyManager) may be creating indexes for the same or different documents at the same
    * time.
    */

    private synchronized void putIndex(DocumentInfo doc, int keyFingerprint, Object index, Controller controller) {
        if (docIndexes==null) {
            // it's transient, so it will be null when reloading a compiled stylesheet
            docIndexes = new WeakHashMap();
        }
        WeakReference indexRef = (WeakReference)docIndexes.get(doc);
        HashMap indexList;
        if (indexRef==null || indexRef.get()==null) {
            indexList = new HashMap();
            // ensure there is a firm reference to the indexList for the duration of a transformation
            controller.setUserData(doc, "key-index-list", indexList);
            docIndexes.put(doc, new WeakReference(indexList));
        } else {
            indexList = (HashMap)indexRef.get();
        }
        indexList.put(new Integer(keyFingerprint), index);
    }

    /**
    * Get the index associated with a particular key and a particular source document
    */

    private synchronized Object getIndex(DocumentInfo doc, int keyFingerprint) {
        if (docIndexes==null) {
            // it's transient, so it will be null when reloading a compiled stylesheet
            docIndexes = new WeakHashMap();
        }
        WeakReference ref = (WeakReference)docIndexes.get(doc);
        if (ref==null) return null;
        HashMap indexList = (HashMap)ref.get();
        if (indexList==null) return null;
        return indexList.get(new Integer(keyFingerprint));
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
