package net.sf.saxon.trans;
import net.sf.saxon.Controller;
import net.sf.saxon.Configuration;
import net.sf.saxon.pattern.Pattern;
import net.sf.saxon.pattern.NoNodeTest;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Navigator;
import net.sf.saxon.value.Type;
//import java.util.*;
import javax.xml.transform.TransformerException;
import net.sf.saxon.xpath.XPathException;
import java.io.Serializable;

    /**
    * A Mode is a collection of rules; the selection of a rule to apply to a given element
    * is determined by a Pattern.
    *
    * @author <A HREF="michael.h.kay@ntlworld.com>Michael H. Kay</A>
    */

public class Mode implements Serializable {

    public final static int DEFAULT_MODE = -1;
    public final static int ALL_MODES = -2;

    private Rule[] ruleDict = new Rule[101 + Type.MAX_NODE_TYPE];
    private String modeName;    // name of this mode in display form (the QName as originally written)
    private int sequence = 0;   // sequence number for the rules in this Mode

    /**
     * Default constructor - creates a Mode containing no rules
     */

    public Mode() {}

    /**
     * Construct a new Mode, copying the contents of an existing Mode
     * @param omniMode the existing mode. May be null, in which case it is not copied
     */

    public Mode(Mode omniMode) {
        if (omniMode != null) {
            for (int i=0; i<this.ruleDict.length; i++) {
                if (omniMode.ruleDict[i] != null) {
                    this.ruleDict[i] = new Rule(omniMode.ruleDict[i]);
                }
            }
            this.sequence = omniMode.sequence;
        }
    }

    /**
    * Set the name of this mode (for diagnostic output).
    */

    public void setModeName(String name) {
        modeName = name;
    }

	/**
	* Get the name of this mode (for diagnostic output). Note that this will be a QName, in display
	* form. If different prefixes have been used for the mode at different places in the stylesheet
	* it is unpredictable which prefix will be used in the tracing output.
    * @return the name of the mode, as a lexical QName, or "#default" for the default mode.
	*/

	public String getModeName() {
		return (modeName==null ? "#default" : modeName);
	}

    /**
    * Add a rule to the Mode. <br>
    * The rule effectively replaces any other rule for the same pattern/mode at the same or a lower
    * priority.
    * @param p a Pattern
    * @param obj the Object to return from getRule() when the supplied node matches this Pattern
    * @param precedence the import precedence of the rule
    * @param priority the explicit or implicit priority of the rule
    */

    public void addRule(Pattern p, Object obj, int precedence, double priority) {

        // System.err.println("Add rule, pattern = " + p.toString() + " class " + p.getClass() + ", priority=" + priority);

		// Ignore a pattern that will never match, e.g. "@comment"

		if (p instanceof NoNodeTest) {
			return;
		}

        // for fast lookup, we maintain one list for each element name for patterns that can only
        // match elements of a given name, one list for each node type for patterns that can only
        // match one kind of non-element node, and one generic list.
        // Each list is sorted in precedence/priority order so we find the highest-priority rule first

        int fingerprint = p.getFingerprint();
        int type = p.getItemType();

        int key = getList(fingerprint, type);
        // System.err.println("Fingerprint " + fingerprint + " key " + key + " type " + type);

		Rule newRule = new Rule(p, obj, precedence, priority, sequence++);

        Rule rule = ruleDict[key];
        if (rule==null) {
            ruleDict[key] = newRule;
            return;
        }

        // insert the new rule into this list before others of the same precedence/priority

        Rule prev = null;
        while (rule != null) {
            if ((rule.precedence < precedence) ||
                	( rule.precedence == precedence && rule.priority <= priority)) {
                newRule.next = rule;
                if (prev==null) {
                	ruleDict[key] = newRule;
                } else {
                	prev.next = newRule;
                }
                break;
            } else {
            	prev = rule;
            	rule = rule.next;
            }
        }
        if (rule==null) {
        	prev.next = newRule;
        	newRule.next = null;
        }
    }

    /**
    * Determine which list to use for a given pattern (we must also search the generic list)
    */

    public int getList(int fingerprint, int type) {

        if (type==Type.ELEMENT) {
            if (fingerprint==-1) {
                return Type.NODE;   // the generic list
            } else {
                return Type.MAX_NODE_TYPE +
                	(fingerprint % 101);
            }
        } else {
            return type;
        }
    }

    /**
    * Get the rule corresponding to a given Node, by finding the best Pattern match.
    * @param node the NodeInfo referring to the node to be matched
    * @return the object (e.g. a NodeHandler) registered for that element, if any (otherwise null).
    */

    public Object getRule(NodeInfo node, Controller controller) throws TransformerException {
    	// System.err.println("Get rule for " + Navigator.getPath(node));
        int fingerprint = node.getFingerprint();
        int type = node.getItemType();
        int key = getList(fingerprint, type);
        int policy = controller.getRecoveryPolicy();

        Rule specificRule = null;
        Rule generalRule = null;
        int specificPrecedence = -1;
        double specificPriority = Double.NEGATIVE_INFINITY;

        // search the specific list for this node type / node name

		// System.err.println("Hash key = " + key);

        if (key!=Type.NODE) {
            Rule r = ruleDict[key];
            while(r!=null) {
            	// if we already have a match, and the precedence or priority of this
            	// rule is lower, quit the search for a second match
            	if (specificRule != null) {
            		if (r.precedence < specificPrecedence ||
            		     (r.precedence==specificPrecedence && r.priority < specificPriority)) {
            			break;
            		}
            	}
            	//System.err.println("Testing " + Navigator.getPath(node) + " against " + r.pattern);
                if (r.pattern.matches(node, controller)) {
                	//System.err.println("Matches");

                    // is this a second match?
                    if (specificRule != null) {
                        if (r.precedence==specificPrecedence && r.priority==specificPriority) {
                            reportAmbiguity(node, specificRule, r, controller);
                        }
                        break;
                    }
                    specificRule = r;
                    specificPrecedence = r.precedence;
                    specificPriority = r.priority;
                    if (policy==Configuration.RECOVER_SILENTLY) {
                        break;                      // find the first; they are in priority order
                    }
                }
                r = r.next;
            }
        }

        // search the general list

        Rule r2 = ruleDict[Type.NODE];
        while (r2 != null) {
            if (r2.precedence < specificPrecedence ||
                 (r2.precedence == specificPrecedence && r2.priority < specificPriority)) {
                break;      // no point in looking at a lower priority rule than the one we've got
            }
            if (r2.pattern.matches(node, controller)) {
                // is it a second match?
                if (generalRule != null) {
                    if (r2.precedence == generalRule.precedence && r2.priority ==generalRule.priority) {
                        reportAmbiguity(node, r2, generalRule, controller);
                    }
                    break;
                } else {
                    generalRule = r2;
                    if (policy==Configuration.RECOVER_SILENTLY) {
                        break;                      // find only the first; they are in priority order
                    }
                }
            }
            r2 = r2.next;
        }

        if (specificRule!=null && generalRule==null)
            return specificRule.object;
        if (specificRule==null && generalRule!=null)
            return generalRule.object;
        if (specificRule!=null && generalRule!=null) {
            if (specificRule.precedence == generalRule.precedence &&
                specificRule.priority == generalRule.priority ) {
                // This situation is exceptional: we have a "specific" pattern and
                // a "general" pattern with the same priority. We have to select
                // the one that was added last
                // (Bug reported by Norman Walsh Jan 2002)
                Object result = (specificRule.sequence > generalRule.sequence ?
                                    specificRule.object :
                                    generalRule.object);

                if (policy!=Configuration.RECOVER_SILENTLY) {
                    reportAmbiguity(node, specificRule, generalRule, controller);
                }
                return result;
            }
            if (specificRule.precedence > generalRule.precedence ||
                 (specificRule.precedence == generalRule.precedence &&
                    specificRule.priority >= generalRule.priority)) {
                return specificRule.object;
            } else {
                return generalRule.object;
            }
        }
        return null;
    }

    /**
    * Get the rule corresponding to a given Node, by finding the best Pattern match, subject to a minimum
    * and maximum precedence. (This supports xsl:apply-imports)
    * @param node the NodeInfo referring to the node to be matched
    * @return the object (e.g. a NodeHandler) registered for that element, if any (otherwise null).
    */

    public Object getRule(NodeInfo node, int min, int max, Controller controller) throws XPathException {
        int fing = node.getFingerprint();
        int type = node.getItemType();
        int key = getList(fing, type);

        Rule specificRule = null;
        Rule generalRule = null;

        // search the the specific list for this node type / name

        if (key!=Type.NODE) {
            Rule r = ruleDict[key];
            while (r!=null) {
                if (r.precedence >= min && r.precedence <= max &&
                         r.pattern.matches(node, controller)) {
                    specificRule = r;
                    break;                      // find the first; they are in priority order
                }
                r = r.next;
            }
        }

        // search the generic list

        Rule r2 = ruleDict[Type.NODE];
        while (r2!=null) {
            if (r2.precedence >= min && r2.precedence <= max && r2.pattern.matches(node, controller)) {
                generalRule = r2;
                break;                      // find only the first; they are in priority order
            }
            r2 = r2.next;
        }
        if (specificRule!=null && generalRule==null)
            return specificRule.object;
        if (specificRule==null && generalRule!=null)
            return generalRule.object;
        if (specificRule!=null && generalRule!=null) {
            if (specificRule.precedence > generalRule.precedence ||
                 (specificRule.precedence == generalRule.precedence &&
                    specificRule.priority >= generalRule.priority)) {
                return specificRule.object;
            } else {
                return generalRule.object;
            }
        }
        return null;
    }

    /**
    * Get the rule corresponding to a given Node, by finding the next-best Pattern match
    * after the specified object.
    * @param node the NodeInfo referring to the node to be matched
    * @return the object (e.g. a NodeHandler) registered for that element, if any (otherwise null).
    */

    public Object getNextMatchRule(NodeInfo node, Object currentHandler, Controller controller) throws TransformerException {
    	//System.err.println("Get next-match rule for " + Navigator.getPath(node));
    	//System.err.println("Current handler is " + currentHandler);
        int fingerprint = node.getFingerprint();
        int type = node.getItemType();
        int key = getList(fingerprint, type);
        int policy = controller.getRecoveryPolicy();

        int currentPrecedence = -1;
        double currentPriority = -1.0;
        int currentSequence = -1;

        // First find the Rule object corresponding to the current handler

        Rule r = ruleDict[key];
        while (r!=null) {
            if (r.object==currentHandler) {
                currentPrecedence = r.precedence;
                currentPriority = r.priority;
                currentSequence = r.sequence;
                break;
            } else {
                r = r.next;
            }
        }
        if (r==null) {
            r = ruleDict[Type.NODE];
            while (r!=null) {
                if (r.object==currentHandler) {
                    currentPrecedence = r.precedence;
                    currentPriority = r.priority;
                    currentSequence = r.sequence;
                    break;
                } else {
                    r = r.next;
                }
            }
            if (r==null) {
                throw new TransformerException("Current template doesn't match current node");
            }
        }

        Rule specificRule = null;
        Rule generalRule = null;
        int specificPrecedence = -1;
        double specificPriority = Double.NEGATIVE_INFINITY;

        // search the specific list for this node type / node name

		// System.err.println("Hash key = " + key);

        if (key!=Type.NODE) {
            r = ruleDict[key];
            while(r!=null) {
            	// skip this rule unless it's "below" the current rule in search order
            	if ((r.precedence > currentPrecedence) ||
            	     (r.precedence == currentPrecedence &&
            	        (r.priority > currentPriority ||
            	            (r.priority == currentPriority && r.sequence >= currentSequence)))) {
            	     // skip this rule
            	} else {
            	     // quit the search on finding the second (recoverable error) match
                	if (specificRule != null) {
                		if (r.precedence < specificPrecedence ||
                		     (r.precedence==specificPrecedence && r.priority < specificPriority)) {
                			break;
                		}
                	}
                	//System.err.println("Testing " + Navigator.getPath(node) + " against " + r.pattern);
                    if (r.pattern.matches(node, controller)) {
                    	//System.err.println("Matches");

                        // is this a second match?
                        if (specificRule != null) {
                            if (r.precedence==specificPrecedence && r.priority==specificPriority) {
                                reportAmbiguity(node, specificRule, r, controller);
                            }
                            break;
                        }
                        specificRule = r;
                        specificPrecedence = r.precedence;
                        specificPriority = r.priority;
                        if (policy==Configuration.RECOVER_SILENTLY) {
                            break;                      // find the first; they are in priority order
                        }
                    }
                }
                r = r.next;
            }
        }

        // search the general list

        Rule r2 = ruleDict[Type.NODE];
        while (r2 != null) {
        	// skip this rule unless it's "after" the current rule in search order
        	if ((r2.precedence > currentPrecedence) ||
        	     (r2.precedence == currentPrecedence &&
        	        (r2.priority > currentPriority ||
        	            (r2.priority == currentPriority && r2.sequence >= currentSequence)))) {
        	     // skip this rule
        	} else {
                if (r2.precedence < specificPrecedence ||
                     (r2.precedence == specificPrecedence && r2.priority < specificPriority)) {
                    break;      // no point in looking at a lower priority rule than the one we've got
                }
                if (r2.pattern.matches(node, controller)) {
                    // is it a second match?
                    if (generalRule != null) {
                        if (r2.precedence == generalRule.precedence && r2.priority ==generalRule.priority) {
                            reportAmbiguity(node, r2, generalRule, controller);
                        }
                        break;
                    } else {
                        generalRule = r2;
                        if (policy==Configuration.RECOVER_SILENTLY) {
                            break;                      // find only the first; they are in priority order
                        }
                    }
                }
            }
            r2 = r2.next;
        }

        if (specificRule!=null && generalRule==null)
            return specificRule.object;
        if (specificRule==null && generalRule!=null)
            return generalRule.object;
        if (specificRule!=null && generalRule!=null) {
            if (specificRule.precedence == generalRule.precedence &&
                specificRule.priority == generalRule.priority ) {
                // This situation is exceptional: we have a "specific" pattern and
                // a "general" pattern with the same priority. We have to select
                // the one that was added last
                // (Bug reported by Norman Walsh Jan 2002)
                Object result = (specificRule.sequence > generalRule.sequence ?
                                    specificRule.object :
                                    generalRule.object);

                if (policy!=Configuration.RECOVER_SILENTLY) {
                    reportAmbiguity(node, specificRule, generalRule, controller);
                }
                return result;
            }
            if (specificRule.precedence > generalRule.precedence ||
                 (specificRule.precedence == generalRule.precedence &&
                    specificRule.priority >= generalRule.priority)) {
                return specificRule.object;
            } else {
                return generalRule.object;
            }
        }
        return null;
    }

    /**
    * Report an ambiguity, that is, the situation where two rules of the same
     * precedence and priority match the same node
     * @param node     The node that matches two or more rules
     * @param r1       The first rule that the node matches
     * @param r2       The second rule that the node matches
     * @param c        The controller for the transformation
    */

    private void reportAmbiguity(NodeInfo node, Rule r1, Rule r2, Controller c)
        throws TransformerException
    {
    	// don't report an error if the conflict is between two branches of the same
    	// Union pattern
    	if (r1.object == r2.object) {
    		return;
    	}
    	Pattern pat1 = r1.pattern;
    	Pattern pat2 = r2.pattern;

    	String path = "node";
    	try {
    	    path = Navigator.getPath(node);
    	} catch (Exception err) {
    	    // can fail if the node is incomplete, e.g. during whitespace stripping
    	}

        c.recoverableError(
            "Ambiguous rule match for " + path + "\n" +
            "Matches both \"" + pat1 + "\" on line " + pat1.getLineNumber() + " of " + pat1.getSystemId() +
            "\nand \"" + pat2 + "\" on line " + pat2.getLineNumber() + " of " + pat2.getSystemId(), null);

    }


    /**
    * Inner class Rule used to support the implementation
    */

    private static class Rule implements Serializable {
        public Pattern pattern;
        public Object object;
        public int precedence;
        public double priority;
        public int sequence;
        public Rule next;

        /**
         * Create a Rule
         * @param p the pattern that this rule matches
         * @param o the object invoked by this rule (usually a Template)
         * @param prec the precedence of the rule
         * @param prio the priority of the rule
         * @param seq a sequence number for ordering of rules
         */

        public Rule( Pattern p, Object o, int prec, double prio, int seq ) {
            pattern = p;
            object = o;
            precedence = prec;
            priority = prio;
            next = null;
            sequence = seq;
        }

        /**
         * Copy a rule, including the chain of rules linked to it
         * @param r
         */

        public Rule(Rule r) {
            this.pattern = r.pattern;
            this.object = r.object;
            this.precedence = r.precedence;
            this.priority = r.priority;
            this.sequence = r.sequence;
            if (r.next==null) {
                this.next = null;
            } else {
                this.next = new Rule(r.next);
            }
        }

    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
