package net.sf.saxon.tree;
import net.sf.saxon.event.Builder;
import net.sf.saxon.event.ReceiverOptions;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.DocumentInfo;
import net.sf.saxon.value.Type;

import org.xml.sax.Locator;

import javax.xml.transform.TransformerException;

import java.util.ArrayList;


/**
  * The Builder class is responsible for taking a stream of SAX events and constructing
  * a Document tree.
  * @author Michael H. Kay (michael.h.kay@ntlworld.com)
  */

public class TreeBuilder extends Builder

{
    private static AttributeCollection emptyAttributeCollection =
    				new AttributeCollection(null);

    private ParentNodeImpl currentNode;

    private NodeFactory nodeFactory;
    private int[] size = new int[100];          // stack of number of children for each open node
    private int depth = 0;
    private ArrayList arrays = new ArrayList();       // reusable arrays for creating nodes
    private int pendingElement;
    private AttributeCollection attributes;
    private int[] namespaces;
    private int namespacesUsed;

    private int nextNodeNumber = 1;

    /**
    * create a Builder and initialise variables
    */

    public TreeBuilder() {
        nodeFactory = new DefaultNodeFactory();
        // System.err.println("new TreeBuilder " + this);
    }

    /**
    * Set the Node Factory to use. If none is specified, the Builder uses its own.
    */

    public void setNodeFactory(NodeFactory factory) {
        nodeFactory = factory;
    }

  ////////////////////////////////////////////////////////////////////////////////////////
  // Implement the org.xml.sax.ContentHandler interface.
  ////////////////////////////////////////////////////////////////////////////////////////

    /**
    * Callback interface for SAX: not for application use
    */

    public void startDocument () throws TransformerException
    {
        // System.err.println("TreeBuilder: " + this + " Start document depth=" + depth);
        //failed = false;
        started = true;

        DocumentImpl doc;
        if (currentDocument==null) {
            // normal case
            doc = new DocumentImpl();
            currentDocument = doc;
        } else {
            // document node supplied by user
            if (!(currentDocument instanceof DocumentImpl)) {
                throw new TransformerException("Root node supplied is of wrong type");
            }
            doc = (DocumentImpl)currentDocument;
            if (doc.getFirstChild()!=null) {
                throw new TransformerException("Supplied document is not empty");
            }

        }

        doc.setSystemId(getSystemId());
        doc.setNamePool(namePool);
        doc.setNodeFactory(nodeFactory);
        currentNode = doc;
        depth = 0;
        size[depth] = 0;
        doc.sequence = 0;
        //charBuffer = new StringBuffer(4096);
        //doc.setCharacterBuffer(charBuffer);
        if (lineNumbering) {
            doc.setLineNumbering();
        }

        super.startDocument();
    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void endDocument () throws TransformerException
    {
        // System.err.println("TreeBuilder: " + this + " End document");
        if (currentNode==null) return;	// can be called twice on an error path
        currentNode.compact(size[depth]);
        currentNode = null;

        // we're not going to use this Builder again so give the garbage collector
        // something to play with
        arrays = null;

        super.endDocument();

    }

    /**
    * Callback interface for SAX: not for application use
    */

    public void setDocumentLocator (Locator locator) {
        // System.err.println("TreeBuilder#setLocator systemId=" + locator.getSystemId());
        this.locator = locator;
    }

    /**
    * Notify the start of an element
    */

    // TODO: splitting the namespaces and attributes into separate calls means that the
    // way the NodeFactory works should probably change. Rather than accumulate all the attributes
    // and namespaces in memory, these ought to be supplied to the newly constructed element node
    // one by one. But this would cause some problems, e.g. in deciding whether to use ElementImpl
    // or ElementWithAttributesImpl.

    public void startElement (int nameCode, int typeCode, int properties) throws TransformerException {
        // System.err.println("TreeBuilder: " + this + " Start element depth=" + depth);

        pendingElement = nameCode;
        namespaces = null;
        namespacesUsed = 0;
        attributes = null;

    }

    public void namespace (int namespaceCode, int properties) {
        if (namespaces==null) {
            namespaces = new int[5];
        }
        if (namespacesUsed == namespaces.length) {
            int[] ns2 = new int[namespaces.length * 2];
            System.arraycopy(namespaces, 0, ns2, 0, namespacesUsed);
            namespaces = ns2;
        }
        namespaces[namespacesUsed++] = namespaceCode;
    }

    public void attribute(int nameCode, int typeCode, CharSequence value, int properties)
    throws TransformerException {

        if ((properties & ReceiverOptions.DISABLE_ESCAPING) != 0) {
            throw new TransformerException("Cannot disable output escaping when writing a tree");
        }

        if (attributes==null) {
            attributes = new AttributeCollection(namePool);
        }
        attributes.addAttribute(nameCode, (typeCode==Type.ID ? "ID" : "CDATA"), value.toString());
    }

    public void startContent() throws TransformerException {
        // System.err.println("TreeBuilder: " + this + " startContent()");
        if (attributes == null) {
            attributes = emptyAttributeCollection;
        } else {
            attributes.compact();
        }

        if (namespaces == null) {
            namespaces = new int[0];
        }

        ElementImpl elem = nodeFactory.makeElementNode( currentNode,
                                                        pendingElement,
                                                        attributes,
                                                        namespaces,
                                                        namespacesUsed,
                                                        locator,
                                                        nextNodeNumber++);

        namespaces = null;
        namespacesUsed = 0;
        attributes = null;

        // the initial array used for pointing to children will be discarded when the exact number
        // of children in known. Therefore, it can be reused. So we allocate an initial array from
        // a pool of reusable arrays. A nesting depth of >20 is so rare that we don't bother.

        while (depth >= arrays.size()) {
            arrays.add(new NodeImpl[20]);
        }
        elem.useChildrenArray((NodeImpl[])arrays.get(depth));

        currentNode.addChild(elem, size[depth]++);
        if (depth >= size.length - 1) {
            int[] newsize = new int[size.length * 2];
            System.arraycopy(size, 0, newsize, 0, size.length);
            size = newsize;
        }
        size[++depth] = 0;
        namespacesUsed = 0;

    	if (currentNode instanceof DocumentInfo) {
    	    ((DocumentImpl)currentNode).setDocumentElement(elem);
    	}

        currentNode = elem;
    }

    /**
    * Notify the end of an element
    */

    public void endElement () throws TransformerException
    {
        // System.err.println("End element depth=" + depth);
        currentNode.compact(size[depth]);
        depth--;
        currentNode = (ParentNodeImpl)currentNode.getParentNode();
    }

    /**
    * Notify a text node. Adjacent text nodes must have already been merged
    */

    public void characters (CharSequence chars, int properties) throws TransformerException
    {
        // System.err.println("Characters: " + chars.toString() + " depth=" + depth);
        if (chars.length()>0) {

            if ((properties & ReceiverOptions.DISABLE_ESCAPING) != 0) {
                throw new TransformerException("Cannot disable output escaping when writing a tree");
            }

			// we rely on adjacent chunks of text having already been merged
            //TextImpl n = new TextImpl(currentNode, bufferStart, length);
			TextImpl n = new TextImpl(currentNode, chars.toString());
            currentNode.addChild(n, size[depth]++);
        }
    }

    /**
    * Notify a processing instruction
    */

    public void processingInstruction (String name, CharSequence remainder, int properties)
    {
    	int nameCode = namePool.allocate("", "", name);
        ProcInstImpl pi = new ProcInstImpl(nameCode, remainder.toString());
        currentNode.addChild(pi, size[depth]++);
        if (locator!=null) {
            pi.setLocation(locator.getSystemId(), locator.getLineNumber());
        }
    }

    /**
    * Notify a comment
    */

    public void comment (CharSequence chars, int properties) throws TransformerException {
        CommentImpl comment = new CommentImpl(chars.toString());
        currentNode.addChild(comment, size[depth]++);
    }


    /**
    * graftElement() allows an element node to be transferred from one tree to another.
    * This is a dangerous internal interface which is used only to contruct a stylesheet
    * tree from a stylesheet using the "literal result element as stylesheet" syntax.
    * The supplied element is grafted onto the current element as its only child.
    */

    public void graftElement(ElementImpl element) throws TransformerException {
        currentNode.addChild(element, size[depth]++);
    }

    /**
    * Set an unparsed entity URI for the document
    */

    public void setUnparsedEntity(String name, String uri, String publicId) {
        ((DocumentImpl)currentDocument).setUnparsedEntity(name, uri, publicId);
    }


    //////////////////////////////////////////////////////////////////////////////
    // Inner class DefaultNodeFactory. This creates the nodes in the tree.
    // It can be overridden, e.g. when building the stylesheet tree
    //////////////////////////////////////////////////////////////////////////////

    private class DefaultNodeFactory implements NodeFactory {

        public ElementImpl makeElementNode(
                NodeInfo parent,
                int nameCode,
                AttributeCollection attlist,
                int[] namespaces,
                int namespacesUsed,
                Locator locator,
                int sequenceNumber)

        {
            if (attlist.getLength()==0 && namespacesUsed==0) {

                // for economy, use a simple ElementImpl node

                ElementImpl e = new ElementImpl();
                String baseURI = null;
                int lineNumber = -1;

                if (locator!=null) {
                    baseURI = locator.getSystemId();
                    lineNumber = locator.getLineNumber();
                }

                e.initialise(nameCode, attlist, parent, baseURI, lineNumber, sequenceNumber);

                return e;

            } else {
                ElementWithAttributes e = new ElementWithAttributes();
                String baseURI = null;
                int lineNumber = -1;

                if (locator!=null) {
                    baseURI = locator.getSystemId();
                    lineNumber = locator.getLineNumber();
                }

                e.setNamespaceDeclarations(namespaces, namespacesUsed);

                e.initialise(nameCode, attlist, parent, baseURI, lineNumber, sequenceNumber);

                return e;
            }
        }
    }


}   // end of outer class TreeBuilder

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
