package net.sf.saxon.value;

import net.sf.saxon.expr.StaticProperty;

/**
* This class contains static methods to manipulate the cardinality
* property of a type.
* Cardinality of expressions is denoted by one of the values ONE_OR_MORE, ZERO_OR_MORE,
* ZERO_OR_ONE, EXACTLY_ONE, or EMPTY. These are combinations of the three bit-significant
* values ALLOWS_ZERO, ALLOWS_ONE, and ALLOWS_MANY.
*/

public final class Cardinality {

    /**
    * Private constructor: no instances allowed
    */

    private Cardinality() {}

    /**
     * Determine whether multiple occurrences are allowed
     */

    public final static boolean allowsMany(int cardinality) {
        return (cardinality & StaticProperty.CARDINALITY_ALLOWS_MANY) != 0;
    }

    /**
     * Determine whether empty sequence is allowed
     */

    public final static boolean allowsZero(int cardinality) {
        return (cardinality & StaticProperty.CARDINALITY_ALLOWS_ZERO) != 0;
    }

    /**
    * Form the union of two cardinalities. The cardinality of the expression "if (c) then e1 else e2"
    * is the union of the cardinalities of e1 and e2.
    * @param c1 a cardinality
    * @param c2 another cardinality
    * @return the cardinality that allows both c1 and c2
    */

    public final static int union(int c1, int c2) {
        int r = c1 | c2;
        // eliminate disallowed options
        if (r == (StaticProperty.CARDINALITY_ALLOWS_MANY |
                    StaticProperty.CARDINALITY_ALLOWS_ZERO ))
            r = StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE;
        return r;
    }

    /**
    * Test if one cardinality subsumes another. Cardinality c1 subsumes c2 if every option permitted
    * by c2 is also permitted by c1.
    * @param c1 a cardinality
    * @param c2 another cardinality
    * @return true if if every option permitted
    * by c2 is also permitted by c1.
    */

    public final static boolean subsumes(int c1, int c2) {
        return (c1|c2)==c1;
    }

    /**
     * Multiply two cardinalities
     */

    public final static int multiply(int c1, int c2) {
        if (c1==StaticProperty.EMPTY || c2==StaticProperty.EMPTY) {
            return StaticProperty.EMPTY;
        }
        if (c2==StaticProperty.CARDINALITY_EXACTLY_ONE) {
            return c1;
        }
        if (c1==StaticProperty.CARDINALITY_EXACTLY_ONE) {
            return c2;
        }
        if (c1==StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE && c2==StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE) {
            return StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE;
        }
        if (c1==StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE && c2==StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE) {
            return StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE;
        }
        return StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE;
    }

    /**
    * Display the cardinality
    */

    public static String toString(int cardinality) {
        switch (cardinality) {
            case StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE:
                return "zero or one";
            case StaticProperty.CARDINALITY_EXACTLY_ONE:
                return "exactly one";
            case StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE:
                return "zero or more";
            case StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE:
                return "one or more";
            default: throw new AssertionError("unknown cardinality value");
        }
    }

    /**
     * Get the occurence indicator representing the cardinality
     */

    public static String getOccurrenceIndicator(int cardinality) {
        switch (cardinality) {
            case StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_ONE:
                return "?";
            case StaticProperty.CARDINALITY_EXACTLY_ONE:
                return "";
            case StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE:
                return "*";
            case StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE:
                return "+";
            default: throw new AssertionError("unknown cardinality value");
        }
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
