package net.sf.saxon.value;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.Controller;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.functions.Component;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;
import java.util.TimeZone;
import java.util.SimpleTimeZone;
import java.util.StringTokenizer;

/**
* A value of type DateTime
*/

public final class DateTimeValue extends CalendarValue implements Comparable {

    // calendar and zoneSpecified together represent the localized value
    private GregorianCalendar calendar;
    private boolean zoneSpecified;

    // UTCDate is held as a redundant representation of the
    // normalized (UTC) value.
    private Date UTCDate = null;   // Always read this using getUTCDate


    /**
    * Constructor: create a dateTime value representing the nominal
    * date/time of this transformation run. Two calls within the same
    * transformation will always return the same answer.
    */

    public DateTimeValue(XPathContext context) {
        Controller c = context.getController();
        if (c==null) {
            // non-XSLT environment
            calendar = new GregorianCalendar();
        } else {
            calendar = c.getCurrentDateTime();
        }
        zoneSpecified = true;
    }

    /**
    * Constructor: create a dateTime value given a Java calendar object
    * @param calendar holds the date and time
    * @param tzSpecified indicates whether the timezone is specified
    */

    public DateTimeValue(GregorianCalendar calendar, boolean tzSpecified) {
        this.calendar = calendar;
        zoneSpecified = tzSpecified;
    }

    /**
    * Constructor: create a dateTime value from a supplied string, in
    * ISO 8601 format
    */

    public DateTimeValue(CharSequence s) throws XPathException {
        // input must have format yyyy-mm-ddThh:mm:ss[.fff*][([+|-]hh:mm | Z)]
        // TODO: rewrite this using regular expressions
        zoneSpecified = false;
        StringTokenizer tok = new StringTokenizer(trimWhitespace(s).toString(), "-:.+TZ", true);
        try {
            if (!tok.hasMoreElements()) badDate("too short");
            String part = (String)tok.nextElement();
            int era = +1;
            if (part.equals("+")) {
                part = (String)tok.nextElement();
            } else if (part.equals("-")) {
                era = -1;
                part = (String)tok.nextElement();
            }
            int year = Integer.parseInt(part) * era;
            if (part.length() < 4) badDate("year is less than four digits");
            if (year==0) badDate("year zero is not allowed");
            if (!tok.hasMoreElements()) badDate("too short");
            if (!(tok.nextElement()).equals("-")) badDate("wrong delimiter after year");

            if (!tok.hasMoreElements()) badDate("too short");
            part = (String)tok.nextElement();
            int month = Integer.parseInt(part);
            if (part.length() != 2) badDate("month must be two digits");
            if (month < 1 || month > 12) badDate("month is out of range");
            if (!tok.hasMoreElements()) badDate("too short");
            if (!tok.nextElement().equals("-")) badDate("wrong delimiter after month");

            if (!tok.hasMoreElements()) badDate("too short");
            part = (String)tok.nextElement();
            int day = Integer.parseInt(part);
            if (part.length() != 2) badDate("day must be two digits");
            if (day < 1 || day > 31) badDate("day is out of range");
            if (!tok.hasMoreElements()) badDate("too short");
            if (!tok.nextElement().equals("T")) badDate("wrong delimiter after day");

            if (!tok.hasMoreElements()) badDate("too short");
            part = (String)tok.nextElement();
            int hour = Integer.parseInt(part);
            if (part.length() != 2) badDate("hour must be two digits");
            if (hour > 23) badDate("hour is out of range");
            if (!tok.hasMoreElements()) badDate("too short");
            if (!tok.nextElement().equals(":")) badDate("wrong delimiter after hour");

            if (!tok.hasMoreElements()) badDate("too short");
            part = (String)tok.nextElement();
            int minute = Integer.parseInt(part);
            if (part.length() != 2) badDate("minute must be two digits");
            if (minute > 59) badDate("minute is out of range");
            if (!tok.hasMoreElements()) badDate("too short");
            if (!tok.nextElement().equals(":")) badDate("wrong delimiter after minute");

            if (!tok.hasMoreElements()) badDate("too short");
            part = (String)tok.nextElement();
            int second = Integer.parseInt(part);
            if (part.length() != 2) badDate("second must be two digits");
            if (hour > 61) badDate("second is out of range");

            int millisecond = 0;
            int tz = 0;

            int state = 0;
            while (tok.hasMoreElements()) {
                if (state==9) {
                    badDate("characters after the end");
                }
                String delim = (String)tok.nextElement();
                if (delim.equals(".")) {
                    if (state != 0) {
                        badDate("decimal separator occurs twice");
                    }
                    part = (String)tok.nextElement();
                    double fractionalSeconds = Double.parseDouble("." + part);
                    millisecond = (int)(Math.round(fractionalSeconds * 1000));
                    state = 1;
                } else if (delim.equals("Z")) {
                    if (state > 1) {
                        badDate("Z cannot occur here");
                    }
                    zoneSpecified = true;
                    tz = 0;
                    state = 9;  // we've finished
                } else if (delim.equals("+") || delim.equals("-")) {
                    if (state > 1) {
                        badDate(delim + " cannot occur here");
                    }
                    state = 2;
                    zoneSpecified = true;
                    if (!tok.hasMoreElements()) badDate("missing timezone");
                    part = (String)tok.nextElement();
                    if (part.length() != 2) badDate("timezone hour must be two digits");
                    tz = Integer.parseInt(part) * 60;
                    if (tz > 14*60) badDate("timezone hour is out of range");
                    if (delim.equals("-")) tz = -tz;

                } else if (delim.equals(":")) {
                    if (state != 2) {
                        badDate("colon cannot occur here");
                    }
                    state = 9;
                    part = (String)tok.nextElement();
                    int tzminute = Integer.parseInt(part);
                    if (part.length() != 2) badDate("timezone minute must be two digits");
                    if (tzminute > 59) badDate("timezone minute is out of range");
                    if (tz<0) tzminute = -tzminute;
                    tz += tzminute;
                } else {
                    badDate("timezone format is incorrect");
                }
            }

            if (state == 2 || state == 3) {
                badDate("timezone incomplete");
            }

            // create a calendar that uses the timezone actually specified, or GMT otherwise
            //calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
            TimeZone zone = new SimpleTimeZone(tz*60000, "LLL");
            calendar = new GregorianCalendar(zone);
            calendar.setLenient(false);
            calendar.set(year, month-1, day, hour, minute, second);
            calendar.set(Calendar.MILLISECOND, millisecond);
            calendar.set(Calendar.ZONE_OFFSET, tz*60000);
            calendar.set(Calendar.DST_OFFSET, 0);

            try {
                calendar.getTime();
            } catch (IllegalArgumentException err) {
                badDate("non-existent date");
            }


        } catch (NumberFormatException err) {
            badDate("non-numeric component");
        }
    }

    private void badDate(String msg) throws XPathException {
        throw new XPathException.Dynamic("Invalid dateTime value (" + msg + ")");
    }


    /**
    * Get the UTC date/time corresponding to this dateTime. This normalizes
    * the value to incorporate the timezone information, for example
    * 2002-01-01T07:00:00-05:00 gives the same answer as 2002-01-01T12:00:00Z
    */

    public Date getUTCDate() {
        // implement this as a memo function
        if (UTCDate==null) {
            UTCDate = calendar.getTime();
        }
        return UTCDate;
    }

    /**
     * Get the Calendar object representing the value of this DateTime
     */

    public Calendar getCalendar() {
        return calendar;
    }

    /**
    * Convert to target data type
    * @param requiredType an integer identifying the required atomic type
    * @return an AtomicValue, a value of the required type
    * @throws XPathException if the conversion is not possible
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.DATE_TIME:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;
        case Type.DATE:
            String ds = getStringValue();
            int sep = ds.indexOf('T');
            if (zoneSpecified) {
                int z = ds.indexOf('Z', sep);
                if (z < 0) {
                    z = ds.indexOf('+', sep);
                }
                if (z < 0) {
                    z = ds.indexOf('-', sep);
                }
                if (z < 0) {
                    // something's gone wrong
                    throw new IllegalArgumentException("Internal date formatting error " + ds);
                }
                return new DateValue(ds.substring(0, sep) + ds.substring(z));
            } else {
                return new DateValue(ds.substring(0, sep));
            }
        case Type.TIME:
            ds = getStringValue();
            sep = ds.indexOf('T');
            return new TimeValue(ds.substring(sep+1));

        case Type.G_YEAR:
            return(convert(Type.DATE).convert(Type.G_YEAR));

        case Type.G_YEAR_MONTH:
            return(convert(Type.DATE).convert(Type.G_YEAR_MONTH));

        case Type.G_MONTH:
            return(convert(Type.DATE).convert(Type.G_MONTH));

        case Type.G_MONTH_DAY:
            return(convert(Type.DATE).convert(Type.G_MONTH_DAY));

        case Type.G_DAY:
            return(convert(Type.DATE).convert(Type.G_DAY));

        case Type.STRING:
            return new StringValue(getStringValue());
        default:
            throw new XPathException.Dynamic("Cannot convert dateTime to " +
                                     Type.getTypeName(requiredType));
        }
    }

    /**
    * Convert to string
    * @return ISO 8601 representation. The value returned is the localized representation,
     * that is it uses the timezone contained within the value itself.
    */

    public String getStringValue() {

        StringBuffer sb = new StringBuffer();
        int year = calendar.get(Calendar.YEAR);
        if (year<0) {
            sb.append('-');
            year = -year;
        }
        appendString(sb, year, (year>9999 ? (calendar.get(Calendar.YEAR)+"").length() : 4));
        sb.append('-');
        appendString(sb, calendar.get(Calendar.MONTH)+1, 2);
        sb.append('-');
        appendString(sb, calendar.get(Calendar.DATE), 2);
        sb.append('T');
        appendString(sb, calendar.get(Calendar.HOUR_OF_DAY), 2);
        sb.append(':');
        appendString(sb, calendar.get(Calendar.MINUTE), 2);
        sb.append(':');
        appendSeconds(calendar, sb);

        if (zoneSpecified) {
            appendTimezone(calendar, sb);
        }

        return sb.toString();

    }

    protected static void appendString(StringBuffer sb, int value, int size) {
        String s = "000"+value;
        sb.append( s.substring(s.length()-size) );
    }

    protected static void appendSeconds(Calendar calendar, StringBuffer sb) {
        appendString(sb, calendar.get(Calendar.SECOND), 2);
        int millis = calendar.get(Calendar.MILLISECOND);
        if (millis != 0) {
            sb.append('.');
            String m = calendar.get(Calendar.MILLISECOND) + "";
            while (m.length() < 3) m = "0" + m;
            while (m.endsWith("0")) m = m.substring(0, m.length()-1);
            sb.append(m);
        }
        return;
    }

    /**
     * Add a string representation of the timezone, typically
     * formatted as "Z" or "+03:00" or "-10:00", to the supplied
     * string buffer
     * @param calendar The Calendar whose timezone value is required
     * @param sb The StringBuffer that will be updated with the resulting string
     * representation
     */

    public static void appendTimezone(Calendar calendar, StringBuffer sb) {
        int timeZoneOffset = (calendar.get(Calendar.ZONE_OFFSET) +
                              calendar.get(Calendar.DST_OFFSET)) / 60000;
        appendTimezone(timeZoneOffset, sb);
    }

     /**
     * Add a string representation of the timezone, typically
     * formatted as "Z" or "+03:00" or "-10:00", to the supplied
     * string buffer
     * @param timeZoneOffset The timezone offset in minutes
     * @param sb The StringBuffer that will be updated with the resulting string
     * representation
     */

    protected static void appendTimezone(int timeZoneOffset, StringBuffer sb) {
        if (timeZoneOffset == 0) {
            sb.append('Z');
        } else {
            sb.append((timeZoneOffset<0 ? "-" : "+"));
            int tzo = timeZoneOffset;
            if (tzo < 0) tzo = -tzo;
            int tzhours = tzo / 60;
            appendString(sb, tzhours, 2);
            sb.append(':');
            int tzminutes = tzo % 60;
            appendString(sb, tzminutes, 2);
        }
        return;
    }

    /**
    * Determine the data type of the exprssion
    * @return Type.DATE_TIME,
    */

    public int getItemType() {
        return Type.DATE_TIME;
    }

    /**
     * Return a dateTime with the same localized value, but
     * without the timezone component
     * @return the result of removing the timezone
     */

    public CalendarValue removeTimezone() {
        return new DateTimeValue(calendar, false);
    }

    /**
     * Return a date, time, or dateTime with the same normalized value, but
     * in a different timezone
     * @return the date/time in the new timezone
     * @throws XPathException
     */

    public CalendarValue setTimezone(SecondsDurationValue tz) throws XPathException {
        if (zoneSpecified) {
            SimpleTimeZone stz = new SimpleTimeZone((int)(tz.getLengthInSeconds()*1000), "xxz");
            GregorianCalendar cal = new GregorianCalendar(stz);
            cal.setTime(getUTCDate());
            return new DateTimeValue(cal, true);
        } else {
            StringBuffer sb = new StringBuffer();
            sb.append(getStringValue());
            appendTimezone((int)(tz.getLengthInSeconds()/60.0), sb);
            return new DateTimeValue(sb);
        }
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(DateTimeValue.class)) return 0;
        if (required.isAssignableFrom(Date.class)) return 1;
        if (required==String.class || required==CharSequence.class) return 3;

        return Integer.MAX_VALUE;
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target.isAssignableFrom(Date.class)) {
            return getUTCDate();
        } else if (target.isAssignableFrom(DateTimeValue.class)) {
            return this;
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==Object.class) {
            return getStringValue();
        } else {
            throw new XPathException.Dynamic("Conversion of dateTime to " + target.getName() +
                        " is not supported");
        }
    }

    /**
    * Get a component of the value. Returns null if the timezone component is
    * requested and is not present.
    */

    public AtomicValue getComponent(int component) throws XPathException {
        switch (component) {
        case Component.YEAR:
            return new IntegerValue(calendar.get(Calendar.YEAR));
        case Component.MONTH:
            return new IntegerValue(calendar.get(Calendar.MONTH) + 1);
        case Component.DAY:
            return new IntegerValue(calendar.get(Calendar.DATE));
        case Component.HOURS:
            return new IntegerValue(calendar.get(Calendar.HOUR_OF_DAY));
        case Component.MINUTES:
            return new IntegerValue(calendar.get(Calendar.MINUTE));
        case Component.SECONDS:
            StringBuffer sb = new StringBuffer();
            appendSeconds(calendar, sb);
            return new DecimalValue(sb.toString());
        case Component.TIMEZONE:
            if (zoneSpecified) {
                int tzsecs = (calendar.get(Calendar.ZONE_OFFSET) +
                              calendar.get(Calendar.DST_OFFSET)) / 1000;
                return SecondsDurationValue.fromSeconds((double)tzsecs);
                //StringBuffer sb2 = new StringBuffer();
                //appendTimezone(calendar, sb2);
                //return new StringValue(sb2);
            } else {
                return null;
            }
        default:
            throw new IllegalArgumentException("Unknown component for dateTime: " + component);
        }
    }


    /**
    * Compare the value to another dateTime value
    * @param other The other dateTime value
    * @return negative value if this one is the earler, 0 if they are chronologically equal,
    * positive value if this one is the later. For this purpose, dateTime values with an unknown
    * timezone are considered to be UTC values (the Comparable interface requires
    * a total ordering).
    * @throws ClassCastException if the other value is not a DateTimeValue (the parameter
    * is declared as Object to satisfy the Comparable interface)
    */

    public int compareTo(Object other) {
        if (!(other instanceof DateTimeValue)) {
            throw new ClassCastException("DateTime values are not comparable to " + other.getClass());
        }
        return getUTCDate().compareTo(((DateTimeValue)other).getUTCDate());
    }

    public boolean equals(Object other) {
        return compareTo(other) == 0;
    }

    public int hashCode() {
        return getUTCDate().hashCode();
    }
    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "dateTime (" + getStringValue() + ")" );
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

