package net.sf.saxon.value;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.om.NamePool;

import java.math.BigDecimal;

/**
* A decimal value
*/

public final class DecimalValue extends NumericValue {

    private final static int DIVIDE_PRECISION = 18;

    private BigDecimal value;

    /**
    * Constructor supplying a BigDecimal
    * @param value the value of the DecimalValue
    */

    public DecimalValue(BigDecimal value) {
        this.value = value;
        loseTrailingZeros();
    }

    /**
    * Constructor supplying a String
    * @param in the value of the DecimalValue
    */

    public DecimalValue(CharSequence in) throws XPathException {
        try {
            this.value = new BigDecimal(trimWhitespace(in).toString());
            loseTrailingZeros();
        } catch (NumberFormatException err) {
            throw new XPathException.Dynamic("Cannot convert string '" + in + "' to decimal");
        }
    }

    /**
    * Constructor supplying a double
    * @param in the value of the DecimalValue
    */

    public DecimalValue(double in) throws XPathException {
        try {
            this.value = new BigDecimal(in);
            loseTrailingZeros();
        } catch (NumberFormatException err) {
            // Must be a special value such as NaN or infinity
            throw new XPathException.Dynamic("Cannot convert double [" + in + "] to decimal");
        }
    }

    /**
    * Constructor supplying a long integer
    * @param in the value of the DecimalValue
    */

    public DecimalValue(long in) throws XPathException {
        this.value = new BigDecimal(in);
    }

    /**
    * Remove insignificant trailing zeros (the Java BigDecimal class retains trailing zeros,
    * but the XPath 2.0 xs:decimal type does not)
    */

    private void loseTrailingZeros() {
        // this is an odd way to do it, but it seems the simplest method available
        if (value.scale() > 0) {
            String s = value.toString();
            int len = s.length();
            if (s.charAt(len-1) != '0') return;
            while (s.charAt(len-1) == '0') {
                len--;
            }
            value = new BigDecimal(s.substring(0, len));
        }
    }

    /**
    * Get the value
    */

    public BigDecimal getValue() {
        return value;
    }

    /**
     * Get the hashCode. This must conform to the rules for other NumericValue hashcodes
     * @see NumericValue#hashCode
     */

    public int hashCode() {
        BigDecimal round = value.setScale(0, BigDecimal.ROUND_DOWN);
        long value = round.longValue();
        if (value > Integer.MIN_VALUE && value < Integer.MAX_VALUE) {
            return (int)value;
        } else {
            return new Double(this.asDouble()).hashCode();
        }
    }

    public boolean effectiveBooleanValue(XPathContext context) {
        return value.signum() != 0;
    }

    /**
    * Convert to target data type
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.BOOLEAN:
                // 0.0 => false, anything else => true
            return BooleanValue.get(value.signum()!=0);
        case Type.NUMBER:
        case Type.DECIMAL:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;
        case Type.INTEGER:
            BigDecimal round = value.setScale(0, BigDecimal.ROUND_DOWN);
            return new IntegerValue(round.longValue());
        case Type.DOUBLE:
            return new DoubleValue(value.doubleValue());
        case Type.FLOAT:
            return new FloatValue(value.floatValue());
        case Type.STRING:
            return new StringValue(getStringValue());
        default:
            throw new XPathException.Dynamic("Cannot convert decimal to " +
                                     Type.getTypeName(requiredType));
        }
    }

    /**
    * Get the value as a String
    * @return a String representation of the value
    */

    public String getStringValue() {
        String s = value.toString();
        //if (value.scale()==0) {
        //    s += ".0";
        //}
        return s;
    }

    /**
    * Determine the data type of the expression
    * @return Type.DECIMAL
    */

    public int getItemType() {
        return Type.DECIMAL;
    }

    /**
    * Negate the value
    */

    public NumericValue negate() {
        return new DecimalValue(value.negate());
    }

    /**
    * Implement the XPath floor() function
    */

    public NumericValue floor() {
        return new DecimalValue(value.setScale(0, BigDecimal.ROUND_FLOOR));
    }

    /**
    * Implement the XPath ceiling() function
    */

    public NumericValue ceiling() {
        return new DecimalValue(value.setScale(0, BigDecimal.ROUND_CEILING));
    }

    /**
    * Implement the XPath round() function
    */

    public NumericValue round() {
        // The XPath rules say that we should round to the nearest integer, with .5 rounding towards
        // positive infinity. Unfortunately this is not one of the rounding modes that the Java BigDecimal
        // class supports.

        // If the value is positive, we use ROUND_HALF_UP; if it is negative, we use ROUND_HALF_DOWN (here "UP"
        // means "away from zero")

        switch (value.signum()) {
            case -1:
                return new DecimalValue(value.setScale(0, BigDecimal.ROUND_HALF_DOWN));
            case 0:
                return this;
            case +1:
                return new DecimalValue(value.setScale(0, BigDecimal.ROUND_HALF_UP));
            default:
                // can't happen
                return this;
        }

    }

    /**
    * Implement the XPath round-to-half-even() function
    */

    public NumericValue roundToHalfEven(int scale) {
        if (scale<0) {
            try {
                return (DecimalValue)((IntegerValue)convert(Type.INTEGER)).roundToHalfEven(scale).convert(Type.DECIMAL);
            } catch (XPathException err) {
                throw new IllegalArgumentException("internal error in integer-decimal conversion");
            }
        } else {
            return new DecimalValue(value.setScale(scale, BigDecimal.ROUND_HALF_EVEN));
        }
    }

    /**
    * Determine whether the value is a whole number, that is, whether it compares
    * equal to some integer
    */

    public boolean isWholeNumber() {
        return value.scale()==0 ||
               value.equals(value.setScale(0, BigDecimal.ROUND_DOWN));
    }

    /**
    * Evaluate a binary arithmetic operator.
    */

    public NumericValue arithmetic(int operator, NumericValue other) throws XPathException {
        if (other instanceof DecimalValue) {
            try {
                switch(operator) {
                    case Tokenizer.PLUS:
                        return new DecimalValue(value.add(((DecimalValue)other).value));
                    case Tokenizer.MINUS:
                        return new DecimalValue(value.subtract(((DecimalValue)other).value));
                    case Tokenizer.MULT:
                        return new DecimalValue(value.multiply(((DecimalValue)other).value));
                    case Tokenizer.DIV:
                        int scale = value.scale() + ((DecimalValue)other).value.scale() + DIVIDE_PRECISION;
                        BigDecimal result = value.divide(((DecimalValue)other).value, scale, BigDecimal.ROUND_HALF_DOWN);
                        return new DecimalValue(result);
                    case Tokenizer.MOD:
                        BigDecimal quotient = value.divide(((DecimalValue)other).value, ((DecimalValue)other).value.scale(), BigDecimal.ROUND_DOWN);
                        BigDecimal remainder = value.subtract(quotient.multiply(((DecimalValue)other).value));
                        return new DecimalValue(remainder);
                    default:
                        throw new AssertionError("Unknown operator");
                }
            } catch (ArithmeticException err) {
                throw new XPathException.Dynamic(err);
            }
        } else if (other instanceof IntegerValue) {
            return arithmetic(operator, (DecimalValue)other.convert(Type.DECIMAL));
        } else {
            return ((NumericValue)convert(other.getItemType())).arithmetic(operator, other);
        }
    }

    /**
    * Compare the value to another numeric value
    */

    public int compareTo(Object other) {
        if (other instanceof IntegerValue) {
            try {
                return compareTo(((NumericValue)other).convert(Type.DECIMAL));
            } catch (XPathException err) {
                throw new AssertionError("Conversion of integer to decimal should never fail");
            }
        } else if (other instanceof DecimalValue) {
            return value.compareTo(((DecimalValue)other).value);
        } else {
            return super.compareTo(other);
        }
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(DecimalValue.class)) return 0;

        if (required==BigDecimal.class) return 1;
        if (required==boolean.class) return 16;
        if (required==Boolean.class) return 17;
        if (required==byte.class) return 14;
        if (required==Byte.class) return 15;
        if (required==char.class) return 12;
        if (required==Character.class) return 13;
        if (required==double.class) return 2;
        if (required==Double.class) return 3;
        if (required==float.class) return 4;
        if (required==Float.class) return 5;
        if (required==int.class) return 8;
        if (required==Integer.class) return 9;
        if (required==long.class) return 6;
        if (required==Long.class) return 7;
        if (required==short.class) return 10;
        if (required==Short.class) return 11;
        if (required==String.class || required==CharSequence.class) return 18;

        return Integer.MAX_VALUE;
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target==Object.class || target.isAssignableFrom(BigDecimal.class)) {
            return value;
        } else if (target.isAssignableFrom(DecimalValue.class)) {
            return this;
        } else if (target==boolean.class) {
            BooleanValue bval = (BooleanValue)convert(Type.BOOLEAN);
            return new Boolean(bval.getValue());
        } else if (target==Boolean.class) {
            BooleanValue bval = (BooleanValue)convert(Type.BOOLEAN);
            return new Boolean(bval.getValue());
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class || target==Double.class) {
            return new Double(value.doubleValue());
        } else if (target==float.class || target==Float.class) {
            return new Float(value.floatValue());
        } else if (target==long.class || target==Long.class) {
            return new Long(value.longValue());
        } else if (target==int.class || target==Integer.class) {
            return new Integer(value.intValue());
        } else if (target==short.class || target==Short.class) {
            return new Short(value.shortValue());
        } else if (target==byte.class || target==Byte.class) {
            return new Byte(value.byteValue());
        } else if (target==char.class || target==Character.class) {
            return new Character((char)value.intValue());
        } else {
            throw new XPathException.Dynamic("Conversion of decimal to " + target.getName() +
                        " is not supported");
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "decimal (" + getStringValue() + ")" );
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file except the asStringXT() and zeros() methods (not currently used).
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (xt) are Copyright (C) (James Clark). All Rights Reserved.
//
// Contributor(s): none.
//

