package net.sf.saxon.value;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.Tokenizer;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.pattern.Pattern;
import net.sf.saxon.om.NamePool;

import java.lang.Math;
import java.util.regex.Matcher;

/**
* A numeric (double precision floating point) value
*/

public final class DoubleValue extends NumericValue {

    public final static DoubleValue ZERO = new DoubleValue(0.0);
    public final static DoubleValue NaN = new DoubleValue(Double.NaN);

    private double value;

    /**
    * Constructor supplying a double
    * @param value the value of the NumericValue
    */

    public DoubleValue(double value) {
        this.value = value;
    }

    /**
    * Constructor supplying a String
    * @param s the numeric value expressed as a String
    */

    public DoubleValue(String s) {
        this.value = Value.stringToNumber(s);
    }

    /**
    * Get the value
    */

    public double getValue() {
        return value;
    }

    public double asDouble() {
        return value;
    }

    /**
     * Get the hashCode. This must conform to the rules for other NumericValue hashcodes
     * @see NumericValue#hashCode
     */

    public int hashCode() {
        if (value > Integer.MIN_VALUE && value < Integer.MAX_VALUE) {
            return (int)value;
        } else {
            return new Double(value).hashCode();
        }
    }

    /**
     * Test whether the value is the double/float value NaN
     */

    public boolean isNaN() {
        return Double.isNaN(value);
    }

    public boolean effectiveBooleanValue(XPathContext context) {
        return (value!=0.0 && !Double.isNaN(value));
    }

    /**
    * Convert to target data type
    * @param requiredType an integer identifying the required atomic type
    * @return an AtomicValue, a value of the required type
    * @throws XPathException if the conversion is not possible
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.BOOLEAN:
            return BooleanValue.get(asBoolean());
        case Type.DOUBLE:
        case Type.NUMBER:
        case Type.ATOMIC:
        case Type.ITEM:
            return this;
        case Type.INTEGER:
            if (Double.isNaN(value)) {
                throw new XPathException.Dynamic("Cannot convert double NaN to an integer");
            }
            if (value > Long.MAX_VALUE || value < Long.MIN_VALUE) {
                throw new XPathException.Dynamic("Double " + value + " is too big to convert to xs:integer");
            }
            return new IntegerValue((long)value);
        case Type.DECIMAL:
            return new DecimalValue(value);
        case Type.FLOAT:
            return new FloatValue((float)value);
        case Type.STRING:
            return new StringValue(getStringValue());
        default:
            throw new XPathException.Dynamic("Cannot convert double to " +
                                     Type.getTypeName(requiredType));
        }
    }

    private static java.util.regex.Pattern nonExponentialPattern =
            java.util.regex.Pattern.compile(
                    "(-?[0-9])([0-9]+?)(0*)\\.([0-9]*)");

    public String getStringValue() {
        if (value==0.0) {
            return "0";
        }
        if (Double.isInfinite(value)) {
            return (value > 0 ? "INF" : "-INF");
        }
        if (Double.isNaN(value)) {
            return "NaN";
        }
        double absval = Math.abs(value);
        String s = Double.toString(value);
        if (absval < 1.0e-6 || absval >= 1.0e+6) {
            if (s.indexOf('E')<0) {
                // need to use scientific notation, but Java isn't using it
                // (Java's cutoff is 1.0E7, while XPath's is 1.0E6)
                // So we have for example -2000000.0 rather than -2.0e6
                StringBuffer sb = new StringBuffer();
                Matcher matcher = nonExponentialPattern.matcher(s);
                if (matcher.matches()) {
                    sb.append(matcher.group(1));
                    sb.append('.');
                    sb.append(matcher.group(2));
                    String fraction = matcher.group(4);
                    if (fraction.equals("0")) {
                        sb.append('E');
                        sb.append(matcher.group(2).length() + matcher.group(3).length());
                        return sb.toString();
                    } else {
                        sb.append(matcher.group(3));
                        sb.append(matcher.group(4));
                        sb.append('E');
                        sb.append(matcher.group(2).length() + matcher.group(3).length());
                        return sb.toString();
                    }
                } else {
                    // fallback, this shouldn't happen
                    return s;
                }
            } else {
                return s;
            }
        }
        int len = s.length();
        if (s.endsWith("E0")) {
            s = s.substring(0, len - 2);
        }
        if (s.endsWith(".0")) {
            return s.substring(0, len - 2);
        }
        int e = s.indexOf('E');
        if (e < 0) {
            //return s;
            // For some reason, Double.toString() in Java can return strings such as "0.0040"
            // so we remove any trailing zeros
            while (s.charAt(len - 1) == '0' && s.charAt(len - 2) != '.') {
                s = s.substring(0, --len);
            }
            return s;
        }
        int exp = Integer.parseInt(s.substring(e + 1));
        String sign;
        if (s.charAt(0) == '-') {
            sign = "-";
            s = s.substring(1);
            --e;
        }
        else
            sign = "";
        int nDigits = e - 2;
        if (exp >= nDigits) {
            return sign + s.substring(0, 1) + s.substring(2, e) + zeros(exp - nDigits);
        } else if (exp > 0) {
            return sign + s.substring(0, 1) + s.substring(2, 2 + exp) + "." + s.substring(2 + exp, e);
        } else {
            while (s.charAt(e-1) == '0') e--;
            return sign + "0." + zeros(-1 - exp) + s.substring(0, 1) + s.substring(2, e);
        }
//        return new java.math.BigDecimal(value).toString();
    }

    static private String zeros(int n) {
        char[] buf = new char[n];
        for (int i = 0; i < n; i++)
            buf[i] = '0';
        return new String(buf);
    }

    /**
    * Convert the value to a boolean
    * @return false if zero, true otherwise
    */

    private boolean asBoolean() {
        return (value!=0.0 && !Double.isNaN(value));
    }

    /**
    * Determine the data type of the expression
    * @return Type.DOUBLE
    */

    public int getItemType() {
        return Type.DOUBLE;
    }

    /**
    * Negate the value
    */

    public NumericValue negate() {
        return new DoubleValue(-value);
    }

    /**
    * Implement the XPath floor() function
    */

    public NumericValue floor() {
        return new DoubleValue(Math.floor(value));
    }

    /**
    * Implement the XPath ceiling() function
    */

    public NumericValue ceiling() {
        return new DoubleValue(Math.ceil(value));
    }

    /**
    * Implement the XPath round() function
    */

    public NumericValue round() {
        if (Double.isNaN(value)) {
            return this;
        }
        if (Double.isInfinite(value)) {
            return this;
        }
        if (value == 0.0) {
            return this;    // handles the negative zero case
        }
        if (value > -0.5 && value < 0.0) {
            return new DoubleValue(-0.0);
        }
        if (value > Long.MIN_VALUE && value < Long.MAX_VALUE) {
            return new DoubleValue(Math.round(value));
        }

        // A double holds fewer significant digits than a long. Therefore,
        // if the double is outside the range of a long, it cannot have
        // any signficant digits after the decimal point. So in this
        // case, we return the original value unchanged

        return this;
    }

    /**
    * Implement the XPath round-to-half-even() function
    */

    public NumericValue roundToHalfEven(int scale) {
        if (Double.isNaN(value)) return this;
        if (Double.isInfinite(value)) return this;
        if (value==0.0) return this;    // handles the negative zero case

        // Convert to a scaled integer, by multiplying by 10^scale

        double factor = Math.pow(10, scale+1);
        double d = Math.abs(value * factor);

        // Now apply any rounding needed, using the "round half to even" rule

        double rem = d % 10;
        if (rem > 5) {
            d += (10-rem);
        } else if (rem < 5){
            d -= rem;
        } else {
            // round half to even - check the last bit
            if ((d % 20) == 15) {
                d +=5 ;
            } else {
                d -=5;
            }
        }

        // Now convert back to the original magnitude

        d /= factor;
        if (value < 0) {
            d = 0.0 -d;
        }
        return new DoubleValue(d);

    }

    /**
    * Determine whether the value is a whole number, that is, whether it compares
    * equal to some integer
    */

    public boolean isWholeNumber() {
        return value == Math.floor(value);
    }

    /**
    * Evaluate a binary arithmetic operator.
    */

    public NumericValue arithmetic(int operator, NumericValue other) throws XPathException {
        if (other instanceof DoubleValue) {
            switch(operator) {
                case Tokenizer.PLUS:
                    return new DoubleValue(value + ((DoubleValue)other).value);
                case Tokenizer.MINUS:
                    return new DoubleValue(value - ((DoubleValue)other).value);
                case Tokenizer.MULT:
                    return new DoubleValue(value * ((DoubleValue)other).value);
                case Tokenizer.DIV:
                    return new DoubleValue(value / ((DoubleValue)other).value);
                case Tokenizer.MOD:
                    return new DoubleValue(value % ((DoubleValue)other).value);
                default:
                    throw new UnsupportedOperationException("Unknown operator");
            }
        } else {
            return arithmetic(operator, (DoubleValue)other.convert(Type.DOUBLE));
        }
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(DoubleValue.class)) return 0;

        if (required==boolean.class) return 14;
        if (required==Boolean.class) return 15;
        if (required==byte.class) return 12;
        if (required==Byte.class) return 13;
        if (required==char.class) return 10;
        if (required==Character.class) return 11;
        if (required==double.class) return 0;
        if (required==Double.class) return 1;
        if (required==float.class) return 2;
        if (required==Float.class) return 3;
        if (required==int.class) return 6;
        if (required==Integer.class) return 7;
        if (required==long.class) return 4;
        if (required==Long.class) return 5;
        if (required==short.class) return 8;
        if (required==Short.class) return 9;
        if (required==String.class || required==CharSequence.class) return 16;

        return Integer.MAX_VALUE;
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target==Object.class) {
            return new Double(value);
        } else if (target.isAssignableFrom(DoubleValue.class)) {
            return this;
        } else if (target==boolean.class) {
            return new Boolean(asBoolean());
        } else if (target==Boolean.class) {
            return new Boolean(asBoolean());
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class || target==Double.class) {
            return new Double(value);
        } else if (target==float.class ||target==Float.class ) {
            return new Float(value);
        } else if (target==long.class || target==Long.class) {
            return new Long((long)value);
        } else if (target==int.class || target==Integer.class) {
            return new Integer((int)value);
        } else if (target==short.class || target==Short.class) {
            return new Short((short)value);
        } else if (target==byte.class || target==Byte.class) {
            return new Byte((byte)value);
        } else if (target==char.class || target==Character.class) {
            return new Character((char)value);
        } else {
            throw new XPathException.Dynamic("Conversion of number to " + target.getName() +
                        " is not supported");
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "number (" + getStringValue() + ")" );
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file except the asStringXT() and zeros() methods (not currently used).
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (xt) are Copyright (C) (James Clark). All Rights Reserved.
//
// Contributor(s): none.
//

