package net.sf.saxon.value;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.ComputedExpression;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.om.NamePool;


/**
* An XPath value that encapsulates a Java object. Such a value can only be obtained by
* calling an extension function that returns it.
*/

public class ObjectValue extends AtomicValue {
    private Object value;

    /**
    * Constructor
    * @param object the object to be encapsulated
    */

    public ObjectValue(Object object) {
        this.value = object;
    }

    /**
    * Convert to target data type
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.ATOMIC:
        case Type.OBJECT:
        case Type.ITEM:
            return this;
        case Type.BOOLEAN:
            return BooleanValue.get(
                    (value==null ? false : value.toString().length() > 0));
        case Type.STRING:
            return new StringValue(getStringValue());
        default:
            return new StringValue(getStringValue()).convert(requiredType);
        }
    }

    /**
    * Get the value as a String
    * @return a String representation of the value
    */

    public String getStringValue() {
        return (value==null ? "" : value.toString());
    }

    /**
    * Determine the data type of the expression
    * @return Type.OBJECT
    */

    public int getItemType() {
        return Type.OBJECT;
    }

    /**
    * Get the encapsulated object
    */

    public Object getObject() {
        return value;
    }

    /**
    * Determine if two ObjectValues are equal
    * @throws ClassCastException if they are not comparable
    */

    public boolean equals(Object other) {
        return this.value.equals(((ObjectValue)other).value);
    }

    public int hashCode() {
        return value.hashCode();
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {
        if (required==value.getClass()) return 0;
        if (required==Object.class) return 2;
        if (required==String.class || required==CharSequence.class) return 3;
        if (required==boolean.class) return 4;
        if (required==Boolean.class) return 5;
        if (required==byte.class) return 6;
        if (required==Byte.class) return 7;
        if (required==char.class) return 8;
        if (required==Character.class) return 9;
        if (required==double.class) return 10;
        if (required==Double.class) return 11;
        if (required==float.class) return 12;
        if (required==Float.class) return 13;
        if (required==int.class) return 14;
        if (required==Integer.class) return 15;
        if (required==long.class) return 16;
        if (required==Long.class) return 17;
        if (required==short.class) return 18;
        if (required==Short.class) return 19;
        if (required.isAssignableFrom(value.getClass())) return 1;
        return Integer.MAX_VALUE;
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {

        if (value==null) return null;

        if (target.isAssignableFrom(value.getClass())) {
            return value;
        } else if (target==Value.class || target==ObjectValue.class) {
            return this;
        } else if (target==boolean.class || target==Boolean.class) {
            BooleanValue bval = (BooleanValue)convert(Type.BOOLEAN);
            return new Boolean(bval.getValue());
        } else if (target==String.class || target==CharSequence.class) {
            return getStringValue();
        } else if (target==double.class || target==Double.class) {
            DoubleValue bval = (DoubleValue)convert(Type.DOUBLE);
            return new Double(bval.getValue());
        } else if (target==float.class || target==Float.class) {
            DoubleValue bval = (DoubleValue)convert(Type.FLOAT);
            return new Float(bval.getValue());
        } else if (target==long.class || target==Long.class) {
            IntegerValue bval = (IntegerValue)convert(Type.INTEGER);
            return new Long(bval.getValue());
        } else if (target==int.class || target==Integer.class) {
            IntegerValue bval = (IntegerValue)convert(Type.INTEGER);
            return new Integer((int)bval.getValue());
        } else if (target==short.class || target==Short.class) {
            IntegerValue bval = (IntegerValue)convert(Type.INTEGER);
            return new Short((short)bval.getValue());
        } else if (target==byte.class || target==Byte.class) {
            IntegerValue bval = (IntegerValue)convert(Type.INTEGER);
            return new Byte((byte)bval.getValue());
        } else if (target==char.class || target==Character.class) {
            String s = getStringValue();
            if (s.length()==1) {
                return new Character(s.charAt(0));
            } else {
                throw new XPathException.Dynamic("Cannot convert string to Java char unless length is 1");
            }
        } else {
            throw new XPathException.Dynamic("Conversion of external object to " + target.getName() +
                        " is not supported");
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "** external object **" );
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

