package net.sf.saxon.value;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.om.XMLChar;
import net.sf.saxon.om.NamePool;
import net.sf.saxon.functions.Component;
import net.sf.saxon.expr.ExpressionTool;


/**
* A QName value
*/

public final class QNameValue extends AtomicValue {

    private String uri;           // null indicates "no namespace"
    private String localName;

    /**
    * Constructor
    * @param uri The namespace part of the QName
    * @param localName The local part of the QName.
    */

    public QNameValue(String uri, String localName) throws XPathException {
        this.uri = uri;
        this.localName = localName;
        if (!XMLChar.isValidNCName(localName)) {
            throw new XPathException.Dynamic("Malformed local name in QName: '" + localName + "'");
        }
    }

    /**
    * Get the string value as a String. QNames are the only kind of values that cannot be converted
    * to a string.
    */

    public String getStringValue() throws XPathException {
        throw new XPathException.Dynamic("QName cannot be converted to string");
    }

    /**
    * Get the local part
    */

    public String getLocalName() {
        return localName;
    }

    /**
    * Get the namespace part (null means no namespace)
    */

    public String getNamespaceURI() {
        return uri;
    }

    /**
    * Get a component. Returns null if the namespace-uri component is
    * requested and is not present.
    * @param part either Component.LOCALNAME or Componenet.NAMESPACE indicating which
    * component of the value is required
    * @return either the local name or the namespace URI, in each case as a StringValue
    */

    public AtomicValue getComponent(int part) {
        if (part==Component.LOCALNAME) {
            return new StringValue(localName);
        } else if (part==Component.NAMESPACE) {
            if (uri==null) {
                return null;
            } else {
                return new StringValue(uri);
            }
        } else {
            throw new UnsupportedOperationException("Component of QName must be URI or Local Name");
        }
    }

    /**
    * Convert to target data type
    * @param requiredType an integer identifying the required atomic type
    * @return an AtomicValue, a value of the required type
    * @throws XPathException if the conversion is not possible
    */

    public AtomicValue convert(int requiredType) throws XPathException {
        switch(requiredType) {
        case Type.ATOMIC:
        case Type.ITEM:
        case Type.QNAME:
            return this;
        default:
            throw new XPathException.Dynamic("Cannot convert QName to " +
                                     Type.getTypeName(requiredType));
        }
    }

    /**
    * Return the type of the expression
    * @return Type.QNAME (always)
    */

    public int getItemType() {
        return Type.QNAME;
    }


    /**
    * Determine if two anyURI values are equal
    * @throws ClassCastException if they are not comparable
    */

    public boolean equals(Object other) {
        QNameValue val = (QNameValue)other;
        return this.uri.equals(val.uri) && this.localName.equals(val.localName);
    }

    public int hashCode() {
        return uri.hashCode() ^ localName.hashCode();
    }

    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {

        if (required==Object.class) return 50;
        if (required.isAssignableFrom(QNameValue.class)) return 0;
        return Integer.MAX_VALUE;
    }


    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {
        if (target.isAssignableFrom(QNameValue.class)) {
            return this;
        } else {
            throw new XPathException.Dynamic("Conversion of anyURI to " + target.getName() +
                        " is not supported");
        }
    }

    /**
    * Diagnostic print of expression structure
    */

    public void display(int level, NamePool pool) {
        System.err.println(ExpressionTool.indent(level) + "QName (\"{" + uri + "}" + localName + "\")" );
    }

    public String toString() {
        return "\"{" + uri + "}" + localName + "\"";
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//

