package net.sf.saxon.value;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.expr.ComputedExpression;
import net.sf.saxon.expr.XPathContext;
import net.sf.saxon.expr.ExpressionTool;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.Item;
import net.sf.saxon.dom.DOMNodeList;
import net.sf.saxon.om.NamePool;

import java.util.List;
import java.util.ArrayList;

import org.w3c.dom.*;

/**
* A SequenceValue object represents a sequence whose members
* are all AtomicValues or Nodes.
*/


public abstract class SequenceValue extends Value {

    /**
    * Determine the data type of the items in the expression, if possible
    * @return Type.ITEM (not known)
    */

    public int getItemType() {
        return Type.ITEM;
    }

    /**
    * Determine the cardinality
    */

    public int getCardinality() {
        try {
            SequenceIterator iter = iterate(null);
            if (iter.hasNext()) {
                iter.next();
                if (iter.hasNext()) {
                    return StaticProperty.CARDINALITY_ALLOWS_ONE_OR_MORE;
                } else {
                    return StaticProperty.CARDINALITY_EXACTLY_ONE;
                }
            } else {
                return StaticProperty.EMPTY;
            }
        } catch (XPathException err) {
            // can't actually happen
            return StaticProperty.CARDINALITY_ALLOWS_ZERO_OR_MORE;
        }
    }

    /**
    * Evaluate as a string. Returns the string value of the first value in the sequence
    * @return the string value of the expression
    */

//    public String getStringValue() throws XPathException {
//        SequenceIterator iter = iterate(null);
//        if (iter.hasNext()) {
//            return iter.next().getStringValue();
//        } else {
//            return "";
//        }
//    }

    /**
    * Get the n'th item in the sequence (starting from 0). This is defined for all
    * SequenceValues, but its real benefits come for a SequenceValue stored extensionally
    */

    public Item itemAt(int n) throws XPathException {
        if (n<0) return null;
        int i=0;        // indexing is zero-based
        SequenceIterator iter = iterate(null);
        while (iter.hasNext()) {
            Item item = iter.next();
            if (i++ == n) {
                return item;
            }
        }
        return null;
    }

    /**
     * Evaluate as a singleton item (or empty sequence)
     */

    public Item evaluateItem(XPathContext context) throws XPathException {
        SequenceIterator iter = iterate(context);
        if (iter.hasNext()) {
            return iter.next();
            // No check to see if there is a second item; this method shouldn't be called
            // it there is.
        } else {
            return null;
        }
    }

    /**
     * Evaluate an expression as a String. This function must only be called in contexts
     * where it is known that the expression will return a single string (or where an empty sequence
     * is to be treated as a zero-length string). Implementations should not attempt to convert
     * the result to a string, other than converting () to "". This method is used mainly to
     * evaluate expressions produced by compiling an attribute value template.
     *
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @exception ClassCastException if the result type of the
     *     expression is not xs:string?
     * @param context The context in which the expression is to be evaluated
     * @return the value of the expression, evaluated in the current context.
     *     The expression must return a string or (); if the value of the
     *     expression is (), this method returns "".
     */

    public String evaluateAsString(XPathContext context) throws XPathException {
        StringValue value = (StringValue)evaluateItem(context);
        if (value==null) return "";
        return value.getStringValue();
    }


    /**
     * Get the effective boolean value of the expression. This returns false if the value
     * is the empty sequence, a zero-length string, a number equal to zero, or the boolean
     * false. Otherwise it returns true.
     *
     * @param context The context in which the expression is to be evaluated
     * @exception XPathException if any dynamic error occurs evaluating the
     *     expression
     * @return the effective boolean value
     */

    public boolean effectiveBooleanValue(XPathContext context) throws XPathException {
        SequenceIterator it = iterate(context);
        if (!it.hasNext()) {
            return false;
        }
        Item first = it.next();
        if (first instanceof NodeInfo) {
            return true;
        } else {
            if (first instanceof BooleanValue) {
                return ((BooleanValue)first).getValue() || it.hasNext();
            } else if (first instanceof StringValue) {
                return (!first.getStringValue().equals("")) || it.hasNext();
            } else if (first instanceof NumericValue) {
                // first==first is a test for NaN
                return it.hasNext() || (!(first.equals(DoubleValue.ZERO)) && first.equals(first));
            } else {
                return true;
            }
        }
    }


    /**
    * Get conversion preference for this value to a Java class. A low result
    * indicates higher preference.
    */

    public int conversionPreference(Class required) {
        if (required==Object.class) return 50;
        if (required.isAssignableFrom(SequenceIterator.class)) return 0;
        if (required.isAssignableFrom(SequenceValue.class)) return 1;
        if (required.isAssignableFrom(List.class)) return 2;
        if (required.isAssignableFrom(NodeList.class)) return 3;

        // If an atomic type is required, base the decision on the type of the
        // first item in the sequence.

        try {
            SequenceIterator iter = iterate(null);
            if (iter.hasNext()) {
                Item item = iter.next();
                if (item instanceof NodeInfo) {
                    return nodeToJavaPreference((NodeInfo)item, required);
                } else {
                    return ((AtomicValue)item).conversionPreference(required);
                }
            }
        } catch (XPathException err) {}
        return Integer.MAX_VALUE;
    }

    public static int nodeToJavaPreference(NodeInfo node, Class required) {
        if (required==Object.class) return 50;
        if (NodeInfo.class.isAssignableFrom(required)) return 6;
        if (Node.class.isAssignableFrom(required)) return 7;

        try {
            Value value = node.getTypedValue();
            return value.conversionPreference(required);
        } catch (XPathException err) {
            return Integer.MAX_VALUE;
        }
    }

    /**
    * Convert to Java object (for passing to external functions)
    */

    public Object convertToJava(Class target) throws XPathException {

        if (target==Object.class) {
            return iterate(null);

        } else if (target.isAssignableFrom(SequenceValue.class)) {
            return this;

        } else if (target.isAssignableFrom(SequenceIterator.class)) {
            return iterate(null);

        } else if (target.isAssignableFrom(List.class)) {
            ArrayList list = new ArrayList();
            SequenceIterator iter = iterate(null);
            while (iter.hasNext()) {
                list.add(iter.next());
            }
            return list;
        } else if (target.isAssignableFrom(NodeList.class)) {
            if (this instanceof SequenceExtent) {
                return new DOMNodeList((SequenceExtent)this);
            } else {
                SequenceExtent ext = new SequenceExtent(iterate(null));
                return new DOMNodeList(ext);
            }
        } else {
            // try converting the first item in the sequence
            SequenceIterator iter = iterate(null);
            while (iter.hasNext()) {
                Item next = iter.next();
                if (next instanceof NodeInfo) {
                    if (target.isAssignableFrom(next.getClass())) {
                        // should cover both NodeInfo and DOM Node
                        return next;
                    } else {
                        return next.getTypedValue().convertToJava(target);
                    }
                } else {
                    return ((AtomicValue)next).convertToJava(target);
                }
            }
            // value is an empty sequence: pass it as a Java null
            return null;
        }

    }

    /**
    * Diagnostic display of the expression
    */

    public void display(int level, NamePool pool) {
        try {
            System.err.println(ExpressionTool.indent(level) + "sequence of " + Type.getTypeName(getItemType()) + " (");
            SequenceIterator iter = iterate(null);
            while (iter.hasNext()) {
                ((ComputedExpression)iter.next()).display(level+1, pool);
            }
            System.err.println(ExpressionTool.indent(level) + ")");
        } catch (Exception err) {}
    }
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of International Computers Limited (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
