package net.sf.saxon.value;
import net.sf.saxon.xpath.XPathException;
import net.sf.saxon.om.NamespaceConstant;
import java.util.HashMap;
import java.io.Serializable;

/**
 * This class contains static information about types and methods for constructing type codes.
 * The class is never instantiated.
 */

public abstract class Type implements Serializable {

    private static HashMap hierarchy = new HashMap(100);

    /**
     * Type representing an element node - element()
     */

    public static final short ELEMENT = 1;
    /**
     * Item type representing an attribute node - attribute()
     */
    public static final short ATTRIBUTE = 2;
    /**
     * Item type representing a text node - text()
     */
    public static final short TEXT = 3;
    /**
     * Item type representing a processing-instruction node
     */
    public static final short PROCESSING_INSTRUCTION = 7;
    /**
     * Item type representing a comment node
     */
    public static final short COMMENT = 8;
    /**
     * Item type representing a document node
     */
    public static final short DOCUMENT = 9;
    /**
     * Item type representing a namespace node
     */
    public static final short NAMESPACE = 13;

    /**
     * An item type that matches any node
     */

    public static final short NODE = 0;

    /**
     * An item type that matches any item
     */

    public static final short ITEM = 88;



    /**
     * A type that matches nothing
     */

    public static final short MAX_NODE_TYPE = 13;
    /**
     * Item type that matches no items (corresponds to SequenceType empty())
     */
    public static final short EMPTY = 15;    // a test for this type will never be satisfied

    /**
     * Test whether a given type is (some subtype of) node()
     *
     * @param type The type to be tested
     * @return true if the item type is node() or a subtype of node()
     */

    public static boolean isNodeType(int type) {
        return type < MAX_NODE_TYPE;
    }

    /**
     * Test whether a given type is (some subtype of) ATOMIC
     *
     * @param type the type to be tested
     * @return true if the item type is xdt:atomicValue or a subtype thereof
     */

//    public static boolean isAtomicType(int type) {
//        return type >= 90;
//    }

    /**
     * Constant denoting any atomic type (the union of all primitive types and types
     * derived from primitive types by restriction or by union)
     */

    public final static int ATOMIC          = 90;

    /**
     * Constant denoting any numeric type (the union of float, double, and decimal)
     */

    public final static int NUMBER          = 91;

    /**
     * Constant denoting the type of a simple-valued element or attribute whose actual
     * type is unknown
     */

    public final static int ANY_SIMPLE_TYPE = 92;

    /**
     * Constant denoting the type of a simple-or-complex-valued element that is
     * allowed to have any content
     */

    public final static int ANY_TYPE = 93;

    /**
     * Constant denoting the type of a simple-or-complex-valued element whose actual
     * type is unknown
     */

    public final static int UNTYPED_ANY = 94;

    /**
     * Constants representing primitive data types defined in Schema Part 2
     */

    public final static int STRING          = 101;
    /**
     * Item type representing the type xs:boolean
     */
    public final static int BOOLEAN         = 102;
    /**
     * Item type representing the type xs:decimal
     */
    public final static int DECIMAL         = 103;
    /**
     * Item type representing the type xs:float
     */
    public final static int FLOAT           = 104;
    /**
     * Item type representing the type xs:double
     */
    public final static int DOUBLE          = 105;
    /**
     * Item type representing the type xs:duration
     */
    public final static int DURATION        = 106;
    /**
     * Item type representing the type xs:dateTime
     */
    public final static int DATE_TIME       = 107;
    /**
     * Item type representing the type xs:time
     */
    public final static int TIME            = 108;
    /**
     * Item type representing the type xs:date
     */
    public final static int DATE            = 109;
    /**
     * Item type representing the type xs:gYearMonth
     */
    public final static int G_YEAR_MONTH    = 110;
    /**
     * Item type representing the type xs:gYear
     */
    public final static int G_YEAR          = 111;
    /**
     * Item type representing the type xs:monthDay
     */
    public final static int G_MONTH_DAY     = 112;
    /**
     * Item type representing the type xs:gDay
     */
    public final static int G_DAY           = 113;
    /**
     * Item type representing the type xs:gMonth
     */
    public final static int G_MONTH         = 114;
    /**
     * Item type representing the type xs:hexBinary
     */
    public final static int HEX_BINARY      = 115;
    /**
     * Item type representing the type xs:base64Binary
     */
    public final static int BASE64_BINARY   = 116;
    /**
     * Item type representing the type xs:anyURI
     */
    public final static int ANY_URI         = 117;
    /**
     * Item type representing the type xs:QName
     */
    public final static int QNAME           = 118;
    /**
     * Item type representing the type xs:NOTATION
     *
     */
    public final static int NOTATION        = 119;

    /**
     * Item type representing the type xdt:untypedAtomic
     * (the type of the content of a schema-less node)
     */

    public final static int UNTYPED_ATOMIC  = 125;

    /**
     * Constant representing the type of an external object (for use by extension functions)
     */

    public final static int OBJECT          = 130;

    /**
     * Item type representing the type xs:integer
     */

    public final static int INTEGER                 = 200;
    /**
     * Item type representing the type xs:nonPositiveInteger
     */
    public final static int NON_POSITIVE_INTEGER    = 201;
    /**
     * Item type representing the type xs:negativeInteger
     */
    public final static int NEGATIVE_INTEGER        = 202;
    /**
     * Item type representing the type xs:long
     */
    public final static int LONG                    = 203;
    /**
     * Item type representing the type xs:int
     */
    public final static int INT                     = 204;
    /**
     * Item type representing the type xs:short
     */
    public final static int SHORT                   = 205;
    /**
     * Item type representing the type xs:byte
     */
    public final static int BYTE                    = 206;
    /**
     * Item type representing the type xs:nonNegativeInteger
     */
    public final static int NON_NEGATIVE_INTEGER    = 207;
    /**
     * Item type representing the type xs:positiveInteger
     */
    public final static int POSITIVE_INTEGER        = 208;
    /**
     * Item type representing the type xs:unsignedLong
     */
    public final static int UNSIGNED_LONG           = 209;
    /**
     * Item type representing the type xs:unsignedInt
     */
    public final static int UNSIGNED_INT            = 210;
    /**
     * Item type representing the type xs:unsignedShort
     */
    public final static int UNSIGNED_SHORT          = 211;
    /**
     * Item type representing the type xs:unsignedByte
     */
    public final static int UNSIGNED_BYTE           = 212;

    /**
     * Item type representing the type xs:normalizedString
     */
    public final static int NORMALIZED_STRING       = 220;
    /**
     * Item type representing the type xs:token
     */
    public final static int TOKEN                   = 221;
    /**
     * Item type representing the type xs:language
     */
    public final static int LANGUAGE                = 222;
    /**
     * Item type representing the type xs:NMTOKEN
     */
    public final static int NMTOKEN                 = 223;
    /**
     * Content type representing the complex type xs:NMTOKENS
     */
    public final static int NMTOKENS                = 224;      // NB: list type
    /**
     * Item type representing the type xs:NAME
     */
    public final static int NAME                    = 225;
    /**
     * Item type representing the type xs:NCNAME
     */
    public final static int NCNAME                  = 226;
    /**
     * Item type representing the type xs:ID
     */
    public final static int ID                      = 227;
    /**
     * Item type representing the type xs:IDREF
     */
    public final static int IDREF                   = 228;
    /**
     * Content type representing the complex type xs:IDREFS
     */
    public final static int IDREFS                  = 229;      // NB: list type
    /**
     *
     * Item type representing the type xs:ENTITY
     */
    public final static int ENTITY                  = 230;
    /**
     * Content type representing the complex type xs:ENTITIES
     */
    public final static int ENTITIES                = 231;      // NB: list type

    /**
     *
     * Item type representing the type xdt:yearMonthDuration
     */
    public final static int YEAR_MONTH_DURATION     = 240;
    /**
     *
     * Item type representing the type xdt:dayTimeDuration
     */
    public final static int DAY_TIME_DURATION       = 241;

    /**
     * Pseudo-type; constant used in function signatures to indicate that
     * the return item type of the function is the same as the actual item
     * type of the first argument
     */
    public final static int SAME_AS_FIRST_ARGUMENT  = 999999;


    static String[] typeName = new String[255];
    static HashMap typeTable = new HashMap(255);

    /**
    * Define the names of the built-in schema types
    */

    static {
        defineType(ITEM,                        "item");
        defineType(NODE,                        "node");
        defineType(ELEMENT,                         "element");
        defineType(DOCUMENT,                        "document");
        defineType(ATTRIBUTE,                       "attribute");
        defineType(TEXT,                            "text");
        defineType(COMMENT,                         "comment");
        defineType(PROCESSING_INSTRUCTION,          "processing-instruction");
        defineType(NAMESPACE,                       "namespace");

        defineType(EMPTY,                           "empty");
        defineType(NUMBER,                          "number");

        defineBuiltInType(UNTYPED_ATOMIC,           "xdt:untypedAtomic");
        defineBuiltInType(ATOMIC,                   "xdt:anyAtomicType");
        defineBuiltInType(YEAR_MONTH_DURATION,     "xdt:yearMonthDuration");
        defineBuiltInType(DAY_TIME_DURATION,       "xdt:dayTimeDuration");

        defineBuiltInType(STRING,                  "xs:string");
        defineBuiltInType(BOOLEAN,                 "xs:boolean");
        defineBuiltInType(DECIMAL,                 "xs:decimal");
        defineBuiltInType(FLOAT,                   "xs:float");
        defineBuiltInType(DOUBLE,                  "xs:double");
        defineBuiltInType(DURATION,                "xs:duration");
        defineBuiltInType(DATE_TIME,               "xs:dateTime");
        defineBuiltInType(TIME,                    "xs:time");
        defineBuiltInType(DATE,                    "xs:date");
        defineBuiltInType(G_YEAR_MONTH,            "xs:gYearMonth");
        defineBuiltInType(G_YEAR,                  "xs:gYear");
        defineBuiltInType(G_MONTH_DAY,             "xs:gMonthDay");
        defineBuiltInType(G_DAY,                   "xs:gDay");
        defineBuiltInType(G_MONTH,                 "xs:gMonth");
        defineBuiltInType(HEX_BINARY,              "xs:hexBinary");
        defineBuiltInType(BASE64_BINARY,           "xs:base64Binary");
        defineBuiltInType(ANY_URI,                 "xs:anyURI");
        defineBuiltInType(QNAME,                   "xs:QName");
        defineBuiltInType(NOTATION,                "xs:NOTATION");
        defineBuiltInType(INTEGER,                 "xs:integer");
        defineBuiltInType(NON_POSITIVE_INTEGER,    "xs:nonPositiveInteger");
        defineBuiltInType(NEGATIVE_INTEGER,        "xs:negativeInteger");
        defineBuiltInType(LONG,                    "xs:long");
        defineBuiltInType(INT,                     "xs:int");
        defineBuiltInType(SHORT,                   "xs:short");
        defineBuiltInType(BYTE,                    "xs:byte");
        defineBuiltInType(NON_NEGATIVE_INTEGER,    "xs:nonNegativeInteger");
        defineBuiltInType(POSITIVE_INTEGER,        "xs:positiveInteger");
        defineBuiltInType(UNSIGNED_LONG,           "xs:unsignedLong");
        defineBuiltInType(UNSIGNED_INT,            "xs:unsignedInt");
        defineBuiltInType(UNSIGNED_SHORT,          "xs:unsignedShort");
        defineBuiltInType(UNSIGNED_BYTE,           "xs:unsignedByte");
        defineBuiltInType(NORMALIZED_STRING,       "xs:normalizedString");
        defineBuiltInType(TOKEN ,                  "xs:token");
        defineBuiltInType(LANGUAGE,                "xs:language");
        defineBuiltInType(NMTOKEN,                 "xs:NMTOKEN");
        defineBuiltInType(NMTOKENS,                "xs:NMTOKENS");
        defineBuiltInType(NAME,                    "xs:Name");
        defineBuiltInType(NCNAME,                  "xs:NCName");
        defineBuiltInType(ID,                      "xs:ID");
        defineBuiltInType(IDREF,                   "xs:IDREF");
        defineBuiltInType(IDREFS,                  "xs:IDREFS");
        defineBuiltInType(ENTITY,                  "xs:ENTITY");
        defineBuiltInType(ENTITIES,                "xs:ENTITIES");

        defineBuiltInType(OBJECT,                  "java:java.lang.Object");
    }

    /**
     * Add a type to the table of built-in types
     *
     * @param type the integer code of the type
     * @param name the name of the type (using a conventional prefix,
     *     eg. xs: or xdt:)
     */
    static void defineBuiltInType(int type, String name) {
        // TODO: this will need to be changed when user-defined types are supported
        typeName[type] = name;
        typeTable.put(name, new Integer(type));
    }

    /**
     * Register a type that is identified by a QName
     *
     * @param type the integer code of the type
     * @param name the name of the type
     */

    static void defineType(int type, String name) {
        typeName[type] = name;
        typeTable.put(name, new Integer(type));
    }

    /**
     * Get the name of the item type with a given numeric type code
     *
     * @param type the numeric code of the required item type
     * @return the name of the type, as a QName using a conventional prefix
     *     such as "xs" or "xdt" or "java"
     */
    public final static String getTypeName(int type) {
        if (type<=255) {
            return typeName[type];
        } else {
            return ""+type;
        }
    }

    /**
     * Get the integer code for a type with a given name.
     * At this level, built-in types must be prefixed "xs:" or "xdt:" or "java:"
     *
     * @param name the name of the type, using a conventional prefix such as
     *     "xs" or "xdt"
     * @exception XPathException if the type is not known
     * @return the integer code for the specified type
     */

    private final static int getType(String name) throws XPathException {
        Integer t = (Integer)typeTable.get(name);
        if (t==null) {
            throw new XPathException.Static("Unknown type " + name);
        }
        return t.intValue();
    }

    /**
    * Define the type hierarchy
    */

    static {
        defineSubType(ITEM, NODE);

        defineSubType(NODE, ELEMENT);
        defineSubType(NODE, ATTRIBUTE);
        defineSubType(NODE, DOCUMENT);
        defineSubType(NODE, TEXT);
        defineSubType(NODE, COMMENT);
        defineSubType(NODE, PROCESSING_INSTRUCTION);
        defineSubType(NODE, NAMESPACE);

        defineSubType(ITEM, ATOMIC);

        defineSubType(ATOMIC, STRING);
        defineSubType(ATOMIC, BOOLEAN);
        defineSubType(ATOMIC, NUMBER);
        defineSubType(ATOMIC, DURATION);
        defineSubType(ATOMIC, DATE_TIME);
        defineSubType(ATOMIC, TIME);
        defineSubType(ATOMIC, DATE);
        defineSubType(ATOMIC, G_YEAR_MONTH);
        defineSubType(ATOMIC, G_YEAR);
        defineSubType(ATOMIC, G_MONTH_DAY);
        defineSubType(ATOMIC, G_DAY);
        defineSubType(ATOMIC, G_MONTH);
        defineSubType(ATOMIC, HEX_BINARY);
        defineSubType(ATOMIC, BASE64_BINARY);
        defineSubType(ATOMIC, ANY_URI);
        defineSubType(ATOMIC, QNAME);
        defineSubType(ATOMIC, NOTATION);

        defineSubType(ATOMIC, UNTYPED_ATOMIC);

        defineSubType(ATOMIC, OBJECT);

        defineSubType(NUMBER, DECIMAL);
        defineSubType(NUMBER, DOUBLE);
        defineSubType(NUMBER, FLOAT);

        defineSubType(DECIMAL, INTEGER);

        defineSubType(INTEGER, NON_POSITIVE_INTEGER);
        defineSubType(NON_POSITIVE_INTEGER, NEGATIVE_INTEGER);
        defineSubType(INTEGER, LONG);
        defineSubType(LONG, INT);
        defineSubType(INT, SHORT);
        defineSubType(SHORT, BYTE);
        defineSubType(INTEGER, NON_NEGATIVE_INTEGER);
        defineSubType(NON_NEGATIVE_INTEGER, POSITIVE_INTEGER);
        defineSubType(NON_NEGATIVE_INTEGER, UNSIGNED_LONG);
        defineSubType(UNSIGNED_LONG, UNSIGNED_INT);
        defineSubType(UNSIGNED_INT, UNSIGNED_SHORT);
        defineSubType(UNSIGNED_SHORT, UNSIGNED_BYTE);

        defineSubType(DURATION, YEAR_MONTH_DURATION);
        defineSubType(DURATION, DAY_TIME_DURATION);

        defineSubType(STRING, NORMALIZED_STRING);
        defineSubType(NORMALIZED_STRING, TOKEN);
        defineSubType(TOKEN, LANGUAGE);
        defineSubType(TOKEN, NAME);
        defineSubType(TOKEN, NMTOKEN);
        defineSubType(NAME, NCNAME);
        defineSubType(NCNAME, ID);
        defineSubType(NCNAME, IDREF);
        defineSubType(NCNAME, ENTITY);
    }

    /**
     * Register a subtype/supertype relationship between two item types
     *
     * @param supertype the numeric code of the supertype
     * @param subtype the numeric code of the subtype
     */
    private static void defineSubType(int supertype, int subtype) {
        hierarchy.put(new Integer(subtype), new Integer(supertype));
    }

    /**
     * Get the immediate supertype of a given type.
     *
     * @param type the type whose supertype is required. This must not be the
     *     most general type ITEM
     * @return the numeric code of the supertype
     */
    private static int getSuperType(int type) {
        Object s = hierarchy.get(new Integer(type));
        if (s==null) {
            throw new IllegalArgumentException("No supertype for type " + type);
        }
        return ((Integer)s).intValue();
    }

    /**
     * Determine whether type A is type B or one of its subtypes, recursively
     *
     * @param subtype numeric code identifying the first type
     * @param supertype numeric code identifying the second type
     * @return true if the first type is the second type or a (direct or
     *     indirect) subtype of the second type
     */

    public static boolean isSubType(int subtype, int supertype) {
        // System.err.println("isSubtype(" + (subtype) + ", " + (supertype) + ")");
        // System.err.println("isSubtype(" + getTypeName(subtype) + ", " + getTypeName(supertype) + ")");
        if (subtype==supertype) return true;
        if (supertype==ITEM) return true;
        if (subtype==ITEM) return false;
        if (subtype==UNTYPED_ANY) return false;
        if (subtype==EMPTY) return true;        // an empty sequence can satisfy any required item type
        int s = getSuperType(subtype);
        return isSubType(s, supertype);
    }

    /**
     * Determine whether type A is numerically promotable to type B, for example
     * xs:integer is promotable to xs:double
     */

    public static boolean isPromotable(int sourceType, int targetType) {
        if (isSubType(sourceType, DECIMAL)) {
            return (targetType==FLOAT || targetType==DOUBLE);
        }
        if (isSubType(targetType, FLOAT)) {
            return (targetType==DOUBLE);
        }
        return false;
    }

    /**
     * For an atomic type, get the primitive type (treating NUMBER as a primitive type)
     *
     * @param type the item type whose supertype is required
     * @return the primitive type that is a supertype of the given item type
     */

    public static int getPrimitiveType(int type) {
        if (type==ITEM) {
            throw new IllegalArgumentException("Supplied type " + type + " is not atomic");
        }
        int s = getSuperType(type);
        if (s == ATOMIC) {
            return type;
        } else {
            return getPrimitiveType(s);
        }
    }

    /**
     * Get a type that is a common supertype of two given types
     *
     * @param t1 the first item type
     * @param t2 the second item type
     * @return the numeric code of the item type that is a supertype of both
     *     the supplied item types
     */

    public final static int getCommonSuperType(int t1, int t2) {
        if (t1==Type.EMPTY) {
            return t2;
        }
        if (t2==Type.EMPTY) {
            return t1;
        }
        if (t1==t2) {
            return t1;
        } else if (Type.isSubType(t1, t2)) {
            return t2;
        } else if (Type.isSubType(t2, t1)) {
            return t1;
        } else {
            return getCommonSuperType(getSuperType(t2), t1);
            // eventually we will hit a type that is a supertype of t2. We reverse
            // the arguments so we go up each branch of the tree alternately.
            // If we hit the root of the tree, one of the earlier conditions will be satisfied,
            // so the recursion will stop.
        }
    }


    /**
     * Get the type with a given expanded name
     *
     * @param uri the URI part of the type name
     * @param localName the local part of the type name
     * @exception XPathException if the type is unknown
     * @return the numeric type code of the specified type
     */
    public static int getNamedType(String uri, String localName) throws XPathException {
        if (uri.equals(NamespaceConstant.SCHEMA) || uri.equals(NamespaceConstant.SCHEMA_DATATYPES)) {
            // schema built-in types
            return getType("xs:" + localName);
        } else if (uri.equals(NamespaceConstant.XDT)) {
            return getType("xdt:" + localName);
        } else if (uri.equals(NamespaceConstant.JAVA_TYPE)) {
            return getType("java:" + localName);
        } else {
            throw new XPathException.Static("Unknown type " +
                        (uri.equals("") ? "" : "{"+uri+"}") + localName);
        }
    }

    /**
     * Get a system type - that is, one whose name is a keyword rather than a QName. This includes the node
     * kinds such as element and attribute, the generic types node() and item(), and the pseudo-type empty()
     *
     * @param name
     * @exception XPathException
     */
    public static int getSystemType(String name) throws XPathException {
        if (name.equals("item"))                   return Type.ITEM;
        else if (name.equals("document-node"))     return Type.DOCUMENT;
        else if (name.equals("element"))           return Type.ELEMENT;
        else if (name.equals("attribute"))         return Type.ATTRIBUTE;
        else if (name.equals("text"))              return Type.TEXT;
        else if (name.equals("comment"))           return Type.COMMENT;
        else if (name.equals("processing-instruction"))
                                                   return Type.PROCESSING_INSTRUCTION;
        else if (name.equals("namespace"))         return Type.NAMESPACE;
        else if (name.equals("node"))              return Type.NODE;
        else if (name.equals("empty"))             return Type.EMPTY;
        else throw new XPathException.Static("Unknown type " + name);
    }


}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
