package net.sf.saxon.xpath;
import net.sf.saxon.expr.CastExpression;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.FunctionCall;
import net.sf.saxon.expr.FunctionProxy;
import net.sf.saxon.expr.StaticContext;
import net.sf.saxon.expr.StaticProperty;
import net.sf.saxon.expr.VariableDeclaration;
import net.sf.saxon.functions.SystemFunction;
import net.sf.saxon.functions.XSLTFunction;
import net.sf.saxon.instruct.FunctionSignature;
import net.sf.saxon.om.*;
import net.sf.saxon.sort.CodepointCollator;
import net.sf.saxon.sort.CollationFactory;
import net.sf.saxon.value.SequenceType;
import net.sf.saxon.value.Type;

import javax.xml.transform.TransformerException;
import java.util.Comparator;
import java.util.HashMap;

/**
* A StandaloneContext provides a context for parsing an expression or pattern appearing
* in a context other than a stylesheet.
*/

public class StandaloneContext implements StaticContext {

	private NamePool namePool;
	private HashMap namespaces = new HashMap();
	private HashMap collations = new HashMap();
	private HashMap variables = new HashMap();
	private String defaultCollationName = null;

	/**
	* Create a StandaloneContext using the default NamePool
	*/

	public StandaloneContext() {
	    this(NamePool.getDefaultNamePool());
	}

	/**
	* Create a StandaloneContext using a specific NamePool
	*/

	public StandaloneContext(NamePool pool) {
		namePool = pool;
		clearNamespaces();
	}

	/**
	* Create a StandaloneContext using a specific Node. This node is used to initialize
	* the NamePool and also to establish the initial set of in-scope namespaces
	*/

	public StandaloneContext(NodeInfo node) {
	    DocumentInfo doc = node.getDocumentRoot();
	    if (doc==null) {
	        throw new IllegalArgumentException(
	                    "The node used to establish a standalone context must be in a tree whose root is a document node");
	    }
	    namePool = doc.getNamePool();
	    setNamespaces(node);
	}

	/**
	* Declare a namespace whose prefix can be used in expressions
	* @param prefix The namespace prefix. Must not be null. Must not be the empty string
	* ("") - unqualified names in an XPath expression always refer to the null namespace.
	* @param uri The namespace URI. Must not be null.
	*/

	public void declareNamespace(String prefix, String uri) {
	    if (prefix==null) {
	        throw new NullPointerException("Null prefix supplied to declareNamespace()");
	    }
	    if (uri==null) {
	        throw new NullPointerException("Null namespace URI supplied to declareNamespace()");
	    }
		namespaces.put(prefix, uri);
		namePool.allocateNamespaceCode(prefix, uri);
	}

	/**
	* Clear all the declared namespaces, except for the standard ones (xml, xslt, saxon)
	*/

	public void clearNamespaces() {
	    namespaces.clear();
		declareNamespace("xml", NamespaceConstant.XML);
		declareNamespace("xsl", NamespaceConstant.XSLT);
		declareNamespace("saxon", NamespaceConstant.SAXON);
		declareNamespace("xs", NamespaceConstant.SCHEMA);
		declareNamespace("", "");
	}

	/**
	* Set all the declared namespaces to be the namespaces that are in-scope for a given node.
	* In addition, the standard namespaces (xml, xslt, saxon) are declared.
	* @param node The node whose in-scope namespaces are to be used as the context namespaces.
	* Note that this will have no effect unless this node is an element.
	*/

	public void setNamespaces(NodeInfo node) {
	    namespaces.clear();
	    AxisIterator iter = node.iterateAxis(Axis.NAMESPACE);
	    while (iter.hasNext()) {
	        NodeInfo ns = (NodeInfo)iter.next();
	        declareNamespace(ns.getLocalPart(), ns.getStringValue());
	    }
	}

    /**
    * Declare a named collation
    * @param name The name of the collation (technically, a URI)
    * @param comparator The Java Comparator used to implement the collating sequence
    * @param isDefault True if this is to be used as the default collation
    */

    public void declareCollation(String name, Comparator comparator, boolean isDefault) {
        collations.put(name, comparator);
        if (isDefault) {
            defaultCollationName = name;
        }
    }

    /**
    * Declare a variable. A variable must be declared before an expression referring
    * to it is compiled.
    */

    public Variable declareVariable(String qname, Object initialValue) throws XPathException {
        String prefix;
        String localName;
        try {
            String[] parts = Name.getQNameParts(qname);
            prefix = parts[0];
            localName = parts[1];
        } catch (QNameException err) {
            throw new XPathException.Static("Invalid QName for variable: " + qname);
        }
        String uri = "";
        if (!("".equals(prefix))) {
            uri = getURIForPrefix(prefix);
        }
        Variable var = Variable.make(qname);
        var.setValue(initialValue);
        //var.setSlotNumber(variables.size());
        int fingerprint = namePool.allocate(prefix, uri, localName) & 0xfffff;
        variables.put(new Integer(fingerprint), var);
        return var;
    }

    /**
    * Get the NamePool used for compiling expressions
    */

    public NamePool getNamePool() {
        return namePool;
    }

    /**
    * Issue a compile-time warning. This method is used during XPath expression compilation to
    * output warning conditions. The default implementation writes the message to System.err. To
    * change the destination of messages, create a subclass of StandaloneContext that overrides
    * this method.
    */

    public void issueWarning(String s) {
        System.err.println(s);
    }

    /**
    * Get the system ID of the container of the expression. Used to construct error messages.
    * @return "" always
    */

    public String getSystemId() {
        return "";
    }

    /**
    * Get the Base URI of the stylesheet element, for resolving any relative URI's used
    * in the expression.
    * Used by the document() function.
    * @return "" always
    */

    public String getBaseURI() {
        return "";
    }

    /**
    * Get the line number of the expression within that container.
    * Used to construct error messages.
    * @return -1 always
    */

    public int getLineNumber() {
        return -1;
    }

    /**
    * Get the URI for a prefix, using this Element as the context for namespace resolution.
    * This method is provided for use by the XPath parser.
    * @param prefix The prefix
    * @throws XPathException if the prefix is not declared
    */

    public String getURIForPrefix(String prefix) throws XPathException {
    	String uri = (String)namespaces.get(prefix);
    	if (uri==null) {
    		throw new XPathException.Static("Prefix " + prefix + " has not been declared");
    	}
    	return uri;
    }

    /**
    * Bind a variable used in an XPath Expression to the XSLVariable element in which it is declared.
    * This method is provided for use by the XPath parser, and it should not be called by the user of
    * the API, or overridden, unless variables are to be declared using a mechanism other than the
    * declareVariable method of this class.
    */

    public VariableDeclaration bindVariable(int fingerprint) throws XPathException {

        Variable var = (Variable)variables.get(new Integer(fingerprint));
        if (var==null) {
            throw new XPathException.Static("Undeclared variable in a standalone expression");
        } else {
            return var;
        }
    }

    /**
    * Identify a (namespace-prefixed) function appearing in the expression
    * @return an Expression representing the function call. This will normally be
    * a FunctionCall, but it may be rewritten as some other expression.
    * @throws XPathException if the function is not in scope (note, this doesn't
    * happen in XSLT, where a reference to an unknown function is a dynamic error).
    */

    public Expression bindFunction(String qname, Expression[] arguments) throws XPathException {
        int colon = qname.indexOf(":");
        if (colon<0) {
            return bindSystemFunction(qname, arguments);

        } else {        // the function name is prefixed

            String prefix = qname.substring(0, colon);
            String uri = getURIForPrefix(prefix);
            String lname = qname.substring(colon+1);

            int namecode = namePool.allocate(prefix, uri, lname);
            int fingerprint = namecode & 0xfffff;

            if (uri.equals(NamespaceConstant.FN)) {
                return bindSystemFunction(lname, arguments);
            }

            // Some extension functions are specially recognized

            if (fingerprint==namePool.getFingerprint(NamespaceConstant.SAXON, "evaluate")) {
                FunctionCall f = SystemFunction.makeSystemFunction("saxon:evaluate");
                f.setArguments(arguments);
                return f;
            }

            if (fingerprint==namePool.getFingerprint(NamespaceConstant.SAXON, "expression")) {
                FunctionCall f = SystemFunction.makeSystemFunction("saxon:expression");
                f.setArguments(arguments);
                return f;
            }

            if (fingerprint==namePool.getFingerprint(NamespaceConstant.SAXON, "parse")) {
                FunctionCall f = SystemFunction.makeSystemFunction("saxon:parse");
                f.setArguments(arguments);
                return f;
            }

            if (fingerprint==namePool.getFingerprint(NamespaceConstant.SAXON, "serialize")) {
                FunctionCall f = SystemFunction.makeSystemFunction("saxon:serialize");
                f.setArguments(arguments);
                return f;
            }

            // Now try for a constructor function for a built-in type

            if (uri.equals(NamespaceConstant.SCHEMA) || uri.equals(NamespaceConstant.SCHEMA_DATATYPES)
                    || uri.equals(NamespaceConstant.XDT)) {
                // it's a constructor function: treat it as shorthand for a cast expression
                if (arguments.length != 1) {
                    throw new XPathException.Static("A constructor function must have exactly one argument");
                }
                SequenceType type =
                    new SequenceType(Type.getNamedType(uri, lname), Type.ITEM, StaticProperty.CARDINALITY_EXACTLY_ONE);
                return new CastExpression(arguments[0], type);

            }

            // Next, see if we can bind a Java method

			Class className = null;
			try {
			    className = getExternalJavaClass(uri);
			} catch (Exception err) {
			    throw new XPathException.Static("Cannot load external Java class", err);
			}

			if (className!=null) {
                FunctionProxy fp = new FunctionProxy();
                fp.setArguments(arguments);
                fp.setFunctionName(className, lname);
                return fp;
			}

            // We've failed - no implementation of this function was found

            throw new XPathException.Static("Unknown function: " + qname);

        }

    }

    private FunctionCall bindSystemFunction(String functionName, Expression[] arguments) throws XPathException {
        FunctionCall f = SystemFunction.makeSystemFunction(functionName);
        if (f==null) {
            throw new XPathException.Static("Unknown system function: " + functionName);
        }
        if (f instanceof XSLTFunction) {
            throw new XPathException.Static("Function " + functionName + " is available only in XSLT");
        }
        f.setArguments(arguments);
        return f;
    }

//    public FunctionSignature getFunctionSignature(int fingerprint, int arity) {
//        throw new UnsupportedOperationException("method not yet implemented");
//    }

    /**
    * Get an external Java class corresponding to a given namespace prefix, if there is
    * one.
    * @param uri The namespace URI corresponding to the prefix used in the function call.
    * @return the Java class name if a suitable class exists, otherwise return null.
    */

    private Class getExternalJavaClass(String uri) {

        // Try to get a Saxon or EXSLT extension function
        Class c = FunctionProxy.getVendorExtensionClass(uri);

        if (c != null) {
            return c;
        }

        // Failing that, try to identify a class directly from the URI
        return FunctionProxy.getImplicitJavaClass(uri);
    }

    /**
    * Get a named collation.
    * @return the collation identified by the given name, as set previously using declareCollation.
    * Return null if no collation with this name is found.
    */

    public Comparator getCollation(String name) {
        try {
            return CollationFactory.makeCollationFromURI(name);
        } catch (TransformerException e) {
            return null;
        }
    }

    /**
    * Get the name of the default collation.
    * @return the name of the default collation; or the name of the codepoint collation
    * if no default collation has been defined
    */

    public String getDefaultCollationName() {
        if (defaultCollationName != null) {
            return defaultCollationName;
        } else {
            return CodepointCollator.URI;
        }
    }

    /**
    * Get the default XPath namespace, as a namespace code that can be looked up in the NamePool
    */

    public short getDefaultElementNamespace() {
        return NamespaceConstant.NULL_CODE;
    }

    /**
     * Determine whether Backwards Compatible Mode is used
     * @return false; XPath 1.0 compatibility mode is not supported in the standalone
     * XPath API
     */

    public boolean isInBackwardsCompatibleMode() {
        return false;
    }

}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is
// Michael Kay of Software AG (michael.h.kay@ntlworld.com).
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
